"""
 ************************************************************************************************
 **                                                                                            **
 **  File: DoTranslationNow.py                                                                 **
 **  Copyright (c) December 2015 and September, 2018 by Leidos                                 **
 **  Produced under U.S. Government contract W5J9CQ-12-D-0004  Task Order: 053 and 073         **
 **  As per DFARS, the U.S. Government is granted unlimited rights to use this work            **
 **  in any manner and for any purpose.                                                        **
 **                                                                                            **
 **  Summary:   This file is the primary processing entry point for translations               **
 **                                                                                            **
 **  Developed by Leidos, Inc., High Resolution ISR Systems Division, Geospatial Programs      **
 **                                                                                            **
 **  Development History:                                                                      **
 **  December 2015 / Annette Janett Filer                                                      **
 **  9/27/2018 - Updates made to support new capabilities                                      **
 **                                                                                            **
 **                                                                                            **
 ************************************************************************************************
"""

import sys
import os
import copy

import arcpy
import uuid
import glob
import traceback

import gc

from TranslateConfig import *
from FieldValueMappings import *
from PrimaryTranslationFunctions import *
from FieldTransformerFunctions import *
from ExcelMappingFileAPI import *
from ExcelResultsOut import APIExcelResultsOutput


def report_xlate_issue(settings, errstmt, errortype="INFO", detail=None, exc_info=None):
    if errstmt is None or errstmt == "" or errstmt == " ":
        return
    if detail is None:
        outdetail = ""
    else:
        outdetail = "\n\t\t{0}".format(detail)
    if errortype.lower() in ['error', 'warning']:
        oneinst = '\n\t\tAdditional issues of this type are summarized with a counter.'
    else:
        oneinst = ""
    if exc_info is None:
        outexc = ""
    else:
        outexc = "\n\t\t{0}".format(traceback.format_exc(5))

    if errortype.lower() in ["error", "err"]:
        outstmt = "ERROR: {0}".format(errstmt)
        settings.my_logger.error("\t\t{0}{1}{2}{3}".format(errstmt, outdetail, oneinst, outexc))

    elif errortype.lower() in ["warning", "warn"]:
        outstmt = "WARNING: {0}".format(errstmt)
        settings.my_logger.warn("\t\t{0}{1}{2}{3}".format(errstmt, outdetail, oneinst, outexc))

    elif errortype.lower() == "debug":
        # outstmt = "DEBUG: {0}".format(errstmt)
        settings.my_logger.debug("\t\t{0}{1}".format(errstmt, outdetail))

    else:  # errortype.lower() == "info":
        # outstmt = "INFO: {0}".format(errstmt)
        settings.my_logger.info("\t\t{0}{1}".format(errstmt, outdetail))


def dump_garbage(settings, whereat=""):
    """
    clear out the garbage and report if necessary
    """

    # force collection
    unreachablenum = gc.collect()

    if settings.debug_settings['instrumentation_output'] > 0:
        settings.my_logger.info("GARBAGE OBJECTS: {0}  {1}".format(unreachablenum, whereat))
        for x in gc.garbage:
            s = str(x)
            if len(s) > 80: s = s[:80]
            for rr in gc.get_referrers(x):
                if "arcpy.arcobjects" in str(rr):
                    settings.my_logger.info("Ref: arcpy.arcobjects Type: {0} Name: {1}".format(type(s), s))
                else:
                    settings.my_logger.info("Ref: {2} Type: {0} Name: {1}".format(type(s), s, rr))


def DoTranslationNow(settings):
    if settings.debug_settings['instrumentation'] > 0:
        gc.enable()
        gc.set_debug(gc.DEBUG_STATS)  # gc.DEBUG_LEAK is ALL flags |gc.DEBUG_UNCOLLECTABLE|gc.DEBUG_COLLECTABLE
    settings.current_step = "Initialization"
    settings.sourcedb.gui_updated_info_update_internal_vars()
    settings.targetdb.gui_updated_info_update_internal_vars()
    source_db_version = settings.sourcedb_name_and_version.get()
    target_db_version = settings.targetdb_name_and_version.get()
    my_logger = settings.get_logger()

    if source_db_version.lower() == "unknown":
        settings.sourcedb.geodb_type_unknown = 1

    settings.transform_fc_count = 0
    # pull in the schema definitions for the geodatabases
    settings.sourcedb.load_schema_information()
    settings.targetdb.load_schema_information()

    if settings.sourcedb.geodb_type_unknown == 1 and settings.file_mapping_specification.get() != "":
        my_logger.info(
            "Setting up with source mappings from file {0} and target schema {1}.".\
                format(settings.file_mapping_specification.get(), target_db_version))
        settings.config_setting_flag['flag_set_unique_identifiers_fresh'] = 1
        # This will trigger us to NOT search for UFI/URI and just set uuids
        settings.mapping_specification = APIExcelMappingFile(
            settings.file_mapping_specification.get(), settings.my_logger)
        import Any_2_GGDM30_TranslateFeaturesandFields as fandf
        settings.mapping_specification.prepare_feature_and_field_translation_rules(fandf)

        import Any_2_GGDM30_FeatureClassesToTranslate as transrules
        setattr(transrules, 'fcs', settings.mapping_specification.get_feature_table_translate_rules())

    #   settings.mapping_specification.prepare_fcs_and_trans_rules()
    #   fandf = settings.mapping_specification.translation_rules_for_features_and_fields
    # print "fandf"
    # print fandf.hydro_to_hydroSrf
    #   transrules = settings.mapping_specification.feature_mappings_object
    # print "transrules"
    # print transrules.fcs
    # print "-------------------"

    #################### SET UP TO RUN ###########################                         
    else:
        my_logger.info(
            "Setting up with source schema {0} and target schema {1} specified.".format(
                source_db_version, target_db_version))
        if (source_db_version == "TDS 6.1" and target_db_version == "GGDM 3.0"):
            import TDS61_2_GGDM30_TranslateFeaturesandFields as fandf
        elif (source_db_version == "TDS 7.0" and target_db_version == "GGDM 3.0"):
            import TDS70_2_GGDM30_TranslateFeaturesandFields as fandf
        elif (source_db_version == "TDS 7.1" and target_db_version == "GGDM 3.0"):
            import TDS71_2_GGDM30_TranslateFeaturesandFields as fandf
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 6.1"):
            import GGDM30_2_TDS61_TranslateFeaturesandFields as fandf
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 7.0"):
            import GGDM30_2_TDS70_TranslateFeaturesandFields as fandf
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 7.1"):
            import GGDM30_2_TDS71_TranslateFeaturesandFields as fandf
        elif (source_db_version == "GGDM 3.0" and target_db_version == "GGDM 2.1"):
            import GGDM30_2_GGDM21_TranslateFeaturesandFields as fandf
        elif (source_db_version == "GGDM 3.0" and target_db_version == "GGDM 2.2"):
            import GGDM30_2_GGDM22_TranslateFeaturesandFields as fandf
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 6.0"):
            import GGDM30_2_GGDM22_TranslateFeaturesandFields as fandf  # Same as GGDM 3.0 -> GGDM 2.2
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 4.0"):
            import GGDM30_2_TDS4_TranslateFeaturesandFields as fandf
        elif (source_db_version == "GGDM 2.2" and target_db_version == "GGDM 3.0"):
            import GGDM22_2_GGDM30_TranslateFeaturesandFields as fandf
        elif (source_db_version == "TDS 6.0" and target_db_version == "GGDM 3.0"):
            import GGDM22_2_GGDM30_TranslateFeaturesandFields as fandf  # TDS 6 is a subset of GGDM 2.2
        elif (source_db_version == "GGDM 2.1" and target_db_version == "GGDM 3.0"):
            import GGDM21_2_GGDM30_TranslateFeaturesandFields as fandf
        elif (source_db_version == "TDS 4.0" and target_db_version == "GGDM 3.0"):
            import TDS4_2_GGDM30_TranslateFeaturesandFields as fandf
        else:
            # Error, source and destination specified are not supported
            my_logger.error("Invalid source schema {0} and target schema {1} specified.".format(
                source_db_version, target_db_version))
            return

        # Now get list of feature classes to translate (edit the file referenced if you only want a partial translation)
        if (source_db_version == "TDS 6.1" and target_db_version == "GGDM 3.0"):
            import TDS61_2_GGDM30_FeatureClassesToTranslate as transrules
        elif (source_db_version == "TDS 7.0" and target_db_version == "GGDM 3.0"):
            import TDS70_2_GGDM30_FeatureClassesToTranslate as transrules
        elif (source_db_version == "TDS 7.1" and target_db_version == "GGDM 3.0"):
            import TDS71_2_GGDM30_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 6.1"):
            import GGDM30_2_TDS61_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 7.0"):
            import GGDM30_2_TDS61_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 7.1"):
            import GGDM30_2_TDS61_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 3.0" and target_db_version == "GGDM 2.2"):
            import GGDM30_2_GGDM22_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 6.0"):
            import GGDM30_2_GGDM22_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 3.0" and target_db_version == "GGDM 2.1"):
            import GGDM30_2_GGDM21_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 3.0" and target_db_version == "TDS 4.0"):
            import GGDM30_2_TDS4_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 2.1" and target_db_version == "GGDM 3.0"):
            import GGDM21_2_GGDM30_FeatureClassesToTranslate as transrules
        elif (source_db_version == "GGDM 2.2" and target_db_version == "GGDM 3.0"):
            import GGDM22_2_GGDM30_FeatureClassesToTranslate as transrules
        # elif (source_db_version == "GGDM 2.2 WRDB" and target_db_version == "GGDM 3.0"):
        #    import GGDM22WRDB_2_GGDM30_FeatureClassesToTranslate as transrules
        # elif (source_db_version == "TDS 6.0" and target_db_version == "GGDM 3.0"):
        ##    import GGDM22_2_GGDM30_FeatureClassesToTranslate as transrules
        elif (source_db_version == "TDS 4.0" and target_db_version == "GGDM 3.0"):
            import TDS4_2_GGDM30_FeatureClassesToTranslate as transrules

    # my_logger.addHandler(handler)

    ######################## START THE TRANSFORMATION PROCESS ################################

    # From the import above, set these critical settings values for later operations.

    settings.fandf = fandf
    report_global_settings(settings)

    settings.current_step = "Step 0 Preconditions"
    my_logger.info("Starting data translation \n\t\tfrom {0}\n\t\tto {1}\n".format(
        settings.sourcedb.dbname.get(), settings.targetdb.dbname.get()))

    if source_db_version != transrules.current_include_source:
        my_logger.warning(
            'Specified source DB version {nma} is different than included source translation rules {nmb}'.format(
                nma=source_db_version, nmb=transrules.current_include_source))

    if target_db_version != transrules.current_include_target:
        my_logger.warning(
            'Specified target DB version {nma} is different than included target translation rules {nmb}'.format(
                nma=target_db_version, nmb=transrules.current_include_target))

    #### STEP 0 ######### Verify unique IDs ##################
    my_logger.info(
        "Step 0 Preconditions - Get schema information and verify the UFI/URI field values are unique in source database")

    my_logger.info("Getting source schema information: {0}".format(settings.sourcedb.dbname.get()))
    settings.sourcedb.fromGEODBGetDatasetsTablesandFC()
    #   verify_gdb_dataset_name_valid(settings, settings.sourcedb)
    my_logger.info("Getting target schema information: {0}".format(settings.targetdb.dbname.get()))
    settings.targetdb.fromGEODBGetDatasetsTablesandFC()
    #   verify_gdb_dataset_name_valid(settings, settings.targetdb)

    src_root = settings.sourcedb.get_default_root_for_fc()
    target_root = settings.targetdb.get_default_root_for_fc()

    my_logger.info(
        "Found database root for source: {0} and target: {1}".format(src_root, target_root))

    sourcedb = settings.sourcedb
    targetdb = settings.targetdb

    my_logger.info("")

    # Prepare the excel report output
    settings.results_xl = APIExcelResultsOutput(settings)
    clear_transformer_functions_report(settings)
    init_transformer_report_column_headings(settings)
    settings.results_xl.save_settings_to_xl_report()

    if settings.config_setting_flag['report_all_duplicate_ufi'] > 0 or \
            settings.config_setting_flag['flag_ufi_is_not_unique_generate_new_unique_ufi'] > 0:
        sourcedb.check_ufi = True
        sourcedb.check_oth = False
        sourcedb.check_NullValues = False
        sourcedb.check_illegal_values = False
        sourcedb.check_schema = False
        dresult = sourcedb.perform_gdb_checks()
        output_dictionary_to_logger(my_logger, "Source GDB Check", dresult)

    if settings.config_setting_flag['flag_expect_no_ufi_conflict'] == 0:
        targetdb.check_ufi = True
        targetdb.check_oth = False
        targetdb.check_NullValues = False
        targetdb.check_illegal_values = False
        targetdb.check_schema = False
        dresult2 = targetdb.perform_gdb_checks()
        output_dictionary_to_logger(my_logger, "Target GDB Check (required for UFI conflict check)", dresult2)

    settings.current_step = "Step 1 Restriction Information"
    if settings.config_setting_flag['flag_skip_all_metadata_translation'] == 1:
        my_logger.info("Step 1 Restriction Information metadata processing is being skipped as per specified flag.\n")
    else:
        try:
            #### STEP 1 ######### Deal with Restriction Info #####################
            if (target_db_version == "GGDM 3.0"):

                # for GGDM, start with restriction information. Read any restriction
                # info fields currently in GGDM (target) and save them locally.
                # Then process the set of feature classes and tables (including dataset in ResourceSrf)
                # looking only at the restriction info fields.
                # For each set of restriction info fields, compare them against the existing target restriction
                # info field vectors. If we find
                # a new restriction info vector, add it to the target and retain a copy locally for later steps.
                my_logger.info(
                    "Step 1 Translate and remember Restriction Information for use with " \
                        "remaining features. Add default record if needed.")
                get_existing_meta_restriction_info(settings, settings.targetdb,
                                                   "RESTRICTION_INFO_T",
                                                   fandf.RestrictionInfoTargetDefaultFieldValues)
                # use target_root_for_tables if this is going to a TABLE, otherwise use target_root
                process_metadata_records_spanning_feature_data(settings, "RESTRICTION_INFO_T",
                                                               transrules.fcs_restriction_info_metadata,
                                                               # List of feature classes, and translation rules for each
                                                               fandf.RestrictionInfoTblFields,
                                                               fandf.RestrictionInfoTargetDefaultFieldValues, fandf,
                                                               settings.saved_restrictioninfo, "Restriction", True,
                                                               "Restriction Info")

                #  process_meta_restriction_info(settings,  "RESTRICTION_INFO_T",  # use target_root_for_
                #  tables if this is going to a TABLE, otherwise use target_root
                #                      transrules.fcs_restriction_info_metadata,   # List of feature classes,
                #                      and translation rules for each
                #                      fandf.RestrictionInfoTblFields,
                #                      fandf.RestrictionInfoTargetDefaultFieldValues, fandf,         # this will
                #                      provide the subset field list as a parameter
                #                      True)                           # true here means we must have a default
                #                      record added if there is no record
                my_logger.info("Saved restriction information records available: {0}".format(
                    str(len(settings.saved_restrictioninfo))))
                show_output_saved_meta("Saved restriction information", my_logger.info, settings.saved_restrictioninfo)

                my_logger.info("Step 1 Complete\n")

            elif (source_db_version == "GGDM 3.0"):
                # for GGDM, start with restriction information. Read any restriction info fields
                # currently in GGDM (target) and save them locally.

                my_logger.info("Step 1 Read Restriction Information for use with remaining features. ")
                get_existing_meta_restriction_info(settings, settings.sourcedb,
                                                   "RESTRICTION_INFO_T",
                                                   fandf.RestrictionInfoTargetDefaultFieldValues)
                # use source if this is going to a TABLE, otherwise use target_root
                get_existing_meta_restriction_info(settings, settings.targetdb,
                                                   "RESTRICTION_INFO_T",
                                                   fandf.RestrictionInfoTargetDefaultFieldValues)
                # use target_root_for_tables if this is going to a TABLE, otherwise use target_root
                my_logger.info("Saved restriction information records available: {0}".format(
                    str(len(settings.saved_restrictioninfo))))
                show_output_saved_meta("Saved restriction information", my_logger.info, settings.saved_restrictioninfo)

                my_logger.info("Step 1 Complete\n")

            else:
                my_logger.error("Step 1 - Failure of translation process from {0} to {1}\n".format(
                    source_db_version, target_db_version))
        except:
            report_xlate_issue(settings, "Metadata Step 1 Failure", "ERROR", None, sys.exc_info())
    # End if skip metadata step 

    settings.current_step = "Step 2 Dataset and Entity Collection Metadata"
    if settings.config_setting_flag['flag_skip_all_metadata_translation'] == 1:
        my_logger.info("Step 2 Dataset/Entity Collection metadata transfer is being skipped as per specified flag.\n")
    else:
        try:
            ##### STEP 2 ############# Process fields that are held in each feature class,
            # but are to be saved in metadata table fields ################
            if ((source_db_version in VERSION_IN_GROUP['TDS61+'] or \
                 source_db_version == "GGDM 2.2") and target_db_version == "GGDM 3.0"):
                # Next copy any entity collection records and Dataset records from the source to the target.
                # Prerequisite: Restriction info processing must preceed this step because Dataset records look this up.
                my_logger.info(
                    "Step 2 Translate Entity Collection Metadata (MetadataSrf) " \
                    "and Dataset (ResourceSrf) records from source to target.")
                transform_and_move_data(settings, transrules.fcs_process_entity_collection_and_dataset,
                                        sourcedb, targetdb, fandf, "ECandDataset", "Step 2 Metadata", True)
                del transrules.fcs_process_entity_collection_and_dataset[:]
                my_logger.info("Step 2 Complete\n")

            elif (source_db_version == "GGDM 3.0" and \
                  (target_db_version in VERSION_IN_GROUP['TDS61+'])):
                my_logger.info("Step 2 No action required for this translation step.")
                my_logger.info("Step 2 Complete\n")

            elif ((source_db_version == "GGDM 2.1" or source_db_version == "TDS 4.0") and \
                  target_db_version == "GGDM 3.0"):
                my_logger.info(
                    "Step 2 Translate CollectionSrf and Dataset (DatasetSrf) records from source to target.")
                transform_and_move_data(settings, transrules.fcs_process_dataset,
                                        sourcedb, targetdb, fandf, "CollectionSrfandDataset", "Step 2 Metadata", True)
                del transrules.fcs_process_dataset[:]
                my_logger.info("Step 2 Complete\n")

            elif (source_db_version == "GGDM 3.0" and \
                  (target_db_version == "TDS 4.0" or target_db_version == "GGDM 2.1" or \
                   target_db_version == "GGDM 2.2" or target_db_version == "TDS 6.0")):
                my_logger.info("Step 2 Translate Dataset DatasetSrf/DATASET_T records from source to target.")
                # This will just pull data from the metadata objects themselves, but we also have to traverse all features
                transform_and_move_data(settings, transrules.fcs_process_dataset_FC,
                                        sourcedb, targetdb, fandf, "DatasetSrf", "Step 2 Metadata Dataset Surface")
                # Move dataset records from FC
                transform_and_move_data(settings, transrules.fcs_process_dataset_Table,
                                        sourcedb, targetdb, fandf, "DatasetTbl", "Step 2 Metadata Dataset Table")
                # Move dataset records from Table
                del transrules.fcs_process_dataset_FC[:]
                del transrules.fcs_process_dataset_Table[:]
                my_logger.info("Step 2 Complete\n")

            else:
                my_logger.error("Step 2 - Failure of translation process from {0} to {1}\n". \
                                format(source_db_version, target_db_version))
        except:
            report_xlate_issue(settings, "Metadata Step 2 Failure", "ERROR", None, sys.exc_info())
    # End If skip metdata step

    settings.current_step = "Step 3 Feature EC Metadata"
    if settings.config_setting_flag['flag_skip_all_metadata_translation'] == 1:
        my_logger.info("Step 3 Entity Collection metadata processing is being skipped as per specified flag.\n")
    else:
        try:

            #### STEP 3 ############## Deal with Entity Collection Metadata ################################
            if ((source_db_version in VERSION_IN_GROUP['TDS61+'] or\
                 source_db_version == "GGDM 2.2") and \
                    target_db_version == "GGDM 3.0"):

                my_logger.info(
                    "Read Entity Collection Metadata Table and ENTITY_COLLECTION_METADATA_T, and save to memory.")
                get_existing_entity_collection_meta(settings, settings.targetdb,
                                                    "ENTITY_COLLECTION_METADATA_T",
                                                    fandf.EntityCollectionTargetDefaultFieldValues)
                get_existing_entity_collection_meta(settings, settings.targetdb, "MetadataSrf",
                                                    fandf.EntityCollectionTargetDefaultFieldValues)

                # Next, we either deal with Entity Collection Metadata, saving to the GGDM Table
                # or to the GGDM Surface feature
                if sourcedb.featureclass_exists_in_geodb("MetadataSrf"):
                    my_logger.info(
                        "Step 3 Translate and remember Entity Collection metadata " \
                                "(from MetadataSrf) for use with remaining features.")
                    process_metadata_records_spanning_feature_data(settings,
                                                                   "MetadataSrf",
                                                                   transrules.fcs_entity_collection_metadata,
                                                                   # List of feature classes, and translation rules for each
                                                                   fandf.EntityCollectionFields,
                                                                   fandf.EntityCollectionTargetDefaultFieldValues,
                                                                   fandf,
                                                                   settings.saved_entity_collection_metadata,
                                                                   "ECMetadataSrf",
                                                                   False, "Entity Collection")

                    #  process_entity_collection_meta(settings,
                    #                                  sourcedb, targetdb, target_root, "MetadataSrf",
                    # use target_root_for_tables if this is going to a TABLE, otherwise use target_root
                    #                      transrules.fcs_entity_collection_metadata,
                    # this will provide the subset field list as a parameter
                    #                      fandf, "ECMetadataSrf",
                    #                      False)
                    # true here means we must have a default record added if there is no record
                    my_logger.info("Saved entity collection records available: {0}". \
                                   format(str(len(settings.saved_entity_collection_metadata))))
                    my_logger.info("Step 3 MetadataSrf Complete\n")
                if sourcedb.featureclass_exists_in_geodb("ENTITY_COLLECTION_METADATA_T"):
                    my_logger.info(
                        "Step 3 Translate and remember Entity Collection metadata " \
                                "(from Table) for use with remaining features.")
                    process_metadata_records_spanning_feature_data(settings,
                                                                   "ENTITY_COLLECTION_METADATA_T",
                                                                   transrules.fcs_entity_collection_metadata,
                                                                   # List of feature classes, and translation rules for each
                                                                   fandf.EntityCollectionFields,
                                                                   fandf.EntityCollectionTargetDefaultFieldValues,
                                                                   fandf,
                                                                   settings.saved_entity_collection_metadata,
                                                                   "ECMetadataSrf",
                                                                   False, "Entity Collection")

                    #      process_entity_collection_meta(settings,
                    #                                      sourcedb, targetdb,
                    #                                      settings.targetdb.get_root_for_table(),
                    #                                      "ENTITY_COLLECTION_METADATA_T",
                    # use target_root_for_tables if this is going to a TABLE, otherwise use target_root
                    #                          transrules.fcs_entity_collection_metadata,
                    # this will provide the subset field list as a parameter
                    #                          fandf, "ECMetadataTbl",
                    #                          False)
                    # true here means we must have a default record added if there is no record
                    my_logger.info("Saved entity collection records available: {0}".format(
                        str(len(settings.saved_entity_collection_metadata))))
                    my_logger.info("Step 3 ENTITY_COLLECTION_METADATA_T Complete\n")
                del transrules.fcs_entity_collection_metadata[:]

                show_output_saved_meta("Saved entity collection metadata",
                                       my_logger.info, settings.saved_entity_collection_metadata)

            elif ((source_db_version == "TDS 6.0") and target_db_version == "GGDM 3.0"):
                pass
                my_logger.info("Step 3 Complete\n")

            elif (source_db_version == "GGDM 3.0" and \
                  (target_db_version in VERSION_IN_GROUP['TDS61+'] or \
                   target_db_version == "GGDM 2.2")):
                get_existing_entity_collection_meta(settings, settings.sourcedb,
                                                    "ENTITY_COLLECTION_METADATA_T",
                                                    fandf.EntityCollectionTargetDefaultFieldValues)
                get_existing_entity_collection_meta(settings, settings.sourcedb,
                                                    "MetadataSrf", fandf.EntityCollectionTargetDefaultFieldValues)
                my_logger.info("Step 3 Read Entity Collection Metadata Table and save to memory.")
                show_output_saved_meta("Saved entity collection metadata",
                                       my_logger.info, settings.saved_entity_collection_metadata)
                my_logger.info("Step 3 Complete\n")

            elif ((source_db_version == "GGDM 2.1" or source_db_version == "TDS 4.0") and \
                  target_db_version == "GGDM 3.0"):
                my_logger.info(
                    "Read Entity Collection Metadata Table and ENTITY_COLLECTION_METADATA_T, and save to memory.")
                get_existing_entity_collection_meta(settings, settings.targetdb,
                                                    "ENTITY_COLLECTION_METADATA_T",
                                                    fandf.EntityCollectionTargetDefaultFieldValues)
                get_existing_entity_collection_meta(settings, settings.targetdb, "MetadataSrf", \
                                                    fandf.EntityCollectionTargetDefaultFieldValues)
                # This should have earlier been translated into GGDM 3.0 so we should get it as the target
                # get_existing_entity_collection_meta(settings, settings.sourcedb,
                # "CollectionSrf", fandf.EntityCollectionTargetDefaultFieldValues)

                process_entity_collection_meta_from_21(settings,
                                                       sourcedb, targetdb, "MetadataSrf",
                                                       "ENTITY_COLLECTION_METADATA_T",
                                                       # use target_root_for_tables if this is going to a TABLE,
                                                       # otherwise use target_root
                                                       fandf.EntityCollectionFields,  # field set
                                                       fandf, "ECMetadataSrftoTbl",
                                                       False)
                # true here means we must have a default record added if there is no record
                my_logger.info(
                    "Saved entity collection records available: {0}".format(
                        str(len(settings.saved_entity_collection_metadata))))
                show_output_saved_meta(
                    "Saved entity collection metadata", my_logger.info,
                    settings.saved_entity_collection_metadata)
                my_logger.info("Step 3 Complete\n")

            elif (source_db_version == "GGDM 3.0" and \
                  (target_db_version == "TDS 4.0" or target_db_version == "GGDM 2.1")):
                my_logger.info(
                    "Step 3 Read Entity Collection Metadata Table, process all features for " \
                        "entity collection values, and save to memory.")
                get_existing_entity_collection_meta(settings, settings.sourcedb,
                                                    "ENTITY_COLLECTION_METADATA_T",
                                                    fandf.EntityCollectionTargetDefaultFieldValues)
                get_existing_entity_collection_meta(settings, settings.sourcedb,
                                                    "MetadataSrf", fandf.EntityCollectionTargetDefaultFieldValues)
                if settings.debug_settings['only_transform_no_move'] == 0:
                    settings.debug_settings['only_transform_no_move'] = 1
                    transform_and_move_data(settings, transrules.fcs_process_for_EC,
                                            sourcedb, targetdb, fandf, "EntityColProc",
                                            "Step 3 Entity Collection")
                    # Exit before saving, we mainly want to run the transform and do not want to move data
                    settings.debug_settings['only_transform_no_move'] = 0
                else:
                    transform_and_move_data(settings, transrules.fcs_process_for_EC,
                                            sourcedb, targetdb, fandf, "EntityCol", "Step 3 Entity Collection")
                    # Exit before saving, we mainly want to run the transform and do not want to move data
                #
                del transrules.fcs_process_for_EC[:]
                process_entity_collection_meta_ggdm30_to_21(
                    settings, sourcedb, targetdb, "CollectionSrf",
                    # use target_root_for_tables if this is going to a TABLE, otherwise use target_root
                    None,
                    # This would normally be a feature class and fields specification. This function is hard coded!
                    fandf, "ECMetadataSrf",
                    False)
                show_output_saved_meta("Saved entity collection metadata",
                                       my_logger.info, settings.saved_entity_collection_metadata)
                my_logger.info("Step 3 Complete\n")

            else:
                my_logger.error("Step 3 - Failure in translation process from {0} to {1}\n".format(
                    source_db_version, target_db_version))
        except:
            report_xlate_issue(settings, "Metadata Step 3 Failure", "ERROR", None, sys.exc_info())

    # End if skip metadata step
    if settings.debug_settings['instrumentation'] > 0:
        dump_garbage(settings, "Begin step 4")
    settings.current_step = "Step 4 Feature Classes and Tables"
    ##### STEP 4 ############# Translate feature classes ###############################################
    my_logger.info(
        "Step 4 Translate feature classes and fields given specified list of " \
            "feature classes, fields and transform functions.")
    try:
        #  settings.my_logger.info("NEW OUTPUT")
        #  del settings.saved_restrictioninfo[:]
        #  get_existing_meta_restriction_info(settings, settings.targetdb,
        #                                              "RESTRICTION_INFO_T",
        #                                              fandf.RestrictionInfoTargetDefaultFieldValues)
        # use target_root_for_tables if this is going to a TABLE, otherwise use target_root
        #  del settings.saved_entity_collection_metadata[:]
        #  get_existing_entity_collection_meta(settings, settings.sourcedb,
        #                                              "ENTITY_COLLECTION_METADATA_T",
        #                                              fandf.EntityCollectionTargetDefaultFieldValues)
        #  get_existing_entity_collection_meta(settings, settings.sourcedb,
        #                                              "MetadataSrf", fandf.EntityCollectionTargetDefaultFieldValues)
        transform_and_move_data(settings, transrules.fcs, sourcedb, targetdb, fandf,
                                "Features", "Step 4 Feature Classes and Tables", True)
    except:
        report_xlate_issue(settings,
                           "Translation Failure in transform and move data", "ERROR", None, sys.exc_info())
    my_logger.info("Step 4 Complete\n")

    settings.current_step = "Step 5 Derived Records"
    ##### STEP 5 ############# Save any remaining derived content ###############################################
    my_logger.info("Step 5 Save derived records.")
    try:
        save_final_derived_records(settings, targetdb,
                                   "Derived", "Step 5 Saved Derived Records", True)
    except:
        report_xlate_issue(settings,
                           "Translation Failure in saving derived records", "ERROR", None, sys.exc_info())
    my_logger.info("Step 5 Complete\n")

    my_logger.info("Log file and Excel results available for review: ")
    my_logger.info("\tLog file: {0}\tExcel results: {1}".format(
        settings.name_log_file.get(), settings.file_results_excel.get()))
    #  logfiles = glob.glob('%s*' % settings.name_log_file.get())
    #  for fname in logfiles:
    #      my_logger.info ( fname )

    my_logger.info("All steps complete.\n")
    my_logger.info("***********************")

#### For if we start from this file first
# root = Tk()
# This is required first, we are using some Tk calls in the configuration
# settings area and they will fail unless we start Tk
# root.title('GGDM/TDS Translation')
# settings = MyConfigSettings()
# __DoTranslationNow(settings)
