﻿
""" 
 ************************************************************************************************
 **                                                                                            **
 **  File: GUISet_Special_Flags.py                                                              **
 **  Copyright (c)  September, 2018 by Leidos                                                  **
 **  Produced under U.S. Government contract W5J9CQ-12-D-0004  Task Order: 074                 **
 **  As per DFARS, the U.S. Government is granted unlimited rights to use this work            **
 **  in any manner and for any purpose.                                                        **
 **                                                                                            **
 **  Summary:   This file provides a GUI capability for setting flags, and special options     **
 **                                                                                            **
 **  Developed by Leidos, Inc., High Resolution ISR Systems Division, Geospatial Programs      **
 **                                                                                            **
 **  Development History:                                                                      **
 **  9/27/2018 -  Annette Janett Filer                                                         **
 **                                                                                            **
 **                                                                                            **
 ************************************************************************************************
"""

import Queue
from Tkinter import *
from ScrolledText import *
import tkFileDialog
import ttk
# from ttk import *
import ConfigParser
from TranslateConfig import MyConfigSettings
import sys
import os
import copy

import arcpy
import uuid
import glob


class Tooltip:
    '''
    It creates a tooltip for a given widget as the mouse goes on it.

    see:

    https://stackoverflow.com/questions/3221956/
           what-is-the-simplest-way-to-make-tooltips-
           in-tkinter/36221216#36221216

    http://www.daniweb.com/programming/software-development/
           code/484591/a-tooltip-class-for-tkinter

    - Originally written by vegaseat on 2014.09.09.

    - Modified to include a delay time by Victor Zaccardo on 2016.03.25.

    - Modified
        - to correct extreme right and extreme bottom behavior,
        - to stay inside the screen whenever the tooltip might go out on 
          the top but still the screen is higher than the tooltip,
        - to use the more flexible mouse positioning,
        - to add customizable background color, padding, waittime and
          wraplength on creation
      by Alberto Vassena on 2016.11.05.

      Tested on Ubuntu 16.04/16.10, running Python 3.5.2

    TODO: themes styles support
    '''

    def __init__(self, widget, \
                 bg='#FFFFEA', \
                 pad=(5, 3, 5, 3), \
                 text='widget info', \
                 waittime=400, \
                 wraplength=250):
        self.waittime = waittime  # in miliseconds, originally 500
        self.wraplength = wraplength  # in pixels, originally 180
        self.widget = widget
        self.text = text
        self.widget.bind("<Enter>", self.onEnter)
        self.widget.bind("<Leave>", self.onLeave)
        self.widget.bind("<ButtonPress>", self.onLeave)
        self.bg = bg
        self.pad = pad
        self.id = None
        self.tw = None

    def onEnter(self, event=None):
        self.schedule()

    def onLeave(self, event=None):
        self.unschedule()
        self.hide()

    def schedule(self):
        self.unschedule()
        self.id = self.widget.after(self.waittime, self.show)

    def unschedule(self):
        id_ = self.id
        self.id = None
        if id_:
            self.widget.after_cancel(id_)

    def show(self):
        def tip_pos_calculator(widget, label,
                               tip_delta=(10, 5), pad=(5, 3, 5, 3)):

            w = widget

            s_width, s_height = w.winfo_screenwidth(), w.winfo_screenheight()

            width, height = (pad[0] + label.winfo_reqwidth() + pad[2],
                             pad[1] + label.winfo_reqheight() + pad[3])

            mouse_x, mouse_y = w.winfo_pointerxy()

            x1, y1 = mouse_x + tip_delta[0], mouse_y + tip_delta[1]
            x2, y2 = x1 + width, y1 + height

            x_delta = x2 - s_width
            if x_delta < 0:
                x_delta = 0
            y_delta = y2 - s_height
            if y_delta < 0:
                y_delta = 0

            offscreen = (x_delta, y_delta) != (0, 0)

            if offscreen:

                if x_delta:
                    x1 = mouse_x - tip_delta[0] - width

                if y_delta:
                    y1 = mouse_y - tip_delta[1] - height

            offscreen_again = y1 < 0  # out on the top

            if offscreen_again:
                # No further checks will be done.

                # TIP:
                # A further mod might automagically augment the
                # wraplength when the tooltip is too high to be
                # kept inside the screen.
                y1 = 0

            return x1, y1

        bg = self.bg
        pad = self.pad
        widget = self.widget

        # creates a toplevel window
        self.tw = Toplevel(widget)

        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)

        win = Frame(self.tw,
                       background=bg,
                       borderwidth=0)
        label = ttk.Label(win,
                          text=self.text,
                          justify=LEFT,
                          background=bg,
                          relief=SOLID,
                          borderwidth=0,
                          wraplength=self.wraplength)

        label.grid(padx=(pad[0], pad[2]),
                   pady=(pad[1], pad[3]),
                   sticky=NSEW)
        win.grid()

        x, y = tip_pos_calculator(widget, label)

        self.tw.wm_geometry("+%d+%d" % (x, y))

    def hide(self):
        tw = self.tw
        if tw:
            tw.destroy()
        self.tw = None



class Application_Set_Special_Settings(Frame):
    """Settings and Flags GUI"""

    def __init__(self, master, settings):
        self.master = master
        self.settings = settings
        self.moveCADthreadcount = 0
        self.moveCADthread = None
        self.vvCADthreadcount = 0
        self.vvCADthread = None
        Frame.__init__(self, master)

        self.names_of_fc = StringVar()
        self.names_of_fld = StringVar()
        # self.settings = MyConfigSettings()  # Standalone ops only
        # self.my_logger = self.settings.get_logger()

        # Queue where the logging handler will write
        #self.logconsole = Queue.Queue()           

        self.threadcount = 0
        self.threaditem = None

        self.gobuttonstyle = ttk.Style()
        self.gobuttonstyle.configure('GoButton.TButton', background = 'maroon', foreground = 'black')


        self.getcurrentlistandputtostring(self.settings.debug_fc_list, self.names_of_fc)
        self.getcurrentlistandputtostring(self.settings.debug_field_list, self.names_of_fld)
         ##### CREATE WIDGETS ######################
        self.pack()
        self.createWidgets() 
 
        


    def close_window(self):
        del self.settings
        self.destroy()
        
    def save_grouped_item(self, generalsettingsdict, procsettingsdict):
        for keyx in generalsettingsdict:
            if keyx in procsettingsdict:
                txtkey = procsettingsdict[keyx]['TextVariable'].get()
                # Set the IntVariable based on the TextVariable, and also set the main dictionary integer
                procsettingsdict[keyx]['IntVariable'].set( procsettingsdict[keyx]['OptionCombo'][txtkey] )
                generalsettingsdict[keyx] = procsettingsdict[keyx]['OptionCombo'][txtkey]        
                         
    def save_updated_settings(self):
        # self.master.status_detail_replace("Save configuration")
        self.save_grouped_item( self.settings.debug_settings, self.settings.procdebugsettings)
        self.save_grouped_item( self.settings.maximum_records, self.settings.proclimitsconstraint)
        self.save_grouped_item( self.settings.config_setting_flag, self.settings.procflagsgeneral)
        self.save_grouped_item( self.settings.config_setting_flag, self.settings.procufiflags)
        #self.save_grouped_item( self.settings.config_setting_flag, self.settings.procoperationmode)
        self.save_grouped_item( self.settings.config_setting_flag, self.settings.proctestflags)
        fcs = self.names_of_fc.get().split()
        del self.settings.debug_fc_list[:]
        for fc in fcs:
            self.settings.debug_fc_list.append(fc)
        flds = self.names_of_fld.get().split()
        del self.settings.debug_field_list[:]
        for fld in flds:
            self.settings.debug_field_list.append(fld)
        self.settings.save_config_settings()

    def getcurrentlistandputtostring(self, listin, setvarout):
        tempvar = ""
        for ff in listin:
            if tempvar == "":
                tempvar = "{0}".format(ff)
            else:
                tempvar = "{0},{1}".format(tempvar,ff)
        setvarout.set(tempvar)

    def createsetsubwidg(self, startrow, gensettingsdict, procsettingsdict, guisettingdict, hostingframe):
        indii = 0
        for keyx in gensettingsdict:
            if keyx in procsettingsdict:
                #if procsettingsdict[keyx]['DefaultValue'] != procsettingsdict['TextValue']:
                guisettingdict[keyx] = dict()
                guisettingdict[keyx]['GUILabel'] = Label(hostingframe, text=procsettingsdict[keyx]['LabelName'], width=80)
                guisettingdict[keyx]['GUILabel'].grid(row=startrow + indii, column=2, columnspan=2, sticky = 'E')
                guisettingdict[keyx]['ComboBox'] = ttk.Combobox(hostingframe, textvariable = procsettingsdict[keyx]['TextVariable'], width = 30)
                guisettingdict[keyx]['ComboBox'].grid(row=startrow + indii, column = 10, sticky = 'E', padx = 2, pady = 2)
                guisettingdict[keyx]['ComboBox']['values'] = procsettingsdict[keyx]['ValuesList']
                if len(procsettingsdict[keyx]['ToolTip']) > 0:
                    guisettingdict[keyx]['Tooltip'] = Tooltip(guisettingdict[keyx]['ComboBox'], text=procsettingsdict[keyx]['ToolTip'], wraplength = 250)
                indii += 1
        return startrow + indii


    def createWidgets(self):

        self.master.title("{0} - Flags and Settings".format(self.settings.name_application_title.get()))
        self.QUIT = Button(self)
        self.QUIT["text"] = "Close"
        # self.QUIT["fg"]   = "red"
        self.QUIT["command"] =  self.close_window

        self.QUIT.grid(row=1,column=10)

        # TITLE TITLE TITLE
        self.TTITLE = Label(self, text=self.settings.app_title.get())
        self.TTITLE.grid(row=1, column = 0, sticky='NW', padx = 2, pady = 2)

        self.dblabel_blankaa = Label(self, text="  ")
        self.dblabel_blankaa.grid(row=2,column=1)

        self.frameall = LabelFrame(self, text="Settings, constraints, reporting limits, verbosity, limits")
        self.frameall.grid(row=3, column=0, columnspan=10, sticky='NW')

        self.tabctl = ttk.Notebook(self.frameall)


        ###############################

        self.tab1 = ttk.Frame(self.tabctl)
        self.tabctl.add(self.tab1, text="General")
        self.tabctl.grid(row=1, column=0, columnspan=10)
        self.guisettinggeneral = dict()
        self.pmframe = LabelFrame(self.tab1, text="General Operations")
        self.pmframe.grid(row=1,column=0, columnspan=10, sticky='NEW')
        finrow = self.createsetsubwidg(3, self.settings.config_setting_flag, self.settings.procflagsgeneral, self.guisettinggeneral, self.pmframe)


        self.tab2 = ttk.Frame(self.tabctl)
        self.tabctl.add(self.tab2, text="Unique IDs")
        self.guisettingufi = dict()
        self.pmframe = LabelFrame(self.tab2, text="Unique Identifiers")
        self.pmframe.grid(row=1,column=0, columnspan=10, sticky='NEW')
        finrow = self.createsetsubwidg(3, self.settings.config_setting_flag, self.settings.procufiflags, self.guisettingufi, self.pmframe)       

        #self.guidsettingmode = dict()
        #if self.allowoperationmodeset > 0:
        #    self.tab2aa = ttk.Frame(self.tabctl)
        #    self.tabctl.add(self.tab2aa, text="Operation Mode")
        #    self.guisettingufi = dict()
        #    self.dbsettings = LabelFrame(self.frameall, text="Operation Mode")
        #    self.dbsettings.grid(row=1,column=0, columnspan=10, sticky='NEW')
        #    finrow = self.createsetsubwidg(3, self.settings.debug_settings, self.settings.procoperationmode, self.guidsettingmode, self.dbsettings)


        self.tab3 = ttk.Frame(self.tabctl)
        self.tabctl.add(self.tab3, text="Limits/Constraints")
        self.guisettinglimit = dict()
        self.dbcontst = LabelFrame(self.tab3, text="Limits and Constraints")
        self.dbcontst.grid(row=1,column=0, columnspan=10, sticky='NEW')
        finrow = self.createsetsubwidg(3, self.settings.maximum_records, self.settings.proclimitsconstraint, self.guisettinglimit, self.dbcontst)

        self.tab4 = ttk.Frame(self.tabctl)
        self.tabctl.add(self.tab4, text="Reporting/Debug")
        self.guisettingdebug = dict()
        self.dbcontst = LabelFrame(self.tab4, text="Reports and Debug")
        self.dbcontst.grid(row=1,column=0, columnspan=10, sticky='NEW')
        finrow = self.createsetsubwidg(3, self.settings.debug_settings, self.settings.procdebugsettings, self.guisettinglimit, self.dbcontst)

        self.guisettingprocm = dict()
        if self.settings.testmode > 0:
            self.tab5 = ttk.Frame(self.tabctl)
            self.tabctl.add(self.tab5, text="Test Mode Flags")
            self.pmframe = LabelFrame(self.tab5, text="Test Mode Flags")
            self.pmframe.grid(row=1,column=0, columnspan=10, sticky='NEW')
            finrow = self.createsetsubwidg(3, self.settings.config_setting_flag, self.settings.proctestflags, self.guisettingprocm, self.pmframe)

        self.tab6 = ttk.Frame(self.tabctl)
        self.tabctl.add(self.tab6, text="Debug Feature Class/Fields")
        self.logdebugframe = LabelFrame(self.tab6, text="Features and Fields for logging and debugging")
        self.logdebugframe.grid(row=3, column=0, columnspan=10, sticky='NEW')

        self.dblabel_ll = Label(self.logdebugframe, text="Detailed logging of processing operations restricted to features and field")
        self.dblabel_ll.grid(row=1,column=2, columnspan=10 )

        self.specfclist = Label(self.logdebugframe, text="Specify comma delimited, quoted, feature class or table names here (for example: 'MetadataSrf')")
        self.specfclist.grid(row=2,column=1, columnspan=10 )   
        self.specfclistresult = Entry(self.logdebugframe, textvariable=self.names_of_fc, width = 130  )
        self.specfclistresult.grid(row=3, column=1, columnspan=10, sticky="W", padx = 2, pady = 2)

        self.specfldlist = Label(self.logdebugframe, text="Specify comma delimited, quoted, field names here (for example: 'ACC','TRS')")
        self.specfldlist.grid(row=4,column=1, columnspan=10 ) 
        self.specfldlistresult = Entry(self.logdebugframe, textvariable=self.names_of_fld, width = 130  )
        self.specfldlistresult.grid(row=5, column=1, columnspan=10,  sticky="W", padx = 2, pady = 2)


        self.endframe = LabelFrame(self.frameall, text="Save Settings")
        self.endframe.grid(row=25,column=0, columnspan=5, sticky='N')
        self.dblabel_blankaa = Label(self.endframe, text="  ")
        self.dblabel_blankaa.grid(row=1,column=1)
        self.svsettingbutton = Button(self.endframe)
        self.svsettingbutton["text"] = "Save Updated Settings"
        self.svsettingbutton["command"] = self.save_updated_settings
        self.svsettingbutton.grid(row=2,column=1)
        self.dblabel_blankgg = Label(self.endframe, text="  ")
        self.dblabel_blankgg.grid(row=4,column=1)



        self.dblabel_blankc = Label(self.frameall, text="  ")
        self.dblabel_blankc.grid(row=30,column=1)

 



    def update_widget(self, widget, queue):
        widget.config(state='normal')
        # Read from the Queue and add to the log widger
        while not queue.empty():
            line = queue.get()
            widget.insert('END', line)
            widget.see('END')  # Scroll to the bottom
            widget.update_idletasks()
        widget.config(state='disabled')
        self.logger_alarm = widget.after(10, self.update_widget, widget, queue)
        self.master.after(1)
        self.master.update_idletasks()


                       

        

#root = Tk()
#app = Application_Set_Special_Settings(master=root)
#app.mainloop()
#root.destroy()
