﻿
""" 
 ************************************************************************************************
 **                                                                                            **
 **  File: GeodatabaseSettings.py                                                              **
 **  Copyright (c) January 2017 by Leidos                                                      **
 **  Produced under U.S. Government contract W5J9CQ-12-D-0004  Task Order: 064 and 074         **
 **  As per DFARS, the U.S. Government is granted unlimited rights to use this work            **
 **  in any manner and for any purpose.                                                        **
 **                                                                                            **
 **  Summary:  Provide a python class that can operate on an ArcGIS geodatabase API-like       ** 
 **                                                                                            **
 **  Developed by Leidos, Inc., High Resolution ISR Systems Division, Geospatial Programs      **
 **                                                                                            **
 **  Development History:                                                                      **
 **  January 2017 - Annette Filer
 **  9/27/2017   Updates to support current translations
 **                                                                                            **
 **                                                                                            **
 ************************************************************************************************
"""

from Tkinter import *
import sys
import os
import copy
import uuid
import arcpy
import glob

from ExcelResultsOut import APIExcelResultsOutput
from StandardUtilities import *


class MyGDBSettings():
    def __init__(self, settings, fulldbnameandpath, schemanameandversion, dataset = 'Composite',
                 sourceortarget = 'Source', dbtype = 'FILEGDB', spatialtype = 'ST_GEOMETRY'):
        if settings.debug_settings['instrumentation'] > 0:
            dump_garbage(settings, "Begin gdb init {0}".format(schemanameandversion))

        # These are used for GUI input/output of name, schema, dataset. 
        self.dbname = StringVar() # For GUI
        self.schema_name_and_version = StringVar()   # For GUI
        self.dataset = StringVar()  # For GUI

        self.geodb_type_unknown = 0
        self.geodb_pathname = ""      # Pathname and version must be set to load schema
        self.db_version = ""
        self.db_type = "GEOPACKAGE"      # Options are "FILEGDB", "SQLITE, "GEOPACKAGE"
        self.db_spatial_type = "ST_GEOMETRY"   # Options are "ST_GEOMETRY" (Esri's compatiable) or
        # "SPATIALITE", "GEOPACKAGE"  - not used if db_type is FILEGDB
        self.geodb_dataset = ""       # If this isn't specified, it will be set based on the one dataset name
        self.mode = ""
        self.posstypes = ['FeatureDataset', 'LocalDatabase', 'Workspace']
        self.default_spatialref = \
            "GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]]," \
            "PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]],VERTCS['WGS_1984'," \
            "DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]]," \
            "PARAMETER['Vertical_Shift',0.0],PARAMETER['Direction',1.0]," \
            "UNIT['Meter',1.0]];-400 -400 1000000000;-100000 10000;-100000 10000;" \
            "8.98315284119522E-09;0.001;0.001;IsHighPrecision"
        self.current_fc = None
        self.former_fc = None
        self.current_f_code = None
        self.former_f_code = None
        self.current_fcsubtype = None
        self.updated_fields = list()
        self.row = dict()           # current row of field data
        self.output_row = dict()       # new row of field data for otuput
        self.saved_derived_records = {}   # Dictionary: Feature Class Name, containing dictionary of field: value
        self.db_schema_def_loaded = False

        self.dbname.set(fulldbnameandpath)
        self.schema_name_and_version.set(schemanameandversion)
        self.dataset.set(dataset)

        self.sourceortarget = sourceortarget

        self.geodb_pathname = fulldbnameandpath      # Pathname and version must be set to load schema
        self.db_version = schemanameandversion
        self.geodb_dataset = dataset       # If this isn't specified, it will be set based on the one dataset name

        self.db_type = dbtype      # Options are "FILEGDB", "SQLITE, "GEOPACKAGE""
        self.db_spatial_type = spatialtype
        # Options are "ST_GEOMETRY" (Esri's compatiable) or "SPATIALITE",
        # "GEOPACKAGE"  - not used if db_type is FILEGDB

        self.settings = settings
        self.my_logger = settings.my_logger

        self.field_information = dict()     # Schema definition from the perspective of feature classes or tables
        self.saved_db_field_lists = dict()
        self.geodb_datasets = list()
        self.geodb_tables = list()

        self.geodb_feature_classes = dict()  # Dictionary of all datasets holding lists of feature classes

        self.existing_unique_ids = list()
        self.saved_fc_datasets = dict()
        # Dictionary of all feature classes (and tables) specifying name of dataset holder (or None for table)

        self.domainList = None               # This is an Esri format list
        self.domainDict = None
        self.domainNameList = None

        self.check_ufi = True
        self.check_oth = False
        self.check_NullValues = True
        self.check_illegal_values = True
        self.check_schema = True
        self.fixshapenames = {'shape':'SHAPE@', 'shape_length':'SHAPE@Length', 'shape_area':'SHAPE@Area'}
        self.common_fields_geom_and_id_and_code = ['shape@', 'shape@length', 'shape@area',
                                                   'fcsubtype', 'objectid', 'globalid', 'f_code']
        self.common_fields_geom_only = ['shape@', 'shape@length', 'shape@area']
        self.typical_spcial_case_fields = ['fcsubtype', 'objectid', 'globalid', 'f_code']
        self.typical_identifier_fields = ['ufi', 'uri', 'objectid', 'globalid']

        self.current_cursor = None

    def save_derived_record_for_later_output(self, settings, featureclass, origufi, dictoffields, explanation):
        keyval = "{0}; {1}".format(origufi, explanation)
        if featureclass not in self.saved_derived_records:
            self.saved_derived_records[featureclass] = dict()
        if keyval not in self.saved_derived_records[featureclass]:
            self.saved_derived_records[featureclass][keyval] = dictoffields
        else:
            report_transfomer_issue_or_increment_count(
                        settings, "May have already saved this derived output or no unique id", "WARNING", 
                        "While processing feature class target {0}, UFI {1}".format(
                         featureclass, strout(origufi)))

    def save_derived_records_to_database(self, settings, reportid, stepname, generate_report):
        move_data_error = 0
        for fc in self.saved_derived_records:
            settings.report_values.clear()     # Start with a clean report set for this feature class
            settings.report_errors.clear()       # Start with a clean error log for this feature class
            settings.my_logger.info('{1}; Started appending derived records to table or feature class: {0}'.format(
                fc, stepname))
            currentrec = self.saved_derived_records[fc]
            # open db to this feature class, get fields and then match fields against
            # the saved record content, setting remaining to default
            targetpathname = self.get_root_for_fc_subtypes(fc)
            self.updated_fields = self.get_fields_from_featureclass(fc)
            with arcpy.da.InsertCursor(targetpathname, self.updated_fields) as new_rows:
                append_data_count = 0
                try:
                    # now prepare each row for insertion into this target db
                    for uniqitem in currentrec:
                        outrow = list()
                        for fld in self.updated_fields:
                            if fld in currentrec[uniqitem]:
                                outrow.append( currentrec[uniqitem][fld] )
                            elif fld.lower() in currentrec[uniqitem]:
                                outrow.append( currentrec[uniqitem][fld.lower()] )
                            elif fld.lower() in self.typical_identifier_fields:
                                outrow.append(  self.makerandomufi() )
                            else:
                                outrow.append( self.field_set_to_default(currentrec[uniqitem]['F_CODE'], fld, fc) )
                        #   # end for each field in the row - prepare the row
                        if settings.debug_settings['itemize_each_translation_each_field'] > 1:
                            settings.my_logger.debug("\tAppending row related to original data: {0}; row: {1}".format(
                                uniqitem, strout(outrow)))

                        new_rows.insertRow(outrow)
                        append_data_count += 1
                except:
                    #print (" ")
                    move_data_error +=1
                    report_transfomer_issue_or_increment_count(
                        settings, "Exception while appending to feature class.", "ERROR", 
                        "While processing feature class target {fc}, data processing index: {idx}".format(
                         fc=fc, idx=append_data_count),
                         sys.exc_info())
            #   #   # end with insert cursor, for each unique item                

            if append_data_count > 0:
                settings.report_values['Derived Features - Instances Appended'] = append_data_count
       
            self.output_transformer_report_detail( settings, fc, settings.report_values, "Information")  
            self.output_transformer_report_detail( settings, fc, settings.report_errors, "Issues")

            if generate_report:
                settings.results_xl.save_results_items_to_report(
                    reportid, "NA", fc, "NA", settings.report_values, append_data_count + 1)
                settings.results_xl.save_results_items_to_report(
                    reportid, "NA", fc, "NA", settings.report_errors, append_data_count + 1, True)
                settings.results_xl.done_save()
        # End for each fc

    def output_transformer_report_detail(self, settings, infeatureclass, reportdict, rptheading):
        totalvals = sum(reportdict.itervalues())   # in Python 3 this goes to sum(reportdict.values())
        if totalvals > 0:
            settings.my_logger.info("  Transformer {1} Report for {0}".format(infeatureclass, rptheading) )
            for keys,values in reportdict.items():
                if values > 0:
                    settings.my_logger.info("    {0} : {1}".format(keys,values))
            #my_logger.info("  Completed Issue Report for {0}".format(infeatureclass) )
        return

    def geodb_prepare_insert_cursor(self):
        self.current_cursor = arcpy.da.InsertCursor(self.get_root_for_fc_subtypes(self.get_fc()), self.updated_fields)
        self.output_rows.clear()

    def clean_name(self, inname):
        # Do not use spaces, hyphens, parentheses, brackets, and other characters such as $, %, and #.
        # Max length 52 characters
        # TODO Consider using regular expression: outname = re.sub('!@#$%^&*()~`d-=+[]{} ', '_', inname)
        return (inname.replace('-', '_').replace('~', '_').replace('.', '_'))[:52]

    def generate_feature_layer_profile_fc_and_copy_to_target(self, fc, allfields, outfields,
                                                             layername, targetdb, outputfc, whereclause):
        # fc, fields, outfields, "templayer")
        # Set up to copy, set the workspace to the source physical database
        if outputfc is None:
            outputfc = fc
        outputfc = self.clean_name(outputfc)
        arcpy.env.workspace = self.geodb_pathname
        fcfullpath = self.get_root_for_fc_subtypes(fc)
        sourcedataset = self.get_dataset_for_fc_in_geodatabase(fc)         # If this is a table, this will return None
        self.geodb_dataset = sourcedataset
        workspacebase = os.path.normpath(
            self.geodb_pathname + (("/" + sourcedataset) if sourcedataset is not None else ""))
        # layerfullpath = self.get_root_dataset_for_fc(fc) + "//" + layername
        arcpy.env.overwriteOutput = True

        #fldinfo = arcpy.FieldInfo
        fldinfo = ""
        if outfields != "all":
            for field in allfields:
                # fldinfo.addField(field, field, "VISIBLE" if field in wantedfields else "HIDDEN", "NONE")
                fldinfo = fldinfo + field + " " + field + (
                    " VISIBLE;" if field in outfields else " HIDDEN;")  # (pass fldinfo[:-1])

        #              input feature class,  output layer name, where clause,
        #              input workspace used to validate field names,
        arcpy.env.workspace = self.geodb_pathname

        # Depending on whether all fields are going to be copied, or if there is a where clause, we can adjust the
        # call to MakeFeatureLayer_management - and hopefully some of these will be faster in processing
        if outfields != "all":
            arcpy.MakeFeatureLayer_management(fcfullpath, layername, whereclause, workspacebase, fldinfo[:-1])
        elif outfields == "all" and whereclause != "":
            arcpy.MakeFeatureLayer_management(fcfullpath, layername, whereclause, workspacebase, None)
        elif outfields == "all" and whereclause == "":
            arcpy.MakeFeatureLayer_management(fcfullpath, layername, None, workspacebase, None)

        targetdb.generate_physical_geodatabase()   # Verify that the physical target is open
        targetdb_new_location = False

        # Ensure the dataset in the physical database exists, if needed. If not, create it
        if targetdb.does_db_type_support_spatial() and sourcedataset is not None and (
                not targetdb.dataset_exists_in_geodb()):
            spref = self.get_spatial_reference_frame(fc)
            # I want the spatial ref of the dataset, but Esri won't give me that,
            # it will only give me one from a feature class
            # sref = sourcedb.get_spatial_reference_frame(sourcedb.geodb_
            # feature_classes[sourcedb.geodb_dataset][0])
            # get spatial reference from any feature class in the freaking DB
            targetdb.generate_dataset(spref) 
            targetdb_new_location = True
        # Set up to copy, set the workspace to the source physical database
        else:
            if outputfc not in targetdb.saved_fc_datasets:
                targetdb_new_location = True
          

        if targetdb.does_db_type_support_spatial() and sourcedataset is not None:
            targetfcpath = os.path.normpath(targetdb.geodb_pathname + "/" + targetdb.geodb_dataset + "/" + outputfc)
        else:
            targetfcpath = os.path.normpath(targetdb.geodb_pathname + "/" + outputfc)

        arcpy.env.workspace = self.geodb_pathname
        if targetdb_new_location:
            arcpy.CopyFeatures_management(layername, targetfcpath)
        else:
            arcpy.Append_management(layername, targetfcpath)
        arcpy.Delete_management(layername)
        return targetfcpath

    def copy_feature_from_src_to_target(self, fc, srcdset, allfields, outfields, targdataset,
                                        outputfc, targetdb, whereclause):  # fc, fields, outfields, "templayer")
        # Set up to copy, set the workspace to the source physical database
        if outputfc is None:
            outputfc = fc
        outputfc = self.clean_name(outputfc)
        arcpy.env.workspace = self.geodb_pathname
        #  fcfullpath = self.get_root_for_fc_subtypes(fc)          # source fc full path
        workspacebase = os.path.normpath(self.geodb_pathname + (("/" + srcdset) if srcdset is not None else ""))
        # layerfullpath = self.get_root_dataset_for_fc(fc) + "//" + targdataset
        arcpy.env.overwriteOutput = True

        #fldinfo = arcpy.FieldInfo
        fldinfo = ""
        if outfields != "all":
            for field in allfields:
                # fldinfo.addField(field, field, "VISIBLE" if field in wantedfields else "HIDDEN", "NONE")
                fldinfo = fldinfo + field + " " + field + (
                    " VISIBLE;" if field in outfields else " HIDDEN;")  # (pass fldinfo[:-1])

        #              input feature class,  output layer name, where clause,
        #              input workspace used to validate field names,
        # arcpy.env.workspace = self.geodb_pathname

        # Depending on whether all fields are going to be copied, or if there is a where clause, we can adjust the
        # call to MakeFeatureLayer_management - and hopefully some of these will be faster in processing
        layername = "{0}_lyr".format(fc)
        if outfields != "all":
            arcpy.MakeFeatureLayer_management(fc, layername, whereclause, workspacebase, fldinfo[:-1])
        elif outfields == "all" and whereclause != "":
            arcpy.MakeFeatureLayer_management(fc, layername, whereclause, workspacebase, None)
        elif outfields == "all" and whereclause == "":
            arcpy.MakeFeatureLayer_management(fc, layername, None, workspacebase, None)

        # This call will generate it if it does not exist or it will just try to open it
        targetdb.generate_physical_geodatabase()   # Verify that the physical target is open

        targetdb_new_location = False
        # Ensure the dataset in the physical database exists, if needed. If not, create it
        if targetdb.does_db_type_support_spatial() and targdataset is not None and (
                not targetdb.dataset_exists_in_geodb()):
            spref = self.get_spatial_reference_frame(fc)
            # I want the spatial ref of the dataset, but Esri won't give me that, it
            # will only give me one from a feature class
            # sref = sourcedb.get_spatial_reference_frame(sourcedb.geodb_feature_classes[srcdset])
            # get spatial reference from any feature class in the freaking DB
            targetdb.geodb_dataset = targdataset
            targetdb.generate_dataset(spref) 
            targetdb_new_location = True
        # Set up to copy, set the workspace to the source physical database
        else:
            if outputfc not in targetdb.saved_fc_datasets:
                targetdb_new_location = True

        # Determine the full target fc path, including the dataset if there is one
        if (targetdb.does_db_type_support_spatial()) and targdataset is not None:
            targetfcpath = os.path.normpath(targetdb.geodb_pathname + "/" + targdataset + "/" + outputfc)
        else:
            targetfcpath = os.path.normpath(targetdb.geodb_pathname + "/" + outputfc)

        arcpy.env.workspace = self.geodb_pathname
        # If going into a new DB location, use copy; otherwise use append and perform with "NO_TEST" 
        # so that we don't get errors about not any field mismatch
        if targetdb_new_location:
            arcpy.CopyFeatures_management(layername, targetfcpath)
        else:
            arcpy.Append_management(layername, targetfcpath, "NO_TEST")

        arcpy.Delete_management(layername)
        return targetfcpath

    def clear_layer(self, layername):
        arcpy.Delete_management(layername)

    # Save the current feature code, and also save the former feature code
    # this will allow us to map the feature code to a new value, but still
    # be able to get at the original one in order to set source default values.   
    def set_f_code(self, f_code_in):
        self.current_f_code = f_code_in
        if self.former_f_code is None:
            self.former_f_code = f_code_in

    def get_f_code(self):
        return self.current_f_code

    def get_f_code_original(self):
        return self.former_f_code

    def set_fc(self, fc_in):
        self.current_fc = fc_in
        if self.former_fc is None:
            self.former_fc = fc_in

    def get_fc(self):
        return self.current_fc

    def get_fc_original(self):
        return self.former_fc

    def gui_updated_info_update_internal_vars(self):
        self.geodb_pathname = self.dbname.get()
        self.geodb_dataset = self.dataset.get()
        self.db_version = self.schema_name_and_version.get()

    ## Adds a feature class to the database specification. If necessary,
    # adds the dataset too (if there is none).
    ## Returns False if there is an error such as the dataset cannot be added or if the
    # featureclass is already in the spec.
    def add_featureclass_to_database_spec(self, featureclass, dataset):
        if dataset is None:
            if featureclass not in self.geodb_tables:
                self.geodb_tables.append(featureclass)
                self.saved_fc_datasets[featureclass] = None
                # This one gives us the dataset name for the feature class
        else:
            if len(self.geodb_datasets) == 0:
                self.geodb_datasets.append(dataset)
                self.geodb_dataset = dataset
                self.geodb_feature_classes[dataset] = list()
            elif dataset not in self.geodb_datasets:
                self.settings.my_logger.error(
                    "Specified dataset {0} is not in the database, so we cannot add the feature class requested.".
                        format(dataset))
                return False

            if featureclass in self.geodb_feature_classes[dataset]:
                self.settings.my_logger.info(
                    "Feature class or table {0} requested for addition is already in the geodatabase.".
                        format(featureclass))
                return False
            else:
                self.geodb_feature_classes[dataset].append(featureclass)
                # This one is db structure - dataset contains feature class
                self.saved_fc_datasets[featureclass] = dataset
                # This one gives us the dataset name for the feature class
        return True

    def sdef_prepare_field_info_for_new_featureclass(self, fc):  # Was prepare_fi_for_new_fc
        # Prepare feild_information content for this feature class or table
        if self.db_version not in self.field_information:
            self.field_information[self.db_version] = dict()
        if fc not in self.field_information[self.db_version]:
            self.field_information[self.db_version][fc] = dict()
            # self.field_information[self.db_version][featureclass]['EnumeratedIndexValues'] = dict()

    def sdef_add_featureclass_fields_to_schema_definition(self, fc, fieldsfc):
        self.sdef_prepare_field_info_for_new_featureclass(fc)
        for field in fieldsfc:
            if field not in self.field_information[self.db_version][fc]:
                self.field_information[self.db_version][fc][field] = None
                # need a list: ('Datatype', ('fcode1', 'fcode2', ...))
                # if datatype == 'Enumeration':
                #     self.field_information[self.db_version][fc]['EnumeratedIndexValues'] = dict()
                #     self.field_information[self.db_version][fc]['EnumeratedIndexValues'][field] = dict()
                #     for fcode in self.field_information[self.db_version][fc][field][1]:
                #       self.field_information[
                #       self.db_version][fc]['EnumeratedIndexValues'][field][fcode] = list of valid codes
                #     
                #         self.field_information[self.db_version]['DefaultValues'][fcode][field] = default value

    # Gets pre-defined schema information for a specific schema version.
    # This will specify the allowed attribute fields
    # per feature code, and the data types for fields, and allowed domain values per field
    def load_schema_information(self):
        if self.db_version == "":
            self.db_version = str(self.settings.schema_name_and_version.get())
        self.field_information[self.db_version] = dict()
        # Import the field information content for the appropriate schema
        # that we are concerned with.
        self.my_logger.info("Loading geodatabase: {0}".format(self.geodb_pathname))
        self.my_logger.info("Loading schema information for {0}.".format(self.db_version))
        if (self.db_version == 'GGDM 3.0' ):
           from GGDM30_FCFields import GGDM30AddSchemaInfoFeaturesSubtypes
           if self.settings.config_setting_flag['flag_allow_flat_version_database'] == 1:
                from GGDM30_FCFieldsFeaAsFC import GGDM30AddSchemaInfoFeaturesAsFCTables
           GGDM30AddSchemaInfoFeaturesSubtypes(self.field_information)
           if self.settings.config_setting_flag['flag_allow_flat_version_database'] == 1:
                GGDM30AddSchemaInfoFeaturesAsFCTables(self.field_information)

        elif (self.db_version == 'GGDM 2.1'):
            from GGDM21_FCFields import GGDM21AddSchemaInfoFeaturesSubtypes
            GGDM21AddSchemaInfoFeaturesSubtypes(self.field_information)

        elif (self.db_version == 'GGDM 2.2'):
            from GGDM22_FCFields import GGDM22AddSchemaInfoFeaturesSubtypes
            GGDM22AddSchemaInfoFeaturesSubtypes(self.field_information)

        # elif (self.db_version == 'GGDM 2.2 WRDB'):
        #     from GGDM22WRDB_FCFields import GGDM22WRDBAddSchemaInfoFeaturesSubtypes
        #     GGDM22WRDBAddSchemaInfoFeaturesSubtypes(self.field_information)

        elif (self.db_version == 'TDS 7.0'):
            from TDS70_FCFields import TDS70AddSchemaInfoFeaturesSubtypes
            TDS70AddSchemaInfoFeaturesSubtypes(self.field_information)

        elif (self.db_version == 'TDS 7.1'):
            from TDS71_FCFields import TDS71AddSchemaInfoFeaturesSubtypes
            TDS71AddSchemaInfoFeaturesSubtypes(self.field_information)

        elif (self.db_version == 'TDS 6.1'):
            from TDS61_FCFields import TDS61AddSchemaInfoFeaturesSubtypes
            TDS61AddSchemaInfoFeaturesSubtypes(self.field_information)

        # elif (self.db_version == 'TDS 6.0'):
        #     from TDS6_FCFields import TDS60AddSchemaInfoFeaturesSubtypes
        #     TDS60AddSchemaInfoFeaturesSubtypes(self.field_information)

        elif (self.db_version == 'TDS 4.0'):
            from TDS4_FCFields import TDS40AddSchemaInfoFeaturesSubtypes
            TDS40AddSchemaInfoFeaturesSubtypes(self.field_information)

        elif os.path.exists(self.settings.file_results_excel.get()):
            pass  
        else:
            report_transfomer_issue_or_increment_count(
                self.settings, "Undefined or unknown database version and schema type specified", "WARN",
            "If a mapping excel file was specified, this error should not be seen. DB version is: {0}".format(
                self.db_version))

        self.db_schema_def_loaded = True

    # Returns the default root path for feature classes -
    # meaning it is the geodatabase pathname plus the default dataset name
    def get_default_root_for_fc(self):
        return os.path.normpath(self.geodb_pathname + "/" + self.geodb_dataset)
        # does not like os.path.join for this

    # Returns the root pathname for a specific feature class or table.
    def get_root_dataset_for_fc(self, fc):
        dset = self.get_dataset_for_fc_in_geodatabase(fc)
        if dset is not None:
            return os.path.normpath(self.geodb_pathname + "/" + dset)
            # does not like os.path.join for this
        else:
            return self.geodb_pathname 

    def get_count_of_instances_in_fc(self, fc):
        dset = self.get_dataset_for_fc_in_geodatabase(fc)
        if dset is not None:
            fcbase =  os.path.normpath(self.geodb_pathname + "/" + dset + "/" + fc)
            # does not like os.path.join for this
        else:
            fcbase =  os.path.normpath(self.geodb_pathname + "/" + fc)
        arcpy.MakeTableView_management(fcbase, "tempCountView")
        count_out = int(arcpy.GetCount_management("tempCountView").getOutput(0))
        #  result = arcpy.GetCount_management(lyrfile)
        arcpy.Delete_management("tempCountView")
        return (count_out)

    # Returns the root pathname for tables is the same as the geodatabase path name
    def get_root_for_table(self):
        return self.geodb_pathname

    # get the featureclass dataset path for a specific featureclass (may not be default dataset)
    def get_root_for_fc_subtypes(self, fc):
        dset = self.get_dataset_for_fc_in_geodatabase(fc)
        if dset is not None:
            return os.path.normpath(self.geodb_pathname + "/" + dset + "/" + fc)
            # Does not like to use the os.path.join on the dataset.
        else:
            if fc is not None:
                return os.path.normpath(self.geodb_pathname + "/" + fc)
                # THis may be a table?
            else:
                return self.geodb_pathname

    # This method will get the field list from the DATABSE and it will save it internally because we 
    # typically (in a mapping) will actually get this field list twice and it will be much faster to save
    # it than to query the database twice
    def get_db_actual_field_list(self, fc_in_val):
        if fc_in_val in self.saved_db_field_lists:
            return self.saved_db_field_lists[fc_in_val]
        else:
            src_fc_db_root = self.get_root_for_fc_subtypes(fc_in_val)
            if src_fc_db_root is not None:
                self.saved_db_field_lists[fc_in_val] = sorted(
                    [self.correct_shape_field_name_to_compensate_for_esri_strangeness(f.name) \
                    for f in arcpy.ListFields(src_fc_db_root)])
                return self.saved_db_field_lists[fc_in_val]
            else:
                return None

    # Look at a feature class: fc_in and present a list of potential expected fields: potential_fields_in.
    # Get the actual set of fields for the feature class and prepare the list of intersecting fields and
    # also prepare a list showing the potential input fields that were not intersected and the list of 
    # additional fields in this feature class.
    # called with a feature class, a list of potential fields, and three empty lists
    def intersect_fields_for_fc(self, fc_in, potential_fields_in, intersected_fields, \
        fields_not_intersected, fields_additional_in_fc, leave_them_lowercase = False):
        
        # Get the fields for this feature class and sort them and put all names lower case
        arcfields = self.get_db_actual_field_list(fc_in)    # change to get_fields_from_featureclass
        tblfields = set([field.lower() for field in arcfields])
        # lowercase set of fields from geodatabase

        del intersected_fields[:]
        del fields_not_intersected[:]
        del fields_additional_in_fc[:]

        # Be sure the input list of fields is a set and lower case
        src_fieldsAA = set([field.lower() for field in potential_fields_in])
        # Mapping specification defines these fields, put them into lower case

        # Adjustment - this is for Esri -  when retretriving the fields with get_field_names, 
        # shape@ will be shape, shape@area will be shape_area, and shape@length will be shape_length
        if 'shape' in tblfields:
            tblfields.remove('shape')
            tblfields.add('shape@')
        if 'shape_length' in tblfields:
            tblfields.remove('shape_length')
            tblfields.add('shape@length')
        if 'shape_area' in tblfields:
            tblfields.remove('shape_area')
            tblfields.add('shape@area')

        source_result = src_fieldsAA.intersection(tblfields)
        # This is the set of intersecting fields
        not_in_result = set(src_fieldsAA.difference(tblfields))
        # This is the set of potential fields that were not found in the feature class or tabe
        additional_fields = set(tblfields.difference(src_fieldsAA))
        # This is the set of additional fields in the feature class or table

        if leave_them_lowercase == False:
            # Okay, so now we have to put the field names back to uppercase or mixed case to provide the final result.
            for f in potential_fields_in:
                if f.lower() in source_result:
                    intersected_fields.append(f)
                if f.lower() in not_in_result:
                    fields_not_intersected.append(f)
            for f in arcfields:
                if f.lower() in additional_fields:
                    fields_additional_in_fc.append(f)
        else:
            # Move the set results into lists
            for f in source_result:
                intersected_fields.append(f)
            for f in not_in_result:
                fields_not_intersected.append(f)
            for f in additional_fields:
                fields_additional_in_fc.append(f)

    def set_value_to_default_if_none(self, field, fieldvalue):
        if field is not None and fieldvalue is None and \
            self.settings.debug_settings['leave_unspecified_targets_null_instead_of_defaulting'] == 0:  # \
           # and field.lower() not in self.common_fields_geom_and_id_and_code:
            if self.settings.debug_settings['itemize_each_translation_each_field'] > 1:
                    self.settings.my_logger.debug(
                        "\tSetting field {0} to default for feature code {1} in feature class: {2}".\
                        format(field, self.get_f_code(), self.get_fc()))
            resultval = dict()
            resultval['DefaultVal'] = None
            possDefault = self.field_set_to_default(self.get_f_code(), field, self.get_fc(), resultval, 'DefaultVal')
            if possDefault is not None:
                return possDefault
        return fieldvalue

    def identify_actual_fields_list_matching_schema(self, reporttype, fc, fieldsinput):
        not_in_fields = list()                 # Fields that do not intersect the field list passed as a parameter
        db_has_but_not_in_schema = list()      # Fields in the database that were not listed in the field list
        self.updated_fields = list()
        self.intersect_fields_for_fc(fc, fieldsinput, self.updated_fields, not_in_fields, db_has_but_not_in_schema)
        if self.settings.debug_settings['report_field_schema_vs_database_differences']>0:
            # If we have the capability, output this report to excel as it'll be easier to decipher there.
            if self.settings.results_xl is None:
                if len(fieldsinput) <> len(self.updated_fields):
                    self.settings.my_logger.warn(
                        "Fields in specification are not all found in database feature class or table {0}. \n  Fields expected: {1}\n    Fields intersected from expected list and database: {2}  \n  Fields expected but not in database: {3}".\
                        format( fc, self.updated_fields, fieldsinput, not_in_fields))
                if len(db_has_but_not_in_schema) > 0:
                    self.settings.my_logger.warn(
                        "Fields in database are not all found in schema definition for feature class or table {0}. \n  Fields found in addition to specification: {1}\n  Fields in specification: {2}".\
                        format( fc, db_has_but_not_in_schema, fieldsinput))
            else:
                self.settings.results_xl.save_field_report(
                    fc, reporttype, fieldsinput, self.updated_fields, db_has_but_not_in_schema)

    # Returns true if the feature class is specified in this class
    def is_featureclass_in_geodatabase_spec(self, featureclass):      # was is_fc_in_geodatabase
        if featureclass in self.saved_fc_datasets:
            return True
        else:
            return False

    # Use the saved dataset and feature class information to determine
    # if the specified feature class is present in this geodatabase in
    # any dataset and retrun the name of the dataset if it is.
    def get_dataset_for_fc_in_geodatabase(self, featureclass):
        if featureclass in self.saved_fc_datasets:
            return self.saved_fc_datasets[featureclass]
        else:
            return None

    # Returns true if the geodatabase physically exists
    def geodatabase_exists(self):
        if arcpy.Exists(self.geodb_pathname):
            desc = arcpy.Describe(self.geodb_pathname)
            wrksp_type = desc.workspaceType

            if wrksp_type.lower() in ['workspace', 'localdatabase', ]:
                return True
            else:
                dset_type = desc.datasetType
                if dset_type.lower() in ['featuredataset', ]:
                    return True
        return False

    # Returns true if the default dataset physically exists
    def dataset_exists_in_geodb(self):
        arcpy.env.workspace = self.geodb_pathname
        if arcpy.Exists(self.geodb_dataset):
            return True
        else:
            return False

    # Returns True if the featureclass physically exists
    def featureclass_exists_in_geodb(self, fc):
        arcpy.env.workspace = self.get_root_dataset_for_fc(fc)
        if arcpy.Exists(fc):
            return True
        else:
            return False

    # UNVERIFIED method.
    def featureclass_getsubtypes_defaults_geodb(self, fc, fcsubtypedict):
        arcpy.env.workspace = self.get_root_dataset_for_fc(fc)
        if arcpy.Exists(fc):
            #  fieldNames = [f.name for f in arcpy.ListFields(fc)]
            subtypes = arcpy.da.ListSubtypes(fc)
            if len(subtypes) > 0:
                for subtcode, stdict in list(subtypes.items()):
                    fcsubtypedict[subtcode] = dict()
                    for stkey in list(stdict.keys()):
                        if stkey == 'FieldValues':
                            fields = stdict[stkey]
                            for ff, fieldvals in list(fields.items()):
                                #  per field name allowed in subtype: default value, domain name
                                fcsubtypedict[subtcode][ff] = (fieldvals[0], "{0}".format(fieldvals[1].name))

    # Returns shapetype for feature class if it is a geometric type exists
    def featureclass_geometry_type_geodb(self, fc):
        arcpy.env.workspace = self.get_root_dataset_for_fc(fc)
        if arcpy.Exists(fc):
            try:
                desc = arcpy.Describe(fc)
                # other parameters available: featureType, hasSpatialIndex, hasZ, shapeFieldName
                return desc.shapeType
            except:
                pass
        return None

    def assume_geometry_type_from_featureclass_name(self, src_fc):
        if src_fc[-3:].lower() == "pnt" or src_fc[-4:].lower() == "point" or \
            src_fc[-2:].lower() == "_p" or src_fc[-6:].lower() == "pntext" or \
                src_fc[-5:].lower() == "_pext":
            return "point"
        elif src_fc[-3:].lower() == "srf" or src_fc[-4:].lower() == "area" or \
            src_fc[-7:].lower() == "polygon" or src_fc[-2:].lower() == "_a" or \
            src_fc[-2:].lower() == "_s" or src_fc[-6:].lower() == "aonext"  or \
            src_fc[-6:].lower() == "srfext" or src_fc[-5:].lower() == "_sext":
            return "polygon"
        elif src_fc[-3:].lower() == "crv" or src_fc[-4:].lower() == "line" or \
            src_fc[-4:].lower() == "edge" or src_fc[-2:].lower() == "_l" or \
            src_fc[-2:].lower() == "_c"  or src_fc[-6:].lower() == "crvext" \
            or src_fc[-5:].lower() == "_cext":
            return "polyline"
        else:
            return "table"


    def generate_physical_geodatabase(self):
        if self.geodatabase_exists() == False:
            try:
                self.settings.my_logger.info(
                    "Creating geodatabase: {0}  Type: {1}  Spatial Type: {2}".format(
                        self.geodb_pathname, self.db_type, self.db_spatial_type))
                if self.geodb_pathname[-4:].lower() == ".gdb" and self.db_type.lower() == "filegdb":
                    arcpy.CreateFileGDB_management(
                        os.path.dirname(self.geodb_pathname), os.path.basename(self.geodb_pathname))
                elif self.geodb_pathname[-7:].lower() == ".sqlite" and self.db_type.lower() == "sqlite":     
                    arcpy.gp.CreateSQLiteDatabase(self.geodb_pathname, self.db_spatial_type)
                    # CreateSQLiteDatabase_management (out_database_name, {spatial_type})
                elif self.geodb_pathname[-5:].lower() == ".gpkg" and self.db_type.lower() == "geopackage":
                    arcpy.gp.CreateSQLiteDatabase(self.geodb_pathname, self.db_spatial_type)
                else:
                    errstmt = "Can not create geodatabase {0} of type {1}, mismatch on name vs type.".format(
                        self.geodb_pathname, self.db_type)
                    self.settings.my_logger.error(errstmt)
                self.mode = "CREATED"
            except:
                self.mode = ""
        else:
            self.mode = "EXISTS"

    def does_db_type_support_spatial(self):
        # NEW20 Added test for file type and file must include a dataset
        return True if self.db_type.lower() in ['filegdb', ] and self.geodb_dataset is not None else False

    # Create a dataset if it does not already exist
    def generate_dataset(self, spatialref):
        if self.dataset_exists_in_geodb() == False:
            if self.does_db_type_support_spatial():
                arcpy.CreateFeatureDataset_management(self.geodb_pathname, self.geodb_dataset, spatialref)
            else:
                self.settings.my_logger.warn("Dataset {0} is not allowed in this type of database {1}".format(
                    self.geodb_dataset, self.db_type))

    def getfieldlistfromschemadef(self, infeatureclass):
        return sorted([ ffi for ffi in self.field_information[self.db_version][infeatureclass]])
    
    def getvalidcodesforfeatureclassfield(self, infeatureclass, field_in):
        if infeatureclass in self.field_information[self.db_version]:
            if field_in in self.field_information[self.db_version][infeatureclass]:
                # datatype = self.field_information[self.db_version][infeatureclass][field_in][0].lower()
                # first element of list
                if isinstance(self.field_information[self.db_version][infeatureclass][field_in][1], str):
                    lstout = list()
                    lstout.append(self.field_information[self.db_version][infeatureclass][field_in][1])
                    return lstout
                else:
                    return list(self.field_information[self.db_version][infeatureclass][field_in][1])
                    # second element - valid fcodes
        return None

    def get_primary_fcode_for_featureclass(self, infeatureclass):
        return self.getvalidcodesforfeatureclassfield(infeatureclass, 'F_CODE')[0]

    def field_set_to_default(self, fcode, field_in, featureclass, result=False, result_key='none'):
        # print 'Field: {0} SrcFC: {1} TargetFC: {2}'.format(field_in, src_fc, featureclass)
        if isinstance(result, dict):
            saveresults = True
        else:
            saveresults = False

        debug_set_dict = self.settings.debug_settings

        if debug_set_dict['itemize_each_translation_each_field'] > 1:
            self.settings.my_logger.debug(
                "Setting Field {0} to default for feature class: {1} having F_CODE: {2} for database: {3}".\
                format(field_in, featureclass, fcode, self.db_version))

        if self.geodb_type_unknown == 1 or len(self.field_information) == 0 :
            if saveresults:
                result[result_key] = "Unknown type of source db or schema definition not loaded"
            report_transfomer_issue_or_increment_count(
                self.settings,
                'Default - either no defined schema or it was not loaded. Attempts to set fields to default will not function.',
                            "WARN")
            return None

        # The fields listed in common_fields_geom_and_id_and_code may not be defined
        # within the schema a as most of these are default Esri fields
        if field_in.lower() in self.common_fields_geom_and_id_and_code:
            if field_in.lower() == 'f_code':
                report_transfomer_issue_or_increment_count(
                    self.settings, 'Default - F_CODE field asked to be set to default, setting to NA', "ERROR",
                         "Issue with featureclass or table: {0} and request for database version: {1}".format(
                             featureclass, self.db_version))
                if saveresults:
                    result[result_key] = "Issues with F_CODE, default to NA"
                return 'NA'   # f_code is a text value
            elif field_in.lower() == 'fcsubtype':
                if fcode is not None:
                    fcsubtypemap = self.get_fcode_subtype_mapping_for_all()
                    if fcode not in fcsubtypemap:
                        report_transfomer_issue_or_increment_count(
                            self.settings, 'Default - FCSubtype field asked to be set to default, and unable to set based on F_CODE', "ERROR",
                                  "Issue with featureclass or table: {0} and request for database version: {1} and F_CODE {2}".format(
                                      featureclass, self.db_version,fcode))
                        if saveresults:
                            result[result_key] = "Issues with FCSubtype, default based on F_CODE"
                        return 0
                    else:
                        return fcsubtypemap[fcode]
                else:
                    report_transfomer_issue_or_increment_count(
                        self.settings, 'Default - FCSubtype field asked to be set to default, but we cannot set it, no F_CODE', "ERROR")
                    return 0
            elif field_in.lower() in self.common_fields_geom_only:    # Field is Shape, Shape_Length, Shape_Area, we can set it to null i.e. None
                if debug_set_dict['allow_default_geometry_test_mode'] == 1:
                    return None
                report_transfomer_issue_or_increment_count(
                    self.settings,
                    'Default - Geometry field asked to be set to default, setting to null content', "WARN")
                if saveresults:
                    result[result_key] = "Issues with Shape field, default to null value"
                return None   # Geometry fields can default to nothing
            else:
                report_transfomer_issue_or_increment_count(
                    self.settings, 'Default - {0} field ased to be set to default'.format(field_in), "ERROR",
                         "Issue with featureclass or table: {0} and request for database version: {1}".format(
                             featureclass, self.db_version))
                if saveresults:
                    result[result_key] = "Issues with default identifier"
                return 0           

        # If we have not yet dealt with the default, we will have to look it up in the schema definition. First, is it even valid in this schema definition?
        validcodes = self.getvalidcodesforfeatureclassfield(featureclass, field_in)
        if validcodes is None or len(validcodes) == 0:
            if (debug_set_dict['report_missing_source_data'] > 0 or \
                    debug_set_dict['itemize_each_translation_each_field'] > 0):
                report_transfomer_issue_or_increment_count(
                    self.settings, 'Default Processing- Feature class and field undefined for this schema',
                            "WARNING", "The feature class {0} and field {1} are not found with any feature code list in the schema: {2}".format(featureclass, field_in, self.db_version))
            if saveresults:
                result[result_key] = "Field undefined in schema"
            return None


        if saveresults:
            result[result_key] = "No updates made. Issues found."

        # verify that the feature class is in the database schema, that the field is in the feature class,
        # and get the data type and list of valid f_codes for the field as per the schema
        if featureclass in self.field_information[self.db_version]:
            if field_in in self.field_information[self.db_version][featureclass]:
                datatype = self.field_information[self.db_version][featureclass][field_in][0].lower()
                # first element of list
                #  validfcodes = self.getvalidcodesforfeatureclassfield(featureclass, field_in)
                # second element of list got it above
                if fcode in validcodes:
                    # print "valid fcode detected " + row['F_CODE']
                    # enumeration, integer, boolean, structuredtext, string, real,  codelist, constrained_string
                    if 'Default - Field set to default value' in self.settings.report_values:
                        self.settings.report_values['Default - Field set to default value'] += 1
                    else:
                        self.settings.report_values['Default - Field set to default value']  = 1
                    if saveresults:
                        result[result_key] = "Success"
                    if field_in == 'INVL' or field_in in self.field_information[self.db_version]['interval_fields']:
                        return 5    # Closed Interval
                    elif field_in == 'SAX_RS0':
                        return 'U'  # Resource Classification default UNCLASSIFIED
                    elif datatype == 'enumeration' or datatype == 'integer' or datatype == 'boolean':
                        return -999999
                    elif datatype == 'string' or datatype == 'text':
                        return 'No Information'
                    elif datatype == 'structuredtext' or datatype == 'codelist':
                        return 'noInformation'
                    elif datatype == 'real':
                        return -999999.0
                    elif datatype == 'constrained_string' and field_in.lower() == 'f_code':
                        report_transfomer_issue_or_increment_count(
                            self.settings,
                            'Default - Tried to set a field to default that is not allowed (Feature Code!)',
                             "ERROR", "We cannot set this field {2} for feature class {1} having feature code {0}.".\
                                format(fcode, featureclass, field_in))
                        if saveresults:
                            result[result_key] = \
                                "Error NA attribute and datatype {0} should never be requested for this".format(datatype)
                        return 'none'
                    elif datatype == 'constrained_string':
                        return 'noInfo'
                    # A few special case default values - there is one and only one value
                    # for these crop species attributes
                    elif fcode == 'EA055' and field_in == 'ZI013_CSP':
                        return 18     # Default value crop species for Hop Field
                    elif fcode == 'BH135' and field_in == 'ZI013_CSP':
                        return 28     # Default value crop species for Rice Field
                    elif fcode == 'EA050' and field_in == 'ZI013_CSP':
                        return 16     # Default value crop species for Vineyard of Grape
                    else:
                        report_transfomer_issue_or_increment_count(
                            self.settings,
                            'Default - Tried to set a field to default but we could not determine the value',
                        "ERROR", "We cannot set this field {2} for feature class {1} having feature code {0}.".format(
                                fcode, featureclass, field_in))
                        if saveresults:
                            result[result_key] = "Error datatype {0} has no default value handler".format(datatype)
                        return 'none'
                else:
                    # The fcode is not valid for this feature class, so we have a different kind of default value for that. Instead of No Information, the default is ------ where possible.
                    if 'set field to not applicable for feature' in self.settings.report_values:
                        self.settings.report_values['set field to not applicable for feature'] += 1
                    else:
                        self.settings.report_values['set field to not applicable for feature'] = 1
                    if saveresults:
                        result[result_key] = "NA Attribute"

                    if datatype == 'enumeration' or datatype == 'integer' or datatype == 'boolean':
                        return -999999
                    elif datatype == 'string' or datatype == 'structuredtext' or \
                            datatype == 'codelist' or datatype == 'text':
                        return '------'
                    elif datatype == 'real':
                        return -999999.0
                    elif datatype == 'constrained_string' or field_in.lower() == "f_code":
                        report_transfomer_issue_or_increment_count(
                            self.settings, 'Default - Tried to set a field to default that is not allowed (Feature Code!)',
                                        "ERROR", "We cannot set this field {2} for feature class {1} having feature code {0}.".format(fcode, featureclass, field_in))
                        if saveresults:
                            result[result_key] = "Error NA attribute and datatype {0} should never be requested for this".format(datatype)
                        return '------'
                    elif datatype == 'constrained_string':
                        return '------'
                    else:
                        report_transfomer_issue_or_increment_count(
                            self.settings, 'Default - Tried to set a field to default but we could not determine the value',
                        "ERROR", "We cannot set this field {2} for feature class {1} having feature code {0}.".format(
                                fcode, featureclass, field_in))
                        if saveresults:
                            result[result_key] = "Error NA attribute and datatype {0} has no default value handler".format(datatype)

            else:  # this field is not defined with a default - it is probably a shape field
                if saveresults:
                    result[result_key] = "Not defined for feature class".format(field_in, featureclass)
                else:
                    report_transfomer_issue_or_increment_count(
                        self.settings, "Received request to set default for a field that is not defined for this geodatabase or has no default defined", "WARN",
                            "Received request to set default for field {1} but field is not defined (and may not be legal in schema) feature Class: {0}".format(featureclass, field_in))
                #if debug_set_dict['itemize_each_translation_each_field'] > 1:
                #    self.settings.my_logger.debug("Received request to set default for field {1} but field is not defined (and may not be legal in schema) feature Class: {0}".format(featureclass, field_in))
                    
                return None
        if saveresults:
            result[result_key] = "No schema definition"
        #self.settings.my_logger.warn("Received request to set default for field {1} on feature class: {0} but we do not have a specified data type for F_CODE {2} list!".format(featureclass, field_in))
        return None

    def is_attribute_value_illegal(self, rowval, field_in, fcode, featureclass):
        # Using the schema definition try to determine if the value is illegal or not
        value_illegal = False
        value_fixed = False
        rtnval = None
        if fcode is None:
            self.my_logger.error(
                "No feature code passed to function to determine if attribute value is illegal. Feature Class {0}, field {1}.".format(featureclass, field_in))
            return (True, False, None)

        if featureclass in self.field_information[self.db_version]:
            if field_in in self.field_information[self.db_version][featureclass]:
                datatype = self.field_information[self.db_version][featureclass][field_in][0].lower()
                # first element of list
                validfcodes = self.getvalidcodesforfeatureclassfield(featureclass, field_in)
                # second element of list
                if fcode in validfcodes:
                    # print "valid fcode detected " + row['F_CODE']
                    # enumeration, integer, boolean, structuredtext, string, real,  codelist, constrained_string
                    if datatype == 'enumeration':
                        # This is an enumeration, the value should be within the
                        # enumeratedindexvalues specification for this feature class, field, and fcode
                        if rowval is None or \
                                rowval not in self.field_information[
                                    self.db_version][featureclass]['EnumeratedIndexValues'][field_in][fcode]:
                            # value is not in the list
                            value_illegal = True
                    elif datatype == 'constrained_string' and field_in == 'F_CODE':
                        if rowval is None or rowval not in validfcodes:
                            value_illegal = True
                    elif datatype == 'structuredtext' or  datatype == 'codelist':
                        specificval = ["noinformation", "NoInformation"]
                        eithercase = ["no information", "no_information", "na", "n/a", "n_a", "none", "<null>",
                                      "null", "unknow", "unknown", "unk", "-32768", "-32767"]
                        if rowval is None or rowval in specificval or rowval.lower() in eithercase:
                            value_illegal = True
                            value_fixed = True
                            rtnval = "noInformation"
                    elif datatype == 'boolean':
                        if rowval is None or rowval not in [-999999, 1000, 1001]:
                            # value is not valid
                            value_illegal = True
                            value_fixed = True
                            rtnval = -999999
                    elif datatype == 'integer':
                        if rowval is None or rowval < 0 and rowval <> -999999:   # This may be an illegal value
                            if (rowval <= -32765 and rowval >= -32769) or (rowval <= -3275 and rowval >= -3277):
                                value_illegal = True
                                value_fixed = True
                                rtnval = -999999
                    elif datatype == 'string' or datatype == 'text':
                        specificval = ["no information"]
                        eithercase = ["noinformation", "no_information", "na", "n/a", "n_a", "none", "<null>",
                                      "null", "unknow", "unknown", "unk", "-32768", "-32767"]
                        if rowval is None or rowval in specificval or rowval.lower() in eithercase:
                            value_illegal = True
                            value_fixed = True
                            rtnval = "No Information"
                    elif datatype == 'real':
                        if rowval is None or rowval < 0 and rowval <> -999999:   # This may be an illegal value
                            if rowval is None or (rowval <= -32765 and rowval >= -32769) or \
                                    (rowval <= -3275 and rowval >= -3277):
                                value_illegal = True
                                value_fixed = True
                                rtnval = -999999.0

                else:   # fcode is not valid, attribute should be 'illegal'
                    if datatype == 'integer':
                        pass
                    elif datatype == 'real':
                        pass
                    elif datatype == 'boolean' or datatype == 'enumeration':
                        pass
                    else:
                        pass

        return (value_illegal, value_fixed, rtnval)

    def get_first_featureclass_name(self):
        return self.geodb_feature_classes[self.geodb_dataset][0]

    # Copies a feature class to a database. Will create the database and datsaet if it needs to.
    def copy_featureclass_to_database(self, featureclass, sourcedb):

        if sourcedb.is_featureclass_in_geodatabase_spec(featureclass):
            # Only if the sourceDB has the specified feature class or table, should we continue
            if sourcedb.does_db_type_support_spatial():
                # Copy feature class from the sourceDB to this DB
                # sourcedb is another MyGDBSettings instance
                # Ensure the target physical database exists and if it does not, create it.
                self.generate_physical_geodatabase()
                sourcedataset = sourcedb.get_dataset_for_fc_in_geodatabase(featureclass)
                # If this is a table, this will return None

            
                # Ensure the dataset in the physical database exists, if needed. If not, create it
                if sourcedataset == None or self.dataset_exists_in_geodb() == False:
                    # I want the spatial ref of the dataset, but Esri won't give me that,
                    # it will only give me one from a feature class
                    # sref = sourcedb.get_spatial_reference_frame(
                    # sourcedb.geodb_feature_classes[sourcedb.geodb_dataset][0])
                    # get spatial reference from any feature class in the freaking DB
                    self.generate_dataset(sourcedb.get_default_root_for_fc())
                    # Hope that Esri will allow us to use the spatial reference data we got!

                # Set up to copy, set the workspace to the source physical database
                arcpy.env.workspace = sourcedb.geodb_pathname

                sourceroot = sourcedb.get_root_dataset_for_fc(featureclass)

                # If this feature has not already been saved to this database, then copy it from the source
                if self.add_featureclass_to_database_spec(featureclass, sourcedataset):
                    # We need to get back a True from this call to copy the feature.

                    # If sourcedataset is None, this is a table
                    if sourcedataset == None:
                        arcpy.Copy_management(sourcedb.geodb_pathname + "/" +
                                              featureclass, self.geodb_pathname + "/" + featureclass)

                    # Otherwise, this is a feature class under a dataset
                    else:
                        infc = os.path.join(sourceroot, featureclass)
                        targroot = self.get_root_dataset_for_fc(featureclass) 
                        outfc = os.path.join(targroot, featureclass)
                        arcpy.CopyFeatures_management(infc, outfc)
            else:
                # Copy feature class from the sourceDB to this DB
                # sourcedb is another MyGDBSettings instance
                # Ensure the target physical database exists and if it does not, create it.
                self.generate_physical_geodatabase()
                sourcedataset = sourcedb.get_dataset_for_fc_in_geodatabase(featureclass)
                # If this is a table, this will return None

                # Set up to copy, set the workspace to the source physical database
                arcpy.env.workspace = sourcedb.geodb_pathname
                sourceroot = sourcedb.get_root_dataset_for_fc(featureclass)

                # If this feature has not already been saved to this database, then copy it from the source
                if self.add_featureclass_to_database_spec(featureclass, None):
                    # We need to get back a True from this call to copy the feature.

                    # If sourcedataset is None, this is a table
                    if sourcedataset is None:
                        arcpy.Copy_management(sourcedb.geodb_pathname + "/" +
                                              featureclass, self.geodb_pathname + "/" + featureclass)

                    # Otherwise, this is a feature class under a dataset
                    else:
                        #infc = os.path.join(sourceroot, featureclass)
                        #outfc = os.path.join(self.geodb_pathname, featureclass)
                        #arcpy.CopyFeatures_management(infc, outfc)
                        arcpy.Copy_management(sourceroot + "/" + featureclass, self.geodb_pathname + "/" + featureclass)
        else:
            # the specified feature clss is not in the sourcedb
            self.settings.my_logger.warn(
                "The specified feature class to copy {0} is not in the source database.".format(featureclass))

    def makerandomufi(self):
        return "{" + str(uuid.uuid4()).upper() + "}"

    def check_ufi_work_ufi_uri_existing(self, result, ttUFI, rows, colnames, fieldname, src_fc, rowcountii):
        result['ufi unique'] = True
        result['ufi updated'] = False

        if ttUFI in self.existing_unique_ids or ttUFI == "" or \
            ttUFI.lower() == 'noinformation' or ttUFI.lower() == 'no information':
            # Careful! If there is a duplicate and this setting is true, we will change the SOURCE
            result['ufi invalid in feature class'] = src_fc
            result['ufi invalid in row'] = rowcountii
            result['ufi invalid in feature field'] = fieldname

            if self.settings.config_setting_flag['flag_ufi_is_not_unique_generate_new_unique_ufi'] == 1:
                # Gen new uuid
                                
                if self.settings.config_setting_flag['report_all_duplicate_ufi'] > 0 and \
                        result['ufi count of ufi not unique or in error'] < \
                        self.settings.config_setting_flag['report_all_duplicate_ufi']:
                    self.settings.my_logger.error(
                        "{3} {0} is not unique or invalid. Found in feature class {1}, row {2}  This will be UPDATED.".\
                            format( ttUFI, src_fc, rowcountii, fieldname))

                newUFI = self.makerandomufi()
                if self.settings.config_setting_flag['report_all_duplicate_ufi'] > 0 and \
                        result['ufi count of ufi not unique or in error'] < \
                        self.settings.config_setting_flag['report_all_duplicate_ufi']:
                    self.settings.my_logger.info(
                        "CHANGING SOURCE {0} duplicate {3} found, we are changing it from {1} to {2}.".\
                            format(src_fc, ttUFI, str(newUFI), fieldname))
                rows[colnames.index(fieldname)] = newUFI
                result['ufi updated'] = True
                result['ufi unique'] = False
                self.existing_unique_ids.append(str(newUFI))

            elif self.settings.config_setting_flag['flag_ufi_is_not_unique_generate_new_unique_ufi'] == 2 and \
                'GlobalID' in colnames:  # Look for GlobalID or ObjectID
                if self.settings.config_setting_flag['report_all_duplicate_ufi'] > 0 and \
                        result['ufi count of ufi not unique or in error'] < \
                        self.settings.config_setting_flag['report_all_duplicate_ufi']:
                    self.settings.my_logger.error(
                        "{3} {0} is not unique or invalid. Found in feature class {1}, row {2}  This will be UPDATED to the GlobalID value".\
                            format( ttUFI, src_fc, rowcountii, fieldname))

                newUFI = rows[colnames.index('GlobalID')]
                if newUFI is not None and newUFI not in self.existing_unique_ids:
                    rows[colnames.index(fieldname)] = newUFI
                    result['ufi updated'] = True
                    result['ufi unique'] = False
                    self.existing_unique_ids.append(str(newUFI))
                else:
                    if self.settings.config_setting_flag['report_all_duplicate_ufi'] > 0 and \
                            result['ufi count of ufi not unique or in error'] < \
                            self.settings.config_setting_flag['report_all_duplicate_ufi']:
                        self.settings.my_logger.error("Failure to set it as GlobalID is null")

            elif self.settings.config_setting_flag['flag_ufi_is_not_unique_generate_new_unique_ufi'] == 3 and \
                'ObjectID' in colnames:
                # Look for GlobalID or ObjectID
                if self.settings.config_setting_flag['report_all_duplicate_ufi'] > 0 and \
                        result['ufi count of ufi not unique or in error'] < \
                        self.settings.config_setting_flag['report_all_duplicate_ufi']:
                    self.settings.my_logger.error(
                        "{3} {0} is not unique or invalid. Found in feature class {1}, row {2}  This will be UPDATED to the ObjectID value".\
                            format( ttUFI, src_fc, rowcountii, fieldname))
                newUFI = rows[colnames.index('ObjectID')]
                if newUFI is not None and newUFI not in self.existing_unique_ids:
                    rows[colnames.index(fieldname)] = newUFI
                    result['ufi updated'] = True
                    result['ufi unique'] = False
                    self.existing_unique_ids.append(str(newUFI))
                else:
                    if self.settings.config_setting_flag['report_all_duplicate_ufi'] > 0 and \
                            result['ufi count of ufi not unique or in error'] < \
                            self.settings.config_setting_flag['report_all_duplicate_ufi']:
                        self.settings.my_logger.error("Failure to set it as ObjectID is null")

            # An alterntative is to report all duplicates only
            elif self.settings.config_setting_flag['report_all_duplicate_ufi'] > 0:
                if result['ufi count of ufi not unique or in error'] < \
                        self.settings.config_setting_flag['report_all_duplicate_ufi']:
                    self.settings.my_logger.error(
                        "{3} {0} is not unique or invalid in feature class {1}, row {2}".format(
                            ttUFI, src_fc, rowcountii, fieldname))
                result['ufi unique'] = False
        else:
            self.existing_unique_ids.append(ttUFI)

        return result

    def check_ufi_check_initialize(self, result, update_source_where_incorrect):
        result['ufi performing check'] = True
        result['ufi checker update enabled'] = update_source_where_incorrect
        result['ufi unique'] = True
        result['ufi count of ufi not unique or in error'] = 0
        result['ufi updated'] = False
        result['ufi count of ufi updated'] = 0
        result['ufi updated on row'] = 0
        result['ufi issue description'] = ""
        result['ufi supress duplicate reports'] = True
        if len(self.existing_unique_ids)>0:
            del self.existing_unique_ids[:]           # re-initialize the existing unique ids

    def check_NullValues_check_initialize(self, result, update_source_where_incorrect):
        result['null values performing check'] = True
        result['null values update enabled'] = update_source_where_incorrect
        result['null values found'] = False
        result['null values counter'] = 0
        result['null values were updated'] = False
        result['null values count of updates made'] = 0
        result['null values issue description'] = ""
        result['null values update completed'] = ""
        result['null values row found'] = 0

    def check_illegal_values_check_initialize(self, result, update_source_where_incorrect):
        result['illegal values performing check'] = True
        result['illegal values update enabled'] = update_source_where_incorrect
        result['illegal values found'] = False
        result['illegal values found on row'] = 0
        result['illegal values count found'] = 0
        result['illegal values were updated'] = False
        result['illegal values count of updates made'] = 0
        result['illegal values issue description'] = ""

    def check_schema_check_initialize(self, result):
        result['schema performing check'] = True
        result['schema found to have issues'] = False
        result['schema count of issues found'] = 0
        result['schema issue description'] = ""

    def output_check_results_for_fc(self, result, fc):
        if 'schema issue description' in result and result['schema issue description'] <> "":
            self.settings.my_logger.error(result['schema issue description'])
            result['schema issue description'] = ""
        if 'illegal values issue description' in result and result['illegal values issue description'] <> "":
            self.settings.my_logger.error(result['illegal values issue description'])
            result['illegal values issue description'] = ""
        if 'null values issue description' in result and result['null values issue description'] <> "":
            self.settings.my_logger.error(result['null values issue description'])
            result['null values issue description'] = ""
        if 'ufi issue description' in result and result['ufi issue description'] <> "":
            self.settings.my_logger.error(result['ufi issue description'])
            result['ufi issue description'] = ""

    # For this part of the check, we really only need the UFI/URI field
    def check_ufi_adjust_fields(self, src_fields, final_fields):
        # Look for the unique identifier field, UFI and URI
        if 'UFI' in src_fields and 'UFI' not in final_fields:
            final_fields.append('UFI')
        elif 'ufi' in src_fields and 'ufi' not in final_fields:
            final_fields.append('ufi')
        elif 'URI' in src_fields and 'URI' not in final_fields:
            final_fields.append('URI')
        elif 'uri' in src_fields and 'uri' not in final_fields:
            final_fields.append('uri')
        else:
            pass   # No URI or UFI in the source field set
        return

    # For this check, we need all fields except geometry or Esri specific identifiers
    def check_attribute_values_adjust_fields(self, src_fields, final_fields):
        # We need all fields except the shape or globalID fields
        possShape = ['shape@', 'shape_', 'shape@length', 'shape_length', 'shape@area',
                     'shape@area', 'globalid', 'objectid']
        for ff in src_fields:
            if ff not in final_fields and ff.lower() not in possShape:
                final_fields.append(ff)


    def check_ufi_verify_fix_unique(self, result, rowD, rows, colnames, src_fc, rowcountii):
        # Get the correct field name for the UFI/URI check
        if 'UFI' in rowD:
            ffname = 'UFI'
        elif 'ufi' in rowD:
            ffname = 'ufi'           
        elif 'URI' in rowD:
            ffname = 'URI'
        elif 'uri' in rowD:
            ffname = 'uri'

        # Get the current URI/UFI value for this instance
        ttUFI = str(rowD[ffname])  # current value
        result['ufi updated'] = False
        # Check this value against the current set of unique values and get an update if needed
        uniquechmade = self.check_ufi_work_ufi_uri_existing(result, ttUFI, rows, colnames, ffname, src_fc, rowcountii)

        for k, v in uniquechmade.iteritems():
            result[k] = v

        if result['ufi unique'] == False:
            result['ufi count of ufi not unique or in error']  += 1
            result['ufi issue description'] = "{0} (Feature class {1} has invalid or no UFI/URI at row {2})".\
                    format(result['ufi issue description'], src_fc, rowcountii)

        if result['ufi updated'] == True:
            result['ufi updated on row'] = rowcountii
            result['ufi count of ufi updated'] += 1

        # This bit here will try to restrict output so we don't have thousands of duplicative reports.
        # It is good enough to report only a few of these per feature class
        if ((result['ufi unique'] == False or result['ufi updated']) == True and \
            result['ufi supress duplicate reports'] == False) or \
           ((result['ufi unique'] == False or result['ufi updated']) == True and \
           (result['ufi supress duplicate reports'] == True and \
            result['ufi count of ufi not unique or in error'] < \
            self.settings.config_setting_flag['report_all_duplicate_ufi'])):
            self.settings.results_xl.output_to_excel('ufi', 'ufi ', result)
        return

    def check_NullValues_verify_and_fix(self, result, row_dict, rows, colnames, src_fc, rowcountii):
        # Check for null values in the row
        if 'F_CODE' in row_dict:
            fcode = row_dict['F_CODE']
        else:
            fcode = None
        result['null values were updated'] = False
        for fldname in row_dict:
            rowval = rows[colnames.index(fldname)]
            if rowval == None or rowval == "":
                result['null values found'] = True
                result['null values counter'] += 1
                result['null values row found'] = rowcountii
                result['null values in feature class'] = src_fc
                result['null values in feature with code'] = fcode
                result['null values in feature field'] = fldname



                if result['null values update enabled'] == 1:  # Update the null value, put in the default value
                    result['null values update completed'] = ""    # Reset
                    newvalue = \
                        self.field_set_to_default(fcode, fldname, src_fc, result, 'null values update completed')
                    rows[colnames.index(fldname)] = newvalue
                    result['null values updated to value'] = newvalue
                    result['null values were updated'] = True
                    result['null values count of updates made'] += 1
 
                null_last = "(Feature class {0} F_CODE {1} Field {2} Value is NULL".\
                    format(src_fc, fcode, fldname)
                if null_last in result['null values issue description']:
                    result['null values issue description'] = \
                        "{0} +".format(result['null values issue description'])
                else:
                    result['null values issue description'] = \
                        "{0} (Feature class {1} F_CODE {2} Field {3} Value {4} illegal at row {5})".\
                        format(result['null values issue description'], src_fc, fcode, fldname, rowval, rowcountii)
                    self.settings.results_xl.output_to_excel('nulls', 'null values ', result)

    def check_for_illegal_attribute_values(self, result, row_dict, rows, colnames, src_fc, rowcountii):
        # Check for illegal attribute values in the row
        if 'F_CODE' in row_dict:
            fcode = row_dict['F_CODE']
        else:
            fcode = None
        
        result['illegal values were updated'] = False
        for fldname in row_dict:
            rowval = rows[colnames.index(fldname)]
            valillegal, valfixed, newval = self.is_attribute_value_illegal(rowval, fldname, fcode, src_fc)
            
            if valillegal:
                result['illegal values found'] = True
                result['illegal values found on row'] = rowcountii
                result['illegal values count found'] += 1
                result['illegal values found in feature class'] = src_fc
                result['illegal values found on feature having feature code'] = fcode
                result['illegal values found in field'] = fldname
                result['illegal values identified'] = rowval
                result['illegal values were updated to value'] = ""

                if result['illegal values update enabled'] == 1 and valfixed:  # Update the value
                    rows[colnames.index(fldname)] = newval
                    result['illegal values were updated to value'] = newval
                    result['illegal values were updated'] = True
                    result['illegal values count of updates made'] += 1

                illegval_last = \
                    "(Feature class {0} F_CODE {1} Field {2} Value {3} illegal at row".format(
                        src_fc, fcode, fldname, rowval)
                if illegval_last in result['illegal values issue description']:
                    result['illegal values issue description'] = \
                        "{0} +".format(result['illegal values issue description'])
                else:
                    result['illegal values issue description'] = \
                        "{0} (Feature class {1} F_CODE {2} Field {3} Value {4} illegal at row {5})".\
                        format(result['illegal values issue description'], src_fc, fcode, fldname, rowval, rowcountii)
                    self.settings.results_xl.output_to_excel('values', 'illegal values ', result)

    def check_schema_verify(self, result, row_dict, rows, colnames, src_fc, rowcountii):
        # verify that this feature class or table is in schema
        # verify that all fields are in schema
        if src_fc not in self.field_information[self.db_version]:
            result['schema found to have issues'] = True
            result['schema count of issues found'] += 1
            result['schema issue found in feature class'] = src_fc
            result['schema issue found with field'] = "NA"
            result['schema issue description'] = "{0} (Feature class {1} not valid in schema)".\
                format(result['schema issue description'], src_fc)
        else:
            for fldname in row_dict:
                if fldname.lower() not in ['fcsubtype', 'shape@', 'shape@length',
                                           'shape@area', 'globalid', 'objectid']:   # Esri specific fields
                    # So the fc or table is in the schema, now verify that all fields are
                    if fldname not in self.field_information[self.db_version][src_fc]:
                        result['schema found to have issues'] = True
                        result['schema count of issues found'] += 1
                        result['schema issue found in feature class'] = src_fc
                        result['schema issue found with field'] = fldname
                        result['schema issue description'] = "{0} (Feature Class {1} Field {2} not valid in schema)".\
                            format(result['schema issue description'], src_fc, fldname) 
                        self.settings.results_xl.output_to_excel('schema', 'schema ', result)

    # This function finds all existing unique identifiers in the db, returns False if identifiers were not unique

    def perform_gdb_checks(self):
    
        result = dict()

        fc_examined = set()        # to keep track of what we have already looked at
        mytimer = MyTimers(self.settings.maximum_records['processing_indicator_n_seconds'])

        # Get the list of feature classes and tables in this current geodatabase
        listoffcandtbl = self.geodb_tables + self.geodb_feature_classes[self.geodb_dataset]

        # Initialize checker portions for specific capabilities
        if self.check_schema:
            self.settings.my_logger.info("Checker - Schema check enabled")
            self.check_schema_check_initialize(result)

        if self.check_ufi:
            self.settings.my_logger.info("Checker - Unique Identifier (UFI/URI) check enabled")
            self.check_ufi_check_initialize(
                result, self.settings.config_setting_flag['flag_ufi_is_not_unique_generate_new_unique_ufi'])

        if self.check_NullValues:
            self.settings.my_logger.info("Checker - NULL value check enabled")
            self.check_NullValues_check_initialize(
                result, self.settings.flag_data_is_null_set_to_default.get())

        if self.check_illegal_values:
            self.settings.my_logger.info("Checker - Illegal value check enabled")
            self.check_illegal_values_check_initialize(
                result, self.settings.flag_data_is_invalid_update_if_we_can.get())

        update_source_where_incorrect = False
        if self.settings.config_setting_flag['flag_ufi_is_not_unique_generate_new_unique_ufi'] > 0 or \
            self.settings.flag_data_is_null_set_to_default.get() or \
            self.settings.flag_data_is_invalid_update_if_we_can.get():
            update_source_where_incorrect = True

        for src_fc in listoffcandtbl:
            rowcountii = 0
            mytimer.set_initial_time('checker')
            # The specification may include feature classes that do not exist in this database
            if (src_fc in self.saved_fc_datasets or src_fc in self.geodb_tables):
                # check to see if we have already processed this feature class
                if src_fc in fc_examined:
                    continue

                fc_examined.add(src_fc)   # add the feature class name to the set

                # Get the fields from the feature class or table in this geodatabase
                src_fc_path = self.get_root_dataset_for_fc(src_fc) + "/" + src_fc
                src_fields = self.get_fields_from_featureclass(src_fc)
                final_fields = list()      # Start with no fields in our initial examination

                # Adjust fields to get for each capability
                if self.check_ufi:
                    self.check_ufi_adjust_fields(src_fields, final_fields)

                if self.check_NullValues or self.check_illegal_values or self.check_schema:
                    self.check_attribute_values_adjust_fields(src_fields, final_fields)


                if len(final_fields) > 0:
            
                    if update_source_where_incorrect == 1:
                        cursor = arcpy.da.UpdateCursor(src_fc_path, final_fields, None)
                        self.settings.my_logger.info("\tChecking with update {0}".format(src_fc))
                    else:
                        cursor = arcpy.da.SearchCursor(src_fc_path, final_fields, None)
                        self.settings.my_logger.info("\tChecking {0}".format(src_fc))

                    rowcountii = 0
                    colnames = cursor.fields    
                    for rows in cursor:
                        rowcountii += 1

                        # output something periodically so the operator can tell we are still working on it
                        if mytimer.check_for_elapsed_time_gt_set_interval('checker'):
                            self.settings.my_logger.info("\t\t...processing... current count: {0}".format(rowcountii))


                        row_dict = dict(zip(colnames, rows))
                        # put the data into a dictionary for easier review

                        change_made = False
                        # We probably should do schema checking first because if
                        # there are errors here they may affect the
                        # rest of the checks
                        if self.check_schema:
                            self.check_schema_verify(
                                result, row_dict, rows, colnames, src_fc, rowcountii)
                
                        # Perform the checker capability at the feature class or table level
                        if self.check_ufi:
                            self.check_ufi_verify_fix_unique(
                                result, row_dict, rows, colnames, src_fc, rowcountii)
                            if result['ufi updated'] == True:
                                change_made = True

                        if self.check_NullValues:
                            self.check_NullValues_verify_and_fix(
                                result, row_dict, rows, colnames, src_fc, rowcountii)
                            if result['null values were updated'] == True:
                                change_made = True

                        if self.check_illegal_values:
                            self.check_for_illegal_attribute_values(
                                result, row_dict, rows, colnames, src_fc, rowcountii)
                            if result['illegal values were updated'] == True:
                                change_made = True

                        if update_source_where_incorrect == 1 and change_made:
                            cursor.updateRow(rows)         ### UPDATING THE SOURCE HERE

                        if self.settings.config_setting_flag['break_after_n_instances'] > 0:
                            if rowcountii > self.settings.config_setting_flag['break_after_n_instances']:
                                break
                    # end for rows in cursor
                    del cursor
                # end if there are fields to check
            # end if this feature class is in the geodatabase
            if self.check_schema or self.check_ufi or self.check_NullValues or self.check_illegal_values:
                self.output_check_results_for_fc(result, src_fc)

            self.settings.results_xl.done_save()
            
            mytimer.set_end_time('checker')
        # end for src_fc in list of feature classes or tables
        return result
                        
    ## These unique values are absolutely critical, so here is the alternative.
    ## If there is a GlobalID, hopefully that is unique and we can use that instead.
    ## If there is not a GlobalID, I will try and grab the UFI and append the ObjectID and FCSubtype
    ## CRITICAL FUNCTION: get_row_unique_id
    def get_row_unique_id(self, settings, row, read_only = True):
        if settings.config_setting_flag['flag_set_unique_identifiers_fresh'] == 1:
            return makerandomufi()
        elif 'UFI' in row:
            return str(row['UFI'])
        elif 'URI' in row:
            return str(row['URI'])
        else:
            if 'Unique identifer created' in settings.report_values:
                settings.report_values['Unique identifer created'] += 1
            else:
                settings.report_values['Unique identifer created'] = 1
            if settings.report_values['Unique identifer created'] < settings.maximum_records[
                'maximum_reported_setting_of_unique_identifier']:
                settings.my_logger.warn("No unique identifier on entries found, creating one")
            return self.makerandomufi()

    # This function determins if row UFI or URI is already in the set of existing unique identifiers,
    # meaning it has already been recorded. Requires list of exisitng_unique_ids
    def is_this_row_already_in_target(self, row, my_logger, use_global_id_instead = False):
        # First perform some quick checks for valid parameters, and existing data
        if isinstance(row, dict) == False:
            my_logger.error( "The function is_this_row_already_in_target requires a dictionary as input" )
            return False
        if self.settings.config_setting_flag['flag_expect_no_ufi_conflict'] == 1:
            return False
        if len(self.existing_unique_ids) == 0:
            return False
        # With the quick and easy checks out of the way, now check to see if this UFI is in the row and
        # if the UFI value is in the existing unique identifiers list. Then do the same for URI.
        if self.get_row_unique_id(self.settings, row) in self.existing_unique_ids: ## YIKE HERE
            return True
        else:
            return False

    # This will open the geodatabase, and get the datasets, tables, and feature classes
    def fromGEODBGetDatasetsTablesandFC(self):

        if arcpy.Exists(self.geodb_pathname):
            self.mode = "NA"
            self.type_of_file = "NA"

            desc = arcpy.Describe(self.geodb_pathname)
            try:    
                self.type_of_file = desc.datasetType
            except:
                try:
                    self.type_of_file = desc.workspaceType
                except:
                    pass

            # Be sure the file type is something we can work with
            if self.type_of_file not in self.posstypes:
                self.settings.my_logger.error(
                    "Unable open geodatabase {0} or it is not a type we can operate on".format(self.geodb_pathname))
                return

            self.mode = "EXISTS"

            # set the workspace environment
            arcpy.env.workspace = self.geodb_pathname
            self.saved_fc_datasets.clear()
            self.geodb_feature_classes.clear()
            del self.geodb_tables[:]

            # First, get all the stand alone tables, feature classes and rasters - stuff not in datasets
            dataList = None
            tlist = arcpy.ListTables()
            fclist = arcpy.ListFeatureClasses()
            
            # Get tables that are not under a dataset
            if tlist is not None and fclist is not None:
                dataList = tlist + fclist

            if dataList is not None and len(dataList) > 0:
                for itemName in dataList:
                    self.geodb_tables.append( itemName )
                    self.saved_fc_datasets[itemName] = None

            # Ask ArcGIS for the dataset names
            dsetlist = arcpy.ListDatasets()
            self.db_count_of_datasets = len(dsetlist)
            
            # Next, for feature datasets get all of the datasets and featureclasses
            # from the list and add them to the master list.
            if self.db_count_of_datasets > 0:

                # for each dataset, found, get the names of the feature classes in the dataset
                for ddset in dsetlist:

                    self.geodb_datasets.append (ddset)
                    self.geodb_feature_classes[ddset] = list()


                    # Now we know a dataset name, append the name to the database and get the feature classes
                    # This is the fun way that ArcGIS must use to get the data
                    arcpy.env.workspace = os.path.join(self.geodb_pathname, ddset)
            
                    dataList = None
                    nfclist = arcpy.ListFeatureClasses()
                    ndsetlist = arcpy.ListDatasets()
                    dataList = nfclist + ndsetlist
                    # If the code fails on the above line, it is likely due to a space in the file geodatabase name.
                    if dataList is not None and len(dataList) > 0:
                        for fcName in dataList:
                            self.geodb_feature_classes[ddset].append (fcName)
                            self.saved_fc_datasets[fcName] = ddset
                            # save to a dictionary of feature class names, each specifying dataset
                
                # Now we make a likely assumption, typically there is only one dataset and multiple
                # feature classes under that dataset. This would be the 'default' dataset and we need to be sure (later)
                # that the operator set the correct dataset.
                # If there is only one dataset, we should have the correct name, if we don't fix it.
                if self.geodb_dataset  == "" or self.geodb_dataset not in dsetlist:
                    if len(dsetlist) > 0 and self.geodb_dataset <> dsetlist[0]:
                        self.geodb_dataset = dsetlist[0]

            self.mode = "OPENED"
            arcpy.env.workspace = self.geodb_pathname
        else:
            self.settings.my_logger.error("Unable open geodatabase {0}".format(self.geodb_pathname))

    def correct_shape_field_name_to_compensate_for_esri_strangeness(self, ff):   # FF is a field name
        if ff.lower() in self.fixshapenames:
            return self.fixshapenames[ff.lower()]
        else:
            return ff

    # Use our saved dataset and feature class information to identify
    # the place to start, then call ArcGIS ListFields to get the fields
    # for this feature class.
    def get_fields_from_featureclass(self, featureclass):
        return self.get_db_actual_field_list(featureclass)

    def get_geometrytype(self, featureclass):
        # arcpy.env.workspace = self.geodb_pathname
        fields = list()
        ddset = self.get_root_dataset_for_fc(featureclass)  # If this is a table, it'll return the table root
        if ddset is not None:
            arcpy.env.workspace = os.path.join(self.geodb_pathname, ddset)
            fclassinfo = arcpy.Describe(featureclass)
            return fclassinfo.shapeType           # to return Polygon, Polyline, Point, Multipoint, MultiPatch
        return None

    def get_spatial_reference_frame(self, featureclass):
        # arcpy.env.workspace = self.geodb_pathname
        fields = list()
        ddset = self.get_root_dataset_for_fc(featureclass)
        if ddset is not None:
            arcpy.env.workspace = os.path.join(self.geodb_pathname, ddset)
            fclassinfo = arcpy.Describe(featureclass)
            return fclassinfo.spatialReference.exportToString()
            # to return Polygon, Polyline, Point, Multipoint, MultiPatch
        else:
            if self.type_of_file == 'FeatureDataset':
                # This could be one without a dataset and might have spatial ref directly
                 fclassinfo = arcpy.Describe(self.geodb_pathname)
                 return fclassinfo.spatialReference.exportToString()
                 # to return Polygon, Polyline, Point, Multipoint, MultiPatch
        return None

    def get_domain_names(self):
        arcpy.env.workspace = self.geodb_pathname #  self.get_default_root_for_fc()
        self.domainList = arcpy.da.ListDomains(self.geodb_pathname)

    #def get_domain_descriptions(self,  dataset, domainname):
    #    arcpy.env.workspace = self.geodb_pathname
    #    # Get the dictionary of codes and descriptions for the Floodplain_Rules domain.
    #    desc_lu = [d.codedValues for d in arcpy.da.ListDomains(self.geodb_pathname) if d.name == domainname][0]
    #    with arcpy.da.SearchCursor(os.path.join(gdb, dataset), 'RuleID') as cur:
    #        for row in cur:
    #            print(desc_lu[row[0]])

    def get_all_domains_as_dictionary(self):
        arcpy.env.workspace = self.geodb_pathname
        if self.domainList == None:
            self.domainList = arcpy.da.ListDomains(self.geodb_pathname)
        # domains = desc.domains
        domainnames = list()
        domainout = dict()
        for index, element in enumerate(self.domainList):
            domainName = unicode(element.name)
            domainnames.append(domainName)
            domainout[domainName] = index
        self.domainDict = domainout
        self.domainNameList = domainnames

    def get_domains_as_dictionary(self, dataset, domainname):
        arcpy.env.workspace = self.geodb_pathname
        if self.domainList == None:
            self.domainList = arcpy.da.ListDomains(self.geodb_pathname)
        # domains = desc.domains
        domainout = dict()
        for index, element in enumerate(self.domainList):
            domainName = unicode(element.name)
            domainout[domainName] = index
        self.domainDict = domainout

    def get_coded_values(self, indomainname):
        outvalue = unicode()
        outvalue = ""
        if indomainname in self.domainDict:
            domainIndex = self.domainDict[indomainname]
            if domainIndex > len(self.domainList):
                return outvalue
            domainType = self.domainList[long(domainIndex)].domainType
    
            if domainType == 'CodedValue':
                coded_values = self.domainList[long(domainIndex)].codedValues
                for val, desc in coded_values.iteritems():
                    if outvalue == "":
                        outvalue = u"{0}".format(val)
                    else:
                        outvalue = u"{0}, {1}".format(outvalue, val)
            elif domainType == 'Range':
                outvalue = u"{0}, {1}".format(self.domainList[long(domainIndex)].range[0], 
                                              self.domainList[long(domainIndex)].range[1])
        return outvalue

    def get_coded_values_and_descriptions(self, indomainname):

        outvalue = unicode()
        outvalue = ""
        outdesc = ""
        if indomainname in self.domainDict:
            domainIndex = self.domainDict[indomainname]
            if domainIndex > len(self.domainList):
                return (outvalue, outdesc)
            domainType = self.domainList[long(domainIndex)].domainType
    
            if domainType == 'CodedValue':
                coded_values = self.domainList[long(domainIndex)].codedValues
                for val, desc in coded_values.iteritems():
                    if outvalue == "":
                        outvalue = u"{0}".format(val)
                        outdesc = u"{0}".format(desc)
                    else:
                        outvalue = u"{0}, {1}".format(outvalue, val)
                        outdesc = u"{0}, {1}".format(outdesc, desc)
            elif domainType == 'Range':
                outvalue = u"{0}, {1}".format(self.domainList[long(domainIndex)].range[0], 
                                              self.domainList[long(domainIndex)].range[1])
                outdesc = u"{0}, {1}".format("Range Minimum", "Range Maximum")
        return (outvalue, outdesc)

    # Is this feature class specified in the field_information 'schema' definition dictionary?
    def verify_field_information_fclass(self, fclass):
        return True if (len(self.field_information) > 0 and self.db_version in self.field_information and \
            fclass in self.field_information[self.db_version]) else False

    # Is this geodatabase represented by a field_information 'schema' definition dictionary?
    def verify_field_information(self):
        return True if (len(self.field_information) > 0 and self.db_version in self.field_information) else False

    # Is the field in the list of possible enumerated fields?
    def is_field_in_list_of_possible_enumerated(self, fieldn):
        return True if fieldn.lower() != "f_code" and self.verify_field_information() and fieldn in \
            self.field_information[self.db_version]['SummaryDomainValues'] else False

    # Is the enumerated value index in the list of possible enumerated fields values, ANYWHERE in the geodatabase?
    def is_index_valid_anywhere(self, fieldn, enumindex):
        if self.verify_field_information() and self.is_field_in_list_of_possible_enumerated( fieldn):
            return enumindex in self.field_information[self.db_version]['SummaryDomainValues'][fieldn]
        else:
            return False
    
    # For a given field and enumerated domain value index, get the text name for this value
    def get_enum_text_for_index(self, fieldn, enumindex):
        if self.is_field_in_list_of_possible_enumerated(fieldn):
            if enumindex in self.field_information[self.db_version]['SummaryDomainValues'][fieldn]:
                return self.field_information[self.db_version]['SummaryDomainValues'][fieldn][enumindex]
        return None

    # Is this field in the feature class according to the schema definition information?
    def is_field_in_fclass(self, fclass, fieldn):
        return True if self.verify_field_information_fclass(fclass) and \
                       fieldn in self.field_information[self.db_version][fclass] else False

    def get_field_list_from_fclass_schemadef(self, fclass):
        fldlist = list()
        if self.verify_field_information_fclass( fclass):
            fldlist = [fieldn for fieldn in self.field_information[
                self.db_version][fclass] if fieldn != 'EnumeratedIndexValues']
        return fldlist

    # Is this field an enumerated domain value in this feature class?
    def is_enumeration(self, fclass, fieldn):
        return True if fieldn.lower() not in self.common_fields_geom_and_id_and_code and \
            self.get_field_datatype(fclass, fieldn) == 'enumeration' else False

    # Get the field data type as per the schema definition dictionary
    def get_field_datatype(self, fclass, fieldn):
        if self.is_field_in_fclass( fclass, fieldn):
            if fieldn in self.field_information[self.db_version][fclass]:
                try:
                    return self.field_information[self.db_version][fclass][fieldn][0].lower()   # first element of list
                except:
                    self.settings.my_logger.error(
                        "Failure to get field datatype information for field: {0} feature class: {1}, no element 0 or no lower() db: {2}.".format(fieldn, fclass, self.db_version))
                    return None
        return None
    
    # Get the field fcode list - valid feature codes for this field and feature class, from the schema definition dictionary
    def get_field_valid_fcode_list(self, fclass, fieldn):
        if self.is_field_in_fclass( fclass, fieldn):
            if fieldn in self.field_information[self.db_version][fclass]:
                try:
                    return self.getvalidcodesforfeatureclassfield(fclass, fieldn)   # second element of list
                except:
                    self.settings.my_logger.error(
                        "Failure to get field fcode list information for field: {0} " \
                        "feature class: {1}, no element 0 or no lower() db: {2}.".format(
                            fieldn, fclass, self.db_version))
                    return None
        return None

    # Test if the field is applicable to this feature class
    # and feature code according to the schema definition dictionary
    def is_field_valid_for_fclass_fcode(self, fclass, fcode, fieldn):
        if self.is_field_in_fclass( fclass, fieldn):
            return fcode in self.getvalidcodesforfeatureclassfield(fclass, fieldn)
        else:
            return False

    # Is the specified index value valid for the field
    # (if it is an enumerated value) in this feature class and for this feature code
    def is_index_valid_for_fclass_fcode(self, fclass, fcode, fieldn, enumindex):
        if self.is_field_valid_for_fclass_fcode( fclass, fcode, fieldn) and self.is_enumeration( fclass, fieldn):
            return enumindex in self.field_information[self.db_version][fclass]['EnumeratedIndexValues'][fieldn][fcode]
        else:
            return False

    # Given an input string, search for a possible enumerated domain value
    def search_for_possible_enum_value_given_string(self, fieldn, enumstring):
        if self.verify_field_information():
            if enumstring is not None and fieldn in self.field_information[self.db_version]['SummaryDomainValues']:
                for enumII, enumVV in self.field_information[self.db_version]['SummaryDomainValues'][fieldn].items():
                    if enumVV.lower().strip() == enumstring.lower().strip():
                        # look for a match in the enum values vs the incoming string (all lowercase)
                        return ((enumII, enumVV))
        return None

    # Given an input enumerated value index, look for the possible value
    def search_for_possible_enum_value_given_index(self, fieldn, enumindex):
        if self.verify_field_information():
            if enumindex is not None and fieldn in self.field_information[self.db_version]['SummaryDomainValues']:
                if enumindex in self.field_information[self.db_version]['SummaryDomainValues'][fieldn]:
                    enumVV = self.field_information[self.db_version]['SummaryDomainValues'][fieldn][enumindex]
                    return ((enumindex, enumVV))
        return None

    # Given both an index and a string, look for the possible enumerated domain value for this field
    # If the string is None, look for the value given the index
    def search_for_possible_enumeration_value(self, fieldn, enumindex, enumstring):
        if enumindex is None:
            return self.search_for_possible_enum_value_given_string( fieldn, enumstring)
        else:
            return self.search_for_possible_enum_value_given_index( fieldn, enumindex)

    def search_for_possible_enumeration_value_return_if_valid(self, fclass, fcode, fieldn, enumindex, enumstring):
        if enumindex is None:
            possValue =  self.search_for_possible_enum_value_given_string(fieldn, enumstring)
        else:
            possValue  = self.search_for_possible_enum_value_given_index(fieldn, enumindex)
        if possValue is not None:
            if self.verify_field_information_fclass( fclass) and \
                    possValue[0] in self.field_information[
                        self.db_version][fclass]['EnumeratedIndexValues'][fieldn][fcode]:
                return possValue
        return None

    def get_schema_defined_enumerated_value_possible_list(self, fclass, fcode, fieldn):
        if self.verify_field_information_fclass( fclass) and \
                fcode in self.field_information[self.db_version][fclass]['EnumeratedIndexValues'][fieldn]:
            return self.field_information[self.db_version][fclass]['EnumeratedIndexValues'][fieldn][fcode]
        return None

    def get_schema_defined_default_value(self, fclass, fcode, fieldn):
        if self.verify_field_information_fclass( fclass) and \
                fcode in self.field_information[self.db_version]['DefaultValues']:
            if fieldn in self.field_information[self.db_version]['DefaultValues'][fcode]:
                return self.field_information[self.db_version]['DefaultValues'][fcode][fieldn]
        return None

    def get_fcode_subtype_mapping_for_all(self):
        fcodemapping = dict()
        for lvls in self.field_information[self.db_version]['Levels']:
            for tbln in self.field_information[self.db_version]['Levels'][lvls]:
                fcode, subtype, fclass = self.field_information[self.db_version]['Levels'][lvls][tbln]
                if fcode not in fcodemapping:
                    fcodemapping[fcode] = int(subtype)
        return fcodemapping

    def get_fcode_subtype_mapping_for_level_and_featureclass(self, levelin, featureclass):
        fcodemapping = dict()
        for tbln in self.field_information[self.db_version]['Levels'][levelin]:
            fcode, subtype, fclass = self.field_information[self.db_version]['Levels'][levelin][tbln]
            if fcode not in fcodemapping and featureclass == fclass:
                fcodemapping[fcode] = int(subtype)
        return fcodemapping

    # Is the specified index valid for the field in both this database and in a target database?
    def is_index_valid_for_fclass_fcode_in_both_src_and_targ(self, sourcedb, targetdb,
                                                             fclass, fcode, fieldn, enumindex):
        iiTarg = targetdb.is_index_valid_for_fclass_fcode(fclass, fcode, fieldn, enumindex)
        iiSour = sourcedb.is_index_valid_for_fclass_fcode( fclass, fcode, fieldn, enumindex)
        if iiTarg == iiSour and iiTarg is not None:
            return True
        else:
            return False

    def get_featureclass_list_from_schema_info_def(self):
        return  list(set([fc for fc in self.field_information[self.db_version] if fc.lower() \
                 not in ['defaultvalues', 'levels', "summarydomainvalues", 'interval_fields']]))

    #for domain in domains:
    #    print 'Exporting %s CV to table in %s' % (domain, gdb)
    #    table = os.path.join(gdb, domain)
    #    arcpy.DomainToTable_management(gdb, domain, table,
    #        'field','description', '#')
