﻿"""
 ************************************************************************************************
 **                                                                                            **
 **  File: PrimaryTranslationFunctions.py                                                      **
 **  Copyright (c) December 2015 and September, 2018 by Leidos                                 **
 **  Produced under U.S. Government contract W5J9CQ-12-D-0004  Task Order: 053 and 073         **
 **  As per DFARS, the U.S. Government is granted unlimited rights to use this work            **
 **  in any manner and for any purpose.                                                        **
 **                                                                                            **
 **  Summary:   This file is the main entry point for translating the physical database
 **             content to or from GGDM 3.0.                                                   **
 **
 **                                                                                            **
 **  Developed by Leidos, Inc., High Resolution ISR Systems Division, Geospatial Programs      **
 **                                                                                            **
 **  Development History:                                                                      **
 **  December 2015 / Annette.J.Filer@Leidos.com                                                **
 **  January 2016 / continued development on primary capabilities, added logging
 **  February 2016 / Added the reverse direction, from 3.0 to 6.1 translation
 **  March 2016 / Update translation to check for allowed value that is used in the source     **
 **                 but not allowed in the target and automatically put any content no longer  **
 **                 allowed in the target into the OTH Other Enumerants attribute.             **
 **  March 2016 / Added comparison function
 **  9/27/2018      Significant restructure and updates
 **                                                                                            **
 **                                                                                            **
 ************************************************************************************************
"""

import sys
import os
import copy

import arcpy
import uuid
import glob

from TranslateConfig import *
import parseOTH
from FieldTransformerFunctions import *
from StandardUtilities import *

common_fields_geom_and_id = ['shape@', 'shape@length', 'shape@area', 'fcsubtype', 'objectid', 'globalid']
common_fields_geom_only = ['shape@', 'shape@length', 'shape@area']


# This function provides for the initialization of transformer function report statistics
# Each dictionary key is provided a count value. Upon transformer operation of this type, the count is incremented.
# The values are reset and reported per feature class table.
def clear_transformer_functions_report(settings):
    # fclass_static.clear()
    # fclass_static['original_F_CODE'] = 'NA'
    # fclass_static['translated_F_CODE'] = 'NA'

    settings.report_values.clear()
    settings.report_values['Derived Features - Instances Appended'] = 0
    settings.report_values['Derived Features - Preparing Instance'] = 0
    settings.report_values['Translate and Move - Instances Moved'] = 0
    settings.report_values['Translate Processing - Instances Processed'] = 0
    settings.report_values[
        'Transform no target: Accuracy category (ZI010_ACC) present in source but we have no defined transform content'] = 0
    settings.report_values['Transform started: Aerodrome Surface (ASU) processing'] = 0
    settings.report_values['Transform processing: Aerodrome Surface illegal value 19 = water may be saved to OTH'] = 0
    settings.report_values['OTH Processing - Value is already in OTH so it is not being appended'] = 0
    settings.report_values['MEM Processing, Append value to MEM'] = 0
    settings.report_values['OTH Processing - Appending a value into OTH'] = 0
    settings.report_values['Dataset - Associated dataset was identified for feature instance'] = 0
    settings.report_values['Entity Collection - Associated entity collection was identified for feature'] = 0
    settings.report_values['Restriction Information - Associated restriction information found for feature'] = 0
    settings.report_values['Transform processing: Bridge Opening Type derived from Bridge Structure Type'] = 0
    settings.report_values['Transform processing: Bridge Structure Type derived from Bridge Opening Type'] = 0
    settings.report_values['Transform processing: Closed in Winter derived from Route Weather Restriction'] = 0
    settings.report_values['Transform processing: Control point type (CPA) derived from survey point (SUY)'] = 0
    settings.report_values['Transform processing: Trying to get country code (GENC) from the dataset GE4'] = 0
    settings.report_values['Transform unit conversion: Decimetre converted to metre'] = 0
    settings.report_values['Default - Field set to default value'] = 0
    settings.report_values['Default - Null data detected, assuming default value'] = 0
    settings.report_values['Default - Tried to set a field to default that is not allowed (shape, f_code)'] = 0
    settings.report_values['Density Equivalent Scale from Equivalent Scale Category'] = 0
    settings.report_values['Derive Dam Function from Dam Type'] = 0
    settings.report_values['Derive Dam Type from Dam Function'] = 0
    settings.report_values['EQS can be derived from EQC but in a different feature'] = 0
    settings.report_values['Elevation Surface Category (ESC)'] = 0
    settings.report_values['Equivalent Scale Category from Density Equivalent Scale'] = 0
    settings.report_values['Error transforming OTH, OTH not a field'] = 0
    settings.report_values['Extended shape to include source feature instance shape'] = 0
    settings.report_values['Extraction Mine Strip-mine converted to Area strip mine'] = 0
    settings.report_values['Facility Operational Status (COS) processing'] = 0
    settings.report_values['Farming Method derived from Permanent Irrigation'] = 0
    settings.report_values['Feature code (F_CODE) changed'] = 0
    settings.report_values['Feature Function place of worship detected from Islamic Prayer Hall'] = 0
    settings.report_values['Feature Function value not allowed on Fortified Building and saved to OTH'] = 0
    settings.report_values['Feature Function value implies religious setting (FFN 932 to ZI037_REL)'] = 0
    settings.report_values['GENC Country derived from Dataset IC2'] = 0
    settings.report_values['GENC Country derived from IC2 Country'] = 0
    settings.report_values['Geodatabase version set to specified value'] = 0
    settings.report_values['Groundwater Potential Scale value conversion'] = 0
    settings.report_values['Hydrologic Persistence (HYP) Ephemeral - Intermittent'] = 0
    settings.report_values['IC2 Country derived from GENC'] = 0
    settings.report_values['Interval - Convert interval to single value'] = 0
    settings.report_values['Interval - Convert single value to interval'] = 0
    settings.report_values['Isolated derived from landmark'] = 0
    settings.report_values['Landmark derived from isolated'] = 0
    settings.report_values['Located Underground derived from TRS'] = 0
    settings.report_values['Location Category derived from TRS'] = 0
    settings.report_values['Medical facility transform'] = 0
    settings.report_values['Medical specialty transform'] = 0
    settings.report_values['MEM processing - set MEM to default'] = 0
    settings.report_values['MEM processing - Value saved to MEM'] = 0
    settings.report_values['Metadata from related entity collection'] = 0
    settings.report_values['Transform unit conversion: Metre converted to decimetre'] = 0
    settings.report_values[
        'Dataset - WARNING Multiple associated datasets were identified for one feature instance'] = 0
    settings.report_values['Multiple associated entity collections identified for one feature instance'] = 0
    settings.report_values['Dataset - No associated dataset identified for feature instance'] = 0
    settings.report_values['Entity Collection - No associated entity collection identified for feature'] = 0
    settings.report_values['Restriction Information - No associated restriction information identified for feature'] = 0
    settings.report_values['Nominal range data type change'] = 0
    settings.report_values['OTH content derived and saved directly to target field'] = 0
    settings.report_values['OTH Processing - Setting OTH to default value'] = 0
    settings.report_values['OTH Processing - Saving a value to OTH'] = 0
    settings.report_values['Other used to derive Irrigation'] = 0
    settings.report_values['Permanent Irrigation derived from Farming Method'] = 0
    settings.report_values['Physical Product (PPO) Bio-Solids converted to Sewage'] = 0
    settings.report_values['placeholder transform - source value retained'] = 0
    settings.report_values['placeholder transform returned None'] = 0
    settings.report_values['Pump type detected and may impact pump present'] = 0
    settings.report_values['Pump type may be derived from pump present'] = 0
    settings.report_values['Religious Designation (ZI037_REL) Adjustment from FFN Islamic Prayer Hall'] = 0
    settings.report_values['Restriction info from related table'] = 0
    settings.report_values['Roof shape (SSR) WARNING issued'] = 0
    settings.report_values['Route Designation Route Identification Conversion'] = 0
    settings.report_values['Route Weather Restriction derived from Closed in Winter'] = 0
    settings.report_values['sample complex map'] = 0
    settings.report_values['MEM Processing, Save value to MEM'] = 0
    settings.report_values['Saved to multivalued attribute'] = 0
    settings.report_values['Saved to multivalued attribute 2'] = 0
    settings.report_values['Saved to multivalued attribute 3'] = 0
    settings.report_values['OTH Processing - Putting a value into OTH'] = 0
    settings.report_values['Processing Time Overall (Seconds)'] = 0
    settings.report_values['Processing Time Field Translations (Seconds)'] = 0
    settings.report_values['Schema version set to target geodatabase version'] = 0
    settings.report_values['set field to not applicable for feature'] = 0
    settings.report_values['Source Date - Non-Spatial (NSD)'] = 0
    settings.report_values['Source Date - Spatial (SSD)'] = 0
    settings.report_values['Source date and time (SDV)'] = 0
    settings.report_values['Source Description - Non-Spatial (NSN)'] = 0
    settings.report_values['Source Description - Spatial (SSN)'] = 0
    settings.report_values['Source description (SDP)'] = 0
    settings.report_values['Source spatial (SPS) boolean was derived'] = 0
    settings.report_values['Source Type - Non-Spatial (NSP)'] = 0
    settings.report_values['Source Type - Spatial (SSY)'] = 0
    settings.report_values['Source type (SRT)'] = 0
    settings.report_values['Specified source field not found in source database'] = 0
    settings.report_values['Specified schema defined target field not found in target database'] = 0
    settings.report_values['string default value'] = 0
    settings.report_values['Subsurface stratum (WWM) values mapped'] = 0
    settings.report_values['Survey point type (SUY) derived from control point type (CPA)'] = 0
    settings.report_values['Terrain Subsurface Material values mapped'] = 0
    settings.report_values['Traffic restriction type Sharp Curve or Steep Grade corrected'] = 0
    settings.report_values['transform CPS from codelist to enum'] = 0
    settings.report_values['transform CPS from enum to codelist'] = 0
    settings.report_values[
        'Transform check for missing enum index: the specified index is not allowed in the target and we will save it to OTH'] = 0
    settings.report_values['transform ETS from codelist to enum'] = 0
    settings.report_values['transform ETS from enum to codelist'] = 0
    settings.report_values['transform HZD from codelist to enum'] = 0
    settings.report_values['transform HZD from enum to codelist'] = 0
    settings.report_values['transform RCG from codelist to enum'] = 0
    settings.report_values['transform RCG from enum to codelist'] = 0
    settings.report_values['transform VDT from codelist to enum'] = 0
    settings.report_values['transform VDT from enum to codelist'] = 0
    settings.report_values['transform ZI001_SRT from codelist to enum'] = 0
    settings.report_values['transform ZI001_SRT from enum to codelist'] = 0
    settings.report_values['transform ZI001_VSC from codelist to enum'] = 0
    settings.report_values['transform ZI001_VSC from enum to codelist'] = 0
    settings.report_values['transform ZI001_WTC value Closed in Winter to Fair Weather'] = 0
    settings.report_values['transform ZI004_RCG from codelist to enum'] = 0
    settings.report_values['transform ZI004_RCG from enum to codelist'] = 0
    settings.report_values['transform ZVH_VDT from enum to codelist'] = 0
    settings.report_values['Trees Widely Scattered (TWS) value derived from Vegetation Type (VEG)'] = 0
    settings.report_values['TRS'] = 0
    settings.report_values['TRS-Aeronautical'] = 0
    settings.report_values['TRS-Automotive'] = 0
    settings.report_values['TRS-Bus'] = 0
    settings.report_values['TRS-Canal'] = 0
    settings.report_values['TRS-Caravan Route'] = 0
    settings.report_values['TRS-Inland Waterway'] = 0
    settings.report_values['TRS-Inland Waterway'] = 0
    settings.report_values['TRS-Maritime'] = 0
    settings.report_values['TRS-No Transportation System'] = 0
    settings.report_values['TRS-Pedestrian'] = 0
    settings.report_values['TRS-Pipeline Maintenance'] = 0
    settings.report_values['TRS-Portage'] = 0
    settings.report_values['TRS-Portage sets PTG'] = 0
    settings.report_values['TRS-Powerline Maintenance'] = 0
    settings.report_values['TRS-Taxiway'] = 0
    settings.report_values['TRS-Truck'] = 0
    settings.report_values['TRS-Underground Railway'] = 0
    settings.report_values['TRS-Underground Railway sets LUN'] = 0
    settings.report_values['Underbridge Clearance derived from Bridge Structure'] = 0
    settings.report_values['Unique identifer created'] = 0
    settings.report_values['Unique identifier already in target'] = 0
    settings.report_values['Vegetation Type (VEG)-Grassland with Trees'] = 0
    settings.report_values['Vegetation Type (VEG)-Grassland Without Trees'] = 0
    settings.report_values['Vegetation Type (VEG)-Meadow'] = 0
    settings.report_values['WRDB Area of Numerous Detected, Subtype Mapped'] = 0
    settings.report_values['WRDB Area of Numerous Detected, Feature Code Mapped'] = 0
    settings.report_values['WRDB Area of Numerous Detected, Feature not retained in original feature class'] = 0
    settings.report_values['WRDB Extension - Water Nitrate'] = 0
    settings.report_values['WRDB Extension - Water Flow Minimum'] = 0
    settings.report_values['WRDB Extension - Water Flow Maximum'] = 0
    settings.report_values['WRDB Extension - Strahler Number'] = 0
    settings.report_values['WRDB Extension - FURL multiple value'] = 0
    settings.report_values['Feature Metadata Extension - COD saved to OTH'] = 0
    settings.report_values['Feature Metadata Extension - COE saved to OTH'] = 0
    settings.report_values['Feature Metadata Extension - ZI010_ACC saved to OTH'] = 0
    return


def init_transformer_report_column_headings(settings):
    settings.results_xl.prepare_results_report(settings.report_values)


# This function provides for the reporting of transformer function operations
def output_transformer_report_detail(my_logger, infeatureclass, reportdict, rptheading):
    totalvals = sum(reportdict.itervalues())  # in Python 3 this goes to sum(reportdict.values())
    if totalvals > 0:
        my_logger.info("  Transformer {1} Report for {0}".format(infeatureclass, rptheading))
        for keys, values in reportdict.items():
            if values > 0:
                my_logger.info("    {0} : {1}".format(keys, values))
        # my_logger.info("  Completed Issue Report for {0}".format(infeatureclass) )
    return


# This function copies a source feature class and field to a destination feature class
# This function has slight modification but is based on GitHub arcpy_attribute_based_etl.py
def copy_field(my_logger, src_fc, src_field, dstn_fc):
    field = [field for field in arcpy.Describe(src_fc).Fields if field.Name.lower() == src_field.lower()][0]
    if field.Name not in [dstnField.Name for dstnField in arcpy.Describe(dstn_fc).Fields]:
        my_logger.debug('copy_field copying %s' % field.Name)
        arcpy.AddField_management(dstn_fc, field.Name, field.Type, field.Precision, field.Scale, field.Length,
                                  field.AliasName, field.IsNullable, field.Required)
    else:
        my_logger.warning('field could not be copied because it already exists %s' % field.Name)


def show_output_saved_meta(introduction, printfunction, contentin):
    printfunction(introduction)
    if isinstance(contentin, dict):
        for iikey in contentin:
            if isinstance(iikey, dict):
                for iikey2 in iikey:
                    printfunction("\t\tkey: {0} value: {1}".format(str(iikey2), strout(iikey[iikey2])))
            elif isinstance(iikey, list):
                jj = 0
                for iikey2 in iikey:
                    printfunction("\t\tindex: {0} value: {1}".format(jj, strout(iikey2)))
                    jj += 1
            else:
                printfunction("\tkey: {0} value: {1}".format(str(iikey), strout(contentin[iikey])))
    elif isinstance(contentin, list):
        ii = 0
        for iikey in contentin:
            printfunction("\tindex: {0}".format(ii))
            if isinstance(iikey, dict):
                for iikey2 in iikey:
                    printfunction("\t\tkey: {0} value: {1}".format(str(iikey2), strout(iikey[iikey2])))
            elif isinstance(iikey, list):
                jj = 0
                for iikey2 in iikey:
                    printfunction("\t\tindex: {0} value: {1}".format(jj, strout(iikey2)))
                    jj += 1
            else:
                printfunction("\tindex: {0} value:{1}".format(ii, strout(iikey)))
            ii += 1
    else:
        printfunction("Item is neither a list or dictionary.")


# This function compares current target row content to existing metadata information. Transformations should not be required.
def compare_target_against_existing_metainfo(settings, rowin, saved_metainfo, \
                                             MetaSrcTargetMapperForFields):
    # First, if there is no saved metadata to search we can just return
    if len(saved_metainfo) == 0:
        # print "no saved metadata exists"
        return None

    # Verify that ths incoming row content is a dictionary and not just a list!
    if not isinstance(rowin, dict):
        settings.my_logger.error("compare_target_against_existing_metainfo, incoming content is not a dictionary")
        return None

    for rif in saved_metainfo:
        # We need to have a 'UFI' for this to work
        if 'UFI' not in rif:
            continue

        # rif is a dictionary representing saved information
        newuid = rif['UFI']
        match = True
        failedat = ""
        expectednumberfields = len(rif)

        for src, target, transformer in MetaSrcTargetMapperForFields:
            # ignore the src item, and work with the target. Transformer should not be needed iether
            if target is not None and target.lower not in ['ufi', 'f_code']:
                if target in rowin:
                    expectednumberfields -= 1
                    if rif.has_key(target) == False or rowin[target] != rif[target]:
                        failedat = target
                        match = False
                # If we have already found the match to fail, break out of the loop (for src, target...)
                if not match:
                    break

        if match == True and expectednumberfields < 4:  # if the flag is still set to true, we are done and can return this UFI
            if settings.debug_settings['compare_target_against_existing_metainfo'] > 0:
                settings.my_logger.debug(
                    "compare_target_against_existing_metainfo, Match found: {0}".format(str(newuid)))
            return newuid
    return None


# This function allows us to make some final changes to the target row.
# Basically, this is called immediately before saving this row to the target.
# This has been set up to look for any saved OTH values that may be compatible with this
# target schema and save the values directly. Note: OTH will still contain the original string content
def confirmtargetvalues(settings, target_row_dict, src_row_dict, target_fc):
    #
    targetchanged = False

    # In this first part, we want to look at all of the saved OTH values for this feature class
    # and if any of the saved OTH values (saved per UFI) are applicable to this target row UFI
    # we then want to determine if the OTH value references a field that is in the target database.
    # If the field is in the target database, we may be able to directly save this OTH value content
    # into the target.

    fieldname = ""  # Start with a blank field name
    localufi = get_row_unique_id(settings,
                                 target_row_dict)  # Get the UFI for the current target row (from the target row)
    if settings.debug_settings['report_unique_id_usage'] > 0:
        settings.my_logger.debug("Confirming target values in feature class {1}, ufi: {0}".format(localufi, target_fc))

    # We can't do anything unless the UFI is set to a workable value, so check for that first
    if localufi.lower() == 'noinformation' or localufi is None or localufi.lower() == 'no_information':
        pass
    else:
        # So now we believe we have a UFI that may be valid. Check to see if there are any saved OTH values and if the UFI we got
        # for this row is associated with any OTH settings.
        if len(settings.saved_oth) > 0 and localufi in settings.saved_oth:
            # parse the OTH values. We may want to directly set a field based on these OTH values
            rezz = parseOTH.parseOTHentry(settings.saved_oth[localufi])
            if settings.debug_settings['itemize_each_translation_each_field'] > 0:
                settings.my_logger.info("Parse of OTH with content: {0}, returned {1} items." \
                                        .format(settings.saved_oth[localufi], parseOTH.get_len_OTH_rtn_items(rezz)))

            # Deal with parsing the sections from the OTH. The OTH can be composed of multiple value settings, for each one:
            for ii in range(parseOTH.get_len_OTH_rtn_items(rezz)):
                # Identify the field name, index value (if enum), string value, and also be sure to save the target f_code
                fieldname = parseOTH.get_OTH_rtn_field(rezz, ii)
                indexvalue = parseOTH.get_OTH_rtn_index_value(rezz, ii)
                stringvalue = parseOTH.get_OTH_rtn_string_value(rezz, ii)
                inputvalue = parseOTH.get_OTH_rtn_inputvalue(rezz, ii)
                targfcode = target_row_dict['F_CODE'] if 'F_CODE' in target_row_dict else None

                if settings.debug_settings['itemize_each_translation_each_field'] > 0:
                    settings.my_logger.debug(
                        " {5}: Parse of OTH identified: Inputvalue={3} Field={0} Index={1}  String={2} for F_CODE {4} " \
                            .format(fieldname, str(indexvalue), strout(stringvalue), strout(inputvalue), targfcode,
                                    str(ii)))

                # If we parsed a field name from OTH that is currently in the row being output, then continue to process
                if fieldname in target_row_dict:  # Okay, so we have this field in the target

                    # First check the target field current value, ONLY if it is set to default value or Other can we change it
                    if is_value_unknown(target_row_dict[fieldname], 0, True):
                        if settings.debug_settings['itemize_each_translation_each_field'] > 0:
                            settings.my_logger.debug("current value is default so we can change this.")

                        # if we got an index value, verify it is valid for this target as best we can.
                        if indexvalue is not None:

                            if settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, targfcode, fieldname,
                                                                                 indexvalue):
                                # Get the string description for it for saving to the log
                                possVal = settings.targetdb.search_for_possible_enum_value_given_index(fieldname,
                                                                                                       indexvalue)
                                target_row_dict[fieldname] = indexvalue
                                report_transfomer_issue_or_increment_count(settings,
                                                                           'OTH content derived and saved directly to target field',
                                                                           "INFO",
                                                                           "Parse of OTH saved to field {0}. Enum Index {1} Enum String {2} for F_CODE {3}". \
                                                                           format(fieldname, indexvalue,
                                                                                  strout(possVal[1]), targfcode))
                                targetchanged = True
                        else:  # we could have index value of none but a string value... try to look up appropriate string value
                            # and if we can get the index value that corresponds to the string, hooray!
                            if stringvalue is not None:
                                possVal = settings.targetdb.search_for_possible_enum_value_given_string(fieldname,
                                                                                                        stringvalue)
                                if possVal is not None:
                                    if settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, targfcode,
                                                                                         fieldname, possVal[0]):
                                        target_row_dict[fieldname] = possVal[0]
                                        report_transfomer_issue_or_increment_count(settings,
                                                                                   'OTH content derived and saved directly to target field',
                                                                                   "INFO",
                                                                                   "Parse of OTH saved to field {0} with string match to index. Enum Index {1} Enum String {2} for F_CODE {3}". \
                                                                                   format(fieldname, strout(possVal[0]),
                                                                                          strout(possVal[1]),
                                                                                          targfcode))
                                        targetchanged = True
                                        # break (why a break here)?

        if settings.debug_settings['report_unique_id_usage'] > 0:
            settings.my_logger.debug(
                "Completed confirmation of target values in feature class {1}, ufi: {0}".format(localufi, target_fc))

    return targetchanged


def verify_row(settings, fc, field_list_for_row, new_row_list_prep, new_row_dict_for_prep, report_missing_items):
    for fld in field_list_for_row:
        if fld not in new_row_dict_for_prep:
            if report_missing_items or settings.debug_settings['compare_src_against_existing_metainfo'] > 1:
                report_transfomer_issue_or_increment_count(
                    settings, "In verify_row Expected field was not specified in the row. Setting to None.", "WARN",
                    "Field: {0}".format(fld))
            new_row_list_prep.append(None)
        else:
            new_row_list_prep.append(new_row_dict_for_prep[fld])
    return True


def adjust_debug_settings_based_on_feature_class(settings, sourcedb, targetdb):
    # Determine if we have identified a feature class to debug within the debug settings. If so, output information about the debug setting and set the
    # debug_setting dictionary entry to a value indicating that we are in debug mode.
    if len(settings.debug_fc_list) > 0:
        src_fc = sourcedb.get_fc()
        target_fc = targetdb.get_fc()
        if (is_item_in_list_nocase(src_fc, settings.debug_fc_list) or \
                is_item_in_list_nocase(target_fc, settings.debug_fc_list)):
            settings.my_logger.debug("At feature class to debug Source: {0} Target: {1}".format(src_fc, target_fc))
            if len(settings.debug_field_list) == 0:
                # BREAKPOINT location might be here if you want to break on a particular feature class with no specific field spec
                settings.debug_settings['itemize_each_translation_each_field'] = 111
        else:
            settings.debug_settings['itemize_each_translation_each_field'] = 0


def adjust_debug_settings_based_on_feature_class_and_field(settings, sourcedb, targetdb, srcfield, targfield):
    if len(settings.debug_field_list) > 0:
        src_fc = sourcedb.get_fc()
        target_fc = targetdb.get_fc()
        if len(settings.debug_fc_list) > 0 and \
                (is_item_in_list_nocase(src_fc, settings.debug_fc_list) or \
                 is_item_in_list_nocase(target_fc, settings.debug_fc_list)):
            if (srcfield is not None and \
                is_item_in_list_nocase(srcfield, settings.debug_field_list)) or \
                    (targfield is not None and is_item_in_list_nocase(targfield, settings.debug_field_list)):
                # BREAKPOINT location might be here if you want to break on a particular feature class and field
                settings.my_logger.debug(
                    "At feature class and field to debug Source: {0} Target: {1} Source Field: {2} Target Field: {3}".format(
                        src_fc, target_fc, srcfield, targfield))
                settings.debug_settings['itemize_each_translation_each_field'] = 111
            else:
                settings.debug_settings['itemize_each_translation_each_field'] = 0
        else:
            settings.debug_settings['itemize_each_translation_each_field'] = 0


def report_output_schema_differences(settings, type_of_fc, fc, fields, updated_fields, not_in_fields,
                                     db_has_but_not_in_schema):
    # This is a debug output block that allows for specific output of information about field differences found in the geodatabase vs. in the
    # mapping specification.
    if settings.debug_settings['report_field_schema_vs_database_differences'] > 0:
        if len(fields) <> len(updated_fields):
            settings.my_logger.warn(
                "Fields in specification are not all found in database {0} Feature {1}. \n  Fields intersected db and definition: {2}  \n  Fields missing from db: {3}". \
                    format(type_of_fc, fc, updated_fields, not_in_fields))
        if len(fields) <> len(db_has_but_not_in_schema):
            settings.my_logger.warn(
                "Fields in database are not all found in schema definition {0} Feature {1}. \n  Fields found in addition to specification: {2}  \n  Fields in specification: {3}". \
                    format(type_of_fc, fc, db_has_but_not_in_schema, fields))


def translate_set_of_fields(settings, src, target, transformer, sourcedb, targetdb, rownumber):
    move_data_error = 0
    src_fc = sourcedb.get_fc()
    target_fc = targetdb.get_fc()

    # Assumptions: 
    # src_fields have been verified to be in the source database
    # target_fields have been verified to be in the target database

    try:

        # # # Verification of inputs # # #  Preparation # # # 

        # Deal with potential input data set as strings for None rather than the value None.
        # Normalize these so that when we test for the values later we can do so consistently.
        src = normalize_field(src)
        target = normalize_field(target)
        transformer = normalize_field(transformer)

        target_row = targetdb.output_row
        source_row = sourcedb.row

        output_itemized_info = settings.debug_settings['itemize_each_translation_each_field']

        # Enable debug logging if it is set up for this feature class and field, or just set in general
        adjust_debug_settings_based_on_feature_class_and_field(settings, sourcedb, targetdb, src, target)

        # If debug settings are enabled, output a line for this field transformation
        if output_itemized_info > 0:
            settings.my_logger.debug(
                "Process source field {0} to target field {1} using transformer {2}".format(src, target, transformer))

            # If the target AND the transformer are None, then we can skip the rest here and continue the loop (This is okay)
        if target is None and transformer is None:
            if output_itemized_info > 0:
                settings.my_logger.debug(
                    "\tSource present but no target and no transform, skipping remaining processing")
            return 0  # Return value is the number of errors/issues

        # In the case of situations where we have NO mapping specification, we need to be able to handle
        # source and target information that may simply map to each other.
        # if source_db_version.lower() == "unknown" and src is not None and is_item_in_list_nocase(src, common_fields_geom_and_id) == False:

        if src is not None:
            if src in source_row and source_row[src] is None:
                # If the source is there, check for it being a value of none or strange default - but don't worry about it if it is a special case field
                source_row[src] = sourcedb.set_value_to_default_if_none(src, source_row[src])
                if output_itemized_info > 0:
                    settings.my_logger.debug(
                        "\tSource field {0} was not in row, prepared a default value {1} for this source.".format(src,
                                                                                                                  strout(
                                                                                                                      source_row[
                                                                                                                          src])))
            elif src not in source_row:
                source_row[src] = sourcedb.field_set_to_default(sourcedb.get_f_code(), src, src_fc)

        # Prepare a location in the target row if it is not already there
        if target is not None and target not in target_row:
            target_row[target] = None

        #   #   # END # # # Verification of inputs # # #  Preparation # # #
    except:
        report_transfomer_issue_or_increment_count(
            settings, "Exception while preparing to translate row of data.", "ERROR",
            "Source table: {0}, Source field: {1}, Target table: {2}, Target field: {3}, Transform: {4}, Rule index or row: {5}". \
                format(src_fc, src, target_fc, target, transformer, rownumber),
            sys.exc_info())
        return 1

    try:
        #### --------------- #### ---------------  ### Done with verification of inputs, now translate ### --------------- ### ----------------- ###
        #### No target field but there is a transformer - hence we are probably saving to OTH
        if target is None and src is not None and transformer is not None:
            # there may be a reason to call a transformer function - perhaps to save an element to OTH, note the return value
            # from this transform will not be saved within the output row. Basically this is used only to save to OTH or MEM.
            if output_itemized_info > 1:
                settings.my_logger.debug(
                    "\t\tTarget is None. Transformer {0} was called presumably to save data to OTH.".format(
                        transformer))
            try:
                valueNA = call_transform_function[transformer](settings, source_row, src, target, src_fc, target_fc)
            except:
                move_data_error += 1
                report_transfomer_issue_or_increment_count(
                    settings, "  Specified transformer {0} had an exception.".format(transformer),
                    "ERROR",
                    "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                        format(transformer, src, target,
                               strout(source_row[src]) if src in source_row else "",
                               src_fc, target_fc, "Value is NA"),
                    sys.exc_info())
                #
        #                  
        elif target is not None and src is None and transformer is None:

            #####   Src is None AND transformer is None - Set target to default value  # since we set it to None first, this one should set it to default
            target_row[target] = targetdb.set_value_to_default_if_none(target, target_row[target])

            if output_itemized_info > 1:
                settings.my_logger.debug(
                    "\t\tSource is None. Transform is None. Setting target field {0} to default {1}".format(target,
                                                                                                            strout(
                                                                                                                target_row[
                                                                                                                    target])))
        #
        elif target is not None and src is None and transformer is not None:
            #####   Src is None AND transformer is NOT None - Call transformer which hopefully will set the target anyway
            if transformer in call_transform_function:
                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\tSource is None. Applying transform function {0} for target {1}".format(transformer,
                                                                                                    target))
                try:
                    target_row[target] = call_transform_function[transformer](settings, source_row, src, target, src_fc,
                                                                              target_fc)
                except:
                    move_data_error += 1
                    report_transfomer_issue_or_increment_count(
                        settings, "  Specified transformer {0} had an exception.".format(transformer),
                        "ERROR",
                        "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                            format(transformer, src, target,
                                   strout(source_row[src]) if src in source_row else "",
                                   src_fc, target_fc, strout(target_row[target]) if target in target_row else ""),
                        sys.exc_info())
                # end except

                if target_row[target] is None:
                    target_row[target] = targetdb.set_value_to_default_if_none(target, target_row[target])

                if output_itemized_info > 1:
                    settings.my_logger.debug("\t\t\tTransform returned: {0}".format(strout(target_row[target])))
            # transformer exists, else:
            else:
                move_data_error += 1
                report_transfomer_issue_or_increment_count(
                    settings, "  Specified transformer {0} does not exist.". \
                        format(transformer), "ERROR",
                    "Transformer {0} for field {1} does not exist".format(transformer, target))
                if target_row[target] is None:
                    target_row[target] = targetdb.set_value_to_default_if_none(target, target_row[target])
                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\tSource is None. Transform {0} does not exist. Target {1} set to default.".format(
                            transformer, target))
            # Report error that the specified transformer does not exist
        #
        elif target is not None and src is not None and transformer is None:

            #  We know there is a source (src) field - but it may not be in the row, transformer present
            if src not in source_row:
                source_row[src] = None

            #  We know there is a source (src) field - but it may not be in the row, transformer is None, and source is a f_code or shape field
            #  Watch out for @SHAPE and FCSubtype and that sort of thing
            if is_item_in_list_nocase(src, sourcedb.common_fields_geom_and_id_and_code):
                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\tSource and Target exist. No Transform. Processing shape or db specific identifier.")
                target_row[target] = source_row[src]

            #  So next check for these two fields being enumerated data types, in which case we have an automated check for differences
            elif source_row[src] is not None and \
                    sourcedb.is_enumeration(src_fc, src) and targetdb.is_enumeration(target_fc, target):
                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\tSource and Target exist. No Transform. Enumerated Values. Automated check for missing values. Source value: {0}". \
                            format(strout(source_row[src])))
                try:
                    target_row[target] = call_transform_function[
                        'check_for_and_transform_missing_enum_index_target_diff'](
                        settings, source_row, src, target, src_fc, target_fc)
                except:
                    move_data_error += 1
                    report_transfomer_issue_or_increment_count(
                        settings, "  Specified transformer {0} had an exception.".format(transformer),
                        "ERROR",
                        "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                            format('check_for_and_transform_missing_enum_index_target_diff', src, target,
                                   strout(source_row[src]) if src in source_row else "",
                                   src_fc, target_fc, strout(target_row[target]) if target in target_row else ""),
                        sys.exc_info())
                # end except
                if target_row[target] is None:
                    target_row[target] = targetdb.set_value_to_default_if_none(target, target_row[target])
                if output_itemized_info > 1:
                    settings.my_logger.debug("\t\t\tTarget value set to: {0}". \
                                             format(strout(target_row[target])))
            # 
            else:  # we probably just need to move the data over
                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\tSource and Target exist. No Transform. Move data: {0}.".format(strout(source_row[src])))
                target_row[target] = source_row[src]
                if target_row[target] is None:
                    target_row[target] = targetdb.set_value_to_default_if_none(target, target_row[target])
            #  end else this is not a shape field
        # 
        elif target is not None and src is not None and transformer is not None:

            if src not in source_row:
                source_row[src] = None

            # print('   {0} to {1} via transformer {2}'.format(src, target, transformer))
            # Verify that the transform function is valid and if it is, run the data through the transform function
            if transformer in call_transform_function:  # transformer valid, src not none

                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\tSource, Target, Transform exist. Calling transform {0}.".format(transformer))

                try:
                    target_row[target] = call_transform_function[transformer](settings, source_row, src, target, src_fc,
                                                                              target_fc)
                except:
                    move_data_error += 1
                    report_transfomer_issue_or_increment_count(
                        settings, "  Specified transformer {0} had an exception.".format(transformer),
                        "ERROR",
                        "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                            format(transformer, src, target,
                                   strout(source_row[src]) if src in source_row else "",
                                   src_fc, target_fc, strout(target_row[target]) if target in target_row else ""),
                        sys.exc_info())
                # end except
                if target_row[target] is None:
                    target_row[target] = targetdb.set_value_to_default_if_none(target, target_row[target])

                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\t\tSource value {0}, Transform returned: {1}".format(strout(source_row[src]),
                                                                                 strout(target_row[target])))


            else:
                move_data_error += 1
                report_transfomer_issue_or_increment_count(
                    settings, "  Specified transformer {0} does not exist.". \
                        format(transformer), True,
                    "Transformer {0} for field {1} does not exist".format(transformer, target), "WARN")
                if target_row[target] is None:
                    target_row[target] = targetdb.set_value_to_default_if_none(target, target_row[target])
                if output_itemized_info > 1:
                    settings.my_logger.debug(
                        "\t\t\tSource, Target, Transform exist. Transform {0} failed. Set field {1} to default.".format(
                            transformer, target))
            #
        #
        # elif target is None and src is not None and transformer is None:
        #    pass   # this one is okay, sometimes we just need the source field to be present for OTHER transforms  This was actually covered earlier
        #
        elif target is None and src is None:
            pass  # Should never see this one really, but whatever - it is here because we don't need to see an error for it I guess
        # 
        else:  # We should never get to the next option
            move_data_error += 1
            report_transfomer_issue_or_increment_count(
                settings, "  Processing combination for src {0} target {1} transformer {2} unknown". \
                    format(src, target, transformer), True, "Option for processing this combination unknown", "WARN")
            target_row[target] = None

        # If we ran a transform that changed the feature code, we need to check for that
        if target is not None and target.lower() == 'f_code':
            targetdb.set_f_code(target_row[target])

    except:
        move_data_error += 1
        report_transfomer_issue_or_increment_count(
            settings, "Exception while translating row.", "ERROR",
            "Source table: {0}, Source field: {1}, Target table: {2}, Target field: {3}, Transform: {4}, Rule index or row: {5}". \
                format(src_fc, src, target_fc, target, transformer, rownumber),
            sys.exc_info())

    return move_data_error


# Check for dictionary item, any case and return the value
def get_dict_item_anycase(indict, inkey):
    if inkey in indict:
        return indict[inkey]
    elif inkey.lower() in indict:
        return indict[inkey.lower()]
    else:
        # Check all of the keys
        for kk in indict:
            if kk.lower() == inkey.lower():
                return indict[kk]
    return None


# This is the primary translation function that moves data from one database to another.
# It operates on a set of feature classes or tables, taking an mapping specification
# as input that may include a specific transformation of the data. If a specific data
# transformation function is specified, it is called to transform the data.
# For each feature class mapping table definition:
#    Validate stuff
#    Get each source row from each source feature class or table and with that source row:
#        Initialize a target row
#        For each field mapping defined, perform the field mapping to get a target row value
#        Write the target row to the target database
# 
def transform_and_move_data(settings, listoffcstomove, sourcedb, targetdb, fandf, reportid, stepname,
                            generate_report=False):
    settings.current_step = stepname
    mytimer = MyTimers(settings.maximum_records['processing_indicator_n_seconds'])

    total_data_count = 0
    transform_fc_count = 0

    # for each mapping item set to Source Feature Class or Table, Target FC or Tbl, Mapping in the list:
    for src_fc, target_fc, mapping, where in listoffcstomove:
        if settings.debug_settings['instrumentation'] > 0:
            dump_garbage(settings, "for fc beginning")
        mytimer.set_initial_time('mappingrule')
        move_data_count = 0  # track how many records we move
        move_data_error = 0
        proc_data_count = 0
        fields_processing_time = 0
        transform_fc_count += 1
        mapping_rules_count = 0

        # Ensure the items in this loop are valid
        if where is None or where.lower() == 'none':
            where = ''

        settings.current_mapping_group = mapping  # Name of the mapping group

        sourcedb.set_fc(src_fc)  # Set the source database to the current feature class from the source
        targetdb.set_fc(target_fc)  # Set the target database to the current target feature class
        sourcepathname = sourcedb.get_root_for_fc_subtypes(
            src_fc)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
        targetpathname = targetdb.get_root_for_fc_subtypes(
            target_fc)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc

        clear_transformer_functions_report(settings)  # Start with a clean report set for this feature class
        clear_transformer_issues_report(settings)  # Start with a clean error log for this feature class

        adjust_debug_settings_based_on_feature_class(settings, sourcedb, targetdb)

        # If the source feature class (table) and the target feature class (table) both exist, we are okay to continue the translation of this content.
        if (sourcedb.is_featureclass_in_geodatabase_spec(src_fc) and targetdb.is_featureclass_in_geodatabase_spec(
                target_fc)):

            # Output a line to the logger stating we are going to prcess this source and target feature class
            settings.my_logger.info('{3}; Started processing table or feature class: {0} with target: {1}  {2}'.format(
                src_fc, target_fc, "where clause: {0}".format(where) if where <> '' else "", stepname))

            # Verify that we actually got a mapping list of fields specified correctly as a list python type.  This extra check
            # is needed for the possiblity of a excel specified mapping. The value has to be corrected a bit to work in python.
            if not isinstance(mapping, list):
                mapping = getattr(fandf, mapping)

            # Get the source fields and target fields from the mapping specification 
            # we will intersect these against the actual database.
            # src_fields = [src for src, target, transformer in mapping if src is not None and src != 'None'] 
            # target_fields = [target for src, target, transformer in mapping if target is not None and target != 'None']
            sourcedb.identify_actual_fields_list_matching_schema(
                "Source", src_fc, [src for src, target, transformer in mapping if src is not None and src != 'None'])
            targetdb.identify_actual_fields_list_matching_schema(
                "Target", target_fc,
                [target for src, target, transformer in mapping if target is not None and target != 'None'])

            if settings.debug_settings['itemize_each_translation_each_field'] > 1:
                settings.my_logger.debug("\tSource fields {0}".format(sourcedb.updated_fields))
                settings.my_logger.debug("\tTarget fields {0}".format(targetdb.updated_fields))

            # Set the source and target fields to the updated lists since these represent the fields actually in the geodatabase vs the schema specification
            # src_fields = sourcedb.updated_fields
            # target_fields = targetdb.updated_fields

            if len(sourcedb.updated_fields) == 0 or len(targetdb.updated_fields) == 0:
                report_transfomer_issue_or_increment_count(
                    settings, "Source or target has no fields to translate.", "ERROR",
                    "Source fields list length {0}, target fields list length {1}".format(
                        len(sourcedb.updated_fields), len(targetdb.updated_fields)))

            else:

                # Get the set of source rows for this feature class and field set
                with arcpy.da.SearchCursor(sourcepathname, sourcedb.updated_fields, where) as rows:
                    # Convert the row to a dictionary
                    srcdicts = rows_as_dicts(rows)

                    # Prepare to insert into the target feature class, the output_rows
                    with arcpy.da.InsertCursor(targetpathname, targetdb.updated_fields) as new_rows:

                        if settings.debug_settings['instrumentation'] > 0:
                            dump_garbage(settings, "fc have gathered rows for processing")
                        # for each source row feature instance/ This loop has been checked and found no mem leaks in it
                        for sourcedb.row in srcdicts:

                            proc_data_count += 1
                            settings.saved_mem.clear()  # We can start with a clean dictionary of MEM content for each row
                            settings.saved_oth.clear()  # We can start with a clean dictionary of OTH content for each row
                            #### now for the row based translation:
                            #  we have a dictionary row from the source: row, and we have a target row: output_rows
                            #  we have a source feature class: source_fc, and target feature class: target_fc
                            #  and we have a source geodatabase and target geodatabase.

                            # Determine the source and target feature code value as this critical information
                            # is required to deal with complex mappings, and subtype identification
                            sourcedb.set_f_code(get_value_from_dict_nocase("F_CODE", sourcedb.row))
                            targetdb.set_f_code(sourcedb.get_f_code())

                            # Check to see if this feature instance has already been moved and if not, transform it. Note: This depends on having unique identifiers
                            if settings.config_setting_flag[
                                'flag_expect_no_ufi_conflict'] == 0 and targetdb.is_this_row_already_in_target(
                                sourcedb.row, settings.my_logger, True):
                                settings.report_values['Unique identifier already in target'] += 1
                            else:

                                targetdb.output_row.clear()  # this is a dictionary represntation of the target row values

                                try:
                                    # From the mapping definition, get source field, target field, transformer name
                                    mytimer.set_initial_time('fields')
                                    ruleindex = 0
                                    move_field_error = 0
                                    # We really need to process any transforms or information related to f_code first
                                    for src, target, transformer in mapping:
                                        if target is not None and target.lower() == 'f_code':
                                            move_field_error += translate_set_of_fields(settings, src, target,
                                                                                        transformer, sourcedb, targetdb,
                                                                                        ruleindex)
                                            ruleindex += 1
                                            break

                                    for src, target, transformer in mapping:
                                        if target is None or target.lower() != 'f_code':
                                            move_field_error += translate_set_of_fields(settings, src, target,
                                                                                        transformer, sourcedb, targetdb,
                                                                                        ruleindex)
                                            ruleindex += 1

                                    # End for src, target, transformer in mapping
                                    mytimer.set_end_time('fields')
                                    fields_processing_time += mytimer.get_elapsedtime('fields')

                                    if settings.debug_settings['only_transform_no_move'] == 0:
                                        if ('F_CODE' in targetdb.updated_fields and (
                                                get_dict_item_anycase(targetdb.output_row, 'f_code') is None)):
                                            move_data_error += 1
                                            report_transfomer_issue_or_increment_count(
                                                settings, "Caution! No feature code has been specified for target row.",
                                                "WARN", "Feature class: {0}".format(target_fc))
                                        if ('FCSubtype' in targetdb.updated_fields and (
                                                get_dict_item_anycase(targetdb.output_row, 'fcsubtype') is None)):
                                            move_data_error += 1
                                            report_transfomer_issue_or_increment_count(
                                                settings,
                                                "Caution! No feature subtype has been specified for target row.",
                                                "WARN", "Feature class: {0}".format(target_fc))

                                            # Give one final chance to update the target row values based on possible OTH values that have been saved previously
                                        if (settings.config_setting_flag[
                                            'flag_evaluate_oth_content_and_parse_for_possible_values_to_set']) == 1:
                                            changesmade = confirmtargetvalues(settings, targetdb.output_row,
                                                                              sourcedb.row,
                                                                              target_fc)  # give it the new target row, target row dictionary, and source row dictionary

                                        # Set these values to indicate we are done 'processing', so that if we get an error we will know it is after the processing has completed.
                                        src = "Source Row"
                                        target = "Target Row"
                                        transform = "Row insert"

                                        # The call to verify_row will verify and will fill the provided list with the target_fields in order for insertion
                                        output_row_out = list()  # this is the returned list, a new row ready for insertion into the target
                                        if verify_row(settings, target_fc, targetdb.updated_fields, output_row_out,
                                                      targetdb.output_row, True):
                                            #  This call will write the output_row into the target Esri Geodatabase
                                            new_rows.insertRow(output_row_out)
                                            move_data_count += 1
                                    # End If insert the row to the target

                                    if mytimer.check_for_elapsed_time_gt_set_interval('mappingrule'):
                                        settings.my_logger.info(
                                            "\t\t...processing... current count: {0}".format(proc_data_count))

                                        # END Try
                                except:
                                    # print (" ")
                                    move_data_error += 1
                                    report_transfomer_issue_or_increment_count(
                                        settings, "Exception while processing feature class.", "ERROR",
                                        "While processing feature class src {0}, feature class target {1}, data processing index: {2}".format(
                                            src_fc, target_fc, proc_data_count),
                                        sys.exc_info())

                                # End Try / Except

                            # End if targetdb.is_this_row_already_in_target

                            if settings.config_setting_flag['break_after_n_instances'] > 0 and \
                                    (settings.config_setting_flag['break_after_n_instances'] <= move_data_count or \
                                     settings.config_setting_flag['break_after_n_instances'] <= proc_data_count):
                                break

                            if move_data_error > settings.maximum_records[
                                'maximum_issues_encountered_per_featureclass']:
                                break
                        # End for row in srcdic
                    # End  with arcpy.da.InsertCursor(targetpathname, target_fields)
                # End SearchCursor 

            # End Else the field lists are not 0 length

        else:  # Source and target feature classes in database

            if settings.debug_settings['report_feature_classes_defined_in_schema_but_not_in_database'] > 0:
                settings.my_logger.info(
                    'Specified mapping but no feature class or table.  Source {0} - {1} Target {2} - {3} '.format( \
                        src_fc, (sourcedb.is_featureclass_in_geodatabase_spec(src_fc)), target_fc,
                        (targetdb.is_featureclass_in_geodatabase_spec(target_fc))))

        # End else either the source or target is missing

        # --- Finish up for the processing on each source to target mapping
        mytimer.set_end_time('mappingrule')

        # print(" ")
        settings.report_values['Translate and Move - Instances Moved'] = move_data_count
        settings.report_values['Translate Processing - Instances Processed'] = proc_data_count
        if proc_data_count > 0:
            settings.report_values['Processing Time Overall (Seconds)'] = mytimer.get_elapsedtime('mappingrule')
            settings.report_values['Processing Time Field Translations (Seconds)'] = fields_processing_time

        output_transformer_report_detail(settings.my_logger, src_fc, settings.report_values, "Information")
        output_transformer_report_detail(settings.my_logger, src_fc, settings.report_errors, "Issues")

        if generate_report:
            # settings.results_xl.save_results_item("FC", src_fc, target_fc, settings.report_values, transform_fc_count + 1)
            settings.results_xl.save_results_items_to_report(reportid, src_fc, target_fc, where, settings.report_values,
                                                             transform_fc_count + 1)
            settings.results_xl.save_results_items_to_report(reportid, src_fc, target_fc, where, settings.report_errors,
                                                             transform_fc_count + 1, True)
            settings.results_xl.done_save()

        settings.my_logger.info(
            'Completed processing {0} to {1}. Processed {2} records and {3} records moved to target.\n'.format(
                src_fc, target_fc, proc_data_count, move_data_count))

        # End For loop on feature classes


# This function - I believe is not currently used - was an attempt to work with metadata in a more generic fashion.
# I believe the specific Restriction Information and Entity Collection processing have replaced this.
def retrieve_target_meta_info(settings, inputtargetmetatable,
                              MetaSrcTargetMapperForFields,
                              saved_metainfo):
    #
    if settings.targetdb.is_featureclass_in_geodatabase_spec(inputtargetmetatable):
        target_obj_path = settings.targetdb.get_root_for_fc_subtypes(inputtargetmetatable)
        target_fields = [target for src, target, transformer in MetaSrcTargetMapperForFields if target]
        metainfo_count = 0

        # Get all current target restriction info content
        with arcpy.da.SearchCursor(target_obj_path, target_fields, None) as rows:
            for row in rows_as_dicts(rows):
                metainfo_count += 1
                if compare_target_against_existing_metainfo(settings, row, settings.saved_metainfo,
                                                            MetaSrcTargetMapperForFields) is None:
                    settings.saved_metainfo.append(row)


def get_existing_meta_restriction_info(settings, fromdb, inputtargetmetatable, inputTableFields, reportvals=True):
    if fromdb.is_featureclass_in_geodatabase_spec(inputtargetmetatable):
        targetpathname = fromdb.get_root_for_fc_subtypes(
            inputtargetmetatable)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc

        target_fields_for_row = [target for target in inputTableFields if target is not None and target != 'None']
        metainfo_count = 0
        # clear_transformer_functions_report(settings)
        # clear_transformer_issues_report(settings)
        #  fromdb.identify_actual_fields_list_matching_schema("Target Restriction Info", inputtargetmetatable, target_fields)

        # Get all current target metadata info content
        with arcpy.da.SearchCursor(targetpathname, target_fields_for_row, None) as rowsinputri:
            for rowinri in rows_as_dicts(rowsinputri):
                if rowinri not in settings.saved_restrictioninfo:
                    metainfo_count += 1
                    settings.saved_restrictioninfo.append(rowinri)

        # if reportvals:
        #    output_transformer_report_detail(settings.my_logger, "Restriction Information", settings.report_values, "Information")
        #    output_transformer_report_detail(settings.my_logger, "Restriction Information", settings.report_errors, "Issues")

    return None


def process_fields_metadata_target(settings, src, target, transformer, newdict, row, MetaTargetDefaultFieldValues,
                                   src_fc, target_fc):
    data_err_count = 0
    if target:  # may be None such as for source fields that no longer exist
        if target.lower() == 'ufi':
            getufi = makerandomufi()
            newdict[target] = getufi
        elif target.lower() in settings.targetdb.common_fields_geom_only:
            if target in row:
                newdict[target] = row[target]
            else:
                newdict[target] = None
            # YIKE - actually a shape may need to be extended to be a union of all shape values coming in
        elif target == 'FCSubType':
            newdict[target] = MetaTargetDefaultFieldValues[target]
        elif target == 'F_CODE':
            newdict[target] = MetaTargetDefaultFieldValues[target]
        else:
            if transformer is None and target in row:
                newdict[target] = row[target]
            elif transformer is None and src in row:
                newdict[target] = row[src]
            elif transformer is not None:
                if transformer in call_transform_function:
                    try:
                        if src in row and row[src] is not None:
                            xform_val = call_transform_function[transformer](settings, row, src, target, src_fc,
                                                                             target_fc)
                        elif target in row and row[target] is not None:
                            xform_val = call_transform_function[transformer](settings, row, target, target, src_fc,
                                                                             target_fc)
                        elif target is not None and target in MetaTargetDefaultFieldValues:
                            xform_val = MetaTargetDefaultFieldValues[target]
                        else:
                            xform_val = None
                    except:
                        data_err_count += 1
                        report_transfomer_issue_or_increment_count(
                            settings,
                            "  Specified transformer {0} had an exception or target field problem".format(transformer),
                            "ERROR",
                            "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}.". \
                                format(transformer, src, target,
                                       settings.sourcedb.row[src] if src in row else "",
                                       src_fc, target_fc, newdict[target] if target in newdict else ""),
                            sys.exc_info())
                        xform_val = None
                    #
                    if target is not None:
                        newdict[target] = xform_val
                else:
                    data_err_count += 1
                    report_transfomer_issue_or_increment_count(
                        settings, "metadata transformer error failed on field.", "WARN",
                        "Meta transformer {0} for field {1} is missing.".format(transformer, target), True)
                # end if then else transformer in call function set
            # end if elif elif esle
        # end if target
    return data_err_count


def merge_src_target_transformer(mappingA, mappingB):
    # Merge the src, target, transformer items between mappingA and mapping B
    dictitems = dict()
    for src, target, transform in mappingA:
        itemkey = "{0}{1}".format(src, target)
        if itemkey not in dictitems:
            dictitems[itemkey] = (src, target, transform)
    for src, target, transform in mappingB:
        itemkey = "{0}{1}".format(src, target)
        if itemkey not in dictitems:
            dictitems[itemkey] = (src, target, transform)
    if "NoneF_CODE" not in dictitems:
        dictitems["NoneF_CODE"] = (
            None, 'F_CODE', 'transform_placeholder_return_None')  # Transform doesn't matter because it is defaulted
    if "NoneUFI" not in dictitems:
        dictitems["NoneUFI"] = (
            None, 'UFI', 'transform_set_UFI')  # Transform doesn't matter because we will set a new UFI
    outlist = list()
    for dditem in dictitems:
        outlist.append(dictitems[dditem])
    return outlist


# This function processes only restriction information metadata. It assumes data model schema
# in which metadata has been stored on every feature and needs to be distributed to alternative target
# locations. Additionally, we want to save the mdatata information locally so that the unique identifier for
# it can be recorded in the feature. This function was developed to operate specifically on
# Restriction Information. 
def process_meta_restriction_info(settings, inputtargetmetatable,
                                  featureclasssettoprocess, RestrictionInfoTblFields,
                                  RestrictionInfoTargetDefaultFieldValues, fandf,
                                  add_default_record_if_target_empty=False):
    settings.current_step = "Restriction Information"
    mytimer = MyTimers(settings.maximum_records['processing_indicator_n_seconds'])
    try:
        ######
        ###### METADATA INFO - Get information records and save them to specific target
        ######
        targetpathname = settings.targetdb.get_root_for_fc_subtypes(
            inputtargetmetatable)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
        target_fields = [target for src, target, transformer in RestrictionInfoTblFields if
                         target is not None and target != 'None']
        metainfo_count = 0
        transform_rest_count = 0
        settings.targetdb.identify_actual_fields_list_matching_schema("Target Restriction Info", inputtargetmetatable,
                                                                      target_fields)
        # settings.targetdb.updated_fields
        if len(settings.targetdb.updated_fields) == 0:
            return

        # Get all current target metadata info content, save locally into settings.saved_restrictioninfo
        with arcpy.da.SearchCursor(targetpathname, settings.targetdb.updated_fields, None) as rows:
            for row in rows_as_dicts(rows):
                metainfo_count += 1
                settings.saved_restrictioninfo.append(row)
    except:
        report_transfomer_issue_or_increment_count(
            settings, "Exception while in metadata restriction info preparation.", "ERROR",
            None,
            sys.exc_info())
        return
    #
    move_data_count = 0
    proc_data_error = 0
    proc_data_count = 0
    #
    # Now if we have no meta info default row
    with arcpy.da.InsertCursor(targetpathname, settings.targetdb.updated_fields) as new_rows:
        new_row = []

        # If there were no records, add a default record
        if metainfo_count == 0 and add_default_record_if_target_empty == True:
            settings.my_logger.info("Adding default metadata record to: {0}".format(targetpathname))
            newdict = {}
            for field in settings.targetdb.updated_fields:
                new_row.append(RestrictionInfoTblFields[field])
                newdict[field] = RestrictionInfoTblFields[field]
            settings.saved_restrictioninfo.append(newdict)
            new_rows.insertRow(new_row)

        # with a new search cursor, go through the feature classes
        # and record any metadata information records needed
        # adding them to an the target and to an internal dictionary
        # The mapping may or may not be a large list and we will likely have to supplement with defaults on many fields
        # expected in the final set up required fields: settings.targetdb.updated_fields
        for src_fc, target_fc, mapping, where in featureclasssettoprocess:
            target_fc = inputtargetmetatable  # Override the mapping spec, the target is always specified as per the function call parameter
            mytimer.set_initial_time('restrictioninfo')

            try:
                # Ensure the items in this loop are valid
                if where is None or where == 'None':
                    where = ''

                settings.current_mapping_group = "Restriction Info FC"  # Name of the mapping group

                settings.sourcedb.set_fc(src_fc)  # Set the source database to the current feature class from the source
                settings.targetdb.set_fc(target_fc)  # Set the target database to the current target feature class
                sourcepathname = settings.sourcedb.get_root_for_fc_subtypes(
                    src_fc)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
                targetpathname = settings.targetdb.get_root_for_fc_subtypes(
                    target_fc)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc

                clear_transformer_functions_report(settings)
                clear_transformer_issues_report(settings)

                transform_rest_count += 1

                if len(settings.saved_restrictioninfo) > settings.maximum_records['saved_restrictioninfo']:
                    proc_data_error += 1
                    settings.my_logger.error(
                        "Error processing restriction restriction info portion of features or feature classes. Maximum of {0} records have been exceeded!". \
                            format(str(settings.maximum_records['saved_restrictioninfo'])))
                    continue

                if (settings.sourcedb.is_featureclass_in_geodatabase_spec(
                        src_fc) == False or settings.targetdb.is_featureclass_in_geodatabase_spec(target_fc) == False):
                    if settings.debug_settings['report_feature_classes_defined_in_schema_but_not_in_database'] > 0:
                        proc_data_error += 1
                        report_transfomer_issue_or_increment_count(settings,
                                                                   "Restriction Info subset translation, source or target is not present. ",
                                                                   "WARN",
                                                                   "Source {0} {2}; target {1} {3}. Skipping this mapping rule set.".format(
                                                                       src_fc, target_fc, (
                                                                           settings.sourcedb.is_featureclass_in_geodatabase_spec(
                                                                               src_fc)), (
                                                                           settings.targetdb.is_featureclass_in_geodatabase_spec(
                                                                               target_fc))))
                else:
                    src_fc_path = settings.sourcedb.get_root_for_fc_subtypes(src_fc)
                    src_fields = [src for src, target, transformer in mapping if
                                  src is not None and src.lower() != 'none']
                    settings.sourcedb.identify_actual_fields_list_matching_schema("Source", src_fc, src_fields)
                    # settings.sourcedb.updated_fields

                    # Set the source and target fields to the updated lists since these represent the fields in the DATABASE vs the schema specification
                    # and the database may be a GRLS level while the schema specification is a GGDM Composite.
                    src_fields = settings.sourcedb.updated_fields

                    settings.my_logger.info(
                        'Processing restriction metadata portion of table or feature class: {0} with target {1}'.format(
                            src_fc, target_fc))
                    if len(src_fields) == 0:
                        report_transfomer_issue_or_increment_count(
                            settings,
                            "No fields found in source database in common with specified geodatabase schema type.",
                            "WARN", " Feature {0}".format(src_fc))
                        continue

                    # search for this feature class and get fields and rows for the subset metadata
                    with arcpy.da.SearchCursor(src_fc_path, src_fields, where) as rows:

                        # For each feature instance in the source geodatabase - we should now have a subset of metadata fields in a row
                        for row in rows_as_dicts(rows):
                            # current_static_ufi = "0"    # If we are generating UFIs this is a trigger to generate a new UFI
                            settings.sourcedb.set_f_code(get_value_from_dict_nocase("F_CODE", row))
                            settings.targetdb.set_f_code(RestrictionInfoTargetDefaultFieldValues['F_CODE'])

                            proc_data_count += 1  # BREAK HERE

                            # start add of new metadata record
                            newdict = settings.targetdb.output_row
                            newdict.clear()

                            # We may have a mapping from the mapping, plus we have defaults from the Restriction Information Table mapping. Merge them into one
                            merged_mapping = merge_src_target_transformer(mapping, RestrictionInfoTblFields)

                            for src, target, transformer in merged_mapping:
                                proc_data_error += process_fields_metadata_target(settings,
                                                                                  src, target, transformer,
                                                                                  newdict, row,
                                                                                  RestrictionInfoTargetDefaultFieldValues,
                                                                                  src_fc, target_fc)

                            # End src, target, transformer 

                            # Now see if we already have this metadata, and if so we will use the existing UFI, otherwise we will use this new one.
                            # So this function will compare this newdict set of information against ALL of the currently stored saved_restrictioninfo
                            # looking for a potential match (with the exception of UFI and F_CODE fields).
                            posUFI = compare_against_existing_metainfo_noxform(settings, newdict,
                                                                               settings.saved_restrictioninfo)

                            # posUFI = compare_src_against_existing_metainfo(
                            #    settings, newdict, settings.saved_restrictioninfo,'RESTRICTION_INFO_T', 'RESTRICTION_INFO_T', RestrictionInfoTblFields)

                            if posUFI is None:

                                # Verify length of target fields and row prior to inserting row into target
                                output_row_out = list()  # this is the returned list, a new row ready for insertion into the target
                                if verify_row(settings, target_fc, settings.targetdb.updated_fields, output_row_out,
                                              newdict, False):
                                    #  This call will write the output_row into the target Esri Geodatabase
                                    settings.saved_restrictioninfo.append(newdict)
                                    new_rows.insertRow(output_row_out)
                                    move_data_count += 1
                                else:
                                    proc_data_error += 1
                                    report_transfomer_issue_or_increment_count(settings,
                                                                               "Restriction info processing did not generate correct number of fields.",
                                                                               "ERROR",
                                                                               "Need {0} fields, but {1} generated. Target fields: {2} current row {3}". \
                                                                               format(len(target_fields), len(new_row),
                                                                                      str(target_fields), str(new_row)))

                            # End if posUFI

                            if mytimer.check_for_elapsed_time_gt_set_interval('restrictioninfo'):
                                settings.my_logger.info(
                                    "\t\t...processing... current count: {0}".format(proc_data_count))

                            if settings.config_setting_flag['break_after_n_instances'] > 0 and \
                                    (move_data_count >= settings.config_setting_flag['break_after_n_instances'] or \
                                     settings.config_setting_flag['break_after_n_instances'] <= proc_data_count):
                                break

                        # End for source row in dictionary (perform translations)
                    # End with

                # End Else


            except:
                proc_data_error += 1
                report_transfomer_issue_or_increment_count(
                    settings, "Exception while translating row.", "ERROR",
                    "Source feature: {0}, Target feature: {1}.".format(src_fc, target_fc),
                    sys.exc_info())
            #
            mytimer.set_end_time('restrictioninfo')
            if move_data_count > 0:
                settings.report_values['Processing Time Overall (Seconds)'] = mytimer.get_elapsedtime('restrictioninfo')

            output_transformer_report_detail(settings.my_logger,
                                             "Restriction Information {0} to {1}".format(src_fc, target_fc),
                                             settings.report_values, "Information")
            output_transformer_report_detail(settings.my_logger,
                                             "Restriction Information {0} to {1}".format(src_fc, target_fc),
                                             settings.report_errors, "Issues")
            settings.results_xl.save_results_items_to_report("Restriction", src_fc, target_fc, where,
                                                             settings.report_values, transform_rest_count + 1)
            settings.results_xl.save_results_items_to_report("Restriction", src_fc, target_fc, where,
                                                             settings.report_errors, transform_rest_count + 1, True)

        # End for src_fc, target_fc, mapping, where

    # End with insert
    settings.results_xl.done_save()

    ###### RESTRICTION METADATA INFO - END


def get_current_meta_records(settings, inputtargetmetatable, MetaTblFieldsDefaultMapping, resultstype, outputType):
    # if clearexisting:
    if resultstype == "Restriction":
        del settings.saved_restrictioninfo[:]
    else:
        del settings.saved_entity_collection_metadata[:]
    metainfo_count = 0
    try:
        ######
        ###### METADATA INFO - Get information records and save them to specific target
        ######
        targetpathname = settings.targetdb.get_root_for_fc_subtypes(
            inputtargetmetatable)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
        target_fields = [targetff for src, targetff, transformer in MetaTblFieldsDefaultMapping if
                         targetff is not None and targetff != 'None']

        settings.targetdb.identify_actual_fields_list_matching_schema(outputType, inputtargetmetatable, target_fields)
        # settings.targetdb.updated_fields
        if len(settings.targetdb.updated_fields) == 0:
            return

        # Get all current target metadata info content, from the TARGET db save to an internally stored list
        with arcpy.da.SearchCursor(targetpathname, settings.targetdb.updated_fields, None) as rows:
            for row in rows_as_dicts(rows):
                metainfo_count += 1
                if resultstype == "Restriction":
                    settings.saved_restrictioninfo.append(row)
                else:
                    settings.saved_entity_collection_metadata.append(row)

    except:
        report_transfomer_issue_or_increment_count(
            settings, "Exception while in metadata {0} preparation.".format(outputType), "ERROR",
            None,
            sys.exc_info())
        return


# This function processes metadata from a source in which the metadata may be allocated across features
# to a target in which metadata is combined into the unique set of attribute values and stored in a table with
# an identifier and is then referenced from the features based on the unique identifier.
def process_metadata_records_spanning_feature_data(settings, inputtargetmetatable,
                                                   featureclasssettoprocess, MetaTblFieldsDefaultMapping,
                                                   TargetDefaultFieldValues,
                                                   fandf, saved_meta_records_list, resultstype="Restriction",
                                                   add_default_record_if_target_empty=False,
                                                   outputType="Restriction Info"):
    settings.current_step = "{0} Feature Metadata".format(resultstype)
    mytimer = MyTimers(settings.maximum_records['processing_indicator_n_seconds'])

    if outputType == "Restriction Info":
        maximumforthistype = settings.maximum_records['saved_restrictioninfo']
    else:
        maximumforthistype = settings.maximum_records['saved_entity_collection_metadata']

    ######
    targetpathname = settings.targetdb.get_root_for_fc_subtypes(
        inputtargetmetatable)
    # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
    metatarget_fields = [targetff for targetff in TargetDefaultFieldValues if
                         targetff is not None and targetff != 'None']

    # This slows things down a LOT, if only we could just use the defined fields!
    ###settings.targetdb.identify_actual_fields_list_matching_schema(outputType, inputtargetmetatable, starting_fields)
    # settings.targetdb.updated_fields
    # We are not going to count on the targetdb.updated_fields staying set
    ##metatarget_fields = copy.deepcopy(settings.targetdb.updated_fields)   # We can't depend on the target to stay the target
    if len(metatarget_fields) == 0:
        return

    if resultstype != "Restriction":
        metainfo_begin = len(settings.saved_entity_collection_metadata)
        # Get all current target metadata info content, from the TARGET db save to an internally stored list
    #   with arcpy.da.SearchCursor(targetpathname, metatarget_fields, None) as rowssaved:
    #       for row in rows_as_dicts(rowssaved):
    #           metainfo_count += 1
    #           settings.saved_entity_collection_metadata.append(row)
    else:
        metainfo_begin = len(settings.saved_restrictioninfo)

    metainfo_count = metainfo_begin
    #   #   #

    #
    transform_rest_count = 0

    move_data_count = 0
    proc_data_error = 0
    proc_data_count = 0
    #
    # Now if we have no meta info default row
    with arcpy.da.InsertCursor(targetpathname, metatarget_fields) as new_rows_meta:

        #
        # If there were no records, add a default record
        #
        if metainfo_count == 0 and add_default_record_if_target_empty == True:
            settings.my_logger.info("Adding default metadata record to: {0}".format(targetpathname))
            newdictmeta = dict()
            newrowmeta = list()
            for fielddef in metatarget_fields:
                if fielddef not in TargetDefaultFieldValues:
                    report_transfomer_issue_or_increment_count(
                        settings, "Metadata field default undefined in mapping tables.", "WARN",
                        "Field {0}, Target feature: {1}.".format(fielddef, inputtargetmetatable))
                else:
                    newrowmeta.append(TargetDefaultFieldValues[fielddef])
                    newdictmeta[fielddef] = TargetDefaultFieldValues[fielddef]

            if resultstype == "Restriction":
                possmatch = compare_against_existing_metainfo_noxform(settings, newdictmeta,
                                                                      settings.saved_restrictioninfo)
                if possmatch is None:
                    new_rows_meta.insertRow(newrowmeta)
                    settings.saved_restrictioninfo.append(newdictmeta)
            else:
                possmatch = compare_against_existing_metainfo_noxform(settings, newdictmeta,
                                                                      settings.saved_entity_collection_metadata)
                if possmatch is None:
                    new_rows_meta.insertRow(newrowmeta)
                    settings.saved_entity_collection_metadata.append(newdictmeta)
        # end processin for default record addition

        # with a new search cursor, go through the feature classes
        # and record any metadata information records needed
        # adding them to an the target and to an internal dictionary
        # The mapping may or may not be a large list and we will likely 
        # have to supplement with defaults on many fields
        # expected in the final set up required fields: metatarget_fields
        #
        # The results are put into the target metadata table AND
        # are retained in memory in a list for lookup later on.
        # Upon feature processing and translation, a transform function will
        # be used to examine the list for a match to the metadata in the
        # feature and at that time, the UFI will be pulled and saved to
        # the target feature as the reference to the metadata
        for src_fc, target_fc, mapping, where in featureclasssettoprocess:
            target_fc = inputtargetmetatable  # Override the mapping spec, the target is always specified as per the function call parameter
            mytimer.set_initial_time(outputType)

            try:
                # Ensure the items in this loop are valid
                if where is None or where == 'None':
                    where = ''

                settings.current_mapping_group = outputType  # Name of the mapping group

                settings.sourcedb.set_fc(src_fc)  # Set the source database to the current feature class from the source
                settings.targetdb.set_fc(target_fc)  # Set the target database to the current target feature class
                sourcepathname = settings.sourcedb.get_root_for_fc_subtypes(
                    src_fc)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
                targetpathname = settings.targetdb.get_root_for_fc_subtypes(
                    target_fc)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc

                clear_transformer_functions_report(settings)
                clear_transformer_issues_report(settings)

                transform_rest_count += 1

                if outputType == "Restriction Info":
                    currentlen = len(settings.saved_restrictioninfo)
                else:
                    currentlen = len(settings.saved_entity_collection_metadata)

                if currentlen > maximumforthistype:
                    proc_data_error += 1
                    settings.my_logger.error(
                        "Error processing {1} portion of features or feature classes. Maximum of {0} records have been exceeded!". \
                            format(str(maximumforthistype), outputType))
                    break

                if (settings.sourcedb.is_featureclass_in_geodatabase_spec(
                        src_fc) == False or settings.targetdb.is_featureclass_in_geodatabase_spec(target_fc) == False):
                    if settings.debug_settings['report_feature_classes_defined_in_schema_but_not_in_database'] > 0:
                        proc_data_error += 1
                        report_transfomer_issue_or_increment_count(settings,
                                                                   "{0} subset translation, source or target is not present. ".format(
                                                                       outputType), "WARN",
                                                                   "Source {0} {2}; target {1} {3}. Skipping this mapping rule set.".format(
                                                                       src_fc, target_fc, (
                                                                           settings.sourcedb.is_featureclass_in_geodatabase_spec(
                                                                               src_fc)), (
                                                                           settings.targetdb.is_featureclass_in_geodatabase_spec(
                                                                               target_fc))))
                else:
                    src_fc_path = settings.sourcedb.get_root_for_fc_subtypes(src_fc)
                    src_fieldsBB = [src for src, target, transformer in mapping if
                                    src is not None and src.lower() != 'none']
                    settings.sourcedb.identify_actual_fields_list_matching_schema("Source", src_fc, src_fieldsBB)
                    # settings.sourcedb.updated_fields

                    # Set the source and target fields to the updated lists since these represent the fields in the DATABASE vs the schema specification
                    # and the database may be a GRLS level while the schema specification is a GGDM Composite.
                    src_fields = copy.deepcopy(settings.sourcedb.updated_fields)

                    settings.my_logger.info(
                        'Processing {0} metadata portion of table or feature class: {1} with target {2}'.format(
                            outputType, src_fc, target_fc))
                    if len(src_fields) == 0:
                        report_transfomer_issue_or_increment_count(
                            settings,
                            "No fields found in source database in common with specified geodatabase schema type.",
                            "WARN", " Feature {0}".format(src_fc))
                        continue

                    # search for this feature class and get fields and rows for the subset metadata
                    with arcpy.da.SearchCursor(src_fc_path, src_fields, '') as rows_metaproc:

                        # For each feature instance in the source geodatabase - we should now have a subset of metadata fields in a row
                        for rowAA in rows_as_dicts(rows_metaproc):
                            # current_static_ufi = "0"    # If we are generating UFIs this is a trigger to generate a new UFI
                            settings.sourcedb.set_f_code(get_value_from_dict_nocase("F_CODE", rowAA))
                            settings.targetdb.set_f_code(TargetDefaultFieldValues['F_CODE'])

                            proc_data_count += 1

                            # start add of new metadata record
                            newdictmeta = dict()

                            # We may have a mapping from the mapping, plus we have defaults from the metadata Table mapping. Merge them into one
                            for srcmm, targetmm, transformermm in MetaTblFieldsDefaultMapping:
                                proc_data_error += process_fields_metadata_target(settings,
                                                                                  srcmm, targetmm, transformermm,
                                                                                  newdictmeta, rowAA,
                                                                                  TargetDefaultFieldValues,
                                                                                  src_fc, target_fc)
                            # Be sure we have any remaining defaults that are required
                            for ffldx in TargetDefaultFieldValues:
                                if ffldx not in newdictmeta:
                                    newdictmeta[ffldx] = TargetDefaultFieldValues[ffldx]

                            if 'SCVN' in newdictmeta and newdictmeta['SCVN'] == 'No Information':
                                newdictmeta['SCVN'] = settings.schema_version[settings.gettarget_db_version()]

                                # Now see if we already have this metadata, and if so we will use the existing UFI, otherwise we will use this new one.
                            # So this function will compare this newdictmeta set of information against ALL of the currently stored metadata
                            # looking for a potential match (with the exception of UFI and F_CODE fields and shape).
                            if resultstype == "Restriction":
                                posUFI = compare_against_existing_metainfo_noxform(settings, newdictmeta,
                                                                                   settings.saved_restrictioninfo)
                            else:
                                posUFI = compare_against_existing_metainfo_noxform(settings, newdictmeta,
                                                                                   settings.saved_entity_collection_metadata)

                            if posUFI is None:
                                # Verify length of target fields and row prior to inserting row into target
                                output_row_out = list()  # this is the returned list, a new row ready for insertion into the target
                                if verify_row(settings, target_fc, metatarget_fields, output_row_out, newdictmeta,
                                              False):
                                    #  This call will write the output_row into the target Esri Geodatabase
                                    if resultstype == "Restriction":
                                        metainfo_count += 1
                                        new_rows_meta.insertRow(output_row_out)
                                        settings.saved_restrictioninfo.append(newdictmeta)
                                    else:
                                        metainfo_count += 1
                                        new_rows_meta.insertRow(output_row_out)
                                        settings.saved_entity_collection_metadata.append(newdictmeta)
                                    move_data_count += 1
                                else:
                                    proc_data_error += 1
                                    report_transfomer_issue_or_increment_count(settings,
                                                                               "{0} processing did not generate correct number of fields.".format(
                                                                                   outputType), "ERROR",
                                                                               "Need {0} fields, but {1} generated. Target fields: {2} current row {3}". \
                                                                               format(len(metatarget_fields),
                                                                                      len(output_row_out),
                                                                                      str(metatarget_fields),
                                                                                      str(output_row_out)))

                                # We are busy adding records, be sure we are not getting carried away!
                                if outputType == "Restriction Info":
                                    currentlen = len(settings.saved_restrictioninfo)
                                else:
                                    currentlen = len(settings.saved_entity_collection_metadata)

                                if currentlen > maximumforthistype:
                                    proc_data_error += 1
                                    settings.my_logger.error(
                                        "Error processing {1} rows for metadata in features or feature classes. Maximum of {0} records have been exceeded!". \
                                            format(str(maximumforthistype), outputType))
                                    break

                            # End if posUFI

                            if mytimer.check_for_elapsed_time_gt_set_interval(outputType):
                                settings.my_logger.info(
                                    "\t\t...processing... current count: {0}".format(proc_data_count))

                            if settings.config_setting_flag['break_after_n_instances'] > 0 and \
                                    (move_data_count >= settings.config_setting_flag['break_after_n_instances'] or \
                                     settings.config_setting_flag['break_after_n_instances'] <= proc_data_count):
                                break

                        # End for source row in dictionary (perform translations)
                    # End with

                # End Else


            except:
                proc_data_error += 1
                report_transfomer_issue_or_increment_count(
                    settings, "Exception while translating row.", "ERROR",
                    "Source feature: {0}, Target feature: {1}.".format(src_fc, target_fc),
                    sys.exc_info())
            #
            mytimer.set_end_time(outputType)
            if move_data_count > 0 or proc_data_error > 0 or metainfo_count > 0:
                settings.report_values['Processing Time Overall (Seconds)'] = mytimer.get_elapsedtime(outputType)

            output_transformer_report_detail(settings.my_logger, "{2} {0} to {1}".format(src_fc, target_fc, outputType),
                                             settings.report_values, "Information")
            output_transformer_report_detail(settings.my_logger, "{2} {0} to {1}".format(src_fc, target_fc, outputType),
                                             settings.report_errors, "Issues")
            settings.results_xl.save_results_items_to_report(resultstype, src_fc, target_fc, where,
                                                             settings.report_values, transform_rest_count + 1)
            settings.results_xl.save_results_items_to_report(resultstype, src_fc, target_fc, where,
                                                             settings.report_errors, transform_rest_count + 1, True)

        # End for src_fc, target_fc, mapping, where

    # End with insert

    settings.results_xl.done_save()

    ###### ANY METADATA INFO - END


def get_existing_entity_collection_meta(settings, lookindb, inputtargetmetatable, EntityCollectionFields,
                                        reportvals=True):
    # if EntityCollectionFields not in globals():
    #    settings.my_logger.error("EntityCollectionFields is undefined but required for this step.")
    #    return

    if lookindb.is_featureclass_in_geodatabase_spec(inputtargetmetatable):
        targetpathname = lookindb.get_root_for_fc_subtypes(
            inputtargetmetatable)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
        target_fields_for_ec = [target for target in EntityCollectionFields if target is not None and target != 'None']
        metainfo_count = 0
        # clear_transformer_functions_report(settings)
        # clear_transformer_issues_report(settings)

        # be sure we have the correct set of fields for this feature
        lookindb.identify_actual_fields_list_matching_schema("Target EC", inputtargetmetatable, target_fields_for_ec)
        # lookindb.updated_fields is set
        metafields_to_get = copy.deepcopy(lookindb.updated_fields)
        if len(metafields_to_get) > 0:
            # Get all current target metadata info content
            with arcpy.da.SearchCursor(targetpathname, metafields_to_get, None) as rowsfromECtable:
                for rowinEC in rows_as_dicts(rowsfromECtable):
                    if rowinEC not in settings.saved_entity_collection_metadata:
                        metainfo_count += 1
                        settings.saved_entity_collection_metadata.append(rowinEC)

    #  if reportvals:
    #      settings.my_logger.info( "Pre-existing entity collection metadata records in {1}: {0}".\
    #          format(len(settings.saved_entity_collection_metadata), inputtargetmetatable))
    #      if (len(settings.saved_entity_collection_metadata) > 0):
    #          show_output_saved_meta( "Pre-existing entity collection metadata", settings.my_logger.info, settings.saved_entity_collection_metadata)

    return None


# This function processes only entity collection metadata. It assumes data model schema
# in which metadata has been stored on every feature and needs to be distributed to alternative target
# locations. Additionally, we want to save the mdatata information locally so that the unique identifier for
# it can be recorded in the feature. This function was developed to operate specifically on
# entity collection information.
# For each feature class mapping table definition:
#    Validate stuff
#    Prepare an insertion row into the target:
#    Get each source row from each source feature class or table and with that source row:
#        Initialize a target row
#        For each field mapping defined, perform the field mapping to get a target row value
#        Write the target row to the target database
# 
def process_entity_collection_meta(settings, sourcedb, targetdb, inputtargetmetatable,
                                   featureclasssettoprocess, fandf, reportid, add_default_record_if_target_empty=False):
    if not settings.targetdb.is_featureclass_in_geodatabase_spec(inputtargetmetatable):
        return
    ######
    ###### METADATA INFO - Get information records and save them to specific target
    ######
    settings.current_step = "Entity Collection Metadata"
    mytimer = MyTimers(settings.maximum_records['processing_indicator_n_seconds'])

    targetdb.set_fc(inputtargetmetatable)
    targetpathname = targetdb.get_root_for_fc_subtypes(
        inputtargetmetatable)  # os.path.join(inputtargetbasedir, inputtargetmetatable)
    target_fields = [target for src, target, transformer in fandf.EntityCollectionFields if
                     target is not None and target.lower() != 'none']

    clear_transformer_functions_report(settings)
    clear_transformer_issues_report(settings)

    metainfo_count = 0
    transform_ec_count = 0

    # be sure we have the correct set of fields for this feature
    targetdb.identify_actual_fields_list_matching_schema("Target EC Processing", inputtargetmetatable, target_fields)
    # targetdb.updated_fields has been set
    target_fields_final = targetdb.updated_fields

    # Get all current target metadata info content, from the current target
    with arcpy.da.SearchCursor(targetpathname, target_fields_final, None) as rows:
        for row in rows_as_dicts(rows):
            metainfo_count += 1
            settings.saved_entity_collection_metadata.append(row)

    settings.my_logger.info("Pre-existing entity collection metadata records in {1}: {0}". \
                            format(len(settings.saved_entity_collection_metadata), inputtargetmetatable))
    if len(settings.saved_entity_collection_metadata) > 0:
        show_output_saved_meta("Pre-existing entity collection metadata",
                               settings.my_logger.info, settings.saved_entity_collection_metadata)

    #
    # 
    # Prepare an insertion Cursor to the target table - the inputtargetmetatable
    # 
    with arcpy.da.InsertCursor(targetpathname, target_fields_final) as new_rows:
        targetdb.output_row.clear()  # this is a dictionary represntation of the target row values

        # If there were no records, add a default record  Now if we have no meta info default row
        if metainfo_count == 0 and add_default_record_if_target_empty == True:

            settings.my_logger.info("Adding default entity collection metadata record to: {0}". \
                                    format(targetpathname))

            for field in target_fields_final:
                targetdb.output_row[field] = fandf.EntityCollectionTargetDefaultFieldValues[field]

            new_row = list()
            if verify_row(settings, inputtargetmetatable, target_fields_final, new_row, targetdb.output_row,
                          True):
                if len(settings.saved_entity_collection_metadata) < settings.maximum_records[
                    'saved_entity_collection_metadata']:
                    settings.saved_entity_collection_metadata.append(targetdb.output_row)
                    new_rows.insertRow(new_row)
                else:
                    report_transfomer_issue_or_increment_count(
                        settings, "Maximum saved records in entity collection has been reached: {0}". \
                            format(settings.maximum_records['saved_entity_collection_metadata']), "ERROR")
                    return

        # with a new search cursor, go through the feature classes
        # and record any metadata information records needed
        # adding them to an the target and to an internal dictionary
        #
        #  EntityCollectionMetaConvertSubsetFields
        #

        for src_fc, target_fc, mapping, where in featureclasssettoprocess:

            mytimer.set_initial_time('entitycollection')
            move_data_count = 0  # track how many records we move
            proc_error_count = 0
            proc_data_count = 0
            transform_ec_count += 1

            # Ensure the items in this loop are valid
            if where is None or where.lower() == 'none':
                where = ''

            settings.current_mapping_group = "Entity Collection Feature Subset"  # Name of the mapping group

            sourcedb.set_fc(src_fc)  # Set the source database to the current feature class from the source
            # targetdb.set_fc( target_fc )                                 # Set the target database to the current target feature class
            sourcepathname = sourcedb.get_root_for_fc_subtypes(
                src_fc)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
            # targetpathname = targetdb.get_root_for_fc_subtypes(target_fc)    # Will return the dataset root if this is a feature class, otherwise will return a table root + fc

            # settings.saved_mem.clear()           # We can start with a clean dictionary of MEM content for each feature class
            # settings.saved_oth.clear()           # We can start with a clean dictionary of OTH content for each feature class

            clear_transformer_functions_report(settings)
            clear_transformer_issues_report(settings)

            adjust_debug_settings_based_on_feature_class(settings, sourcedb, targetdb)

            if (settings.sourcedb.is_featureclass_in_geodatabase_spec(
                    src_fc) and settings.targetdb.is_featureclass_in_geodatabase_spec(inputtargetmetatable)):

                # Output a line to the logger stating we are going to prcess this source and target feature class
                settings.my_logger.info(
                    'Processing entity collection metadata subset for table or feature class: {0} with target: {1}'. \
                        format(src_fc, inputtargetmetatable))

                # Verify that we actually got a mapping list of fields specified correctly as a list python type.  This extra check
                # is needed for the possiblity of a excel specified mapping. The value has to be corrected a bit to work in python.
                if not isinstance(mapping, list):
                    mapping = getattr(fandf, mapping)

                # Get the source fields and target fields from the mapping specification - bascially these lists are hard coded here. 
                # But we will intersect these against the actual database.
                src_fields = [src for src, target, transformer in mapping if src is not None and src.lower() != 'none']
                # target_fields = [target for src, target, transformer in mapping if target is not None and target != 'None']
                sourcedb.identify_actual_fields_list_matching_schema("Source EC", src_fc, src_fields)
                # targetdb.identify_actual_fields_list_matching_schema("Target EC", inputtargetmetatable, target_fields)
                # sourcedb.updated_fields is set
                # targetdb.updated_fields is set

                if len(sourcedb.updated_fields) == 0:
                    report_transfomer_issue_or_increment_count(settings,
                                                               "Source or target has no fields to translate.", "ERROR",
                                                               "Source fields list length {0}, target fields list length {1}".format(
                                                                   len(sourcedb.updated_fields),
                                                                   len(targetdb.updated_fields)))
                    break

                if len(settings.saved_entity_collection_metadata) > settings.maximum_records[
                    'saved_entity_collection_metadata']:
                    report_transfomer_issue_or_increment_count(
                        settings,
                        "Error Processing entity collection metadata portion of features or feature classes. Maximum of {0} records have been exceeded!".format(
                            str(settings.maximum_records['saved_entity_collection_metadata'])), "ERROR")
                    break

                # search for this feature class or table and get fields and rows for the subset metadata
                with arcpy.da.SearchCursor(sourcepathname, sourcedb.updated_fields, where) as rows:

                    # for each row in this feature class
                    for sourcedb.row in rows_as_dicts(rows):

                        try:
                            sourcedb.set_f_code(get_value_from_dict_nocase("F_CODE", sourcedb.row))
                            targetdb.set_f_code(targetdb.get_primary_fcode_for_featureclass(inputtargetmetatable))
                            proc_data_count += 1

                            # current_static_ufi = "0"    # If we are generating UFIs this is a trigger to generate a new UFI
                            newtemprow = {}
                            for src, target, transformer in fandf.EntityCollectionFCSubsetFields:
                                if sourcedb.row.has_key(src):
                                    newtemprow[src] = sourcedb.row[src]

                            if len(settings.saved_entity_collection_metadata) == 0:
                                posUFI = None
                            else:  # BREAK HERE
                                posUFI = compare_src_against_existing_metainfo(
                                    settings, newtemprow, settings.saved_entity_collection_metadata,
                                    'ENTITY_COLLECTION_METADATA_T', 'MetadataSrf',
                                    settings.fandf.EntityCollectionFields)
                                if settings.debug_settings['report_entity_collection_findings'] > 0:
                                    settings.my_logger.debug(
                                        "Entity collection meta initiate, possible UFI found: {0} for temprow: {1}". \
                                            format(str(posUFI), str(newtemprow)))

                            # As long as this UFI has not already been used,    
                            if posUFI is None:
                                # start add of new metadata record
                                targetdb.output_row.clear()  # Target is still Entity Collection Metadata

                                getufi = ""
                                for src, target, transformer in fandf.EntityCollectionFields:
                                    # same as: for target in target_fields:
                                    # We are really mainly dealing with the target and transformer here
                                    if target and target in target_fields_final:  # may be None such as for source fields that no longer exist
                                        if target.lower() == 'ufi':
                                            getufi = makerandomufi()
                                            targetdb.output_row[target] = getufi
                                        elif target.lower() in common_fields_geom_only:
                                            pass
                                            # settings.my_logger.debug( "YIKE: do something with metadata SHAPE, need to extend metadata with this shape?" )
                                        elif target.lower() == 'fcsubtype':
                                            targetdb.output_row['FCSubType'] = \
                                                fandf.EntityCollectionTargetDefaultFieldValues['FCSubType']
                                        elif target.lower() == 'f_code':
                                            if transformer is None:
                                                targetdb.output_row[target] = \
                                                    fandf.EntityCollectionTargetDefaultFieldValues[target]
                                            else:
                                                # NEW CODE NEW CODE
                                                # newfcode = call_transform_function[transformer](settings, sourcedb.row, target, src_fc, target_fc)
                                                newfcode = "NA"
                                                if target in sourcedb.row and sourcedb.row[target] is not None:
                                                    try:
                                                        newfcode = call_transform_function[transformer](settings,
                                                                                                        sourcedb.row,
                                                                                                        target, target,
                                                                                                        src_fc,
                                                                                                        inputtargetmetatable)
                                                    except:
                                                        report_transfomer_issue_or_increment_count(
                                                            settings,
                                                            "  Specified transformer {0} had an exception.".format(
                                                                transformer),
                                                            "ERROR",
                                                            "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                                                                format(transformer, target, target,
                                                                       sourcedb.row[
                                                                           target] if target in sourcedb.row else "",
                                                                       src_fc, inputtargetmetatable,
                                                                       targetdb.output_row[
                                                                           target] if target in targetdb.output_row else ""),
                                                            sys.exc_info())
                                                        proc_error_count += 1
                                                        newfcode = fandf.EntityCollectionTargetDefaultFieldValues[
                                                            target]
                                                    #
                                                # new - attempt to work with src:
                                                elif src in sourcedb.row and sourcedb.row[src] is not None:
                                                    try:
                                                        newfcode = call_transform_function[transformer](settings,
                                                                                                        sourcedb.row,
                                                                                                        src, target,
                                                                                                        src_fc,
                                                                                                        inputtargetmetatable)
                                                    except:
                                                        report_transfomer_issue_or_increment_count(
                                                            settings,
                                                            "  Specified transformer {0} had an exception.".format(
                                                                transformer),
                                                            "ERROR",
                                                            "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                                                                format(transformer, src, target,
                                                                       sourcedb.row[src] if src in sourcedb.row else "",
                                                                       src_fc, inputtargetmetatable,
                                                                       targetdb.output_row[
                                                                           target] if target in targetdb.output_row else ""),
                                                            sys.exc_info())
                                                        proc_error_count += 1
                                                        newfcode = fandf.EntityCollectionTargetDefaultFieldValues[
                                                            target]
                                                    #
                                                #
                                                if newfcode is None or newfcode == "NA":
                                                    targetdb.output_row[target] = \
                                                        fandf.EntityCollectionTargetDefaultFieldValues[target]
                                                else:
                                                    targetdb.output_row[target] = newfcode
                                                settings.targetdb.set_f_code(targetdb.output_row[target])
                                            #   #  Done with f_code
                                        else:
                                            # Check for the target field present in the source row
                                            if transformer is None and target in sourcedb.row:
                                                targetdb.output_row[target] = sourcedb.row[target]
                                            # Check for the target field not in the source row
                                            elif transformer is None and target not in sourcedb.row:
                                                targetdb.output_row[target] = None
                                                if settings.debug_settings['report_missing_source_data'] > 0:
                                                    report_transfomer_issue_or_increment_count(settings,
                                                                                               "Entity collection field processing issue",
                                                                                               "WARN",
                                                                                               "Entity collection field expected but not in source data. Source FC: {0} Field: {1} Target FC: {2} Field: {3}". \
                                                                                               format(src_fc, target,
                                                                                                      inputtargetmetatable,
                                                                                                      target))
                                            elif transformer is not None:
                                                if transformer in call_transform_function:
                                                    if src in sourcedb.row and sourcedb.row[src] is not None:
                                                        try:
                                                            targetdb.output_row[target] = call_transform_function[
                                                                transformer](settings, sourcedb.row, src, target,
                                                                             src_fc, inputtargetmetatable)
                                                        except:
                                                            report_transfomer_issue_or_increment_count(
                                                                settings,
                                                                "  Specified transformer {0} had an exception.".format(
                                                                    transformer),
                                                                "ERROR",
                                                                "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                                                                    format(transformer, src, target,
                                                                           sourcedb.row[
                                                                               src] if src in sourcedb.row else "",
                                                                           src_fc, inputtargetmetatable,
                                                                           targetdb.output_row[
                                                                               target] if target in targetdb.output_row else ""),
                                                                sys.exc_info())
                                                            proc_error_count += 1
                                                            targetdb.output_row[target] = \
                                                                fandf.EntityCollectionTargetDefaultFieldValues[target]
                                                        #
                                                    elif target in targetdb.updated_fields:
                                                        try:
                                                            targetdb.output_row[target] = call_transform_function[
                                                                transformer](settings, sourcedb.row, target, target,
                                                                             src_fc, inputtargetmetatable)
                                                        except:
                                                            report_transfomer_issue_or_increment_count(
                                                                settings,
                                                                "  Specified transformer {0} had an exception.".format(
                                                                    transformer),
                                                                "ERROR",
                                                                "Transformer {0} for source {1} to target {2} failed. Setting target to default". \
                                                                    format(transformer, "NA", target))
                                                            proc_error_count += 1
                                                            targetdb.output_row[target] = \
                                                                fandf.EntityCollectionTargetDefaultFieldValues[target]
                                                    else:
                                                        targetdb.output_row[target] = \
                                                            fandf.EntityCollectionTargetDefaultFieldValues[target]
                                                        report_transfomer_issue_or_increment_count(
                                                            settings,
                                                            "  Specified transformer {0} defined but no source or target. Setting to default.".format(
                                                                transformer),
                                                            "ERROR",
                                                            "Transformer {0} for source {1} to target {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                                                                format(transformer, src, target,
                                                                       sourcedb.row[src] if src in sourcedb.row else "",
                                                                       src_fc, inputtargetmetatable,
                                                                       targetdb.output_row[
                                                                           target] if target in targetdb.output_row else ""))

                                                    #
                                                else:
                                                    report_transfomer_issue_or_increment_count(settings,
                                                                                               "Entity collection transformer failure.",
                                                                                               "WARN",
                                                                                               "Entity collection transformer {0} for source {1} to target {2} is missing.".format(
                                                                                                   transformer, src,
                                                                                                   target), True)

                                            # End if elif

                                        # end field level checks
                                    # end if target is in target fields
                                    # Update the target f_code if it was set
                                    if target is not None and target.lower() == 'f_code' and target in targetdb.output_row:
                                        settings.targetdb.set_f_code(targetdb.output_row[target])
                                # end for src, target, transformer

                                if len(settings.saved_entity_collection_metadata) < settings.maximum_records[
                                    'saved_entity_collection_metadata']:
                                    settings.saved_entity_collection_metadata.append(targetdb.output_row)
                                    new_row_out = list()
                                    if verify_row(settings, inputtargetmetatable, target_fields_final, new_row_out,
                                                  targetdb.output_row, True):
                                        new_rows.insertRow(new_row_out)
                                        # Indicate we are writing data
                                        move_data_count += 1

                                    if settings.debug_settings['report_entity_collection_findings'] > 0:
                                        settings.my_logger.debug(
                                            "Entity collection meta initiate, added new row to saved meta {0}".format(
                                                str(targetdb.output_row)))

                                else:
                                    report_transfomer_issue_or_increment_count(settings,
                                                                               "Maximum saved records in entity collection has been reached: {0}".format(
                                                                                   settings.maximum_records[
                                                                                       'saved_entity_collection_metadata']),
                                                                               "ERROR")
                                # end else len less than max records

                                if settings.config_setting_flag['break_after_n_instances'] > 0 and \
                                        (settings.config_setting_flag['break_after_n_instances'] <= move_data_count or \
                                         settings.config_setting_flag['break_after_n_instances'] <= proc_data_count):
                                    break

                                if mytimer.check_for_elapsed_time_gt_set_interval('entitycollection'):
                                    settings.my_logger.info(
                                        "\t\t...processing... current count: {0}".format(proc_data_count))

                                    # end if possUFI is None

                            if settings.config_setting_flag['break_after_n_instances'] > 0 and \
                                    settings.config_setting_flag['break_after_n_instances'] < proc_data_count:
                                break


                        except:
                            report_transfomer_issue_or_increment_count(
                                settings, "  Failure processing entity collection metadata",
                                "ERROR", "Source {0} Target {1} Source row: {2} failed processing". \
                                    format(src_fc, inputtargetmetatable,
                                           str(sourcedb.row)),
                                sys.exc_info())

                    # end for each row

            # if not fields_are_missing: else this feature class is not in the data
            else:
                if settings.debug_settings['report_feature_classes_defined_in_schema_but_not_in_database'] > 0:
                    report_transfomer_issue_or_increment_count(settings, "No feature class or table", "WARN",
                                                               "Source {0} {2}; Target {1} {3}. Skipping this mapping rule set.".format(
                                                                   src_fc, inputtargetmetatable, (
                                                                       settings.sourcedb.is_featureclass_in_geodatabase_spec(
                                                                           src_fc)),
                                                                   (
                                                                       settings.targetdb.is_featureclass_in_geodatabase_spec(
                                                                           inputtargetmetatable))))

            mytimer.set_end_time('entitycollection')
            if proc_data_count > 0:
                settings.report_values['Processing Time Overall (Seconds)'] = mytimer.get_elapsedtime(
                    'entitycollection')

            output_transformer_report_detail(settings.my_logger,
                                             "Entity Collection {0} to {1}".format(src_fc, inputtargetmetatable),
                                             settings.report_values, "Information")
            output_transformer_report_detail(settings.my_logger,
                                             "Entity Collection {0} to {1}".format(src_fc, inputtargetmetatable),
                                             settings.report_errors, "Issues")

            settings.results_xl.save_results_items_to_report(reportid, src_fc, inputtargetmetatable, where,
                                                             settings.report_values, transform_ec_count + 1)
            settings.results_xl.save_results_items_to_report(reportid, src_fc, inputtargetmetatable, where,
                                                             settings.report_errors, transform_ec_count + 1, True)

            # end for src_fc
        # end with
        settings.results_xl.done_save()
        # End if ec in target
    ###### ENTITY COLLECTION META - END


def process_entity_collection_meta_ggdm30_to_21(settings, sourcedb, targetdb, inputtargetmetatable,
                                                featureclasssettoprocess, fandf, reportid,
                                                add_default_record_if_target_empty=False):
    ######
    ###### METADATA INFO - Get information records and save them to specific target
    ######
    if not targetdb.is_featureclass_in_geodatabase_spec(inputtargetmetatable):
        return
    settings.current_step = "Entity Collection Metadata"
    proc_data_count = 0
    move_data_count = 0
    mytimer = MyTimers(settings.maximum_records['processing_indicator_n_seconds'])
    targetpathname = targetdb.get_root_for_fc_subtypes(
        inputtargetmetatable)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
    # target_obj_path = os.path.join(inputtargetbasedir, inputtargetmetatable)
    target_fields = [target for src, target, transformer in fandf.EntityCollectionTargetFields if
                     target is not None and target != 'None']
    targetdb.identify_actual_fields_list_matching_schema("Target EC", inputtargetmetatable, target_fields)
    # targetdb.updated_fields is set
    metainfo_count = len(settings.saved_entity_collection_metadata)

    #
    # Now with this one we have to process each insert row based on the 
    # Saved Entity Collection Metadata.
    with arcpy.da.InsertCursor(targetpathname, targetdb.updated_fields) as new_rows:

        # If there were no records, add a default record
        if metainfo_count > 0:

            for row_src in settings.saved_entity_collection_metadata:
                proc_data_count += 1
                mytimer.set_initial_time('entitycollection')
                # if we did not get a source type or source date set by parsing all feature instances, set them to default
                # If we do not do this, then we may end up with null values as source type and date which is not optimal.
                if 'ZI001_SRT' not in row_src:
                    row_src['ZI001_SRT'] = -999999  # Default
                if 'ZI001_SDV' not in row_src:
                    row_src['ZI001_SDV'] = 'noInformation'  # Default

                settings.my_logger.info(
                    "Adding entity collection metadata records with translations to: {0}".format(targetpathname))
                targetdb.output_row.clear()

                for field in targetdb.updated_fields:
                    target_fcode_ec = ""
                    # THis is going to be nasty, we have to look at each possible field and manually translate it.
                    value = None
                    origvalue = None
                    if field in row_src:
                        origvalue = row_src[field]
                    if field in row_src and field.lower() in common_fields_geom_only:
                        value = row_src[field]
                    if field in row_src and field in ['UFI', 'ZI001_SPS', 'ZI001_VSD', 'ZI001_VSN', 'ZI004_PRE']:
                        value = row_src[field]

                    else:  # assumptions about source fields and required target fields
                        if field in ['AHA', 'AVA', 'DQS', 'EQS', 'EVA', 'HVA', 'MDE', 'RCG']:
                            value = call_transform_function['transform_set_EntityCollection_field_to_default'](settings,
                                                                                                               row_src,
                                                                                                               field,
                                                                                                               field,
                                                                                                               'Saved Entity Collection',
                                                                                                               'Target Entity Collection')

                        elif field == 'ZI001_VSC':  # Vertical Source Category - convert from enum to code list
                            # do something
                            value = call_transform_function['transformZI001_VSCfromenumtocodelist'](settings, row_src,
                                                                                                    field, field,
                                                                                                    'Saved Entity Collection',
                                                                                                    'Target Entity Collection')
                            if value is None:
                                value = "noInformation"

                        elif field == 'ZI001_NSP':  # Non spatial Source Type
                            # do something
                            value = call_transform_function['transform_ZI001_NSP'](settings, row_src, field, field,
                                                                                   'Saved Entity Collection',
                                                                                   'Target Entity Collection')
                            if value is None:
                                value = "noInformation"

                        elif field == 'ZI001_SSY':  # Source Type for Spatial MD
                            # do something
                            value = call_transform_function['transform_ZI001_SSY'](settings, row_src, field, field,
                                                                                   'Saved Entity Collection',
                                                                                   'Target Entity Collection')
                            if value is None:
                                value = "noInformation"

                        elif field == 'ZI001_NSD' or field == 'ZI001_SSD':  # Source Date Spatial and Non-Spatial
                            # do something
                            if 'ZI001_SDV' in row_src:
                                value = row_src['ZI001_SDV']
                            if value is None:
                                value = "No Information"

                        elif field == 'ZI001_NSN' or field == 'ZI001_SSN':  # Source Description Spatial and Non-Spatial
                            # do something
                            if 'ZI001_SDP' in row_src:
                                value = row_src['ZI001_SDP']
                            if value is None:
                                value = "No Information"

                        elif field == 'F_CODE':
                            # do something
                            value = call_transform_function['transformMetadataSrfCollectionSrf_F_CODE'](settings,
                                                                                                        row_src, field,
                                                                                                        field,
                                                                                                        'Saved Entity Collection',
                                                                                                        'Target Entity Collection')
                            target_fcode_ec = value

                        elif field == 'FCSubtype':
                            # do something
                            value = call_transform_function['transform_FCSubtype'](settings, row_src, field, field,
                                                                                   'Saved Entity Collection',
                                                                                   'Target Entity Collection')

                        elif field == 'OTH':
                            # do something
                            value = call_transform_function['transform_OTH'](settings, row_src, field, field,
                                                                             'Saved Entity Collection',
                                                                             'Target Entity Collection')

                        else:
                            # put field in with default value
                            value = settings.targetdb.field_set_to_default(target_fcode_ec, field, 'CollectionSrf')
                        # end if

                        if value is None:
                            report_transfomer_issue_or_increment_count(settings,
                                                                       "Entity Collection processing, value is None. ",
                                                                       "WARN",
                                                                       "Field {0} original value {1}. Will try to set to default.".format(
                                                                           field, origvalue))
                            value = settings.targetdb.field_set_to_default(target_fcode_ec, field, 'CollectionSrf')
                    # end if field in target

                    if value is not None:
                        targetdb.output_row[field] = value

                    if settings.debug_settings['report_entity_collection_findings'] > 0:
                        settings.my_logger.info(
                            "Entity Collection Translation field {0} from {2} to {1}".format(field, strout(value),
                                                                                             strout(origvalue)))

                # end for each field
                new_row_out = list()  # this is the returned list, a new row ready for insertion into the target
                if verify_row(settings, inputtargetmetatable, targetdb.updated_fields, new_row_out, targetdb.output_row,
                              True):
                    #  This call will write the new_row into the target Esri Geodatabase
                    new_rows.insertRow(new_row_out)
                    move_data_count += 1

                if mytimer.check_for_elapsed_time_gt_set_interval('entitycollection'):
                    settings.my_logger.info("\t\t...processing... current count: {0}".format(proc_data_count))

            # end for each source row

        # end if metadata count > 0   
    # end with Insert Cursor
    mytimer.set_end_time('entitycollection')
    settings.report_values['Translate and Move - Instances Moved'] = move_data_count
    settings.report_values['Translate Processing - Instances Processed'] = proc_data_count
    if proc_data_count > 0:
        settings.report_values['Processing Time Overall (Seconds)'] = mytimer.get_elapsedtime('entitycollection')

    output_transformer_report_detail(settings.my_logger, inputtargetmetatable, settings.report_values, "Information")
    output_transformer_report_detail(settings.my_logger, inputtargetmetatable, settings.report_errors, "Issues")


# This function processes only entity collection metadata. It assumes data model schema
# in which metadata has been stored in two separate features Spatial Metadata Collection and Non-Spatial Metadata Collect.
# This one does not look at ALL geometric features, just the two needed.
# Additionally, we want to save the mdatata information locally so that the unique identifier for
# it can be recorded in the feature. This function was developed to operate specifically on
# entity collection information.

def process_entity_collection_meta_from_21(settings, sourcedb, targetdb, inputtargetmetatable, tableoutname,
                                           fieldstoprocess, fandf, reportid, add_default_record_if_target_empty=False):
    ######
    ###### METADATA INFO - Get information records and save them to specific target
    ######
    defaultPoly = arcpy.Polygon(
        arcpy.Array([arcpy.Point(0.0, 0.0, 1), arcpy.Point(0.0, 1.0, 1), arcpy.Point(1.0, 1.0, 1)]))

    # With each saved entity collection row,
    # Go through the fields, and if any of them are translated or saved to a new destination field, 
    # save the translation result and save new destination fields
    for row in settings.saved_entity_collection_metadata:

        if 'transformed' not in row:
            row['save_as_table'] = True  # default to save this record into the table
            for srcfield, target, transformer in fieldstoprocess:

                if srcfield is not None and srcfield in row:
                    newval = row[srcfield]
                    # The srcfield is defined and in the collected data
                    if transformer is not None:
                        newval = call_transform_function[transformer](settings, row, srcfield, target, "CollectionSrf",
                                                                      "MetadataSrf")
                        if newval is not None:
                            row[srcfield] = newval

                    # be sure we have the defined target field (it may be different)
                    if target is not None and target != srcfield and target not in row:
                        row[target] = row[srcfield]

                elif srcfield is not None and srcfield not in row:
                    row[srcfield] = None

                elif srcfield is None and target is not None:
                    if transformer is not None:
                        newval = call_transform_function[transformer](settings, row, target, target, "CollectionSrf",
                                                                      "MetadataSrf")
                        if newval is not None:
                            row[target] = newval
                    else:
                        row[target] = None

                if srcfield == 'SHAPE@' and row[srcfield] is not None:
                    row['save_as_table'] = False  # looks like there is a geometry for this
                    polarea = row['SHAPE@Area']
                    pollen = row['SHAPE@Length']
                    if defaultPoly.equals(row[srcfield]) == True or (
                            (polarea > 0.5 and polarea < 0.6) and (pollen > 3.4 and pollen < 3.5)):
                        # This is the default polygon, the entity collection should be saved as a TABLE.
                        row['save_as_table'] = True  # nope. the geometry is just the default
                # end if srcfield == 'SHAPE@'
            # end for srcfield, target, transformer
            row['transformed'] = True
        # end if transformed in row  (i.e. have we already transformed this row?)
    # end for row in entity collection meta

    if not targetdb.is_featureclass_in_geodatabase_spec(inputtargetmetatable):
        return

    # Save to feature class (spatial data)
    targetpathname = targetdb.get_root_for_fc_subtypes(
        inputtargetmetatable)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
    # target_obj_path = os.path.join(inputtargetbasedir, inputtargetmetatable)
    target_fields = [target for src, target, transformer in fandf.EntityCollectionFields if
                     target is not None and target != 'None']
    metainfo_count = 0
    clear_transformer_functions_report(settings)
    clear_transformer_issues_report(settings)

    # be sure we have the correct set of fields for this feature
    targetdb.identify_actual_fields_list_matching_schema("Target EC", inputtargetmetatable, target_fields)
    # targetdb.updated_fields

    with arcpy.da.InsertCursor(targetpathname, targetdb.updated_fields) as new_rows:

        for row in settings.saved_entity_collection_metadata:

            if not row['save_as_table']:
                #  Now save this row data to the target list and then to the target database.
                new_row = []
                for field in targetdb.updated_fields:
                    if field in row:
                        new_row.append(row[field])
                    else:
                        new_row.append('0')
                # insert the row data
                new_rows.insertRow(new_row)

            # end if save as table is FALSE
        # end for row in saved entity collection

    # Save to Table
    targetpathname = targetdb.get_root_for_fc_subtypes(
        tableoutname)  # Will return the dataset root if this is a feature class, otherwise will return a table root + fc
    # target_obj_path = os.path.join(tablebasedir, tableoutname)
    target_fields = [target for src, target, transformer in fandf.EntityCollectionFields if
                     target is not None and target != 'None']
    metainfo_count = 0
    clear_transformer_functions_report(settings)
    clear_transformer_issues_report(settings)

    # be sure we have the correct set of fields for this feature
    targetdb.identify_actual_fields_list_matching_schema("Target EC", tableoutname, target_fields)
    # targetdb.updated_fields

    with arcpy.da.InsertCursor(targetpathname, targetdb.updated_fields) as new_rows:

        for row in settings.saved_entity_collection_metadata:

            if row['save_as_table']:
                #  Now save this row data to the target list and then to the target database.
                new_row = []
                for field in targetdb.updated_fields:
                    if field in row:
                        new_row.append(row[field])
                    else:
                        new_row.append('0')
                # insert the row data
                new_rows.insertRow(new_row)

            # end if save as table is true
        # end for row in saved entity collection

    ## Just report the entity catalog saved content
    output_transformer_report_detail(settings.my_logger, "Entity Collection", settings.report_values, "Information")
    output_transformer_report_detail(settings.my_logger, "Entity Collection", settings.report_errors, "Issues")

    ###### ENTITY COLLECTION META 21 - END


def save_final_derived_records(settings, targetdb, reportid, stepname, generate_report=False):
    if len(targetdb.saved_derived_records) == 0:
        return
    targetdb.save_derived_records_to_database(settings, reportid, stepname, generate_report)

##
## Compare source to target, data content
