﻿
""" 
 ************************************************************************************************
 **                                                                                            **
 **  File: StandardUtilities.py                                                                 **
 **  Copyright (c) September 2018 by Leidos                                                     **
 **  Produced under U.S. Government contract W5J9CQ-12-D-0004  Task Order: 073                 **
 **  As per DFARS, the U.S. Government is granted unlimited rights to use this work            **
 **  in any manner and for any purpose.                                                        **
 **                                                                                            **
 **  Summary:   This file provides for standard functions that may be utilized by this tool.   **
 **                                                                                            **
 **  Developed by Leidos, Inc., High Resolution ISR Systems Division, Geospatial Programs      **
 **                                                                                            **
 **  Development History:                                                                      **
 **  September 2018 / Annette Janett Filer                                                     **
 **                                                                                            **
 ************************************************************************************************
"""

import sys
import os
import copy

import arcpy
import uuid
import glob
import traceback

import datetime
import time
import logging

import gc


def dump_garbage(settings, whereat = ""):
    """
    clear out the garbage and report if necessary
    """
        
    # force collection
    unreachablenum = gc.collect()

    if settings.debug_settings['instrumentation_output'] > 0:
        settings.my_logger.info("GARBAGE OBJECTS: {0}  {1}".format(unreachablenum, whereat))
        for x in gc.garbage:
            s = str(x)
            if len(s) > 80: s = s[:80]
            for rr in gc.get_referrers(x):
                if "arcpy.arcobjects" in str(rr):
                    settings.my_logger.info("Ref: arcpy.arcobjects Type: {0} Name: {1}".format(type(s), s))
                else:
                    settings.my_logger.info("Ref: {2} Type: {0} Name: {1}".format(type(s), s, rr))


class MyTimers():
    def __init__(self, reportintervalseconds):
        self.numtimers = 0
        self.timerdictstart = dict()
        self.timerdictend = dict()
        self.timerdictcheck = dict()
        self.reportseconds = reportintervalseconds

    def set_initial_time(self, timername):
        self.timerdictstart[timername] = time.time()
        self.timerdictcheck[timername] = self.timerdictstart[timername]

    def set_end_time(self, timername):
        self.timerdictend[timername] = time.time()

    def check_for_elapsed_time_gt_set_interval(self, timername):
        curtimeval = time.time()
        if timername not in self.timerdictcheck:
            return False
        else:
            etimeval =  float("%.4f" % (curtimeval - self.timerdictcheck[timername]))
            if etimeval > self.reportseconds:
                self.timerdictcheck[timername] = curtimeval
                return True
        return False

    def get_elapsedtime(self, timername):
        if timername not in self.timerdictend or timername not in self.timerdictstart:
            return -999999.0
        etimeout = float("%.4f" % (self.timerdictend[timername] - self.timerdictstart[timername]))
        self.timerdictstart[timername] = None
        self.timerdictend[timername] = None 
        return etimeout 
    
                   



def normalize_field(field):
    if field is not None:
        if field.lower() == 'none' or field.lower() == 'na' or field == '':
            return None
    return field


def strout(invalue):
    if invalue is None:
        return ""
    if isinstance(invalue, str):
        return invalue
    elif isinstance(invalue, unicode):
        try:
            outval = str(invalue)
            return outval
        except:
            return "--- unicode string characters detected ---"
    else:
        return str(invalue)


def output_dictionary_to_logger(my_logger, introduction, contentin):
    my_logger.info(introduction)
    if isinstance(contentin, dict):
        for iikey in contentin:
            if isinstance(iikey, dict):
                for iikey2 in iikey:
                    my_logger.info("\t\tkey: {0} value: {1}".format(strout(iikey2), strout(iikey[iikey2])))
            elif isinstance(iikey, list):
                jj = 0
                for iikey2 in iikey:
                    my_logger.info("\t\tindex: {0} value: {1}".format(jj, strout(iikey2)))
                    jj+=1
            else:
                my_logger.info("\tkey: {0} value: {1}".format(strout(iikey), strout(contentin[iikey])))
    elif isinstance(contentin, list):
        ii = 0
        for iikey in contentin:
            my_logger.info("\tindex: {0}".format(ii))
            if isinstance(iikey, dict):
                for iikey2 in iikey:
                    my_logger.info("\t\tkey: {0} value: {1}".format(strout(iikey2), strout(iikey[iikey2])))
            elif isinstance(iikey, list):
                jj = 0
                for iikey2 in iikey:
                    my_logger.info("\t\tindex: {0} value: {1}".format(jj, strout(iikey2)))
                    jj+=1
            else:
                my_logger.info("\tindex: {0} value:{1}".format(ii, strout(iikey)))
            ii+=1
    else:
        my_logger.info("Item is neither a list or dictionary.")



def is_item_dict(initem):
    if isinstance(initem, dict):
        return True
    else:
        return False


def is_item_str(initem):
    if isinstance(initem, str) or isinstance(initem, unicode):
        return True
    else:
        return False


def get_int_from_str(initem):
    # sometimes the input value is a floating point text field!  Bad.
    try: 
        inval = int(initem)
    except ValueError:
        inval = int(float(initem))
    return inval

def is_item_in_list_nocase(initem, inlist):
    if initem is None:
        return False
    checkset = set([field.lower() for field in inlist])
    if initem.lower() in checkset:
        return True
    else:
        return False

def is_item_in_dict_nocase(initem, indict):
    if initem is None:
        return False
    aalower = {k.lower():v for k,v in indict.items()}
    if initem.lower() in aalower:
        return True
    else:
        return False

def get_value_from_dict_nocase(inkey, indict):
    try:
        aalower = {k.lower():v for k,v in indict.items()}
        return aalower[inkey.lower()]
    except:
        return 'NA'


#def get_source_F_CODE(row):
#    if 'F_CODE' in row:
#        return row['F_CODE']
#    elif 'f_code' in row:
#        return row['f_code']
#    elif 'F_Code' in row:
#        return row['F_Code']
#    elif 'F_code' in row:
#        return row['F_code']
#    else:
#        return 'NA'

## YIKE, we have to accomodate mappings of the F_CODE
#def get_target_F_CODE(settings, row, target_fc):
#    if 'F_CODE' in row:
#        return row['F_CODE']
#    elif 'f_code' in row:
#        return row['f_code']
#    elif 'F_Code' in row:
#        return row['F_Code']
#    elif 'F_code' in row:
#        return row['F_code']
#    elif settings.mapping_specification.get_F_CODE_for_target(target_fc) is not None:      YIKE CHECK THIS YIKE
#        return settings.mapping_specification.get_F_CODE_for_target(target_fc)
#    else:
#        return 'NA'

################ general functions #####################

# This function creates a GUID or UUID
def makerandomufi():
    return  "{" + str(uuid.uuid4()).upper() + "}"



## There are some source databases that do not have unique UFI values.
## These unique values are absolutely critical, so here is the alternative.
## If there is a GlobalID, hopefully that is unique and we can use that instead.
## If there is not a GlobalID, I will try and grab the UFI and append the ObjectID and FCSubtype
## CRITICAL FUNCTION: get_row_unique_id
def get_row_unique_id(settings, row):
        if settings.config_setting_flag['flag_set_unique_identifiers_fresh'] == 1:
            return makerandomufi()
        elif 'UFI' in row:
            return str(row['UFI'])
        elif 'URI' in row:
            return str(row['URI'])
        else:
            settings.report_values['Unique identifer created'] += 1
            if settings.report_values['Unique identifer created'] < settings.maximum_records['maximum_reported_setting_of_unique_identifier']:
                settings.my_logger.warn("No unique identifier on entries found, creating one")
            return makerandomufi()

# This function returns a set of unique field values found for a specific field in a specific feature class. Requires ArcPy
# This function is from GitHub arcpy_attribute_based_etl.py
def get_field_values(fc, field, where=""):
    '''
    @return: A list of the values for the given field in the given feature class
        with the given optional where clause applied.
    @rtype: list
    '''
    field_values = []
    with arcpy.da.SearchCursor(fc, field, where) as rows:
        for row in rows:
            field_values.append(row[0])
    return set(field_values)

# This function converts a arcPy cursor into a set of dictionary rows so that one can process the field and value
# without having to assume the order of the fields. Requires ArcPy
# This function is from GitHub arcpy_attribute_based_etl.py
def rows_as_dicts(cursor):
    '''USAGE:
        with arcpy.da.SearchCursor(r'c:\data\world.gdb\world_cities', '*') as sc:
            for row in rows_as_dicts(sc):
                print row['CITY_NAME']
    Reference: http://arcpy.wordpress.com/2012/07/12/getting-arcpy-da-rows-back-as-dictionaries
    '''
    colnames = cursor.fields
    for row in cursor:
        yield dict(zip(colnames, row))

# This function returns a sorted list of fields given a feature class or table. Requires ArcPy
# This function is from GitHub arcpy_attribute_based_etl.py
def get_field_names(fc):
    return sorted([f.name for f in arcpy.ListFields(fc)])



def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        pass
    try:
        import unicodedata
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        pass
    return False

def is_value_unknown(invalue, include_legacy_stuff = 0, check_for_other = False):
    """ determine if the value is any of the standard default values """
    defpossible = set([u'no information', u'noinformation', u'------', u'-------', u'----------', u'-999999', u'-999999.0'])
    deflegacypossible = set([u'na', u'n/a', u'n_a', u'none', 
                             u'<null>', u'null', u'unknow', u'unknown', 
                             u'unk', u'-32768', u'-32767'])
    defotherstr = set([u'999', u'999.0'])
    defnumericpossible = set([-999999, -999999.0])
    defnumericlegacypossible = set([-32768, -32767, -32768.0, -32767.0])
    defotherpossible=set([999, 999.0])

    if invalue is None:
        return True
    if is_item_str(invalue):
        testval = unicode(invalue).lower()
        if testval in defpossible:
            return True
        if include_legacy_stuff and testval in defpossible:
            return True
        if check_for_other and testval in defotherstr:
            return True
    else:
        if invalue in defnumericpossible:
            return True
        if include_legacy_stuff and invalue in defnumericlegacypossible:
            return True
        if check_for_other and invalue in defotherpossible:
            return True        
    return False





# This function returns the appropriate numeric value given a boolean True/False input
# Note that it does not provide for the No Information value -999999
def get_boolean_enum_value(invalue):
    if invalue is not None and invalue == True:
        return 1001
    else:
        return 1000



def is_UFI_in_restriction_info(inUFI, saved_restrictioninfo):
    if len(saved_restrictioninfo) == 0:
        return False
    for rowx in saved_restrictioninfo:
        if rowx['UFI'] == inUFI:
            return True
    return False
 

def get_restriction_info_field(inUFI, fieldname, saved_restrictioninfo):
    if len(saved_restrictioninfo) == 0:
        return None
    for rowx in saved_restrictioninfo:
        if rowx['UFI'] == inUFI:
            if fieldname in rowx:
                return rowx[fieldname]
            else:
                return None
    return None

def search_for_restriction_info(infield, invalue, saved_restrictioninfo):
    if len(saved_restrictioninfo) == 0:
        return None
    resultset = set()
    for rowx in saved_restrictioninfo:
        if infield in rowx and rowx[infield] == invalue:
            resultset.add(rowx['UFI'])
    return resultset



def get_entity_collection_info_field(inUFI, fieldname, saved_entity_collection_metadata):
    if len(saved_entity_collection_metadata) == 0:
        return None
    for rowx in saved_entity_collection_metadata:
        if rowx['UFI'] == inUFI:
            if fieldname in rowx:
                return rowx[fieldname]
            else:
                return None
    return None

def get_entity_collection_dictionary(inUFI, saved_entity_collection_metadata):
    if len(saved_entity_collection_metadata) == 0:
        return None
    for rowx in saved_entity_collection_metadata:
        if rowx['UFI'] == inUFI:
            return rowx
    return None



def is_UFI_in_entity_collection_info(inUFI, saved_entity_collection_metadata):
    if len(saved_entity_collection_metadata) == 0:
        return False
    for rowx in saved_entity_collection_metadata:
        if rowx['UFI'] == inUFI:
            return True
    return False
 


# Helper function to prepare a field, value, and name to an entry in OTH or MEM
def prepare_string_for_OTH_or_MEM(infield, invalue, invaluename):
    if is_value_unknown(invalue, 0, True) == False:
        # Create a string such as: (infield:invalue=invaluename)
        if invaluename is None or invaluename == "":
            return ( "(" + infield + ":" + unicode(invalue) + ")" )
        elif invalue is None:
            return ( "(" + infield + ":" + invaluename + ")" )
        else:
            return  ( "(" + infield + ":" + unicode(invalue) + "=" + invaluename + ")" )
    else:
        return None

def clear_transformer_issues_report(settings):
    settings.report_errors.clear()



# Save the general error/warning statement to a dictionary and report the detail once if it exists
def report_transfomer_issue_or_increment_count(settings, errstmt, errortype = "INFO", detail = None, exc_info = None):
    if errstmt is None or errstmt == "" or errstmt == " ":
        return
    if detail is None:
        outdetail = ""
    else:
        outdetail = "\n\t\t{0}".format(detail)
    if errortype.lower() in ['error', 'warning']:
        oneinst = '\n\t\tAdditional issues of this type are summarized with a counter.'
    else:
        oneinst = ""
    if exc_info is None:
        outexc = ""
    else:
        outexc = "\n\t\t{0}".format(traceback.format_exc(5))
           # format(str(sys.exc_info()[0]), 
           #        str(sys.exc_info()[1]), 
           #        ''.join(traceback.format_stack()))
    #

    if errortype.lower() in [ "error", "err" ]:
        outstmt = "ERROR: {0}".format(errstmt)
        if errstmt in settings.report_errors:
            settings.report_errors[errstmt] += 1
        else:
            settings.report_errors[errstmt] = 1
            settings.my_logger.error("\t\t{0}{1}{2}{3}".format(errstmt, outdetail, oneinst, outexc))

    elif errortype.lower() in [ "warning" , "warn"]:
        outstmt = "WARNING: {0}".format(errstmt)
        if errstmt in settings.report_errors:
            settings.report_errors[errstmt] += 1
        else:
            settings.report_errors[errstmt] = 1
            settings.my_logger.warn("\t\t{0}{1}{2}{3}".format(errstmt, outdetail, oneinst, outexc))

    elif errortype.lower() == "debug":
        #outstmt = "DEBUG: {0}".format(errstmt)
        settings.my_logger.debug("\t\t{0}{1}".format(errstmt, outdetail))

    else:  # errortype.lower() == "info":
        #outstmt = "INFO: {0}".format(errstmt)
        if errstmt in settings.report_values:
            settings.report_values[errstmt] += 1
        else:
            settings.report_values[errstmt] = 1
            settings.my_logger.info("\t\t{0}{1}".format(errstmt, outdetail))


def search_for_meta_in_dictionary( contentin, searchkey):
    dictout = {}
    if isinstance(contentin, dict):
        for iikey in contentin:
            if isinstance(iikey, dict):
                if searchkey in iikey:
                    if searchkey not in dictout:
                        dictout[searchkey] = list()
                    dictout[searchkey].append(iikey[searchkey])
    return dictout






