"""
 ************************************************************************************************
 **                                                                                            **
 **  File: TranslateConfig.py                                                                  **
 **  Copyright (c) March 2016 and September 2018  by Leidos                                    **
 **  Produced under U.S. Government contract W5J9CQ-12-D-0004  Task Order: 053 and 073.         **
 **  As per DFARS, the U.S. Government is granted unlimited rights to use this work            **
 **  in any manner and for any purpose.                                                        **
 **                                                                                            **
 **  Summary:   This file uses the python ConfigParser to parse and interpret configuration
 **             settings.
 **                                                                                            **
 **  Developed by Leidos, Inc., High Resolution ISR Systems Division, Geospatial Programs      **
 **                                                                                            **
 **  Development History:                                                                      **
 **  March 2016 / Annette Janett Filer                                                          **   
 **  January 2017 / Major Update: Put all content into class structure, organized and cleaned up **
 **  September 2018  Support new flags and settings GUI
 **                                                                                            **
 **                                                                                            **
 ************************************************************************************************
"""

from Tkinter import *
import sys
import os
import datetime
import time
import ConfigParser
import logging
import logging.handlers
from GeodatabaseSettings import *

VERSION_IN_GROUP = {'TDS61+': ['TDS 6.1', 'TDS 7.0', 'TDS 7.1', ],
                    'TDS4': ['TDS 4.0', ],
                    'GGDM3': ['GGDM 3.0', ]}


class MyConfigSettings():
    def __init__(self):
        self.testmode = 0  # Modify variable to enable test mode operation
        self.comparisonmode = 0  # Modify variable to enable comparison mode

        self.config_setting = dict()
        self.configuration_complete = False
        self.config_file_name = "TranslateOptions.ini"
        self.config_start_of_log_file = "Translate_Log"
        self.config_start_of_result_file = "Translate_Results"

        # These get set by the application
        self.fandf = None
        self.transrules = None
        self.mapping_specification = None
        self.current_mapping_group = None
        self.results_xl = None

        self.my_logger = None
        self.app_title = StringVar()
        self.app_title.set("GGDM Data Translator")
        self.version_identifier = StringVar()
        self.version_identifier.set("v2.5, October 2018")
        self.name_application_title = StringVar()
        self.name_application_title.set("{0}, Version: {1}".format(self.app_title.get(), self.version_identifier.get()))

        ## GUI Geodatabase Source/Target Settings
        self.dir_target_geodatabase = StringVar()
        self.dir_source_geodatabase = StringVar()
        self.sourcedb_name_and_version = StringVar()
        self.targetdb_name_and_version = StringVar()
        # self.sourcedb_cgrls = StringVar()
        # self.targetdb_cgrls = StringVar()

        ## End GUI Geodatabase Settings

        ## GUI Log and Excel Results settings
        self.file_results_excel = StringVar()
        self.file_mapping_specification = StringVar()
        self.name_log_file = StringVar()

        ## End GUI Log and Excel Results Settings

        ## Checker settings
        self.flag_data_is_invalid_update_if_we_can = IntVar()
        self.flag_data_is_null_set_to_default = IntVar()

        self.flag_data_is_invalid_update_if_we_can.set(0)
        self.flag_data_is_null_set_to_default.set(0)

        ## End Checker Settings

        # Output strings for progress reporting
        self.progress_title = StringVar()
        self.progress_detail = StringVar()

        self.curdate = StringVar()
        self.setting_log_filename_base = StringVar()

        self.setting_allow_default_geometry = False

        # A temporary default value
        self.setting_datasetURI = 'noInformation'
        self.setting_QAMode = False

        self.report_errors = dict()

        self.logtextout = StringVar()

        self.current_dir = StringVar()
        self.current_dir.set(os.getcwd())

        self.current_step = "TBD"

        self.current_my_docs_dir = StringVar()
        self.current_my_docs_dir.set(os.path.expanduser(r'~\Documents'))

        # Save some lists and dictionaries for later quick look-up. These are initializers.
        self.saved_restrictioninfo = []  # List of entries for saved restriction information
        self.saved_entity_collection_metadata = []  # List of entries for saved entity collection

        self.saved_dataset = []  # List of entries for saved datasets
        self.saved_oth = {}  # Dictionary to hold saved OTH entries
        self.saved_mem = {}  # Dictionary to hold saved MEM entries
        self.existing_unique_ids = []  # List to hold existing unique identifiers so we don't duplicate existing data
        self.report_values = {}  # Dictionary for reporting transformer values to for
        # printing when feature processing is complete

        # This is the dictionary holder for transform functions
        self.call_transform_function = {}

        self.schema_version = {}
        self.schema_version['GGDM 3.0'] = 'GGDM 3.0'
        self.schema_version['GGDM 2.2'] = 'GGDM 2.2'
        self.schema_version['GGDM 2.1'] = 'GGDM 2.1'
        self.schema_version['GGDM 2.0'] = 'GGDM 2.0'
        self.schema_version['TDS 6.1'] = 'TDS 6.1'
        self.schema_version['TDS 7.0'] = 'TDS 7.0'
        self.schema_version['TDS 7.1'] = 'TDS 7.1'
        self.schema_version['TDS 6.0'] = 'TDS 6.0'
        self.schema_version['TDS 4.0'] = 'TDS 4.0'

        self.geodb_version = {}
        self.geodb_version['GGDM 3.0'] = 7
        self.geodb_version['GGDM 2.2'] = 6
        self.geodb_version['GGDM 2.1'] = 5
        self.geodb_version['GGDM 2.0'] = 4
        self.geodb_version['TDS 7.0'] = 8
        self.geodb_version['TDS 7.1'] = 9
        self.geodb_version['TDS 6.1'] = 3
        self.geodb_version['TDS 6.0'] = 2
        self.geodb_version['TDS 4.0'] = 1
        self.geodb_version['Other'] = 999

        self.sourcedb = None
        self.targetdb = None
        # Specification of the allowed source and target version values
        self.target_allowed_for_source = dict()
        self.target_allowed_for_source['GGDM 3.0'] = ['Unknown', 'TDS 4.0', 'TDS 6.1', 'TDS 7.0', 'TDS 7.1', 'GGDM 2.1']
        self.target_allowed_for_source['GGDM 2.1'] = ['GGDM 3.0']
        self.target_allowed_for_source['GGDM 2.2'] = ['GGDM 3.0']
        # self.target_allowed_for_source['GGDM 2.2 WRDB'] = ['GGDM 3.0']
        self.target_allowed_for_source['TDS 6.1'] = ['GGDM 3.0']
        self.target_allowed_for_source['TDS 7.0'] = ['GGDM 3.0']
        self.target_allowed_for_source['TDS 7.1'] = ['GGDM 3.0']
        # self.target_allowed_for_source['TDS 6.0'] = ['GGDM 3.0']
        self.target_allowed_for_source['TDS 4.0'] = ['GGDM 3.0']

        self.nameversionshort = {
            'GGDM 3.0': 'GGDM3',
            'GGDM 2.1': 'GGDM21',
            'GGDM 2.2': 'GGDM22',
            'TDS 4.0': 'TDS4',
            'TDS 7.0': 'TDS70',
            'TDS 7.1': 'TDS71',
            'TDS 6.1': 'TDS61',
            'Unknown': 'UNK'}

        self.config_setting_flag = {
            'flag_save_invalid_values_to_oth_instead_of_dropping_them': 1,
            'flag_skip_all_metadata_translation': 0,
            'flag_evaluate_oth_content_and_parse_for_possible_values_to_set': 1,
            'flag_allow_flat_version_database': 0,

            'flag_expect_no_ufi_conflict': 1,
            'report_all_duplicate_ufi': 0,  # Set this option > 0 to report duplicate UFI content in the source database
            'flag_ufi_is_not_unique_generate_new_unique_ufi': 0,
            'flag_set_unique_identifiers_fresh': 0,
            # If this is a Unknown source, we will automatically set this setting to 1

            'break_after_n_instances': 0,

            'flag_skip_search_for_applicable_dataset_while_testing': 0,
            'flag_skip_search_for_applicable_entity_collection_while_testing': 0
            #   'flag_allow_use_of_default_geometry_while_testing':0
        }
        #  'flag_do_not_perform_unique_ufi_check': 1,   not used
        # 'flag_use_mapping_specification_file': 0,
        #  'setting_superuser':0,   not used
        # 'flag_execution_of_app_is_in_compare_mode':0,

        # The member gui_out_settings is used temporarily to store this information in a quick to edit format.
        # The contents are moved into different variables and this one is destroyed.
        self.gui_out_setting = dict()
        self.gui_out_setting[
            'Flags'] = dict()  # Values: Description, Tool tip, dictionary of options, default value (Note - the default will be updated based on the default specified above, and then from the config file)
        self.gui_out_setting['Flags']['flag_save_invalid_values_to_oth_instead_of_dropping_them'] = (
        "Save invalid values to OTH",
        "Save invalid values to OTH even if they are specified as being something to drop",
        {'Save only valid data': 0, 'Save invalid data to OTH': 1}, 'Save invalid data to OTH')
        self.gui_out_setting['Flags']['flag_skip_all_metadata_translation'] = ("Skip all metadata operations", "",
                                                                               {'Process Metadata': 0,
                                                                                'Skip Metadata': 1}, 'Process Metadata')
        self.gui_out_setting['Flags']['flag_evaluate_oth_content_and_parse_for_possible_values_to_set'] = (
        "Parse OTH source content for possible target values",
        "Parse all OTH values and if a destination attribute can support the value, save it to the specified attribute",
        {'Disabled': 0, 'Parse Source OTH': 1}, 'Parse Source OTH')
        self.gui_out_setting['Flags']['break_after_n_instances'] = (
        "CAREFUL: Halt processing of each feature class after specified number of instances (rows)",
        "", {'Process all data': 0, '2 rows': 2, '10 rows': 10, '500 rows': 500, '2500 rows': 2500}, 'Process all data')
        self.gui_out_setting['Flags']['flag_allow_flat_version_database'] = (
        "Allow for target GGDM 3.0 to be flat version (one feature per feature class)",
        "Flat version of GGDM 3.0 has one feature per feature class with only one subtype. Selecting flat target allowed will allow target to be either standard or flat type. This option works only for the target GGDM 3.0.",
        {'Flat target allowed': 1, 'Feature classes and subypes': 0}, 'Feature classes and subypes')

        self.gui_out_setting['UFIFlags'] = dict()
        self.gui_out_setting['UFIFlags']['flag_expect_no_ufi_conflict'] = (
        "Target UFI: Output to a new target db, no UFI conflicts expected",
        "When enabled, we will not check for the presence of existing data in the target before writing new rows to target. If disabled, we will look at all target rows to see if the UFI value is already present.",
        {'Do not overwrite target data': 0, 'No conflicts expected': 1}, 'No conflicts expected')
        self.gui_out_setting['UFIFlags']['flag_ufi_is_not_unique_generate_new_unique_ufi'] = (
        "Source UFI check and UPDATE source",
        "Recommend you enable the 'Source ufi check and report' as well. On issues with the UFI, options allow for either setting new UFI value, or attempt to use GlobalID or ObjectID if they are available",
        {'Disabled - No UFI check/update': 0, 'Check UFI and correct, set to new uuid': 1,
         'Check UFI and correct, set to GlobalID': 2, 'Check UFI and correct, set to ObjectID': 3},
        'Disabled - No UFI check/update')
        # self.gui_out_setting['UFIFlags']['flag_do_not_perform_unique_ufi_check'] = ("Skip the pre-condition check of unique UFIs",
        #     "Assume all UFI/URI values are unique and valid in the source database.",
        #     {'Disabled':0,'Enabled':1},'Enabled' )
        self.gui_out_setting['UFIFlags']['report_all_duplicate_ufi'] = (
        "Source UFI check and report issues, but limit the issue report",
        "Check for UFI problems with the source data and report issues. Limit the number of issues reported per feature class to the value specified. If you have also specified the option to update the source, this will allow for the reports needed.",
        {'Disabled - No UFI check/report': 0, 'Enabled, limit to 3': 3, 'Enabled, limit to 10': 10,
         'Enabled, limit to 25': 25}, 'Disabled - No UFI check/report')

        # self.gui_out_setting['OpModeFlags'] = dict()
        # self.gui_out_setting['OpModeFlags']['setting_superuser'] = ("Run in special test mode", "",
        #     {'Disabled':0,'Enabled':1},'Disabled' )
        # self.gui_out_setting['OpModeFlags']['flag_execution_of_app_is_in_compare_mode'] = ("Run the application in database compare mode", "",
        #     {'Disabled':0,'Enabled':1},'Disabled' )

        self.gui_out_setting['TestModeFlags'] = dict()
        self.gui_out_setting['TestModeFlags']['flag_skip_search_for_applicable_dataset_while_testing'] = (
        "Do not search for matching Dataset (ResourceSrf) for any features",
        "Do not search for matching Dataset (ResourceSrf) for any features. Use this only if working with generated QA sample data",
        {'Disabled': 0, 'Enabled': 1}, 'Disabled')
        self.gui_out_setting['TestModeFlags']['flag_skip_search_for_applicable_entity_collection_while_testing'] = (
        "Test mode, skip applicable entity collection search", "",
        {'Disabled': 0, 'Enabled': 1}, 'Disabled')
        self.gui_out_setting['TestModeFlags']['flag_allow_use_of_default_geometry_while_testing'] = (
        "Allow the use and specification of default geometry", "",
        {'Disabled': 0, 'Enabled': 1}, 'Disabled')

        # Debug settings and maximums for error checking
        self.debug_settings = {  # 'compare_target_against_existing_metainfo':1,
            'compare_src_against_existing_metainfo': 2,  # Set to 0 to disable, 1 for part report, 2 for max report
            'report_entity_collection_findings': 1,
            # Set to > 0 if you want additional reporting on entity collection processing
            'itemize_each_translation_each_field': 2,  # Set to 0 to disable, 1 for part report, 2 for max report
            # (alternative is to use the debug values below to start this reporting
            # on a specific feature class and field)
            'report_missing_source_data': 1,
            # Set to > 0 to get reporting about missing source data (i.e. does not match schema)
            # This option may no longer be functional since we now need to process GRLS
            'leave_unspecified_targets_null_instead_of_defaulting': 0,
            'report_feature_classes_defined_in_schema_but_not_in_database': 1,
            # With modification for GRLS, this probably is not functional
            'report_field_schema_vs_database_differences': 1,
            # With modification for GRLS, this probably is not functional
            'allow_default_geometry_test_mode': 0,
            'report_unique_id_usage': 1,  # Set this option > 0 to report values of record unique identifiers and
            # OTH and MEM saved for specific unique identified records
            'instrumentation': 0,
            'instrumentation_output': 0,
            'only_transform_no_move': 0
            # Do not move data, only perform the transform functions. This is used for testing
        }

        self.debug_fc_list = []  # Specify the feature classes that you want to enable detail
        # reporting on, for example ['FacilityPnt','InformationCrv']
        self.debug_field_list = []  # For the feature classes in debug_fc_list, now identify the fields
        # that you want detailed reporting on, for example: ['ZI005_FNA', 'ZI006_MEM']

        self.maximum_records = {'saved_restrictioninfo': 100,  # Maximum number of restriction information records
                                # 'processing_indicator_n_rows':500,       # Output an 'working' indicator after processing this many rows in a feature class
                                'processing_indicator_n_seconds': 300,
                                'saved_entity_collection_metadata': 2000,
                                # Maximum number of entity collection metadata records
                                # 'maximum_ufi_errors_per_feature_class':3,   # Maximum number of UFI errors reported per feature class
                                'maximum_reported_transforms_of_none': 3,
                                # Maximum number of detail reports when the transform is none
                                'maximum_reported_source_field_not_found_in_specified_set': 10,
                                # These can happen a LOT if the source is not a composite geodatabase
                                'maximum_reported_target_field_not_found_in_specified_set': 10,
                                # These can happen a LOT if the target is not a composite geodatabase
                                'maximum_reported_setting_of_unique_identifier': 3,
                                # Maximum number of unique identifier creation warnings per feature class
                                'maximum_issues_encountered_per_featureclass': 3
                                # When we process feature classes, and hit this maximum count of errors, processing will cease for that feature class
                                }

        self.gui_out_setting[
            'DebugSetting'] = dict()  # Values: Description, Tool tip, dictionary of options, default value (Note - the default will be updated based on the default specified above, and then from the config file)
        # self.gui_out_setting['DebugSetting']['compare_target_against_existing_metainfo'] = ("Report comparison of target against existing metadata",
        #    "", {'Disabled':0,'Enabled':1},'Disabled' )
        self.gui_out_setting['DebugSetting']['compare_src_against_existing_metainfo'] = (
        "Report comparison of source against existing metadata",
        "", {'Disabled': 0, 'Minimum Report': 1, 'Maximum Report': 2}, 'Disabled')
        self.gui_out_setting['DebugSetting']['report_entity_collection_findings'] = (
        "Report entity collection detailed results",
        "", {'Disabled': 0, 'Enabled': 1}, 'Disabled')
        self.gui_out_setting['DebugSetting']['itemize_each_translation_each_field'] = (
        "Output detailed processing notes to the log",
        "If no feature classes or fields are specified, this will output extensive information. Do not use this with no feature classes or fields if you are processing all records.",
        {'Disabled': 0, 'Minimum Report': 1, 'Maximum Report': 2}, 'Disabled')
        self.gui_out_setting['DebugSetting']['report_missing_source_data'] = (
        "Report source fields specified in the defined schema but not found in database",
        "", {'Disabled': 0, 'Enabled': 1}, 'Disabled')
        self.gui_out_setting['DebugSetting']['leave_unspecified_targets_null_instead_of_defaulting'] = (
        "Leave unspecified or unsettable target fields as null instead of attempting to set to default",
        "", {'Default when possible': 0, 'Leave nulls': 1}, 'Default when possible')
        self.gui_out_setting['DebugSetting']['report_feature_classes_defined_in_schema_but_not_in_database'] = (
        "Report source tables or feature classes specified in the schema and mapping but not found in the database",
        "", {'Disabled': 0, 'Enabled': 1}, 'Disabled')
        self.gui_out_setting['DebugSetting']['allow_default_geometry_test_mode'] = (
        "Allow default (null) geometry and do not report issues with it (this is okay in some test modes)",
        "", {'Disabled': 0, 'Enabled': 1}, 'Disabled')
        self.gui_out_setting['DebugSetting']['only_transform_no_move'] = (
        "CAREFUL: Limit processing to transform and do not output records to the target database",
        "", {'Transform and save': 0, 'Process only-no save': 1}, 'Transform and save')
        self.gui_out_setting['DebugSetting']['report_unique_id_usage'] = (
        "Report the usage of unique identifiers within content such as OTH, MEM and metadata",
        "", {'Disabled': 0, 'Enabled': 1}, 'Disabled')

        self.gui_out_setting[
            'LimitSetting'] = dict()  # Values: Description, Tool tip, dictionary of options, default value (Note - the default will be updated based on the default specified above, and then from the config file)
        self.gui_out_setting['LimitSetting']['saved_restrictioninfo'] = (
        "Limit saved restriction information records to this value and report excess",
        "", {'RI Max 100 records': 100, 'RI Max 200 records': 200, 'RI Max 500 records': 500}, 'RI Max 200 records')
        # self.gui_out_setting['LimitSetting']['processing_indicator_n_rows'] = ("Output a processing indicator after processing this number of records",
        #                   "",{'Indicator at 50': 50, 'Indicator at 250':250, 'Indicator at 500':500, 'Indicator at 1000':1000, 'Indicator at 2000': 2000}, 'Indicator at 500')
        self.gui_out_setting['LimitSetting']['saved_entity_collection_metadata'] = (
        "Limit saved entity collection metadata records to this value and report excess",
        "", {'EC Max 250': 250, 'EC Max 500': 500, 'EC Max 2000': 2000, 'EC Max 10000': 10000}, 'EC Max 2000')
        # self.gui_out_setting['LimitSetting']['maximum_ufi_errors_per_feature_class'] = ("Limit the unique identifier errors reported per feature class or table to this value",
        #                  "",{'UFI Error Max 3': 3, 'UFI Error Max 5': 5, 'UFI Error Max 10':10},'UFI Error Max 3')
        self.gui_out_setting['LimitSetting']['maximum_reported_transforms_of_none'] = (
        "Limit reporting of transform value null result",
        "", {'Xform None Max 3': 3, 'Xform None Max Max 5': 5, 'Xform None Max Max 10': 10}, 'Xform None Max 3')
        self.gui_out_setting['LimitSetting']['maximum_reported_source_field_not_found_in_specified_set'] = (
        "Limit reports of source fields not found as per the schema specified",
        "", {'Source Not Found Max 3': 3, 'Source Not Found Max 5': 5, 'Source Not Found Max 10': 10},
        'Source Not Found Max 5')
        self.gui_out_setting['LimitSetting']['maximum_reported_target_field_not_found_in_specified_set'] = (
        "Limit reports of target fields not found as per the schema specified",
        "", {'Target Not Found Max 3': 3, 'Target Not Found Max 5': 5, 'Target Not Found Max 10': 10},
        'Target Not Found Max 5')
        self.gui_out_setting['LimitSetting']['maximum_reported_setting_of_unique_identifier'] = (
        "Limit reports about the generating new unique identifier values",
        "", {'Set UFI Max 3': 3, 'Set UFI Max 5': 5, 'Set UFI Max 10': 10}, 'Set UFI Max 5')
        self.gui_out_setting['LimitSetting']['maximum_issues_encountered_per_featureclass'] = (
        "Limit reports on issues encountered per feature class or table",
        "", {'Max Feature Issues 3': 3, 'Max Feature Issues 5': 5, 'Max Feature Issues 10': 10}, 'Max Feature Issues 5')
        self.gui_out_setting['LimitSetting']['processing_indicator_n_seconds'] = (
        "Output a processing indicator after processing records for this number of seconds",
        "", {'5 minutes (300 seconds)': 300, '3 minutes (180 seconds)': 180, '1 minute (60 seconds)': 60},
        '5 minutes (300 seconds)')

        self.procflagsgeneral = dict()
        self.prepareguivars(self.config_setting_flag, self.procflagsgeneral, self.gui_out_setting, 'Flags')
        self.procufiflags = dict()
        self.prepareguivars(self.config_setting_flag, self.procufiflags, self.gui_out_setting, 'UFIFlags')
        # self.procoperationmode = dict()
        # self.prepareguivars( self.config_setting_flag, self.procoperationmode, self.gui_out_setting, 'OpModeFlags')
        self.proctestflags = dict()
        self.prepareguivars(self.config_setting_flag, self.proctestflags, self.gui_out_setting, 'TestModeFlags')
        self.procdebugsettings = dict()
        self.prepareguivars(self.debug_settings, self.procdebugsettings, self.gui_out_setting, 'DebugSetting')
        self.proclimitsconstraint = dict()
        self.prepareguivars(self.maximum_records, self.proclimitsconstraint, self.gui_out_setting, 'LimitSetting')

        # Remove gui_out_setting content now, we don't need it anymore as it was moved into the proc variables

        del self.gui_out_setting['Flags']
        del self.gui_out_setting['UFIFlags']
        del self.gui_out_setting['TestModeFlags']
        del self.gui_out_setting['LimitSetting']
        del self.gui_out_setting['DebugSetting']
        del self.gui_out_setting

        self.get_config_settings()
        self.save_config_settings()

        self.configuration_complete = True
        # self.start_logger()

    def prepareguivars(self, settingsdictin, guiprocsettingsdict, guisettings, guioption):
        for keyx in settingsdictin:
            if keyx in guisettings[guioption]:
                # '['compare_src_against_existing_metainfo'] = ("Report comparison of source against existing metadata",
                # "Provides a detailed report comparing source content metadata.", {'Disabled':0,'Minimum Report':1,'Maximum Report':2},0)
                txtname, txtdescrip, dictcombo, default = guisettings[guioption][keyx]
                # self.keysetting[ii].set(txtname)

                guiprocsettingsdict[keyx] = dict()

                guiprocsettingsdict[keyx]['ToolTip'] = str()
                guiprocsettingsdict[keyx]['TextVariable'] = StringVar()
                guiprocsettingsdict[keyx]['IntVariable'] = IntVar()
                guiprocsettingsdict[keyx]['ValuesList'] = list()
                guiprocsettingsdict[keyx]['OptionCombo'] = dict()
                guiprocsettingsdict[keyx]['DefaultValue'] = default  # text
                guiprocsettingsdict[keyx]['LabelName'] = txtname
                guiprocsettingsdict[keyx]['ToolTip'] = txtdescrip
                guiprocsettingsdict[keyx]['OptionCombo'] = dictcombo

                vallist = list()
                for keya in dictcombo:
                    vallist.append(keya)
                    if dictcombo[keya] == settingsdictin[keyx]:
                        guiprocsettingsdict[keyx]['IntVariable'].set(dictcombo[keya])
                        guiprocsettingsdict[keyx]['TextVariable'].set(keya)
                guiprocsettingsdict[keyx]['ValuesList'] = vallist

    def prepgeodatabasesource(self):
        self.sourcedb = MyGDBSettings(self, self.dir_source_geodatabase.get(), self.sourcedb_name_and_version.get(),
                                      'unspecified', 'Source')

    def prepgeodatabasetarget(self):
        self.targetdb = MyGDBSettings(self, self.dir_target_geodatabase.get(), self.targetdb_name_and_version.get(),
                                      'unspecified', 'Target')

    def gettarget_db(self):
        return self.targetdb.dbname.get()

    def getsource_db(self):
        return self.sourcedb.dbname.get()

    def settarget_db(self, invalue):
        self.targetdb.dbname.set(invalue)

    def setsource_db(self, invalue):
        self.sourcedb.dbname.set(invalue)

    def getsource_db_version(self):
        return self.sourcedb_name_and_version.get()

    def gettarget_db_version(self):
        return self.targetdb_name_and_version.get()

    def output_defined_config_section(self, config_handle, flagdictin, procdictin, sectionid):
        for keyx in flagdictin:
            if keyx in procdictin:
                valueout = 0
                for keya in procdictin[keyx]['OptionCombo']:
                    if procdictin[keyx]['TextVariable'].get().lower() == keya.lower():
                        valueout = procdictin[keyx]['OptionCombo'][keya]
                        break
                config_handle.set(sectionid, keyx, valueout)

    def save_config_settings(self):
        cfgfile = open(self.config_file_name, 'w')
        self.config_setting.clear()

        Config = ConfigParser.ConfigParser()
        Config.add_section('Geodatabases')
        Config.set('Geodatabases', 'Source_Database', self.dir_source_geodatabase.get())
        Config.set('Geodatabases', 'Source_Database_Schema', self.sourcedb_name_and_version.get())
        # Config.set('Geodatabases','Source_Database_Dataset', self.sourcedb_cgrls.get())

        Config.set('Geodatabases', 'Target_Database', self.dir_target_geodatabase.get())
        Config.set('Geodatabases', 'Target_Database_Schema', self.targetdb_name_and_version.get())
        # Config.set('Geodatabases','Target_Database_Dataset', self.targetdb_cgrls.get())

        Config.add_section('ExcelFiles')
        Config.set('ExcelFiles', 'Mapping_Specification', self.file_mapping_specification.get())
        Config.set('ExcelFiles', 'Results_Output_Excel', self.file_results_excel.get())

        Config.add_section('TextFiles')
        Config.set('TextFiles', 'Log_File', self.name_log_file.get())

        Config.add_section('Flags')
        self.output_defined_config_section(Config, self.config_setting_flag, self.procflagsgeneral, 'Flags')
        self.output_defined_config_section(Config, self.config_setting_flag, self.procufiflags, 'Flags')
        # self.output_defined_config_section(Config, self.config_setting_flag, self.procoperationmode, 'Flags')
        self.output_defined_config_section(Config, self.config_setting_flag, self.proctestflags, 'Flags')

        Config.add_section('Debug_And_Limits')
        self.output_defined_config_section(Config, self.debug_settings, self.procdebugsettings, 'Debug_And_Limits')
        self.output_defined_config_section(Config, self.maximum_records, self.proclimitsconstraint, 'Debug_And_Limits')

        tempstr = ""
        for itemii in self.debug_fc_list:
            if tempstr == "":
                tempstr = "{0}".format(itemii)
            else:
                tempstr = "{0},{1}".format(tempstr, itemii)
        Config.set('Debug_And_Limits', 'Debug_Feature_Classes', tempstr)
        tempstr = ""
        for itemii in self.debug_field_list:
            if tempstr == "":
                tempstr = "{0}".format(itemii)
            else:
                tempstr = "{0},{1}".format(tempstr, itemii)
        Config.set('Debug_And_Limits', 'Debug_Fields', tempstr)

        Config.write(cfgfile)
        cfgfile.close()
        del Config

    def input_configuration_section(self, config_handle, settingsdict, procsettingsdict, configsection):
        namevals = [name for name, value in config_handle.items(configsection)]
        for keyx in settingsdict:
            if keyx in procsettingsdict:
                if keyx.lower() in namevals:
                    procsettingsdict[keyx]['IntVariable'].set(config_handle.getint(configsection, keyx.lower()))
                    settingsdict[keyx] = procsettingsdict[keyx]['IntVariable'].get()
                else:
                    procsettingsdict[keyx]['IntVariable'].set(settingsdict[keyx])
                #
                procsettingsdict[keyx]['TextVariable'].set(
                    self.getkeyfromvalue(procsettingsdict[keyx]['OptionCombo'], settingsdict[keyx]))

    def get_config_settings(self):

        Config = ConfigParser.ConfigParser()

        try:
            Config.read(self.config_file_name)
            config_section_list = Config.sections()

            if 'Geodatabases' in config_section_list:
                namevals = [name for name, value in Config.items('Geodatabases')]

                if 'Source_Database'.lower() in namevals:
                    self.dir_source_geodatabase.set(Config.get('Geodatabases', 'Source_Database'))
                else:
                    self.dir_source_geodatabase.set(os.path.expanduser(r'~\Documents') + "\\TDS6_1_sample.gdb")

                if 'Source_Database_Schema'.lower() in namevals:
                    self.sourcedb_name_and_version.set(Config.get('Geodatabases', 'Source_Database_Schema'))
                else:
                    self.sourcedb_name_and_version.set("TDS 6.1")

                # if 'Source_Database_Dataset'.lower() in namevals:
                #    self.sourcedb_cgrls.set(Config.get('Geodatabases', 'Source_Database_Dataset'))
                # else:
                #    self.sourcedb_cgrls.set("Composite")

                if 'Target_Database'.lower() in namevals:
                    self.dir_target_geodatabase.set(Config.get('Geodatabases', 'Target_Database'))
                else:
                    self.dir_target_geodatabase.set(os.path.expanduser(r'~\Documents') + "\\GGDM3_0_sample.gdb")

                if 'Target_Database_Schema'.lower() in namevals:
                    self.targetdb_name_and_version.set(Config.get('Geodatabases', 'Target_Database_Schema'))
                else:
                    self.targetdb_name_and_version.set("GGDM 3.0")

                # if 'Target_Database_Dataset'.lower() in namevals:
                #    self.targetdb_cgrls.set(Config.get('Geodatabases', 'Target_Database_Dataset'))
                # else:
                #    self.targetdb_cgrls.set("Composite")

            else:
                print "Expected geodatabases section in config file but it was not there"

                # ExcelFiles
            if 'ExcelFiles' in config_section_list:
                namevals = [name for name, value in Config.items('ExcelFiles')]

                if 'Results_Output_Excel'.lower() in namevals:
                    self.file_results_excel.set(Config.get('ExcelFiles', 'Results_Output_Excel'))
                else:
                    self.file_results_excel.set(os.path.expanduser(r'~\Documents') + "\\XlateResults.xls")

                if 'Mapping_Specification'.lower() in namevals:
                    self.file_mapping_specification.set(Config.get('ExcelFiles', 'Mapping_Specification'))
                else:
                    self.file_mapping_specification.set("")

            # TextFiles
            if 'TextFiles' in config_section_list:
                namevals = [name for name, value in Config.items('TextFiles')]
                if 'Log_File'.lower() in namevals:
                    self.name_log_file.set(Config.get('TextFiles', 'Log_File'))
                else:
                    self.name_log_file.set(os.path.expanduser(r'~\Documents') + "\\XlateLogFile.txt")

            if 'Flags' in config_section_list:

                self.input_configuration_section(Config, self.config_setting_flag, self.procflagsgeneral, 'Flags')
                self.input_configuration_section(Config, self.config_setting_flag, self.procufiflags, 'Flags')
                # self.input_configuration_section(Config, self.config_setting_flag, self.procoperationmode, 'Flags' )
                self.input_configuration_section(Config, self.config_setting_flag, self.proctestflags, 'Flags')
            else:
                print "Expected flags section in config file but it was not there"

            if 'Debug_And_Limits' in config_section_list:
                namevals = [name for name, value in Config.items('Debug_And_Limits')]
                self.input_configuration_section(Config, self.debug_settings, self.procdebugsettings,
                                                 'Debug_And_Limits')
                self.input_configuration_section(Config, self.maximum_records, self.proclimitsconstraint,
                                                 'Debug_And_Limits')

                tempin = Config.get('Debug_And_Limits', 'Debug_Feature_Classes')

                if tempin is not None and tempin != "":
                    del self.debug_fc_list[:]
                    for itemin in tempin.split():
                        if itemin is not None and itemin != "":
                            self.debug_fc_list.append(itemin)

                tempin = Config.get('Debug_And_Limits', 'Debug_Fields')

                if tempin is not None and tempin != "":
                    del self.debug_field_list[:]
                    for itemin in tempin.split():
                        if itemin is not None and itemin != "":
                            self.debug_field_list.append(itemin)

                #
            else:
                print "Expected debug_and_limits section in config file but it was not there"

        except:
            print ("No {0} file or error with content. Rebuilding with defaults".format(self.config_file_name))
            self.save_config_settings()

        self.curdate.set(datetime.datetime.now().strftime("%Y%m%d%H%M"))
        # self.name_log_file.set("{0}_{1}_2_{2}_{3}.txt".\
        #    format("LogTranslations", self.sourcedb.dataset.get(), self.targetdb.dataset.get(), self.curdate.get()))
        del Config

    def getkeyfromvalue(self, dictin, valuein):
        return [key for key, value in dictin.iteritems() if value == valuein][0]

    ### METHODS RELATED TO DATA LOGGING
    def prepare_log_file_name(self, basedirectory):
        if basedirectory is not None:
            if os.path.exists(basedirectory) == False:
                basedirectory = r'~\Documents'
        import datetime
        curdt = datetime.datetime.now().strftime("%Y%m%d%H%M")  # import datetime
        logfnbase = "{0}_{1}_{2}_{3}.txt".format(self.config_start_of_log_file,
                                                 self.nameversionshort[self.sourcedb_name_and_version.get()],
                                                 self.nameversionshort[self.targetdb_name_and_version.get()],
                                                 curdt)
        newlogname = os.path.join(basedirectory, logfnbase)
        self.name_log_file.set(newlogname)

    def prepare_file_results_excel(self, basedirectory):
        if basedirectory is not None:
            if os.path.exists(basedirectory) == False:
                basedirectory = r'~\Documents'
        import datetime
        curdt = datetime.datetime.now().strftime("%Y%m%d%H%M")  # import datetime
        logfnbase = "{0}_{1}_{2}_{3}.xls".format(self.config_start_of_result_file,
                                                 self.nameversionshort[self.sourcedb_name_and_version.get()],
                                                 self.nameversionshort[self.targetdb_name_and_version.get()],
                                                 curdt)
        newfnname = os.path.join(basedirectory, logfnbase)
        self.file_results_excel.set(newfnname)

    def start_logger(self):
        os.system("mode con cols=150 lines=5000")

        if self.configuration_complete == False:
            print "Configuration is not yet complete. Cannot start logger."
            sys.exit()

        setting_log_filename_base = self.name_log_file.get()
        if setting_log_filename_base <> "":
            logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                                datefmt='%Y-%m-%d %H:%M',
                                filename=setting_log_filename_base, filemode='w')
        else:
            print "Error, no log filename identified. Initial configuration failed."
            sys.exit()

        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        # set a format which is simpler for console use
        formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
        # tell the handler to use this format
        console.setFormatter(formatter)
        # add the handler to the root logger
        logging.getLogger('').addHandler(console)
        #### Logging for this file
        self.my_logger = logging.getLogger('Main')
        self.my_logger.setLevel(logging.DEBUG)
        #    self.my_logger.info( "Starting data logger \n\t\ton {0}\n\t\tLog file name: {1}\n\t\tExcel results file: {2}".\
        #       format(self.name_application_title.get(),  self.name_log_file.get(), self.file_results_excel.get()))
        self.my_logger.info(
            "Starting data logger \n\t\ton {0}\n\t\tLog file name: {1}\n\t\tExcel results file name: {2}\n". \
            format(self.name_application_title.get(), self.name_log_file.get(), self.file_results_excel.get()))

    def set_logger_section(self, arg1):
        self.my_logger = logging.getLogger(arg1)

    def get_logger(self):
        if self.my_logger is None:
            print "Error, log file information is not currently defined."
        return self.my_logger


def report_global_settings(settings):
    settings.my_logger.info("Utility {0} ".format(settings.name_application_title.get()))

    settings.my_logger.info("Settings report:")
    settings.my_logger.info("\tSource database directory: {0}".format((settings.sourcedb.dbname.get())))
    settings.my_logger.info("\tSource dataset: {0}".format((settings.sourcedb.dataset.get())))
    settings.my_logger.info("\tTarget database directory: {0}".format((settings.targetdb.dbname.get())))
    settings.my_logger.info("\tTarget dataset: {0}".format((settings.targetdb.dataset.get())))

    if settings.comparisonmode == 1:
        settings.my_logger.info("Application is in COMPARISON Mode")

    if settings.testmode == 1:
        settings.my_logger.info("Application is in TEST Mode")

    settings.my_logger.info("\tLogger filename = {0}".format(settings.name_log_file.get()))

    # settings.my_logger.info("Defaults")
    # settings.my_logger.info("\tSchema Version = {0}".format(schema_version[target_db_version]))
    # settings.my_logger.info("\tGeodbversion = {0}    ".format(str(geodb_version[target_db_version])))

    settings.my_logger.info("Configuration settings")
    for setingii in settings.config_setting_flag:
        if settings.config_setting_flag[setingii] > 0:
            settings.my_logger.info("\tConfiguration Setting {0} specified as {1}".format(setingii, str(
                settings.config_setting_flag[setingii])))

    settings.my_logger.info("Other settings")
    for setingii in settings.debug_settings:
        if settings.debug_settings[setingii] > 0:
            settings.my_logger.info(
                "\tDebug Setting {0} specified as {1}".format(setingii, str(settings.debug_settings[setingii])))

    for setingii in settings.maximum_records:
        settings.my_logger.info(
            "\tLimits setting {0} specified as {1}".format(setingii, str(settings.maximum_records[setingii])))

    if len(settings.debug_fc_list) > 0:
        settings.my_logger.info("Debug feature classes:")
        for settingii in settings.debug_fc_list:
            settings.my_logger.info("\tDebug print starting at {0}".format(settingii))

    if len(settings.debug_field_list) > 0:
        settings.my_logger.info("Debug field names:")
        for settingii in settings.debug_field_list:
            settings.my_logger.info("\tDebug print starting at field {0}".format(settingii))

    settings.my_logger.info("-end settings report\n")
