﻿"""
 ************************************************************************************************
 **                                                                                            **
 **  File: FieldTransformerFunctions.py                                                        **
 **  Copyright (c) December 2015 by Leidos                                                     **
 **  Produced under U.S. Government contract W5J9CQ-12-D-0004  Task Order: 053 and 073         **
 **  As per DFARS, the U.S. Government is granted unlimited rights to use this work            **
 **  in any manner and for any purpose.                                                        **
 **                                                                                            **
 **  Summary:   This file provides for the functions required to map special case attributes   **
 **                                                                                            **
 **  Developed by Leidos, Inc., High Resolution ISR Systems Division, Geospatial Programs      **
 **                                                                                            **
 **  Development History:                                                                      **
 **  December 2015 / Annette Janett Filer                                                      **
 **  January  2016 / added logging                                                             **
 **  May 2016 / added additional translation functions, ensure default behavior where appropriate
 **  9/28/2016 - Corrections made to interval <-> non-interval translations                    **
 **  9/27/2018 - Updates made to support directed target based translations, corrections based **
 **              on testing.                                                                   **
 **                                                                                            **
 **                                                                                            **
 ************************************************************************************************
"""

# Import required system modules
import os, sys
import arcpy
import uuid
import glob
import math

from TranslateConfig import *
from FieldValueMappings import *
from StandardUtilities import *

##Most common useful functions used throughout:
## field_set_to_default(get_target_F_CODE(settings, row, target_fc), field_in, src_fc, target_fc)
## check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)
## settings.sourcedb.get_enum_text_for_index( field_in, row[field_in]) 

fclass_static = dict()


# my_logger = logging.getLogger('FieldTF')
# my_logger.setLevel(logging.DEBUG)

############################# Functions begin here ##################################################
# This function compares current row content (and possibly transforms it if needed) against
# a saved list of dictionaries that hold all metadata. This function is trying to find the
# unique identifier, UFI from the saved metadata that represents the incoming row of metadata.
# The incoming row may be a subset from any feature class, and some of the content may have to
# be transformed in order to match.
# This is one of the most complex functions in here. It is very tricky to manage these fields
# and in some cases, we have to compare against transformed data. We also have to be sure
# to compare all of the fields and if the input row doesn't have ALL of the fields, we need to
# assume defaults for the items it does not have.
def compare_src_against_existing_metainfo(settings, rowinorig, saved_metainfo, \
                                          src_fc, targ_fc, MetaSrcTargetMapperForFields, \
                                          MaxRemainingFields=1, CompareToTargetRunTransforms=True):
    fieldskiplist = ['scvn', 'ufi']
    fieldskiplist.extend(settings.targetdb.common_fields_geom_and_id_and_code)

    # First, if there is no saved metadata to search we can just return
    if len(saved_metainfo) == 0:
        if settings.debug_settings['compare_src_against_existing_metainfo'] > 1:
            settings.my_logger.debug("compare_src_against_existing_metainfo, no saved metadata exists")
        return None

    # Verify that ths incoming row content is a dictionary and not just a list!
    if isinstance(rowinorig, dict) == False:
        settings.my_logger.error("compare_src_against_existing_metainfo, incoming content is not a dictionary")
        return None

    if settings.debug_settings['compare_src_against_existing_metainfo'] > 0:
        settings.my_logger.debug("Saved metadata records:" + str(len(saved_metainfo)))

    rowcountii = 0

    rowincopy_compsrc = copy.deepcopy(rowinorig)
    curUFI = ""
    curF_CODE = ""
    match = True

    # Loop through the saved meatadata and compare the fields from the parameter list to the metadata and when found, return the UFI
    for record_saved_meta in saved_metainfo:
        rowcountii += 1
        groupReport = {}  # create a temporary reporting dictionary to verify we got all of the fields
        if 'UFI' not in record_saved_meta:
            continue
        # record_saved_meta is a dictionary representing saved information
        newuid = record_saved_meta['UFI']  # get the current UFI from the Saved Metadata row
        match = True
        failedat = ""
        # for the incoming row - which is a subset of fields typically found in a metadata table
        # look at all of the possible fields based on the mapper.
        # if the incoming row (which is a subset of fields) is not in the possible fields,
        # then we know outright this is not a match.
        # if the field item values in the incoming row does not match the field item values saved
        # in the global saved metadata, then we know this is not a match.
        expectednumberfields = len(record_saved_meta)  # We need to be sure we compare a minimum number of fields

        # for each field in the subset list to compare rowin against saved meta
        for srcmetamap, targetmetamap, transformermetamap in MetaSrcTargetMapperForFields:
            groupReport[targetmetamap] = 'Comparing'
            # We may have some fields saved in the metadata save area that we don't need to compare against
            # If this is the UFI or F_CODE, don't bother with a compare
            if targetmetamap is None:
                lcfield = None
            else:
                lcfield = targetmetamap.lower()

            if targetmetamap is not None and (lcfield == 'ufi' or lcfield == 'f_code'):
                if lcfield == 'ufi':
                    curUFI = record_saved_meta[targetmetamap]
                else:
                    curF_CODE = record_saved_meta[targetmetamap]
                #
            else:
                # It is possible to have a SOURCE field and no TARGET field. In this case,
                # we can only skip over it.
                if targetmetamap is None:
                    if srcmetamap is not None and srcmetamap in record_saved_meta and srcmetamap in rowincopy_compsrc:
                        if unicode(rowincopy_compsrc[srcmetamap]).lower() == unicode(
                                record_saved_meta[srcmetamap]).lower():
                            match = True
                            groupReport[srcmetamap] = 'Source value matched'
                        else:
                            match = False
                            failedat = srcmetamap  # record the field the test fails at for printing later, if needed
                            groupReport[srcmetamap] = u'Failed to match source content ' + \
                                                      str(rowincopy_compsrc[srcmetamap]) + " with saved: " + \
                                                      str(record_saved_meta[srcmetamap])
                    else:
                        pass

                elif targetmetamap.lower() in fieldskiplist:
                    pass

                # field specified in list is in the input row and in saved instance
                elif targetmetamap in rowincopy_compsrc and targetmetamap in record_saved_meta:
                    strowinfield = unicode(rowincopy_compsrc[targetmetamap])
                    striffield = unicode(record_saved_meta[targetmetamap])

                    expectednumberfields -= 1
                    if (strowinfield.lower() == striffield.lower()) or \
                            (rowincopy_compsrc[targetmetamap] == record_saved_meta[targetmetamap]):
                        match = True
                        groupReport[targetmetamap] = 'Value matched'


                    elif transformermetamap is not None and transformermetamap in call_transform_function:
                        try:
                            if srcmetamap in rowincopy_compsrc and rowincopy_compsrc[srcmetamap] is not None:
                                xformval = call_transform_function[transformermetamap](settings, rowincopy_compsrc,
                                                                                       srcmetamap, targetmetamap,
                                                                                       src_fc, targ_fc)
                            elif targetmetamap in rowincopy_compsrc and rowincopy_compsrc[targetmetamap] is not None:
                                xformval = call_transform_function[transformermetamap](settings, rowincopy_compsrc,
                                                                                       targetmetamap, targetmetamap,
                                                                                       src_fc, targ_fc)
                        except:
                            report_transfomer_issue_or_increment_count(
                                settings, "  Specified transformer {0} had an exception.".format(transformermetamap),
                                "ERROR",
                                "Transformer {0} for source {1} to targetmetamap {2} for source with value {3} for feature {4} to feature {5} failed. Target set to {6}". \
                                    format(transformermetamap, srcmetamap, targetmetamap,
                                           rowincopy_compsrc[srcmetamap] if srcmetamap in settings.sourcedb.row else "",
                                           src_fc, targ_fc, xformval),
                                sys.exc_info())

                        if xformval is None:
                            xformval = ""
                        if unicode(xformval).lower() != unicode(record_saved_meta[targetmetamap]).lower():
                            if is_value_unknown(xformval, 0, False) and is_value_unknown(striffield, 0, False):
                                match = True
                                groupReport[targetmetamap] = u'Transformed value matched as default'

                            else:
                                match = False
                                failedat = targetmetamap  # record the field the test fails at for printing later, if needed
                                groupReport[targetmetamap] = u'Failed to match TRANSFORMED content ' + \
                                                             str(
                                                                 xformval) + " from: " + strowinfield + " with saved: " + striffield
                        else:
                            match = True
                            groupReport[targetmetamap] = 'Transformed value matched'

                    else:  # not a match, and no transformer
                        match = False
                        failedat = targetmetamap  # record the field the test fails at for printing later, if needed
                        groupReport[
                            targetmetamap] = u'Failed to match content ' + strowinfield + " with saved: " + striffield

                # The field is in the row being checked but not in the saved dictionary
                elif targetmetamap in rowincopy_compsrc and targetmetamap not in record_saved_meta:
                    strowinfield = unicode(rowincopy_compsrc[targetmetamap])
                    # Check for field being default, if default, skip
                    if targetmetamap.lower() in fieldskiplist or rowincopy_compsrc[targetmetamap] is None or \
                            is_value_unknown(strowinfield, 0, False):
                        pass
                    else:
                        match = False
                        failedat = targetmetamap
                        groupReport[targetmetamap] = u'Failed to match, row has content (' + strowinfield + \
                                                     ' but saved values does not have this field ' + targetmetamap


                # The field is in the saved dictionary but not in the row being checked
                elif targetmetamap not in rowincopy_compsrc and targetmetamap in record_saved_meta:
                    striffield = unicode(record_saved_meta[targetmetamap])

                    # Check for field being default, if default, skip
                    if targetmetamap.lower() in fieldskiplist or \
                            record_saved_meta[targetmetamap] is None or is_value_unknown(striffield, 0, False):
                        pass
                    else:
                        match = False
                        failedat = targetmetamap
                        groupReport[targetmetamap] = u'Failed to match, saved set has ' + striffield + \
                                                     ' but row does not have this field ' + targetmetamap

                # end if (case like stmt)
            # end if not UFI or F_CODE

            # If we have already found the match to fail, break out of the loop (for src, target...)
            if match == False:
                if settings.debug_settings['compare_src_against_existing_metainfo'] > 1:
                    settings.my_logger.debug(
                        'compare_src_against_existing_metainfo, compare different already at field {0}'.format(
                            failedat))
                break

            # Update the target f_code if it was set
        ####     if targetmetamap is not None and targetmetamap.lower() == 'f_code' and targetmetamap in record_saved_meta:
        ####         settings.targetdb.set_f_code(record_saved_meta[targetmetamap])
        # End for each src, target, transformer
        if match == True:
            break
    # End for each row in saved data

    # If the match value is still true
    # Then we are good and found the correct identifier and can return it.
    if match == True:
        if settings.debug_settings['compare_src_against_existing_metainfo'] > 0:
            settings.my_logger.debug(
                "compare_src_against_existing_metainfo, possible UFI {0}, report {1}".format(str(newuid),
                                                                                             str(groupReport)))
        return newuid
    else:
        if settings.debug_settings['compare_src_against_existing_metainfo'] > 0:
            settings.my_logger.debug(
                "Row {1} No match found for input row. Report on fields: {0}".format(str(groupReport), rowcountii))
    return None


################## HELPER FUNCTIONS ##########################

def get_potential_codelist_values_for_this_field(ffld):
    outlist = ['noInformation']
    if ffld in ['ZI020_GE4', 'ZI020_GE42', 'ZI020_GE43', 'ZI020_GE44', 'ZI020_GE4A', 'ZI020_GE4B']:
        outlist = list(keyval for keyval in GE4codelist2IC2codelist)
    elif ffld == 'CPS':
        outlist = list(keyval for keyval in CPScodelist2CPSenumvalue)
    elif ffld == 'ETS':
        outlist = list(keyval for keyval in ETScodelist2ETSenumvalue)
    elif ffld == 'HZD':
        outlist = list(keyval for keyval in HZDcodelist2HZDenumvalue)
    elif ffld == 'VDT':
        outlist = list(keyval for keyval in VDTcodelist2VDTenumvalue)
    elif ffld == 'RCG':
        outlist = list(keyval for keyval in RCGcodelist2RCGenumvalue)
    elif ffld == 'ZI001_SRT':
        outlist = list(keyval for keyval in ZI001_SRTcodelist2ZI001_SRTenumvalue)
    elif ffld == 'ZI001_VSC':
        outlist = list(keyval for keyval in ZI001_VSCcodelist2ZI001_VSCenumvalue)
    elif ffld == 'ZI004_RCG':
        outlist = list(keyval for keyval in ZI004_RCGcodelist2ZI004_RCGenumvalue)
    elif ffld == 'EQC':
        outlist = list(keyval for keyval in EQCcodelist2numericEQS)
    elif ffld == 'IC2':
        outlist = list(keyval for keyval in IC2codelist2GE4codelist)
    return outlist


# Return true if this is a valid enumerated value
def is_field_and_value_valid_in_source(settings, row, field_in, target_field, src_fc, target_fc):
    # We know coming into this that src is not None
    # If this is an enumerated data type, and this feature class is subtyped and has F_CODE attribute (which we require to get valid attr list)
    # If this field is not applicable to this feature class, these functions will catch that and return false
    curfcode = settings.sourcedb.get_f_code()
    if field_in is not None and src_fc is not None and settings.sourcedb.is_field_in_fclass(src_fc, 'F_CODE'):
        if settings.sourcedb.is_field_valid_for_fclass_fcode(src_fc, curfcode, field_in):
            if settings.sourcedb.is_enumeration(src_fc, field_in):
                # Is the source index row[field_in] NOT valid for this target?
                if not settings.sourcedb.is_index_valid_for_fclass_fcode(src_fc, curfcode, field_in, row[field_in]):
                    return False
                else:
                    return True
            else:
                return None  # Not an enumerated value
        else:
            return False  # Not a valid field
    else:
        return False
    return None  # None of the above, return True


def is_field_and_value_valid_in_target(settings, row, field_in, src_fc, target_fc, value_in):
    # We know coming into this that src is not None
    # If this is an enumerated data type, and this feature class is subtyped and has F_CODE attribute (which we require to get valid attr list)
    # If this field is not applicable to this feature class, these functions will catch that and return false
    if field_in is None:  # If the field is not in the row, it is kind of an issue
        return False
    curfcode = settings.targetdb.get_f_code()
    if field_in is not None and target_fc is not None and settings.targetdb.is_field_in_fclass(target_fc, 'F_CODE'):
        if settings.targetdb.is_field_valid_for_fclass_fcode(target_fc, curfcode, field_in):
            if settings.targetdb.is_enumeration(target_fc, field_in):
                # Is the source index row[field_in] NOT valid for this target?
                if not settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, curfcode, field_in, value_in):
                    return False
                else:
                    return True
            else:
                return None  # Not an enumerated value
        else:
            return False  # Not a valid field
    else:
        return False
    return None  # None of the above, return True


# This is a default transform that will, for an enumerated index, determine if the index
# value is valid in the destination and if not, it will save it to OTH Other Enumerants
def check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc):
    # We know coming into this that src is not None
    # If this is an enumerated data type, and this feature class is subtyped and has F_CODE attribute (which we require to get valid attr list)
    # If this field is not applicable to this feature class, these functions will catch that and return false
    if field_in in row:
        if row[field_in] is not None:
            valid_in_target = is_field_and_value_valid_in_target(settings, row, field_in, src_fc, target_fc,
                                                                 row[field_in])
        else:
            valid_in_target = False

        # If this value is NOT valid in the target, and the value is not the default or other, save the value as other
        if valid_in_target == True:
            # if we get here - valid_in_target is true, return the original value if there is one
            return row[field_in]
        else:
            # Be sure the value we are trying to save is not just default or other!
            if is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == False:
                enumtext = settings.sourcedb.get_enum_text_for_index(field_in, row[
                    field_in])  # Get the enumerated text value for this index
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], enumtext),
                               'OTH')  # Save to OTH, other
                settings.report_values[
                    'Transform check for missing enum index: the specified index is not allowed in the target and we will save it to OTH'] += 1
                # settings.my_logger.warn("{3} Field {0} value {1} = {2} is not allowed in the target.".format(field_in, row[field_in], enumtext, src_fc))
                if settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, settings.targetdb.get_f_code(),
                                                                     field_in, 999):  # Verify 999 is allowed as a value
                    return 999  # Return 999 instead of the original value, because it is not allowed
                else:
                    return row[field_in]  # None of the above, return the original value

        return row[field_in]  # None of the above, return the original value
    else:  # not valid field in row
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                      "Put results dict here", False)
    return -999999


def check_for_and_transform_missing_enum_index_target_diff(settings, row, field_in, target_field, src_fc, target_fc):
    # If this is an enumerated data type, and this feature class is subtyped and has F_CODE attribute (which we require to get valid attr list)
    # If this field is not applicable to this feature class, these functions will catch that and return false
    valid_in_target = is_field_and_value_valid_in_target(settings, row, field_in, src_fc, target_fc, row[field_in])

    # If this value is NOT valid in the target, and the value is not the default or other, save the value as other
    if valid_in_target == False:
        # Be sure the value we are trying to save is not just default or other!
        if field_in in row and is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == False:
            enumtext = settings.sourcedb.get_enum_text_for_index(field_in, row[
                field_in])  # Get the enumerated text value for this index
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], enumtext),
                           'OTH')  # Save to OTH, other

            settings.report_values[
                'Transform check for missing enum index: the specified index is not allowed in the target and we will save it to OTH'] += 1
            # settings.my_logger.warn("{3} Field {0} value {1} = {2} is not allowed in the target.".format(field_in, row[field_in], enumtext, src_fc))
            if settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, settings.targetdb.get_f_code(),
                                                                 target_field, 999):  # Verify 999 is allowed as a value
                if settings.debug_settings['itemize_each_translation_each_field'] > 1:
                    settings.my_logger.debug(
                        "\t\Added field {0} field value {1} from source to OTH. Setting target to 999".format(field_in,
                                                                                                              row[
                                                                                                                  field_in]))
                return 999  # Return 999 instead of the original value noting this value has been saved to OTH
            else:
                if settings.debug_settings['itemize_each_translation_each_field'] > 1:
                    settings.my_logger.debug(
                        "\t\Added field {0} field value {1} from source to OTH. Setting target to default as 999 is not valid for it.".format(
                            field_in, row[field_in]))
                return transform_set_to_default(settings, None, None, target_field, None, target_fc)

            return row[field_in]  # None of the above, return the original value
        else:
            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                          "Put results dict here", False)

    # if we get here - valid_in_target is true, return the original value if there is one
    if field_in in row:
        return row[field_in]
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                      "Put results dict here", False)


# This is a default transform that will, for an enumerated index, determine if the index
# value is valid in the destination and if not, it will save it to OTH Other Enumerants
def transform_missing_enum_index_with_warnings(settings, row, field_in, target_field, src_fc, target_fc):
    # We know coming into this that src is not None
    # If this is an enumerated data type, and this feature class is subtyped and has F_CODE attribute (which we require to get valid attr list)
    # If this field is not applicable to this feature class, these functions will catch that and return false
    if field_in is not None and field_in in row:
        valid_in_target = is_field_and_value_valid_in_target(settings, row, field_in, src_fc, target_fc, row[field_in])

        # If this value is NOT valid in the target, and the value is not the default or other, save the value as other
        if valid_in_target == False:
            # Be sure the value we are trying to save is not just default or other!
            if is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == False:
                enumtext = settings.sourcedb.get_enum_text_for_index(field_in, row[
                    field_in])  # Get the enumerated text value for this index
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], enumtext),
                               'OTH')  # Save to OTH, other
                report_transfomer_issue_or_increment_count(settings,
                                                           'Transform check for missing enum index: the specified index is not allowed in the target and we will save it to OTH',
                                                           "WARN",
                                                           "{3} Field {0} value {1} = {2} is not allowed in the target.".format(
                                                               field_in, strout(row[field_in]), enumtext, src_fc))
                if settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, settings.targetdb.get_f_code(),
                                                                     field_in, 999):  # Verify 999 is allowed as a value
                    return 999  # Return 999 instead of the original value, because it is not allowed
                else:
                    return -999999  # Return -999999 - this field may not be allowed in the feature class, we can't return 999, so the only thing left is -999999

                return row[field_in]  # None of the above, return the original value
            else:  # The field is in the row and is default or other, return the original value
                return row[field_in]

        # if we get here - valid_in_target is true, return the original value if there is one:
        return row[field_in]
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                      "Put results dict here", False)


# Ensure that the specified field is valid for the specified target feature class for the source F_CODE
def is_field_valid_to_set(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.targetdb.is_field_in_fclass(target_fc, target_field):
        validcodes = settings.targetdb.get_field_valid_fcode_list(target_fc, target_field)
        if validcodes is not None and settings.targetdb.get_f_code() in validcodes:
            return True
    return False


# This function allows one to add content to OTH
# It will actually be SAVED to the database at the end of the feature class processing
def add_to_row_OTH(settings, row, value_in, field_in='OTH'):
    if settings.setting_QAMode:
        return
    # We do not allow the 'addition' of a default value into OTH. Check for a default value first.
    if value_in is not None and is_value_unknown(value_in, settings.sourcedb.geodb_type_unknown, True) == False:
        localufi = get_row_unique_id(settings,
                                     row)  # We can not assume UFI will be the unique identifier (use a function to get unique id)
        if settings.debug_settings['report_unique_id_usage'] > 0:
            settings.my_logger.debug("Adding content to OTH for row having unique id: {0}".format(localufi))
        # Verify that we got a unique identifier before going on
        if localufi <> 'noInformation':
            # Not a default, save this OTH with the UFI
            if len(settings.saved_oth) > 0 and localufi in settings.saved_oth:
                # Check to see if we already have this value, and if we do, do not add it again
                if settings.saved_oth[localufi].find(value_in) == -1:  # this new value is not already in OTH
                    settings.report_values['OTH Processing - Appending a value into OTH'] += 1
                    settings.saved_oth[localufi] = settings.saved_oth[localufi] + value_in
                else:
                    settings.report_values['OTH Processing - Value is already in OTH so it is not being appended'] += 1
            # If there are no existing saved OTH data, just save this content
            else:
                settings.report_values['OTH Processing - Putting a value into OTH'] += 1
                settings.saved_oth[localufi] = value_in
            row[field_in] = settings.saved_oth[
                localufi]  # I don't think this will do anything, we can't change row content, but may affect later processing
        else:
            report_transfomer_issue_or_increment_count(settings,
                                                       "No unique identifier for row having OTH value to save", "WARN")
    else:
        row[field_in] = 'noInformation'  # I don't think this will do anything, we can't change row content


# returns a list of OTH entries, for example say we had OTH = (FFN:3=name of 3)(TRS2:Bus)  (TRS:2)
# we want to get out of this:
# ( ('FFN', 3, 'name of 3'), ('TRS2', None, 'Bus'), ('TRS', 2, None) )
def parseOTHentry(inOTHstr):
    # if there are no parens in the OTH string, we cannot parse it
    if '(' not in inOTHstr:
        return ()
    # if there is no colon in the string, we cannot parse it
    if ':' not in inOTHstr:
        return ()
    return ()


# Similar in many ways to OTH processing, this function allows one to save information into MEM
# It will actually be SAVED to the database at the end of the feature class processing
def add_to_row_MEM(settings, row, value_in, field_in='ZI006_MEM'):
    if settings.setting_QAMode:
        return
    if value_in is not None and is_value_unknown(value_in, settings.sourcedb.geodb_type_unknown, True) == False:
        localufi = get_row_unique_id(settings, row)
        if settings.debug_settings['report_unique_id_usage'] > 0:
            settings.my_logger.debug(
                "Adding content to MEM or ZI006_MEM for row having unique id: {0}".format(localufi))
        if localufi == 'noInformation':
            report_transfomer_issue_or_increment_count(settings,
                                                       "No unique identifier for row having valid MEM content", "WARN")
        else:
            # Not a default, save this OTH with the UFI
            if len(settings.saved_mem) > 0 and localufi in settings.saved_mem:
                if settings.saved_mem[localufi].find(value_in) == -1:  # Check to see if this value is not already there
                    settings.report_values['MEM Processing, Append value to MEM'] += 1
                    settings.saved_mem[localufi] = settings.saved_mem[localufi] + value_in
            else:
                settings.report_values['MEM Processing, Save value to MEM'] += 1
                settings.saved_mem[localufi] = value_in
            row[field_in] = settings.saved_mem[localufi]
    else:
        row[field_in] = 'No Information'


# This function is based on a very specific structure for the OTH content. It tries to match incoming field and value
# against the possible OTH content and returns true if the content is found.
# It may be advisable to change this to use the OTH parse capability.
def rowOTHcheck(row, field_in, field_value, possiblestring):
    tempOTH = row['OTH'].lower()  # Get the OTH as all lower case
    if possiblestring and possiblestring != "":  # Ensure we have an input value to look for
        compstring = possiblestring.lower()  # Put our input value to lower case
        genposB = "(" + field_in + ":" + str(
            field_value) + "=" + compstring + ")"  # Create one possible option for an OTH representative of this input value
        genposC = "(" + field_in + ":" + compstring + ")"  # Create another possible option for an OTH representative of this input value
        if tempOTH.find(genposB) != -1 or tempOTH.find(genposC) != -1:  # Look for either of the two options within OTH
            return True
    genposA = "(" + field_in + ":" + str(
        field_value) + ")"  # If we difn't find it yet, try one other option for OTH representation
    if tempOTH.find(genposA) != -1:
        return True
    return False


### This function will look for the optimal multivalued field to set and it will
##def set_multivalued_field_in_row(row, fieldname, value_in, value_name):
##    # examine FFN, FFN2, and FFN3 and add the value to the appropriate location
##    if row[fieldname] == -999999:
##        settings.report_values['Saved to multivalued attribute'] = settings.report_values['Saved to multivalued attribute'] + 1
##        row[fieldname] = value_in
##        return
##    elif row[fieldname] == value_in:
##        return
##    elif row[fieldname + '2'] == -999999:
##        settings.report_values['Saved to multivalued attribute 2'] = settings.report_values['Saved to multivalued attribute 2'] + 1
##        row[fieldname + '2'] = value_in
##        return
##    elif row[fieldname + '2'] == value_in:
##        return
##    elif row[fieldname + '3'] == -999999:
##        settings.report_values['Saved to multivalued attribute 3'] = settings.report_values['Saved to multivalued attribute 3'] + 1
##        row[fieldname + '3'] = value_in
##        return
##    elif row[fieldname + '3'] == value_in:
##        return
##    else:
##        add_to_row_OTH(settings,row, prepare_string_for_OTH_or_MEM(fieldname, value_in, value_name), 'OTH')
##        settings.my_logger.warn( "Not enough places to save field {4} value {0} = {3} on feature fcode {1} with unique id {2}" \
##                        .format(value_in,get_target_F_CODE(settings, row, target_fc),get_row_unique_id(row),value_name,fieldname))
##    return

# This is looking for the first field that is set to no information where the field is a multi-valued field.
def get_first_open_multifield(row, fieldnamebase, targetvalue):
    # examine multi-valued field such as FFN, FFN2, and FFN3 and determine most appropriate field for use
    # If the field is in the row and the value is either No Information or is already set to the target, return the first field
    if fieldnamebase in row and (row[fieldnamebase] == -999999 or row[fieldnamebase] == targetvalue):
        return fieldnamebase
    # If the 2nd field is in the row and the value is either No Information or is already set to the target, return the second field
    elif (fieldnamebase + '2') in row and (row[fieldnamebase + '2'] == -999999 or row[fieldnamebase] == targetvalue):
        return fieldnamebase + '2'
    # If the 3rd field is in the row and the value is either No Information or is already set to the target, return the second field
    elif (fieldnamebase + '3') in row and (row[fieldnamebase + '3'] == -999999 or row[fieldnamebase] == targetvalue):
        return fieldnamebase + '3'
    elif (fieldnamebase + '4') in row and (row[fieldnamebase + '3'] == -999999 or row[fieldnamebase] == targetvalue):
        return fieldnamebase + '4'
    elif (fieldnamebase + '5') in row and (row[fieldnamebase + '3'] == -999999 or row[fieldnamebase] == targetvalue):
        return fieldnamebase + '5'
    # If no field is found for this, return None
    else:
        return None


# Reports the name of the dictionary if this source and destination attribute are represented by a codelist to codelist mapping
def get_codelist2codelistmapping_name(srcattr, destattr):
    keyval = srcattr + "2" + destattr
    if keyval in codelist2codelist_mappings:
        return codelist2codelist_mappings[keyval]
    else:
        return None


# Reports the name of the dictionary if this source and destination attribute are represented by a enumerated value to codelist
def get_enumvalue2codelistmapping_name(srcattr, destattr):
    keyval = srcattr + "2" + destattr
    if keyval in enumvalue2codelist_mappings:
        return enumvalue2codelist_mappings[keyval]
    else:
        return None


# Reports the name of the dictionary if this source and destination attribute are represented by a codelist to enumerated value mapping
def get_codelist2enumvalue_mapping_name(srcattr, destattr):
    keyval = srcattr + "2" + destattr
    if keyval in codelist2enumvalue_mappings:
        return codelist2enumvalue_mappings[keyval]
    else:
        return None


# Return a mapped value for a codelist to codelist, given a source attribute and destination attribute and source value
def get_codelist2codelistmapping_value(srcattr, destattr, srcvalue):
    keyval = srcattr + "2" + destattr
    if keyval in codelist2codelist_mappings:
        if keyval == "GE42IC2":
            if srcvalue in GE4codelist2IC2codelist:
                return GE4codelist2IC2codelist[srcvalue]
            else:
                return None
        elif keyval == "IC22GE4":
            if srcvalue in IC2codelist2GE4codelist:
                return IC2codelist2GE4codelist[srcvalue]
            else:
                return None
    return None


# Return a mapped value for a enumerated value to codelist, given a source attribute and destination attribute and source value
def get_enumvalue2codelistmapping_value(srcattr, destattr, srcvalue):
    keyval = srcattr + "2" + destattr
    if keyval in enumvalue2codelist_mappings:
        if keyval == "CPS2CPS":
            if srcvalue in CPSenumvalue2CPScodelist:
                return CPSenumvalue2CPScodelist[srcvalue]
            else:
                return None
        elif keyval == "ETS2ETS":
            if srcvalue in ETSenumvalue2ETScodelist:
                return ETSenumvalue2ETScodelist[srcvalue]
            else:
                return None
        elif keyval == "HZD2HZD":
            if srcvalue in HZDenumvalue2HZDcodelist:
                return HZDenumvalue2HZDcodelist[srcvalue]
            else:
                return None
        elif keyval == "VDT2VDT":
            if srcvalue in VDTenumvalue2VDTcodelist:
                return VDTenumvalue2VDTcodelist[srcvalue]
            else:
                return None
        elif keyval == "ZI001_SRT2ZI001_SRT":
            if srcvalue in ZI001_SRTenumvalue2ZI001_SRTcodelist:
                return ZI001_SRTenumvalue2ZI001_SRTcodelist[srcvalue]
            else:
                return None
        elif keyval == "ZI001_VSC2ZI001_VSC":
            if srcvalue in ZI001_VSCenumvalue2ZI001_VSCcodelist:
                return ZI001_VSCenumvalue2ZI001_VSCcodelist[srcvalue]
            else:
                return None
    return None


# Return a mapped value for a codelist to enumerated value, given a source attribute and destination attribute and source value
def get_codelist2enumvaluemapping_value(srcattr, destattr, srcvalue):
    keyval = srcattr + "2" + destattr
    if keyval in codelist2enumvalue_mappings:
        if keyval == "CPS2CPS":
            if srcvalue in CPScodelist2CPSenumvalue:
                return CPScodelist2CPSenumvalue[srcvalue]
            else:
                return None
        elif keyval == "ETS2ETS":
            if srcvalue in ETScodelist2ETSenumvalue:
                return ETScodelist2ETSenumvalue[srcvalue]
            else:
                return None
        elif keyval == "HZD2HZD":
            if srcvalue in HZDcodelist2HZDenumvalue:
                return HZDcodelist2HZDenumvalue[srcvalue]
            else:
                return None
        elif keyval == "VDT2VDT":
            if srcvalue in VDTcodelist2VDTenumvalue:
                return VDTcodelist2VDTenumvalue[srcvalue]
            else:
                return None
        elif keyval == "ZI001_SRT2ZI001_SRT":
            if srcvalue in ZI001_SRTcodelist2ZI001_SRTenumvalue:
                return ZI001_SRTcodelist2ZI001_SRTenumvalue[srcvalue]
            else:
                return None
        elif keyval == "ZI001_VSC2ZI001_VSC":
            if srcvalue in ZI001_VSCcodelist2ZI001_VSCenumvalue:
                return ZI001_VSCcodelist2ZI001_VSCenumvalue[srcvalue]
            else:
                return None
    return None


# Examples of use
# get_codelist2codelistmapping_value("IC2","GE4","BEL")
# returns 'ge:GENC:3:1-2:BEL'
#
# get_codelist2codelistmapping_value("GE4","IC2", "ge:GENC:3:1-2:LKA")
# returns 'LKA'
#
# get_enumvalue2codelistmapping_value("VDT", "VDT", "4")
# returns 'navd88'
#
# get_enumvalue2codelistmapping_value("ETS", "ETS", "999")
# returns 'other'
#
# get_enumvalue2codelistmapping_value("CPS", "CPS", "93")   # out of bounds value
# returns None

#################### TRANSFORM FUNCTIONS ############### ################################## ############################

# All of these functions operate using a common parameter set: (settings, row, field_in, target_field, src_fc, target_fc)
# row:       The source row content from the source database. Contains all fields and field values as a dictionary.
# field_in:  The name of the target field currently being processed
# src_fc:    The name of the source feature class
# target_fc: The name of the target feature class
#
# Each function is called during the processing and construction of the TARGET row for the TARGET database.
# Each function needs to return the appropriate target value for this field
#

def transform_APT(settings, row, field_in, target_field, src_fc, target_fc):
    # If target is GGDM 2.1 and the value is 999 = Other, return No Information
    if settings.gettarget_db_version() == "GGDM 2.1":
        if field_in is not None and field_in in row and row[field_in] == 999:
            return -999999
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


# Only if destination is GGDM 3.0. Otherwise, no transform needed for ASU.
# Therefore, use the default capability to look for any potential new/missing value and also check
# for the critical value of 19 being translated, in which case we output a warning
def transform_ASU(settings, row, field_in, target_field, src_fc, target_fc):
    settings.report_values['Transform started: Aerodrome Surface (ASU) processing'] += 1

    if field_in is not None and field_in in row and is_field_and_value_valid_in_target(settings, row, field_in, src_fc,
                                                                                       target_fc, row[field_in]):
        return row[field_in]
    elif field_in is not None and field_in in row:  # The value is not allowed in the destination
        if row[field_in] == 19:  # Water
            report_transfomer_issue_or_increment_count(settings,
                                                       'Transform processing: Aerodrome Surface illegal value 19 = water may be saved to OTH',
                                                       "WARN",
                                                       "Aerodrome Surface (ASU) value 19 = Water is no longer allowed. This value is not reasonable on this feature = {1} {0}.".format(
                                                           target_fc, settings.targetdb.get_f_code()))
            if settings.config_setting_flag['flag_save_invalid_values_to_oth_instead_of_dropping_them'] == 1:
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Water'), 'OTH')
            return -999999
        elif is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == True:
            return row[field_in]
        else:
            enumtext = settings.sourcedb.get_enum_text_for_index(field_in, row[
                field_in])  # Get the enumerated text value for this index
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], enumtext),
                           'OTH')  # Save to OTH, other
            return 999
    return -999999


# Set IC2 based on GE4. If GE4 is in the source row, we can do this, otherwise just return a default value for IC2
def transform_IC2_from_GE4(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI020_GE4' in row:
        settings.report_values['IC2 Country derived from GENC'] += 1
        return get_codelist2codelistmapping_value("GE4", "IC2", row['ZI020_GE4'])
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                      "Put results dict here", False)


def get_GE4_from_matching_dataset(settings, row, field_in, target_field, src_fc, target_fc):
    ''' Find and return URI for dataset having spatial overlap with this feature'''
    # print "   transform_find_dataset_id"

    if (row['SHAPE@']) and settings.config_setting_flag['flag_skip_search_for_applicable_dataset_while_testing'] == 0:
        possibledatasets = ["DatasetSrf", "ResourceSrf"]  # , "DATASET_T"]
        doneflag = False
        returnval = 'noInformation'
        for fcname in possibledatasets:
            target_obj_path = settings.targetdb.get_root_for_fc_subtypes(fcname)

            if settings.targetdb.is_featureclass_in_geodatabase_spec(fcname):
                # for each dataset record, does the shape of this feature overlap the dataset shape?
                with arcpy.da.SearchCursor(target_obj_path, ('URI', 'SHAPE@', 'ZI020_GE4'), None) as row22AA:
                    for row22 in row22AA:
                        if row22[1].disjoint(row['SHAPE@']) == True:
                            pass  # this is a faster test, no points in common, this is not the one we are looking for
                        elif row22[1].contains(row['SHAPE@']) == True:
                            # print 'Dataset URI {0} matches feature {1}'.format(str(row22[0]), str(row['UFI']))
                            returnval = row22[2]
                            doneflag = True
                            break
            if doneflag:
                break  # get out of for loop

    else:  # If there is no shape value to test against, don't bother trying to get dataset URI
        returnval = 'noInformation'
    return returnval


# In this one, it is tricky, we assume that datset records have already been transferred to the target
# Look up the matching dataset based on geometry and get the GE4 value from the dataset feature       
def transform_GE4_from_IC2_Dataset(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
            settings.report_values['Transform processing: Trying to get country code (GENC) from the dataset GE4'] += 1
            return get_GE4_from_matching_dataset(settings, row, field_in, target_field, src_fc, target_fc)
        return row[field_in]
    return -999999


# Set GE4 from IC2. This can be done if IC2 is in the source row, otherwise return the default value
def transform_GE4_from_IC2(settings, row, field_in, target_field, src_fc, target_fc):
    if 'IC2' in row and row['IC2'] is not None:
        settings.report_values['GENC Country derived from IC2 Country'] += 1
        return get_codelist2codelistmapping_value("IC2", "GE4", row['IC2'])
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                      "Put results dict here", False)


# this transform takes the codelist input value and determines the enumerated domain value for the target
def transformZI004_RCGfromcodelisttoenum(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if row[field_in] in ZI004_RCGcodelist2ZI004_RCGenumvalue:
            settings.report_values['transform ZI004_RCG from codelist to enum'] += 1
            return ZI004_RCGcodelist2ZI004_RCGenumvalue[row[field_in]]
        elif row[field_in] is None:
            return -999999
        # If the value is already a valid enum, return it
        elif row[field_in] in RCGenumvalue2RCGcodelist:
            return row[field_in]
        else:
            report_transfomer_issue_or_increment_count(
                settings,
                "Illegal input value. This value cannot be transformed via transformZI004_RCGfromcodelisttoenum",
                "ERROR", "Field: {0}, Value: {1}".format(field_in, strout(row[field_in])))
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], ''), 'OTH')
            # Save to OTH, other
            return 999
    return -999999


def transformRCGcodelist2RCGenumvalue(settings, row, field_in, target_field, src_fc, target_fc):
    # First test if the field has a value and is in the row
    if field_in is not None and field_in in row:
        if row[field_in] in RCGcodelist2RCGenumvalue:
            settings.report_values['transform RCG from codelist to enum'] += 1
            return RCGcodelist2RCGenumvalue[row[field_in]]
        # Now check if the field value is actually already in the correct target format
        if row[field_in] in RCGenumvalue2RCGcodelist:
            return row[field_in]
    return -999999


def transformZI004_RCGenumeration2ZI004_RCGcodelist(settings, row, field_in, target_field, src_fc, target_fc):
    # First test if the field has a value and is valid in the row
    if field_in is not None and field_in in row:
        # Next see if the field value is in the enum list, and return the code list mapping
        if row[field_in] in ZI004_RCGenumvalue2ZI004_RCGcodelist:
            settings.report_values['transform ZI004_RCG from enum to codelist'] += 1
            return ZI004_RCGenumvalue2ZI004_RCGcodelist[row[field_in]]
        # Test to see if the field is already a codelist and return the value
        if row[field_in] in RCGcodelist2RCGenumvalue:
            return row[field_in]
    return 'noInformation'


def transformRCGenumeration2RCGcodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if row[field_in] in ZI004_RCGenumvalue2ZI004_RCGcodelist:
            settings.report_values['transform RCG from enum to codelist'] += 1
            return ZI004_RCGenumvalue2ZI004_RCGcodelist[row[field_in]]
        # Test if the value is already the target codelist and if so, return it
        if row[field_in] in RCGcodelist2RCGenumvalue:
            return row[field_in]
    return 'noInformation'


def transformCPScodelist2CPSenumvalue(settings, row, field_in, target_field, src_fc, target_fc):
    if 'CPS' in row and row['CPS'] is not None:
        if row['CPS'] in CPScodelist2CPSenumvalue:
            settings.report_values['transform CPS from codelist to enum'] += 1
            return CPScodelist2CPSenumvalue[row['CPS']]
        # Test if the value is already an enum
        if row['CPS'] in CPSenumvalue2CPScodelist:
            return row[field_in]
    return -999999


def transformETScodelist2ETSenumvalue(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ETS' in row and row['ETS'] is not None:
        if row['ETS'] in ETScodelist2ETSenumvalue:
            settings.report_values['transform ETS from codelist to enum'] += 1
            return ETScodelist2ETSenumvalue[row['ETS']]
        # Check to see if the ETS value is already in the correct target format
        if row['ETS'] in ETSenumvalue2ETScodelist:
            return row['ETS']
    return -999999


def transformHZDcodelist2HZDenumvalue(settings, row, field_in, target_field, src_fc, target_fc):
    if 'HZD' in row and row['HZD'] is not None:
        if row['HZD'] in HZDcodelist2HZDenumvalue:
            settings.report_values['transform HZD from codelist to enum'] += 1
            return HZDcodelist2HZDenumvalue[row['HZD']]
        # If the attribute is already in the correct target format, return it
        if row['HZD'] in HZDenumvalue2HZDcodelist:
            return row['HZD']
    return -999999


def transformVDTcodelist2VDTenumvalue(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if row[field_in] in VDTcodelist2VDTenumvalue:
            settings.report_values['transform VDT from codelist to enum'] += 1
            return VDTcodelist2VDTenumvalue[row[field_in]]
        # Check to see if the value is already in the correct format
        if row[field_in] in VDTenumvalue2VDTcodelist:
            return row[field_in]
    return -999999


def transformZI001_SRTcodelist2ZI001_SRTenumvalue(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI001_SRT' in row and row['ZI001_SRT'] is not None:
        if row['ZI001_SRT'] in ZI001_SRTcodelist2ZI001_SRTenumvalue:
            settings.report_values['transform ZI001_SRT from codelist to enum'] += 1
            return ZI001_SRTcodelist2ZI001_SRTenumvalue[row['ZI001_SRT']]
        # If the value is already a valid enum value, return that enum value
        if row['ZI001_SRT'] in ZI001_SRTenumvalue2ZI001_SRTcodelist:
            return row['ZI001_SRT']
    return -999999


def transformZI001_VSCcodelist2ZI001_VSCenumvalue(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI001_VSC' in row and row['ZI001_VSC'] is not None:
        if row['ZI001_VSC'] in ZI001_VSCcodelist2ZI001_VSCenumvalue:
            settings.report_values['transform ZI001_VSC from codelist to enum'] += 1
            return ZI001_VSCcodelist2ZI001_VSCenumvalue[row['ZI001_VSC']]
        # If the value is already a valid enum value, return the enum
        if row['ZI001_VSC'] in ZI001_VSCenumvalue2ZI001_VSCcodelist:
            return row['ZI001_VSC']
    return -999999


def transformZI001_VSCenumvalue2ZI001_VSCcodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI001_VSC' in row and row['ZI001_VSC'] is not None:
        if row['ZI001_VSC'] in ZI001_VSCenumvalue2ZI001_VSCcodelist:
            settings.report_values['transform ZI001_VSC from enum to codelist'] += 1
            return ZI001_VSCenumvalue2ZI001_VSCcodelist[row['ZI001_VSC']]
        # If the value is already a valid codelist value, just return it
        if row['ZI001_VSC'] in ZI001_VSCcodelist2ZI001_VSCenumvalue:
            return row['ZI001_VSC']
    return 'noInformation'


def transformZI001_SRTenumeration2ZI001_SRTcodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI001_SRT' in row and row['ZI001_SRT'] is not None:
        if row['ZI001_SRT'] in ZI001_SRTenumvalue2ZI001_SRTcodelist:
            settings.report_values['transform ZI001_SRT from enum to codelist'] += 1
            return ZI001_SRTenumvalue2ZI001_SRTcodelist[row['ZI001_SRT']]
        # If we want a codelist and it is already a codelist, return the value
        if row['ZI001_SRT'] in ZI001_SRTcodelist2ZI001_SRTenumvalue:
            return row['ZI001_SRT']
    return 'noInformation'


def transformZVH_VDTenumeration2ZVH_VDTcodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row and row[field_in] is not None:
        if row[field_in] in VDTenumvalue2VDTcodelist:
            settings.report_values['transform ZVH_VDT from enum to codelist'] += 1
            return VDTenumvalue2VDTcodelist[row[field_in]]
        # Check to see if the value is already in the correct format
        if row[field_in] in VDTcodelist2VDTenumvalue:
            return row[field_in]
    return 'noInformation'


def transformVDTenumeration2VDTcodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row and row[field_in] is not None:
        if row[field_in] in VDTenumvalue2VDTcodelist:
            settings.report_values['transform VDT from enum to codelist'] += 1
            return VDTenumvalue2VDTcodelist[row[field_in]]
        # Check to see if the value is already in the correct format
        if row[field_in] in VDTcodelist2VDTenumvalue:
            return row[field_in]
    return 'noInformation'


def transformHZDenumeration2HZDcodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row and row[field_in] is not None:
        if row[field_in] in HZDenumvalue2HZDcodelist:
            settings.report_values['transform HZD from enum to codelist'] += 1
            return HZDenumvalue2HZDcodelist[row[field_in]]
        # Check to see if the value is already in the correct format
        if row[field_in] in HZDcodelist2HZDenumvalue:
            return row[field_in]
    return 'noInformation'


def transformETSenumeration2ETScodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row and row[field_in] is not None:
        if row[field_in] in ETSenumvalue2ETScodelist:
            settings.report_values['transform ETS from enum to codelist'] += 1
            return ETSenumvalue2ETScodelist[row[field_in]]
        # Check to see if the value is already a codelist
        if row[field_in] in ETScodelist2ETSenumvalue:
            return row[field_in]
    return 'noInformation'


def transformCPSenumeration2CPScodelist(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row and row[field_in] is not None:
        if row[field_in] in CPSenumvalue2CPScodelist:
            settings.report_values['transform CPS from enum to codelist'] += 1
            return CPSenumvalue2CPScodelist[row[field_in]]
        # If the value is already a codelist, return it
        if row[field_in] in CPScodelist2CPSenumvalue:
            return row[field_in]
    return 'noInformation'


def find_matchrestrictioninfoCCNCDR(settings, row, field_in, target_field, src_fc, target_fc):
    return 'noInformation'
    # Look for restriction info having this CCN, CDR and all other values as no information!
    resultsetA = set()
    resultsetB = set()
    finset = set()
    if 'ZI002_CCN' in row and 'ZI002_CDR' in row:
        resultsetA = search_for_restriction_info('CCN', row['ZI002_CCN'], settings.saved_restrictioninfo)
        resultsetB = search_for_restriction_info('CDR', row['ZI002_CDR'], settings.saved_restrictioninfo)
        # We have to get two results back for this to work
        if len(resultsetA) > 0 and len(resultsetB) > 0:
            finset = resultsetA.intersect(resultsetB)
            if len(finset) == 1:
                return finset[0]
    return 'noInformation'


# This transform is specific to GGDM 3.0 -> GGDM 2.1 or TDS 4.0
# It looks for saved restriction information and saves CCN or CDR into OTH if they are not default
# It saves ZSAX_RX0 into OTH if it is not default
# It saves ZI020_GE4 into OTH if it is not default
def transform_SaveSomeRestrictionInfoToOTH(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI002_UFI' in row:
        # If the UFI is saved as a restriction info
        if is_UFI_in_restriction_info(row['ZI002_UFI'], settings.saved_restrictioninfo) == True:
            # Get CCN and CDR attributes
            rCCN = get_restriction_info_field(row['ZI002_UFI'], 'CCN', settings.saved_restrictioninfo)
            rCDR = get_restriction_info_field(row['ZI002_UFI'], 'CDR', settings.saved_restrictioninfo)
            # If CCN/CDR is not default, save it as OTH
            if rCCN is not None and rCCN.lower() != "noinformation":
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CCN', rCCN, ''), 'OTH')
            if rCDR is not None and rCDR.lower() != "noinformation":
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CDR', rCDR, ''), 'OTH')
    # Also save ZSAX_RX0 if it is not default
    if 'ZSAX_RX0' in row:
        if row['ZSAX_RX0'].lower() != "noinformation":
            # save to OTH
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZSAX_RX0', row['ZSAX_RX0'], ''), 'OTH')
    if 'ZI020_GE4' in row:
        if row['ZI020_GE4'].lower() != "noinformation":
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZI020_GE4', row['ZI020_GE4'], ''), 'OTH')
        #   add_to_row_OTH(settings,row, prepare_string_for_OTH_or_MEM('IC2', GE4codelist2IC2codelist[row['ZI020_GE4']], ''), 'OTH')
    return None


def transform_saveECUFIToOther(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI039S_UFI' in row and row['ZI039S_UFI'] is not None and row['ZI039S_UFI'].lower() != 'noinformation':
        # Save the entity collection item UFI so that if we translate this data back, we can possibly populate the UFI again.
        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZI039S_UFI', row['ZI039S_UFI'], ''), 'OTH')


def transform_saveDatasetURItoOther(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI031S_URI' in row and row['ZI031S_URI'] is not None and row['ZI031S_URI'].lower() != 'noinformation':
        # Save the Dataset URI so that if we translate this data back, we can possibly populate the URI again.
        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZI031S_URI', row['ZI031S_URI'], ''), 'OTH')


saved_err_stmt = str()
saved_err_stmt2 = str()


# This transform is specific to GGDM 3.0 -> GGDM 2.1 or TDS 4.0
# It is used to save additional information into the saved Entity Collection records.
# Assumption is that prior to running this transform, all potential entity collection records from source and target are already saved.
# Upon completion of the processing of all features, the EC information will then be saved
def transform_SaveRelatedECInformation(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI039S_UFI' in row and row['ZI039S_UFI'] is not None and row['ZI039S_UFI'].lower() != 'noinformation':

        # If the UFI is saved as a restriction info, get the dictionary
        ecdict = get_entity_collection_dictionary(row['ZI039S_UFI'], settings.saved_entity_collection_metadata)
        if ecdict is not None:
            # We do have an ec in the list having this UFI, save ALL attributes related
            # to this EC if they are not already there.
            # If our source data has Source Date and Time and that structured text is not set to the default, and if
            # it is not already saved with this entity collection entry, save it
            # Source Date
            if 'ZI001_SDV' in row:
                if 'ZI001_SDV' not in ecdict:
                    ecdict['ZI001_SDV'] = row['ZI001_SDV']
                else:
                    if ecdict['ZI001_SDV'].lower() != row['ZI001_SDV'].lower():
                        # Since we tend to get many of these in one set, print one message and + for each identical one
                        curr_err_stmt = "Error in source data, metadata values for ZI001_SDV are not consistent for features having the same entity collection metadata."
                        detail_stmt = "Feature Class {0}, inconsistent data: {1} current ec data: {2}" \
                            .format(src_fc, strout(row['ZI001_SDV']), strout(ecdict['ZI001_SDV']))
                        report_transfomer_issue_or_increment_count(settings, curr_err_stmt, "ERROR", detail_stmt)

            # Source Type
            if 'ZI001_SRT' in row:
                if 'ZI001_SRT' not in ecdict:
                    ecdict['ZI001_SRT'] = row['ZI001_SRT']
                else:
                    if ecdict['ZI001_SRT'] != row['ZI001_SRT']:
                        curr_err_stmt = "Error in source data, metadata values for ZI001_SRT are not consistent for features having the same entity collection metadata."
                        detail_stmt = "Feature Class {0}, inconsistent data: {1} current ec data: {2}." \
                            .format(src_fc, strout(row['ZI001_SRT']), strout(ecdict['ZI001_SRT']))
                        report_transfomer_issue_or_increment_count(settings, curr_err_stmt, "ERROR", detail_stmt)

    return None


def transform_set_UFI(settings, row, field_in, target_field, src_fc, target_fc):
    locu = get_row_unique_id(settings, row)
    if settings.debug_settings['report_unique_id_usage'] > 0:
        settings.my_logger.debug("transform_set_UFI: {0}  was {1}".format(locu, str(locu)))
    return (locu)


def transform_set_URI(settings, row, field_in, target_field, src_fc, target_fc):
    locu = get_row_unique_id(settings, row)
    if settings.debug_settings['report_unique_id_usage'] > 0:
        settings.my_logger.debug("transform_set_URI: {0}  was {1}".format(locu, str(locu)))
    return (locu)


# If the UFI/URI is being set based on globalID, store the original URI or URI into the FUFI field
def transform_set_FUFI_to_UFI_when_needed(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.config_setting_flag['flag_ufi_is_not_unique_generate_new_unique_ufi'] > 1:
        if 'FUFI' in row:
            return row['FUFI']
        elif 'UFI' in row:
            return row['UFI']
        elif 'URI' in row:
            return row['URI']
        else:
            return 'noInformation'
    else:
        if 'FUFI' in row:
            return row['FUFI']
        else:
            return 'noInformation'

        # a default transformation that specifically returns None


def transform_placeholder_return_None(settings, row, field_in, target_field, src_fc, target_fc):
    settings.report_values['placeholder transform returned None'] += 1
    return None


# a default transformation that passes through the source value unchanged
def transform_placeholder_return_source(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        settings.report_values['placeholder transform - source value retained'] += 1
        return row[field_in]
    return None


# this transform sets the schema version to a pre-determined value
def transform_set_schemaversion(settings, row, field_in, target_field, src_fc, target_fc):
    settings.report_values['Schema version set to target geodatabase version'] += 1
    return settings.schema_version[settings.gettarget_db_version()]


def transform_set_to_default(settings, row, field_in, target_field, src_fc, target_fc):
    # print 'Field: {0} SrcFC: {1} TargetFC: {2}'.format(field_in, src_fc, target_fc)
    resultdict = dict()
    resultdict['DefaultVal'] = None
    valuedefault = settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                          resultdict, 'DefaultVal')
    if resultdict['DefaultVal'] in ['Success', 'NA Attribute']:
        return valuedefault
    return None


def transform_set_EntityCollection_field_to_default(settings, row, field_in, target_field, src_fc, target_fc):
    # Most often, the field_in is not relevant
    datatype = settings.targetdb.get_field_datatype(target_fc, target_field)
    if datatype is not None:
        # validfcodes =  targetdb.getvalidcodesforfeatureclassfield('MetadataSrf', field_in)
        vval = settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)
        return vval
    else:
        report_transfomer_issue_or_increment_count(settings,
                                                   "Received request to set entity collection default but field does not exist in schema definition.",
                                                   "WARN", "Feature: MetadataSrf, Field: {0}".format(target_field))
        return None


def transform_GWS(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if settings.getsource_db_version() == "GGDM 3.0" and (
                settings.gettarget_db_version() == "GGDM 2.1" or settings.gettarget_db_version() == "GGDM 2.2"):
            settings.report_values['Groundwater Potential Scale value conversion'] += 1
            if row[field_in] == 1:
                return 0
            elif row[field_in] == 2:
                return 0.01
            elif row[field_in] == 3:
                return 5
            elif row[field_in] == 4:
                return 500
            elif row[field_in] == 999:
                return -999999
            elif row[field_in] == -999999:
                return -999999
            else:
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in],
                                                                            settings.sourcedb.get_enum_text_for_index(
                                                                                field_in, row[field_in])), 'OTH')
                return -999999
        elif settings.getsource_db_version() == "GGDM 2.1" and (
                settings.gettarget_db_version() == "GGDM 3.0" or settings.gettarget_db_version() == "GGDM 2.2"):
            settings.report_values['Groundwater Potential Scale value conversion'] += 1
            if row[field_in] == -999999:
                return -999999
            elif row[field_in] <= 0:
                return 1
            elif row[field_in] > 0 and row[field_in] <= 0.015749:
                return 2
            elif row[field_in] > 0.015749 and row[field_in] <= 400:
                return 3
            elif row[field_in] > 400 and row[field_in] <= 10000:
                return 4
        else:
            return row[field_in]
    return -999999


def transform_report_issue(settings, row, field_in, target_field, src_fc, target_fc):
    report_transfomer_issue_or_increment_count(
        settings, 'EQS may be not be derived from EQC because it is in a different feature',
        "WARNING", "")
    return transform_set_to_default(settings, None, None, target_field, None, target_fc)


def transform_set_geodbversion(settings, row, field_in, target_field, src_fc, target_fc):
    settings.report_values['Geodatabase version set to specified value'] += 1
    return settings.geodb_version[settings.gettarget_db_version()]


# this transform takes the scale enumerated domain value and determines the corresponding scale
# numeric (denominator) based on the enumerated domain value names.
def transformEQCtoEQS(settings, row, field_in, target_field, src_fc, target_fc):
    # Returns an integer value The denominator (for example: 50,000) of the map scale that best represents the 
    # density of feature collection within the geospatial extent of a specified region.   
    # [DatatypeSpecification] An integer domain value that is restricted to be non-negative.
    if 'EQC' in row:
        if row['EQC'] in EQCcodelist2numericEQS:
            settings.report_values['Density Equivalent Scale from Equivalent Scale Category'] += 1
            return EQCcodelist2numericEQS[row['EQC']]
        else:
            report_transfomer_issue_or_increment_count(settings,
                                                       "transform of EQC to EQS illegal codelist value to transform to numeric",
                                                       "WARNING",
                                                       "Illegal value: {0}".format(row['EQC']))
    elif 'EQS' in row:
        return row['EQS']

    return -999999


def transformEQStoEQC(settings, row, field_in, target_field, src_fc, target_fc):
    # Returns an enumerated domain value.
    if 'EQS' in row:
        if row['EQS'] in EQSnumeric2codelistEQC:
            settings.report_values['Equivalent Scale Category from Density Equivalent Scale'] += 1
            return EQSnumeric2codelistEQC[row['EQS']]
        else:
            report_transfomer_issue_or_increment_count(settings,
                                                       "transform of EQS to EQC illegal numeric value to transform to codelist",
                                                       "WARNING",
                                                       "Illegal value: {0}".format(row['EQS']))
    elif 'EQC' in row:
        return row['EQC']

    return -999999


# Extend the shape area
def transform_extend_shape_to_include_source(settings, row, field_in, target_field, src_fc, target_fc):
    settings.report_values['Extended shape to include source feature instance shape'] += 1
    report_transfomer_issue_or_increment_count(settings,
                                               "transform_extend_shape_to_include_source is UNDEFINED and is only a placeholder",
                                               "ERROR")
    return None


def transform_default_value_string(settings, row, field_in, target_field, src_fc, target_fc):
    settings.report_values['string default value'] += 1
    return 'No Information'


def transform_metre_to_decimetre(settings, row, field_in, target_field, src_fc, target_fc):
    # change incoming units from metre to decimetre, if this complex logic fails
    # Baseically, if the feature is ED020 we don't do the conversion or
    # If the source is GGDM 2.1 or TDS 4.0 and the feature is ED010.
    if field_in is not None and field_in in row:
        if settings.sourcedb.get_f_code() == 'ED020' or \
                ((settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]) and \
                 settings.sourcedb.get_f_code() == 'ED010'):
            return row[field_in]

        if is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == False:
            settings.report_values['Transform unit conversion: Metre converted to decimetre'] += 1
            return (row[field_in] * 10)
        else:
            return (row[field_in])
    return -999999


def transform_decimetre_to_metre(settings, row, field_in, target_field, src_fc, target_fc):
    # change incoming units from decimetre to metre
    if field_in is not None and field_in in row:
        if settings.sourcedb.get_f_code() == 'ED020' or \
                ((settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]) and \
                 settings.sourcedb.get_f_code() == 'ED010'):
            return row[field_in]

        if is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == False:
            settings.report_values['Transform unit conversion: Decimetre converted to metre'] += 1
            return (row[field_in] / 10)
        else:
            return (row[field_in])
    return -999999


def transform_gramspercm_to_mgperlitre(settings, row, field_in, target_field, src_fc, target_fc):
    # change - this is actually the same. A gram per cubic meter is equivalent to a milligram per litre
    return (row[field_in])


def transform_mgperlitre_to_gramspercm(settings, row, field_in, target_field, src_fc, target_fc):
    # change this is actually the same. A gram per cubic meter is equivalent to a milligram per litre
    return (row[field_in])


def transform_SRE(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
            if row[field_in] == 4:  # Sharp Curve
                settings.report_values['Traffic restriction type Sharp Curve or Steep Grade corrected'] += 1
                return 8  # Sharp Curves
            elif row[field_in] == 5:  # Steep Grade
                settings.report_values['Traffic restriction type Sharp Curve or Steep Grade corrected'] += 1
                return 9  # Steep Grades
        return row[field_in]
    return -999999


##def transform_find_entity_collection_id(settings, row, field_in, target_field, src_fc, target_fc):
##    # create a temp dictionary containing the restriction info elements from this source feature
##    if len(saved_entity_collection_metadata) == 0:
##        return 'noInformation'
##    newtemprow = {}
##
##    SubsetFields = ['ZI001_SDP','ZI001_SDV','ZI001_SPS','ZI001_SRT','ZI001_VSC','ZI001_VSD','ZI001_VSN','ZI004_PRE','OTH']
##    # Create a temporary fow for comparison or possible insertion
##    # for src, target, transformer in mapping:
##    for field in SubsetFields:
##        if field == 'OTH':
##            newtemprow.append ( (field, call_transform_function['transformMoveOTHOnlyIfMetaAttribute'](settings, row, field_in, target_field, src_fc, target_fc) ) )
##        elif field == 'ZI001_SRT':
##            newtemprow.append ( (field, call_transform_function['transformZI001_SRTfromcodelisttoenum'](settings, row, field_in, target_field, src_fc, target_fc) ) )
##        elif field == 'ZI001_VSC':
##            newtemprow.append ( (field, call_transform_function['transformZI001_VSCfromcodelisttoenum'](settings, row, field_in, target_field, src_fc, target_fc) ) )
##        else:
##            newtemprow.append( (field, row[field]) )
##    tempvalue = compare_src_against_existing_metainfo(newtemprow, saved_entity_collection_metadata, EntityCollectionFields)
##    # settings.my_logger.debug( 'Find entity collection id value {0}'.format(str(tempvalue)))
##    if tempvalue is None:
##        settings.report_values['no associated entity collection'] = settings.report_values['no associated entity collection'] + 1
##        return 'NoInformation'
##    else:
##        settings.report_values['associated entity collection'] = settings.report_values['associated entity collection'] + 1
##        return tempvalue

# This function compares current row content against
# a saved list of dictionaries that hold all metadata. This function is trying to find the
# unique identifier, UFI from the saved metadata that represents the incoming row of metadata.
# The incoming row may be a subset from any feature class, and some of the content may have to
# be transformed in order to match and this function assumes it already has been transformed.
# We also assume the incoming row has already installed default values where necessry.

def compare_against_existing_metainfo_noxform(settings, row_in_to_compare, saved_metainfo):
    # Do not bother to compare these following fields
    fieldskiplist = ['scvn', 'ufi']
    fieldskiplist.extend(settings.targetdb.common_fields_geom_and_id_and_code)

    # First, if there is no saved metadata to search we can just return
    if len(saved_metainfo) <= 0:
        if settings.debug_settings['compare_src_against_existing_metainfo'] > 1:
            settings.my_logger.debug("compare_src_against_existing_metainfo, no saved metadata exists")
        return None

    # Verify that ths incoming row content is a dictionary and not just a list!
    if isinstance(row_in_to_compare, dict) == False:
        settings.my_logger.error("compare_src_against_existing_metainfo, incoming content is not a dictionary")
        return None

    if settings.debug_settings['compare_src_against_existing_metainfo'] > 0:
        settings.my_logger.debug("Saved metadata records:" + str(len(saved_metainfo)))

    # Loop through the saved meatadata and compare the fields from the parameter list to the metadata and when found, return the UFI
    meta_rec_count = 0
    for saved_m_record in saved_metainfo:
        valuematch = True
        if 'UFI' in saved_m_record:
            curUFI = saved_m_record['UFI']
        else:
            curUFI = None
        # for each field in the saved metadata record, check against our incoming row
        for field_meta in saved_m_record:
            if field_meta is not None and field_meta.lower() not in fieldskiplist and field_meta in row_in_to_compare:
                valuematch = (saved_m_record[field_meta] == row_in_to_compare[field_meta]) or (
                        unicode(saved_m_record[field_meta]).lower() == unicode(
                    row_in_to_compare[field_meta]).lower())
                if valuematch == False and settings.debug_settings['compare_src_against_existing_metainfo'] > 0:
                    report_transfomer_issue_or_increment_count(
                        settings, "Metadata not matched to saved metadata", "DEBUG",
                        "Failed to match row to meta record {3} field {0} having value {1} to metadata with value {2}". \
                            format(field_meta, saved_m_record[field_meta], row_in_to_compare[field_meta],
                                   meta_rec_count))
                if valuematch == False:  # If the match failed we are done with the fields
                    break  # Get out of this for loop on fields and go to the next metadata record

        # If we got through all of the fields and got a match, we are done. Return the UFI
        if valuematch:
            return curUFI

        meta_rec_count += 1

    # end for each saved metadata item
    return None


def transform_find_entity_collection_id(settings, row, field_in, target_field, src_fc, target_fc):
    # create a temp dictionary containing the restriction info elements from this source feature
    if len(settings.saved_entity_collection_metadata) == 0:
        return 'noInformation'

    newtemprow = dict()
    # SubsetFields = ['ZI001_SDP','ZI001_SDV','ZI001_SPS','ZI001_SRT','ZI001_VSC','ZI001_VSD','ZI001_VSN','ZI004_PRE','OTH']
    # Create a temporary fow for comparison or possible insertion
    for src_xformFECI, target_xformFECI, transformer_xformFECI in settings.fandf.EntityCollectionFCSubsetFields:
        if target_xformFECI is None and src_xformFECI is not None:
            target_xformFECI = src_xformFECI
        if src_xformFECI is not None and src_xformFECI in row and target_xformFECI is not None and (
                transformer_xformFECI is None or transformer_xformFECI not in call_transform_function):
            newtemprow[target_xformFECI] = row[src_xformFECI]
        elif src_xformFECI is not None and src_xformFECI in row and (
                transformer_xformFECI is not None and transformer_xformFECI in call_transform_function):
            newtemprow[target_xformFECI] = call_transform_function[transformer_xformFECI](settings, row, src_xformFECI,
                                                                                          target_xformFECI, src_fc,
                                                                                          target_fc)
        elif src_xformFECI is None and target_xformFECI is not None and (
                transformer_xformFECI is not None and transformer_xformFECI in call_transform_function):
            newtemprow[target_xformFECI] = call_transform_function[transformer_xformFECI](settings, row,
                                                                                          target_xformFECI,
                                                                                          target_xformFECI, src_fc,
                                                                                          target_fc)
        elif src_xformFECI is None and target_xformFECI is not None:
            newtemprow[target_xformFECI] = settings.fandf.EntityCollectionTargetDefaultFieldValues[target_xformFECI]

    # Set the rest to defaults  
    for f_fieldx in settings.fandf.EntityCollectionTargetDefaultFieldValues:
        if f_fieldx is not None and f_fieldx not in newtemprow:
            newtemprow[f_fieldx] = settings.fandf.EntityCollectionTargetDefaultFieldValues[f_fieldx]

    # def compare_src_against_existing_metainfo(settings, rowinorig, saved_metainfo, src_fc, 
    #     targ_fc, MetaSrcTargetMapperForFields, MaxRemainingFields = 1, CompareToTargetRunTransforms = True):
    # tempvalue = compare_src_against_existing_metainfo(settings, newtemprow,
    #                 settings.saved_entity_collection_metadata, 'ENTITY_COLLECTION_METADATA_T',
    #                 'MetadataSrf',  settings.fandf.EntityCollectionFields)
    tempvalue = compare_against_existing_metainfo_noxform(settings, newtemprow,
                                                          settings.saved_entity_collection_metadata)

    if settings.debug_settings['report_entity_collection_findings'] > 0:
        settings.my_logger.debug('Find entity collection id value {0}'.format(str(tempvalue)))

    if tempvalue is None:
        settings.report_values['Entity Collection - No associated entity collection identified for feature'] += 1
        return 'NoInformation'
    else:
        settings.report_values['Entity Collection - Associated entity collection was identified for feature'] += 1
        return tempvalue


# YIKE ! There are issues and we really cannot rely on finding the correct entity collection identifier
# based on the shape overlap. Currently disabled.
def transform_find_entity_collection_id_by_shape(settings, row, field_in, target_field, src_fc, target_fc):
    ''' Find and return UFI for the entity collection having spatial overlap with this feature'''
    return 'noInformation'
    # This function is not functional.
    foundcount = 0
    if row['SHAPE@'] and not settings.config_setting_flag[
                                 'flag_skip_search_for_applicable_entity_collection_while_testing'] == 1:
        # The feature class name is different depending on the schema version we are working with

        possibleecs = ["CollectionSrf", "MetadataSrf"]  # , "ENTITY_COLLECTION_METADATA_T"]
        doneflag = False
        returnval = 'noInformation'
        for fcname in possibleecs:
            target_obj_path = settings.targetdb.get_root_for_fc_subtypes(fcname)
            if settings.targetdb.is_featureclass_in_geodatabase_spec(fcname):
                # for each entity collection record, does the shape of this feature overlap the dataset shape?
                with arcpy.da.SearchCursor(target_obj_path, ('URI', 'SHAPE@'), None) as row22AA:
                    for row22 in row22AA:
                        if row22[1].disjoint(row['SHAPE@']) == True:
                            pass  # this is a faster test, no points in common, this is not the one we are looking for
                        elif row22[1].contains(row['SHAPE@']) == True:
                            # print 'Entity Collection UFI {0} matches feature {1}'.format(str(row22[0]), str(row['UFI']))
                            returnval = row22[0]
                            doneflag = True
                            foundcount = foundcount + 1
                            # break  # If we are sure there are not going to be any overlaps, we could just break out of this here
                if returnval == 'noInformation':
                    settings.report_values[
                        'Entity Collection - No associated entity collection identified for feature'] += 1
                elif foundcount > 1:
                    if settings.debug_settings['itemize_each_translation_each_field'] > 0:
                        report_transfomer_issue_or_increment_count(settings,
                                                                   "Multiple associated entity collections identified for one feature instance",
                                                                   "WARN",
                                                                   "Found {0} possibilities in feature class {1} having UFI {2}".format(
                                                                       foundcount, src_fc, str(row['UFI'])))
                else:
                    settings.report_values[
                        'Entity Collection - Associated entity collection was identified for feature'] += 1
            if doneflag:
                break
    else:  # If there is no shape value to test against, don't bother trying to get dataset URI
        report_transfomer_issue_or_increment_count(settings,
                                                   "Either there is no geometry or entity collection search has been disabled",
                                                   "WARN")
        returnval = 'noInformation'
    return returnval


# This transform utility looks in our global set of restriction information content to see if this
# feature (or feature class) instance is already represented, and gets the UFI for that representation.
def transform_find_restriction_info_id(settings, row, field_in, target_field, src_fc, target_fc):
    # create a temp dictionary containing the restriction info elements from this source feature  
    if len(settings.saved_restrictioninfo) == 0:
        return 'noInformation'

    # if settings.getsource_db_version() == "GGDM 2.2":      # YIKE GGDM 2.2 return default restriction info always
    #     return settings.saved_restrictioninfo[0]['UFI']

    # List of the fields we are interested in for this restriction information test
    SubsetFields = ['CCN', 'CDR', 'SAX_RS1', 'SAX_RS2', 'SAX_RS4', 'SAX_RS5', 'SAX_RS6',
                    'SAX_RS8', 'SAX_RS9', 'SAX_RX1', 'SAX_RX2', 'SAX_RX5',
                    'SAX_RX6', 'SAX_RX7', 'SAX_RX9', 'SAX_RY0', 'SAX_RY1', 'SAX_RY2']

    RenameFields = {'ZI002_CCN': 'CCN', 'ZI002_CDR': 'CDR'}
    newtemprow = {}
    verificationtest = True
    expectedfieldcount = len(SubsetFields)
    # Create a dictionary from the subset fields and values from the input data
    for field in SubsetFields:
        if field in row:
            expectedfieldcount -= 1
            newtemprow[field] = row[field]
    # There are a couple of fields that might be renamed
    for field in RenameFields:
        if field in row:
            expectedfieldcount -= 1
            newtemprow[RenameFields[field]] = row[field]
    # One last time, fill anything missing with default values
    for field in settings.fandf.RestrictionInfoTargetDefaultFieldValues:
        if field not in newtemprow:
            newtemprow[field] = settings.fandf.RestrictionInfoTargetDefaultFieldValues[field]
    # newtemprow['F_CODE'] = settings.fandf.RestrictionInfoTargetDefaultFieldValues['F_CODE']

    if verificationtest == True:
        # tempvalue = compare_src_against_existing_metainfo(settings, newtemprow, settings.saved_restrictioninfo,
        #                            'RESTRICTION_INFO_T', 'RESTRICTION_INFO_T', settings.fandf.RestrictionInfoTblFields)
        tempvalue = compare_against_existing_metainfo_noxform(settings, newtemprow, settings.saved_restrictioninfo)

        if tempvalue is None:
            report_transfomer_issue_or_increment_count(
                settings,
                'Restriction Information - No associated restriction information identified for feature, insufficient match',
                "WARNING",
                "   Source fields did not sufficiently match saved restriction information")
            return 'noInformation'
        else:
            settings.report_values[
                'Restriction Information - Associated restriction information found for feature'] += 1
            return tempvalue
    else:
        report_transfomer_issue_or_increment_count(
            settings,
            'Restriction Information - No associated restriction information identified for feature, no fields',
            "WARNING",
            "   No fields could be identified in source to match to saved restriction information")
        return 'noInformation'


# YIKE, we cannot do this yet. too many unmatched values
# assuming the matching dataset is in ResoureceSrf feature class
def transform_find_dataset_id(settings, row, field_in, target_field, src_fc, target_fc):
    """Find and return URI for dataset having spatial overlap with this feature"""

    # print "   transform_find_dataset_id"
    foundcount = 0
    if row['SHAPE@'] and settings.config_setting_flag['flag_skip_search_for_applicable_dataset_while_testing'] == 0:
        # The feature class name is different depending on the schema version we are working with
        possibledatasets = ["DatasetSrf", "ResourceSrf"]  # , "DATASET_T"]
        doneflag = False
        returnval = 'noInformation'
        for fcname in possibledatasets:

            if settings.targetdb.is_featureclass_in_geodatabase_spec(fcname):
                target_obj_path = settings.targetdb.get_root_for_fc_subtypes(fcname)
                # for each dataset record, does the shape of this feature overlap the dataset shape?
                with arcpy.da.SearchCursor(target_obj_path, ('URI', 'SHAPE@'), None) as row22AA:
                    for row22 in row22AA:
                        if row22 is None:
                            continue
                        elif len(row22) < 1:
                            continue
                        if row22[1] is None:
                            # TODO Should we print dataset has no geometry?
                            pass
                        elif row22[1].disjoint(row['SHAPE@']):
                            pass  # this is a faster test, no points in common, this is not the one we are looking for
                        elif row22[1].contains(row['SHAPE@']):
                            # print 'Dataset URI {0} matches feature {1}'.format(str(row22[0]), str(row['UFI']))
                            returnval = row22[0]
                            doneflag = True
                            foundcount = foundcount + 1
                            # break  # If we are sure there are not going to be any overlaps,
                            # we could just break out of this here
                    if returnval == 'noInformation':
                        settings.report_values['Dataset - No associated dataset identified for feature instance'] += 1
                    elif foundcount > 1:
                        if settings.debug_settings['itemize_each_translation_each_field'] > 0:
                            report_transfomer_issue_or_increment_count(
                                settings,
                                'Dataset - WARNING Multiple associated datasets were identified for one feature instance',
                                "WARN",
                                "Multiple {0} associated datasets identified for feature instance in feature class {1} having UFI {2}".format(
                                    foundcount, src_fc, str(row['UFI'])))
                    else:
                        settings.report_values['Dataset - Associated dataset was identified for feature instance'] += 1

                    if doneflag:
                        break

    else:  # If there is no shape value to test against, don't bother trying to get dataset URI
        report_transfomer_issue_or_increment_count(settings,
                                                   "Either there is no geometry or dataset search has been disabled.",
                                                   "WARN")
        returnval = 'noInformation'
    return returnval


# This is called with each OTH found in the source data when we are translating from one geometric feature
# to another. Within this call, we need to retain any non-default OTH that is in the source data and also add
# any new OTH content that has been identified during the processing of this feature class.
def transform_OTH(settings, row, field_in, target_field, src_fc, target_fc):
    ''' If OTH is not a default value, make sure to append it if we already have stored a non default'''
    # We have a check to make sure OTH is valid first, just because there are some features that it is not on (metadata)
    if target_field is None:
        settings.report_values['Error transforming OTH, OTH not a field'] += 1
        return None

    if target_field is not None:
        # If we only have a default value as input, then do not append to OTH, replace it if we have content
        if field_in is not None and field_in in row and \
                is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == False:
            localufi = get_row_unique_id(settings, row)
            if settings.debug_settings['report_unique_id_usage'] > 0:
                settings.my_logger.debug("transform_OTH for fc {0}  uniqueID {1}".format(src_fc, localufi))

            if localufi == 'noInformation':
                report_transfomer_issue_or_increment_count(settings, "Trying to process OTH but data has no unique id",
                                                           "WARN")
                return 'noInformation'
            else:
                # Specail OTH processing for WRDB variants of GGDM 2.2 and possibly others-
                # These metadata values may be on each feature. Save if there are values

                # If the data row contains some of these feature metadata values, save them as OTH
                if 'COD' in row and not is_value_unknown(row['COD'], settings.sourcedb.geodb_type_unknown, True):
                    settings.report_values['Feature Metadata Extension - COD saved to OTH'] += 1
                    add_to_row_OTH(
                        settings, row, prepare_string_for_OTH_or_MEM('COD', row['COD'],
                                                                     settings.sourcedb.get_enum_text_for_index(
                                                                         'COD', row['COD'])), 'OTH')
                if 'COE' in row and not is_value_unknown(row['COE'], settings.sourcedb.geodb_type_unknown, True):
                    settings.report_values['Feature Metadata Extension - COE saved to OTH'] += 1
                    add_to_row_OTH(
                        settings, row, prepare_string_for_OTH_or_MEM('COE', row['COE'],
                                                                     settings.sourcedb.get_enum_text_for_index(
                                                                         'COE', row['COE'])), 'OTH')
                if 'ZI010_ACC' in row and not is_value_unknown(row['ZI010_ACC'], settings.sourcedb.geodb_type_unknown,
                                                               True):
                    settings.report_values['Feature Metadata Extension - ZI010_ACC saved to OTH'] += 1
                    add_to_row_OTH(
                        settings, row, prepare_string_for_OTH_or_MEM('ZI010_ACC', row['ZI010_ACC'],
                                                                     settings.sourcedb.get_enum_text_for_index(
                                                                         'ZI010_ACC', row['ZI010_ACC'])),
                        'OTH')

                # If there are any settings.saved_oth values for this UFI
                if len(settings.saved_oth) > 0 and localufi in settings.saved_oth:
                    if settings.saved_oth[localufi].find(row[field_in]) == -1:  # this new value is not already in OTH
                        settings.saved_oth[localufi] = settings.saved_oth[localufi] + row[
                            field_in]  # Append the new OTH
                        settings.report_values[
                            'OTH Processing - Saving a value to OTH'] = settings.report_values[
                                                                            'OTH Processing - Saving a value to OTH'] + 1
                    else:  # Flag that we noticed a duplicate OTH
                        settings.report_values[
                            'OTH Processing - Value is already in OTH so it is not being appended'] += 1
                else:
                    settings.saved_oth[localufi] = row[field_in]  # Len is 0, just save the OTH
                    settings.report_values['OTH Processing - Saving a value to OTH'] = settings.report_values[
                                                                                           'OTH Processing - Saving a value to OTH'] + 1
                if len(settings.saved_oth[localufi]) > 254 and (
                        settings.gettarget_db_version() in ['TDS 4.0', 'TDS 6.0', 'TDS 6.1', 'TDS 7.0', 'TDS 7.1']):
                    report_transfomer_issue_or_increment_count(
                        settings,
                        "The OTH value to be saved to feature exceeds maximum size.",
                        "ERROR",
                        "Size of UFI {0} exceeds 254 characters and will be truncated from: {1} to: {2}" \
                            .format(localufi, strout(settings.saved_oth[localufi]),
                                    strout(settings.saved_oth[localufi][:254])))
                    return (settings.saved_oth[localufi][:254])
                else:
                    return settings.saved_oth[localufi]
        else:
            settings.report_values['OTH Processing - Setting OTH to default value'] += 1
            return 'noInformation'
    return 'noInformation'


def transform_MEM(settings, row, field_in, target_field, src_fc, target_fc):
    ''' If MEM or ZI006_MEM is not a default value, make sure to append it if we already have stored a non default'''
    if field_in is not None and field_in in row and not is_value_unknown(row[field_in],
                                                                     settings.sourcedb.geodb_type_unknown,
                                                                     True):
        localufi = get_row_unique_id(settings, row)
        if settings.debug_settings['report_unique_id_usage'] > 0:
            settings.my_logger.debug("transform_MEM for fc {0}  uniqueID {1}".format(src_fc, localufi))
        if localufi in ['noInformation', 'noinformation', 'No Information']:
            report_transfomer_issue_or_increment_count(
                settings,
               "Trying to transform MEM into row but no unique id found for this row (UFI may be set to 'noInformation').",
               "WARN")
            if 'MEM' in row:
                if row['MEM'] is None or row['MEM'].lower() in ['noinformation', 'no information', '']:
                    return 'No Information'
                else:
                    return row['MEM']
            elif 'ZI006_MEM' in row:
                if row['ZI006_MEM'] is None or row['ZI006_MEM'].lower() in ['noinformation', 'no information', '']:
                    return 'No Information'
                else:
                    return row['ZI006_MEM']
            else:
                return 'No Information'
        else:
            # Not a default, save this MEM with the UFI
            if len(settings.saved_mem) > 0 and localufi in settings.saved_mem:
                if settings.saved_mem[localufi].find(row[field_in]) == -1:  # Add this content if it isn't already there
                    settings.saved_mem[localufi] = settings.saved_mem[localufi] + row[field_in]
            else:
                settings.saved_mem[localufi] = row[field_in]
            settings.report_values['MEM processing - Value saved to MEM'] += 1

        if len(settings.saved_mem[localufi]) > 254 and \
                (settings.gettarget_db_version() in ['TDS 4.0', 'TDS 6.0', 'TDS 6.1', 'TDS 7.0', 'TDS 7.1']):
            report_transfomer_issue_or_increment_count(
                settings,
               "The MEM value to be saved to feature with exceeds maximum size",
               "ERROR",
               "Entry having UFI {0} exceeds 254 characters and will be truncated from: {1} to: {2}" \
               .format(localufi, strout(settings.saved_mem[localufi]),
                       strout(settings.saved_mem[localufi][:254])))
            return (settings.saved_mem[localufi][:254])
        else:
            return settings.saved_mem[localufi]

    else:
        settings.report_values['MEM processing - set MEM to default'] += 1
        return 'No Information'


# There are two transformations used for ZI016_WTC.
# The first one takes the input ZI016_WTC and if the value is closed in winter, it maps it to 2 fair-weather.
# This will return a value for ZI016_WTC
# This is for the source being anything but GGDM 3.0 and the dest is GGDM 3.0
def transformZI016_WTC(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row and row[field_in] is not None:
        if settings.getsource_db_version() != "GGDM 3.0" and settings.gettarget_db_version() == "GGDM 3.0":
            if row[field_in] == 5:
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, 5, 'Closed in Winter'), 'OTH')
                settings.report_values['transform ZI001_WTC value Closed in Winter to Fair Weather'] = \
                    settings.report_values['transform ZI001_WTC value Closed in Winter to Fair Weather'] + 1
                return 2  # fair-weather
            else:
                return row[field_in]
        else:
            report_transfomer_issue_or_increment_count(settings,
                                                       "Incorrect transform specified, transformZI016_WTC should only be used if the source is not GGDM 3.0 and the destiniation is GGDM 3.0",
                                                       "ERROR")
            return row[field_in]
    return -999999


# Call this to set ZI016_WTC in an non-GGDM3 model
# This is for setting WTC when the source is GGDM 3.0
def transform_CIW_to_ZI016_WTC(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if settings.getsource_db_version() == "GGDM 3.0" and settings.gettarget_db_version() != "GGDM 3.0":
            if 'CIW' in row and row['CIW'] is not None and row['CIW'] == 1001:
                settings.report_values['Route Weather Restriction derived from Closed in Winter'] += 1
                if rowOTHcheck(row, 'ZI016_WTC', 5, 'Closed in Winter'):
                    return 5  # Closed in winter
                else:
                    return 2  # fair-weather
            else:
                return row[field_in]
        else:
            report_transfomer_issue_or_increment_count(settings,
                                                       "Incorrect transform specified, transform_CIW_to_ZI016_WTC should only be used if the source is GGDM 3.0 and the destiniation is not GGDM 3.0",
                                                       "ERROR")
            return row[field_in]
    return -999999


# set closed in winter to true if road is fair-weather or closed in winter (2 or 5)
# set closed in winter to false if road is all-weather or winter only (1, 3)
# The second mapping is for the destination CIW Closed in Winter. This one looks at the source
# ZI016_WTC value to set CIW
def transformZI016_WTC_CIW(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI016_WTC' in row and row['ZI016_WTC'] is not None:
        mapping = {1: 1000, 2: 1001, 3: 1000, 4: -999999, 5: 1001, 999: 999,
                   -999999: -999999}  # set closed in winter to true if road is fair-weather
        settings.report_values['Transform processing: Closed in Winter derived from Route Weather Restriction'] += 1
        if row['ZI016_WTC'] in mapping:
            return mapping[row['ZI016_WTC']]  # not a typo, need to use WTC to derive CIW
    return -999999


# Derive all values of RIN_ROI from RTN_ROI. Special processing to handle the multi-valued attribute terms.
def transform_RIN_ROI_from_RTN_ROI(settings, row, field_in, target_field, src_fc, target_fc):
    if 'RTN_ROI' in row and row['RTN_ROI'] is not None:
        settings.report_values['Route Designation Route Identification Conversion'] += 1
        return row['RTN_ROI']
    elif 'RTN_ROI2' in row and row['RTN_ROI2'] is not None:
        settings.report_values['Route Designation Route Identification Conversion'] += 1
        return row['RTN_ROI2']
    elif 'RTN_ROI3' in row and row['RTN_ROI3'] is not None:
        settings.report_values['Route Designation Route Identification Conversion'] += 1
        return row['RTN_ROI3']
    return -999999


# Derive all values of RTN_ROI from RIN_ROI. Special processing to handle the multi-valued attribute terms.
def transform_RTN_ROI_from_RIN_ROI(settings, row, field_in, target_field, src_fc, target_fc):
    if 'RIN_ROI' in row and row['RIN_ROI'] is not None:
        settings.report_values['Route Designation Route Identification Conversion'] += 1
        return row['RIN_ROI']
    elif 'RIN_ROI2' in row and row['RIN_ROI2'] is not None:
        settings.report_values['Route Designation Route Identification Conversion'] += 1
        return row['RIN_ROI2']
    elif 'RIN_ROI3' in row and row['RIN_ROI3'] is not None:
        settings.report_values['Route Designation Route Identification Conversion'] += 1
        return row['RIN_ROI3']
    return -999999


def transform_ZI013_PIG(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if 'ZI013_FMM' in row and row['ZI013_FMM'] is not None:
            if row['ZI013_FMM'] == 5 or row['ZI013_FMM2'] == 5 or row['ZI013_FMM3'] == 5:  # Permanent Irrigation
                settings.report_values['Permanent Irrigation derived from Farming Method'] += 1
                return 1001  # True
        # ZI013_PIG attribute is not in GGDM 2.1 or TDS 4.0, so if it can't be derived, it has to be default
        return transform_set_to_default(settings, None, None, target_field, None, target_fc)
    return -999999


def transform_ZI013_FMM(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
            # We can derive FMM from PIG if PIG is True
            if 'ZI013_PIG' in row and row['ZI013_PIG'] is not None:
                if row['ZI013_PIG'] == 1001:
                    rtnval = row[field_in]
                    # PIG is true, now determine if this particular field can be set to it or not
                    targetval = 5  # Permanent Irrigation
                    openfield = get_first_open_multifield(row, 'ZI013_FMM', targetval)
                    if openfield is None:
                        add_to_row_OTH(settings, row,
                                       prepare_string_for_OTH_or_MEM('ZI013_FMM', targetval, 'Permanent Irrigation'),
                                       'OTH')
                    else:
                        settings.report_values['Farming Method derived from Permanent Irrigation'] += 1
                        if field_in == openfield and targetval != rtnval:
                            return targetval
        return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)
    return -999999


# Communal Bath is in GGDM 3.0 and GGDM 2.2 and is in TDS 61 and TDS 6, just not in TDS 4 and GGDM 2.1
# Islamic Prayer hall is not in GGDM 3.0 but is in all others
def transform_FFN(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        rtnval = row[field_in]
        curfcode = settings.targetdb.get_f_code()
        if row[field_in] == 931 and rowOTHcheck(row, 'FFN', 932, 'Islamic Prayer Hall') and \
                settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, curfcode, field_in, 932):
            settings.report_values['Feature Function place of worship detected from Islamic Prayer Hall'] += 1
            rtnval = 932
        if row[field_in] == 559 and settings.sourcedb.is_index_valid_for_fclass_fcode(src_fc, curfcode, field_in,
                                                                                      row[field_in]) \
                and not settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, curfcode, field_in, row[field_in]):
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Communal Bath'),
                           'OTH')
            rtnval = 999
        if row[field_in] == 915 and settings.sourcedb.is_index_valid_for_fclass_fcode(src_fc, curfcode, field_in,
                                                                                      row[field_in]) \
                and not settings.targetdb.is_index_valid_for_fclass_fcode(target_fc, curfcode, field_in, row[field_in]):
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Water Park'), 'OTH')
            rtnval = 999
        return rtnval
    return -999999


def transform_put_source_to_OTH(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in is not None and field_in in row:
        if is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True) == False:
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], ''), 'OTH')
    return None


def transform_FFN_932_and_TRS(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in not in row:
        return -999999

    rtnval = row[field_in]

    # if we have 999 other to start, see if we can bring in some values from OTH
    if rtnval == 999:
        if rowOTHcheck(row, field_in, 559, 'Communal Bath'):
            rtnval = 559
        elif rowOTHcheck(row, field_in, 915, 'Water Park'):
            rtnval = 915

    # check to see if this value is not valid in the target, and then see if we got index 932  
    if rtnval == 932 and is_field_and_value_valid_in_target(settings, row, field_in, src_fc, target_fc,
                                                            row[field_in]) == False:
        # Removal of index 932 Islamic Prayer Hall
        settings.report_values['Feature Function place of worship detected from Islamic Prayer Hall'] += 1
        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Islamic Prayer Hall'),
                       'OTH')

        # Since the ZI037_REL is not valid for all F_CODES, we did the F_CODE check first
        # this will save the value to OTH since fortified building doesn't allow for this attribute
        if settings.sourcedb.get_f_code() == 'AH055':  # AH055 = Fortified Building
            settings.report_values['Feature Function value not allowed on Fortified Building and saved to OTH'] += 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZI037_REL', 2, 'Islam'), 'OTH')
        rtnval = 931  # return value FFN value 931 Place of Worship

    # consider TRS now for possible values        
    if 'TRS' in row:
        if settings.sourcedb.get_f_code() == 'AQ060':  # AQ060 = Control Tower
            if row['TRS'] == 1 or row['TRS2'] == 1 or row['TRS3'] == 1:
                targetval = 525  # Air Traffic Control
                openfield = get_first_open_multifield(row, 'FFN', targetval)
                if openfield is None:
                    add_to_row_OTH(settings, row,
                                   prepare_string_for_OTH_or_MEM('FFN', targetval, 'Air Traffic Control'), 'OTH')
                else:
                    if field_in == openfield:
                        rtnval = targetval  # 525 = Air Traffic Control

        elif settings.sourcedb.get_f_code() == 'AQ125':  # AQ125 = Transportation Station
            if row['TRS'] == 1 or row['TRS2'] == 1 or row['TRS3'] == 1:  # 1 = Aeronautical
                targetval = 520  # 520 = Air Transport
                openfield = get_first_open_multifield(row, 'FFN', targetval)
                if openfield is None:
                    add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('FFN', targetval, 'Air Transport'),
                                   'OTH')
                else:
                    if field_in == openfield and targetval != rtnval:
                        rtnval = targetval  # 520 = Air Transport

            if row['TRS'] == 16 or row['TRS2'] == 16 or row['TRS3'] == 16:  # 16 = Underground Railway
                targetval = 490  # Railway Transport
                openfield = get_first_open_multifield(row, 'FFN', targetval)
                if openfield is None:
                    add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('FFN', targetval, 'Railway Transport'),
                                   'OTH')
                else:
                    if field_in == openfield and targetval != rtnval:
                        rtnval = targetval  # 490 = Railway Transport

            if row['TRS'] == 5 or row['TRS2'] == 5 or row['TRS3'] == 5:  # 5 == Canal
                targetval = 508  # Canal Transport
                openfield = get_first_open_multifield(row, 'FFN', targetval)
                if openfield is None:
                    add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('FFN', targetval, 'Canal Transport'),
                                   'OTH')
                else:
                    if field_in == openfield and targetval != rtnval:
                        rtnval = targetval  # 508 = Canal Transport

            if row['TRS'] == 17 or row['TRS2'] == 17 or row['TRS3'] == 17:  # 17 == Inland Waterway
                targetval = 507  # Inland Waters
                openfield = get_first_open_multifield(row, 'FFN', targetval)
                if openfield is None:
                    add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('FFN', targetval, 'Inland Waters'),
                                   'OTH')
                else:
                    if field_in == openfield and targetval != rtnval:
                        rtnval = targetval  # 507 = Inland Waters

    return rtnval


# transform_ZI037_REL
# value 14 = Mixed and/or No Designation
# Is not in GGDM 2.1 or TDS 4
def transform_ZI037_REL(settings, row, field_in, target_field, src_fc, target_fc):
    if 'FFN' in row and (row['FFN'] == 932 or row['FFN2'] == 932 or row['FFN3'] == 932):
        settings.report_values['Religious Designation (ZI037_REL) Adjustment from FFN Islamic Prayer Hall'] += 1
        return 2  # Set Religious designation to 2 = Islam
    else:
        if field_in in row:
            return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)
        else:
            return -999999


def transform_TRS_fromGGDM3(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in not in row or row[field_in] is None:
        return -999999
    rtnval = row[field_in]
    curfcode = settings.sourcedb.get_f_code()
    if curfcode == 'AP010' and row[field_in] == 999 and rowOTHcheck(row, field_in, 21, 'Pipeline Maintenance'):
        rtnval = 21
    elif curfcode == 'AP010' and row[field_in] == 999 and rowOTHcheck(row, field_in, 22, 'Powerline Maintenance'):
        rtnval = 22
    elif curfcode == 'AQ060' and row[field_in] == 999 and rowOTHcheck(row, field_in, 1, 'Aeronautical'):
        rtnval = 1
    elif curfcode == 'DB070' and row[field_in] == 999 and rowOTHcheck(row, field_in, 9, 'Pedestrian'):
        rtnval = 9
    elif curfcode == 'DB071' and row[field_in] == 999 and rowOTHcheck(row, field_in, 9, 'Pedestrian'):
        rtnval = 9
    elif curfcode == 'BH070' and row[field_in] == 999 and rowOTHcheck(row, field_in, 6, 'Caravan Route'):
        rtnval = 6
    elif curfcode == 'BH070' and row[field_in] == 999 and rowOTHcheck(row, field_in, 11, 'Portage'):
        rtnval = 11
    elif curfcode == 'BH070' and row[field_in] == 999 and rowOTHcheck(row, field_in, 21, 'Pipeline Maintenance'):
        rtnval = 21
    elif curfcode == 'BH070' and row[field_in] == 999 and rowOTHcheck(row, field_in, 22, 'Powerline Maintenance'):
        rtnval = 22
    elif curfcode == 'AP040' and row[field_in] == 3 and rowOTHcheck(row, field_in, 4, 'Bus'):
        rtnval = 4
    elif curfcode == 'AP040' and row[field_in] == 3 and rowOTHcheck(row, field_in, 15, 'Truck'):
        rtnval = 15
    elif curfcode == 'AH070' and row[field_in] == 13 and rowOTHcheck(row, field_in, 3, 'Automotive'):
        rtnval = 15
    elif curfcode == 'AH070' and row[field_in] == 13 and rowOTHcheck(row, field_in, 4, 'Bus'):
        rtnval = 4
    elif curfcode == 'AH070' and row[field_in] == 13 and rowOTHcheck(row, field_in, 15, 'Truck'):
        rtnval = 15
    elif curfcode == 'AQ125':
        if row[field_in] == 999 and rowOTHcheck(row, field_in, 1, 'Aeronautical'):
            rtnval = 1
        elif row[field_in] == 13 and rowOTHcheck(row, field_in, 3, 'Automotive'):
            rtnval = 3
        elif row[field_in] == 13 and rowOTHcheck(row, field_in, 15, 'Truck'):
            rtnval = 15
        elif row[field_in] == 13 and rowOTHcheck(row, field_in, 4, 'Bus'):
            rtnval = 4
        elif row[field_in] == 999 and rowOTHcheck(row, field_in, 16, 'Underground Railway'):
            rtnval = 16
        elif row[field_in] == 7:  # 7 == Maritime
            settings.report_values['TRS-Maritime'] = settings.report_values['TRS-Maritime'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Maritime'), 'OTH')
            rtnval = 999  # 999 = Other
        elif row[field_in] == 8:  # 8 == No Transportation System
            settings.report_values['TRS-No Transportation System'] = settings.report_values[
                                                                         'TRS-No Transportation System'] + 1
            add_to_row_OTH(settings, row,
                           prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'No Transportation System'), 'OTH')
            rtnval = 999  # 999 = Other
        elif row[field_in] == 17:  # 17 == Inland Waterway
            settings.report_values['TRS-Inland Waterway'] = settings.report_values['TRS-Inland Waterway'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Inland Waterway'),
                           'OTH')
            rtnval = 999  # 999 = Other
        elif row[field_in] == 20:  # 20 == Taxiway
            settings.report_values['TRS-Taxiway'] = settings.report_values['TRS-Taxiway'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Taxiway'), 'OTH')
            rtnval = 999  # 999 = Other
    elif curfcode == 'AQ130':  # AQ130 = Tunnel
        if row[field_in] == 999 and rowOTHcheck(row, field_in, 11, 'Portage'):  # 11 = Portage
            rtnval = 11
        elif row[field_in] == 999 and rowOTHcheck(row, field_in, 17, 'Inland Waterway'):  # 17 = Inland Waterway
            rtnval = 17

    return rtnval


def transform_TRS(settings, row, field_in, target_field, src_fc, target_fc):
    if (settings.getsource_db_version() == 'GGDM 3.0' and (settings.gettarget_db_version() != 'GGDM 3.0')):
        return transform_TRS_fromGGDM3(settings, row, field_in, target_field, src_fc, target_fc)

    if not field_in in row:
        return -999999

    settings.report_values['TRS'] = settings.report_values['TRS'] + 1
    curfcode = settings.sourcedb.get_f_code()
    if curfcode == 'AP010':  # AP010 = Cart Track
        if row[field_in] == 21:  # 21 = Pipeline Maintenance
            settings.report_values['TRS-Pipeline Maintenance'] = settings.report_values['TRS-Pipeline Maintenance'] + 1
            add_to_row_OTH(settings, row,
                           prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Pipeline Maintenance'), 'OTH')
            return 999
        elif row[field_in] == 22:  # 22 = Powerline Maintenance
            settings.report_values['TRS-Powerline Maintenance'] = settings.report_values[
                                                                      'TRS-Powerline Maintenance'] + 1
            add_to_row_OTH(settings, row,
                           prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Powerline Maintenance'), 'OTH')
            return 999
        else:
            return row[field_in]
    elif curfcode == 'AQ060':  # AQ060 = Control Tower
        if row[field_in] == 1:  # 1 = Aeronautical
            settings.report_values['TRS-Aeronautical'] = settings.report_values['TRS-Aeronautical'] + 1
            # set_multivalued_field_in_row(row, 'FFN', 525, 'Air Traffic Control') # Set FFN (or 2 or 3) to 525 Air Traffic Control
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Aeronautical'), 'OTH')
            return 999
        else:
            return row[field_in]
    elif curfcode == 'DB070' or curfcode == 'DB071':  # DB070 = CUT, DB071 = CUT_LINE
        if row[field_in] == 9:  # 9 = Pedestrian
            settings.report_values['TRS-Pedestrian'] = settings.report_values['TRS-Pedestrian'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Pedestrian'), 'OTH')
            return 999
        else:
            return row[field_in]
    elif curfcode == 'BH070':  # BH070 = Ford
        if row[field_in] == 6:  # 6 = Caravan Route
            settings.report_values['TRS-Caravan Route'] = settings.report_values['TRS-Caravan Route'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Caravan Route'),
                           'OTH')
            return 999
        elif row[field_in] == 11:  # 11 = Portage
            settings.report_values['TRS-Portage'] = settings.report_values['TRS-Portage'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Portage'), 'OTH')
            return 999
        elif row[field_in] == 21:  # 21 = Pipeline Maintenance
            settings.report_values['TRS-Pipeline Maintenance'] = settings.report_values['TRS-Pipeline Maintenance'] + 1
            add_to_row_OTH(settings, row,
                           prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Pipeline Maintenance'), 'OTH')
            return 999
        elif row[field_in] == 22:  # 22 = Powerline Maintenance
            settings.report_values['TRS-Powerline Maintenance'] = settings.report_values[
                                                                      'TRS-Powerline Maintenance'] + 1
            add_to_row_OTH(settings, row,
                           prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Powerline Maintenance'), 'OTH')
            return 999
        else:
            return row[field_in]
    elif curfcode == 'AH070':  # AH070 = Checkpoint
        if row[field_in] == 3:  # 3 = Automotive
            # Change the output to 13 = Road AND set Other and put the original value of 3 = Automotive there
            settings.report_values['TRS-Automotive'] = settings.report_values['TRS-Automotive'] + 1
            add_to_row_OTH(row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Automotive'), 'OTH')
            return 13  # 13 = Road
        elif row[field_in] == 4:  # 4 = Bus
            # Change the output to 13 = Road AND set Other and put the original value of 4 = Bus there
            settings.report_values['TRS-Bus'] = settings.report_values['TRS-Bus'] + 1
            add_to_row_OTH(row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Bus'), 'OTH')
            return 13  # 13 = Road
        elif row[field_in] == 15:  # 15 = Truck
            # Change the output to 13 = Road AND set Other and put the original value of 3 = Automotive there
            settings.report_values['TRS-Truck'] = settings.report_values['TRS-Truck'] + 1
            add_to_row_OTH(row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Truck'), 'OTH')
            return 13  # 13 = Road
        else:
            return row[field_in]
    elif curfcode == 'AP040':  # AP040 = Gate
        if row[field_in] == 4:  # 4 = Bus
            # Change the output to 3 = Automotive AND set Other and put the original value of 4 = Bus there
            settings.report_values['TRS-Bus'] = settings.report_values['TRS-Bus'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Bus'), 'OTH')
            return 3  # 3 = Automotive
        elif row[field_in] == 15:  # 15 = Truck
            # Change the output to 3 = Automotive AND set Other and put the original value of 15 = Truck there
            settings.report_values['TRS-Truck'] = settings.report_values['TRS-Truck'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Truck'), 'OTH')
            return 3  # 3 = Automotive
        else:
            return row[field_in]
    elif curfcode == 'AQ125':  # AQ125 = Transportation Station
        if row[field_in] == 1:  # 1 = Aeronautical
            settings.report_values['TRS-Aeronautical'] = settings.report_values['TRS-Aeronautical'] + 1
            # set_multivalued_field_in_row(row, 'FFN', 520, 'Air Transport')
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Aeronautical'), 'OTH')
            return 999
        elif row[field_in] == 3:  # 3 = Automotive
            # Change the output to 13 = Road AND set Other and put the original value of 3 = Automotive there
            settings.report_values['TRS-Automotive'] = settings.report_values['TRS-Automotive'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Automotive'), 'OTH')
            return 13  # 13 = Road
        elif row[field_in] == 4:  # 4 = Bus
            # Change the output to 13 = Road AND set Other and put the original value of 4 = Bus there
            settings.report_values['TRS-Bus'] = settings.report_values['TRS-Bus'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Bus'), 'OTH')
            return 13  # 13 = Road
        elif row[field_in] == 15:  # 15 = Truck
            # Change the output to 13 = Road AND set Other and put the original value of 15 = Truck there
            settings.report_values['TRS-Truck'] = settings.report_values['TRS-Truck'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('TRS', row[field_in], 'Truck'), 'OTH')
            return 13  # 13 = Road
        elif row[field_in] == 16:  # 16 = Underground Railway
            settings.report_values['TRS-Underground Railway'] = settings.report_values['TRS-Underground Railway'] + 1
            # set_multivalued_field_in_row(row, 'FFN', 490, 'Railway Transport')
            # Located Underground (LUN) = True 
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Underground Railway'),
                           'OTH')
            return 999  # 999 = Other
        elif row[field_in] == 5:  # 5 == Canal
            settings.report_values['TRS-Canal'] = settings.report_values['TRS-Canal'] + 1
            # set_multivalued_field_in_row(row, 'FFN', 508, 'Canal Transport')
            return row[field_in]  # do not change field value
        #
        # The following options are for translating from GGDM 3
        #
        elif row[field_in] == 7:  # 7 == Maritime
            settings.report_values['TRS-Maritime'] = settings.report_values['TRS-Maritime'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Maritime'), 'OTH')
            return 999  # 999 = Other
        elif row[field_in] == 8:  # 8 == No Transportation System
            settings.report_values['TRS-No Transportation System'] = settings.report_values[
                                                                         'TRS-No Transportation System'] + 1
            add_to_row_OTH(settings, row,
                           prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'No Transportation System'), 'OTH')
            return 999  # 999 = Other
        elif row[field_in] == 17:  # 17 == Inland Waterway
            settings.report_values['TRS-Inland Waterway'] = settings.report_values['TRS-Inland Waterway'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Inland Waterway'),
                           'OTH')
            return 999  # 999 = Other
        elif row[field_in] == 20:  # 20 == Taxiway
            settings.report_values['TRS-Taxiway'] = settings.report_values['TRS-Taxiway'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Taxiway'), 'OTH')
            return 999  # 999 = Other
        else:
            return row[field_in]
    elif curfcode == 'AQ130':  # AQ130 = Tunnel
        if row[field_in] == 11:  # 11 = Portage
            settings.report_values['TRS-Portage'] = settings.report_values['TRS-Portage'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Portage'), 'OTH')
            return 999
        elif row[field_in] == 17:  # 17 = Inland Waterway
            settings.report_values['TRS-Inland Waterway'] = settings.report_values['TRS-Inland Waterway'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Inland Waterway'),
                           'OTH')
            return 999
        else:
            return row[field_in]
    else:
        return row[field_in]


def transform_TSSM(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row and row[field_in] is not None:
        if settings.getsource_db_version() == "GGDM 3.0":
            mapping = {1: 6, 2: 8, 3: 16, 4: 22, 5: 29, 6: 43, 7: 46, 8: 130, 9: 127, 10: 52, 11: 269, 12: 61, 13: 65,
                       14: 88, 15: 90, 19: 89, 20: 67, 269: 269, 999: 999, -999999: -999999, 16: 999, 17: 999, 18: 999}
            if row[field_in] == 16 or row[field_in] == 17 or row[field_in] == 18:
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in],
                                                                            settings.sourcedb.get_enum_text_for_index(
                                                                                field_in, row[field_in])), 'OTH')
            settings.report_values['Terrain Subsurface Material values mapped'] += 1
            return mapping[row[field_in]]
        elif settings.gettarget_db_version() == "GGDM 3.0":
            mapping = {6: 1, 8: 2, 16: 3, 22: 4, 29: 5, 43: 6, 46: 7, 130: 8, 127: 9, 52: 10, 269: 11, 61: 12, 65: 13,
                       88: 14, 90: 15, 89: 19, 67: 20, 269: 269, 999: 999, -999999: -999999}
            settings.report_values['Terrain Subsurface Material values mapped'] += 1
            return mapping[row[field_in]]
        else:
            return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)
    else:
        return -999999


def transform_TRS_to_PTG(settings, row, field_in, target_field, src_fc, target_fc):
    if 'TRS' in row:
        if is_field_valid_to_set(settings, row, field_in, target_field, src_fc, target_fc):
            # if TRS specifies 11 = Portage, then the Portage Use PTG attribute should be True
            if row['TRS'] == 11 or row['TRS2'] == 11 or row['TRS3'] == 11:
                settings.report_values['TRS-Portage sets PTG'] += 1
                return (1001)  # True
            else:
                if field_in in row and row[field_in]:  # If there is an existing value, use it
                    return row[field_in]
                else:
                    return (-999999)  # No Information
        else:
            return -999999
    else:
        return -999999


def transform_TWS(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() == "GGDM 3.0":
        if is_field_valid_to_set(settings, row, field_in, target_field, src_fc, target_fc):
            if settings.sourcedb.get_f_code() == 'EB010':  # EB010 = Grassland
                # if VEG is set to 9, TWS is True. If VEG is 56, TWS if False
                if row['VEG'] == 9:  # Grassland with Trees
                    settings.report_values['Trees Widely Scattered (TWS) value derived from Vegetation Type (VEG)'] += 1
                    return (1001)  # Trees Widely Scattered = True
                elif row['VEG'] == 56:  # Without Trees
                    settings.report_values['Trees Widely Scattered (TWS) value derived from Vegetation Type (VEG)'] += 1
                    return (1000)  # Trees Widely Scattered = False
                else:
                    return (-999999)  # No information
            if field_in in row and row[
                field_in]:  # We need to see if the field exists because we may need to set it to default
                return row[field_in]
            else:
                return -999999
        else:
            return transform_set_to_default(settings, None, None, target_field, None, target_fc)
    return row[field_in]


def transform_VEG(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() == "GGDM 3.0":
        if settings.sourcedb.get_f_code() == 'EB010':  # EB010 = Grassland
            # if VEG is set to 9, TWS is True. If VEG is 56, TWS if False
            if row['VEG'] == 9:  # Grassland with Trees
                settings.report_values['Vegetation Type (VEG)-Grassland with Trees'] += 1
                add_to_row_OTH(settings, row,
                               prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Grassland with Trees'), 'OTH')
                return (8)  # Grassland
            elif row['VEG'] == 56:  # Without Trees
                settings.report_values['Vegetation Type (VEG)-Grassland Without Trees'] += 1
                add_to_row_OTH(settings, row,
                               prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Grassland Without Trees'), 'OTH')
                return (8)  # Grassland
            elif row['VEG'] == 83:  # Meadow
                settings.report_values['Vegetation Type (VEG)-Meadow'] += 1
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Meadow'), 'OTH')
                return (8)  # Grassland
    elif settings.getsource_db_version() == "GGDM 3.0":
        if settings.sourcedb.get_f_code() == 'EB010':  # EB010 = Grassland
            if row[field_in] == 999 or row[field_in] == -999999:
                if row['TWS'] == 1001 and row[field_in] == 8:
                    settings.report_values['Vegetation Type (VEG)-Grassland with Trees'] += 1
                    return 9  # Grassland With Trees
                elif row['TWS'] == 1000 and row[field_in] == 8:
                    settings.report_values['Vegetation Type (VEG)-Grassland Without Trees'] += 1
                    return 56  # Without Trees
                elif row['TWS'] == -999999 and row[field_in] == 8 and rowOTHcheck(row, field_in, 83, 'Meadow'):
                    settings.report_values['Vegetation Type (VEG)-Meadow'] += 1
                    return 83

    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


# transform_COS ??  All values are in GGDM 3.
# TDS 6 and GGDM 2.2 do not have: 1 = Continuous, 3 = Secheduled
# GGDM 2.1 and TDS 4 in addition to above line, do not have: 2 = Intermittent


def transform_COS(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in not in row:
        return -999999
    if is_field_and_value_valid_in_target(settings, row, field_in, src_fc, target_fc, row[field_in]):
        return row[field_in]
    else:
        # if row[field_in] == 8:    # Not in Operation
        curfcode = settings.sourcedb.get_f_code()
        if row[field_in] == 8 and (
                curfcode == 'GB035' or curfcode == 'GB005' or curfcode == 'GB065'):  # GB035 = Heliport GB005 = Land Aerodrome GB065 = Water Aerodrome
            settings.report_values['Facility Operational Status (COS) processing'] = settings.report_values[
                                                                                         'Facility Operational Status (COS) processing'] + 1

            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Not in Operation'),
                           'OTH')
            return 4  # Non-operational

        return row[field_in]
    return row[field_in]


def transform_WWM(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if settings.getsource_db_version() == "GGDM 2.2":
            if row[field_in] == 89:
                settings.report_values['Subsurface stratum (WWM) values mapped'] += 1
                return 19
            elif row[field_in] == 67:
                settings.report_values['Subsurface stratum (WWM) values mapped'] += 1
                return 20
        elif settings.gettarget_db_version() == "GGDM 2.2":
            if row[field_in] == 19:
                settings.report_values['Subsurface stratum (WWM) values mapped'] += 1
                return 89
            elif row[field_in] == 20:
                settings.report_values['Subsurface stratum (WWM) values mapped'] += 1
                return 67
        return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)
    else:
        return -999999


# Real in GGDM 2.1 (not in TDS), Integer in other models
def transform_LVN(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if settings.getsource_db_version() == settings.schema_version['GGDM 2.1']:
            settings.report_values['Nominal range data type change'] += 1
            intout = int(row[field_in])
            return intout
        elif settings.gettarget_db_version() == settings.schema_version['GGDM 2.1']:
            settings.report_values['Nominal range data type change'] += 1
            fltout = float(row[field_in])
            return fltout
        elif field_in in row:
            return row[field_in]
    else:
        return -999999


# If the destination is GGDM 3.0 we need DFU
def transformDWT_to_DFU(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() == "GGDM 3.0":
        if target_field == 'DFU':  # Only do this for the first one because there was only one DWT
            if 'DWT' in row:
                settings.report_values['Derive Dam Function from Dam Type'] += 1
                return row[
                    'DWT']  # this is called for the DFU field, referencing the DWT field is on purpose here to get the value mapped
        else:
            return -999999
    return -999999


# If the source is GGDM 3.0
def transformDFU_to_DWT(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() == "GGDM 3.0":
        if 'DFU' in row:
            settings.report_values['Derive Dam Type from Dam Function'] = settings.report_values[
                                                                              'Derive Dam Type from Dam Function'] + 1

            if row['DFU2'] <> 999 and row['DFU2'] <> -999999:
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('DFU2', row['DFU2'],
                                                                            settings.sourcedb.get_enum_text_for_index(
                                                                                'DFU2', row['DFU2'])), 'OTH')
            if row['DFU3'] <> 999 and row['DFU3'] <> -999999:
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('DFU3', row['DFU3'],
                                                                            settings.sourcedb.get_enum_text_for_index(
                                                                                'DFU3', row['DFU3'])), 'OTH')
            if row['DFU'] == 1 or row['DFU'] == 6 or row['DFU'] == 7 or row['DFU'] == 8:
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('DFU', row['DFU'], ''), 'OTH')
                return 999  # Other
            else:
                return row['DFU']

    return -999999


def transformCreateUFI(settings, row, field_in, target_field, src_fc, target_fc):
    return str(uuid.uuid4()).upper()


# The idea here is that OTH may be source within a geometric feature instance, but the destination is a metadata feature and
# the OTH content may be irrelevant to the metadata instance. Therefore, only if OTH contains information about a metadata element
# we should save the OTH content, otherwise we must not include OTH.
def transformMoveOTHOnlyIfMetaAttribute(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row and (field_in == 'OTH' or target_field == 'OTH'):
        SubsetFields = ['AHA', 'AVA', 'DQS', 'EQS', 'EVA', 'HVA', 'MDE', 'RCG', 'SCVN', 'ZI001_SDP', 'ZI001_SDV',
                        'ZI001_SPS', 'ZI001_SRT', 'ZI001_VSC', 'ZI001_VSD', 'ZI001_VSN', 'ZI004_PRE', 'COD', 'COE',
                        'ZI010_ACC']
        for possval in SubsetFields:
            if row[field_in].find(possval) != -1:
                return row[field_in]
        return "noInformation"
    else:
        settings.report_values['Error transforming OTH, OTH not a field'] = settings.report_values[
                                                                                'Error transforming OTH, OTH not a field'] + 1


# If dest is GGDM 2.1 or TDS 4 then:
def transform_AQP(settings, row, field_in, target_field, src_fc, target_fc):
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


# Value Coral is only in GGDM 3.0
def transform_BH141_SLT(settings, row, field_in, target_field, src_fc, target_fc):
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


# 3 = Deck only in GGDM 3
# 28 = Viaduct only in GGDM 3
# ***************   30 Submersible is only in GGDM 21 and TDS 4 
def transform_BSC(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if 'BOT' in row:
            rtnval = row[field_in]
            if row['BOT'] == 14:  # Submersible
                targetval = 30  # Submersible
                openfield = get_first_open_multifield(row, 'BSC', targetval)
                if openfield is not None:
                    if field_in == openfield and targetval != rtnval:
                        settings.report_values[
                            'Transform processing: Bridge Structure Type derived from Bridge Opening Type'] += 1
                        return targetval
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


# If the source is GGDM 2.1 or TDS 4.0
def transform_BOT(settings, row, field_in, target_field, src_fc, target_fc):
    if 'BSC' in row:
        if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
            if row['BSC'] == 30 or row['BSC2'] == 30 or row['BSC3'] == 30 and field_in == 'BOT':
                settings.report_values[
                    'Transform processing: Bridge Opening Type derived from Bridge Structure Type'] += 1
                return 14  # Submersible  / BOT is not in GGDM 2.1 or TDS 4.0 so we do not have to look for current values being default
            else:
                return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)

        return row[field_in]
    else:
        return -999999


##def transform_COA(settings, row, field_in, target_field, src_fc, target_fc):
##    if settings.getsource_db_version() == "GGDM 2.1":
##        if row[field_in] == 1:
##            add_to_row_OTH(settings,row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Visual and Radar'), 'OTH')
##            return 999   # Other
##        elif row[field_in] == 3:
##            add_to_row_OTH(settings,row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Radar'), 'OTH')
##            return 999   # Other
##        ## Be sure there is a transform on COC to set these if the source is GGDM 2.1
##    return row[field_in]
##
##
### If the source is GGDM 2.1 then COA can be derived from the value of COC for two specific values
##def transform_COC(settings, row, field_in, target_field, src_fc, target_fc):
##    if settings.getsource_db_version() == "GGDM 2.1":
##        if row['COA'] == 1 or row['COA'] == 3:   # Visual and Radar, Radar
##            return row['COA']   # COC needs to be set to this value
##    return row[field_in]

# this 48 value is only in GGDM 3
def transform_LOC(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if (row[field_in] == -999999 or row[field_in] == 999) and \
                'TRS' in row:  # If the existing value is default or no information
            if row['TRS'] == 16 or row['TRS2'] == 16 or row['TRS3'] == 16:  # Underground Railway
                settings.report_values['Location Category derived from TRS'] = settings.report_values[
                                                                                   'Location Category derived from TRS'] + 1
                return 40  # Completely Below Ground Surface
        return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc,
                                                          target_fc)  # Otherwise, just do the default
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)


# This function is not used, we use the other one trs_to_lun
def transform_LUN(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if (row[field_in] == -999999 or row[field_in] == 999) and 'TRS' in row:
            if row['TRS'] == 16 or row['TRS2'] == 16 or row['TRS3'] == 16:  # Underground Railway
                settings.report_values['Located Underground derived from TRS'] += 1
                return 1001  # Located Underground
        return row[field_in]
    else:
        return -999999


def transform_TRS_to_LUN(settings, row, field_in, target_field, src_fc, target_fc):
    if 'TRS' in row:
        if is_field_valid_to_set(settings, row, field_in, target_field, src_fc, target_fc):
            if row['TRS'] == 16 or row['TRS2'] == 16 or row['TRS3'] == 16:  # 16 = Underground Railway
                settings.report_values['TRS-Underground Railway sets LUN'] += 1
                return 1001  # Located Underground (LUN) = True
            else:
                if field_in in row and row[field_in] is not None:  # If there is an existing value, use it
                    return row[field_in]
                else:
                    return -999999  # No information
        else:
            return -999999
    else:
        return -999999


# Change Strip Mine to Area strip Mine.
def transform_MZN(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if row[field_in] == 12:  # Strip-mine
            settings.report_values['Extraction Mine Strip-mine converted to Area strip mine'] += 1
            return 6  # Area strip mine
        return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)
    else:
        return -999999


# All GGDM has but no TDS has
#  23 = Copper
#  24 = Coral
#  54 = Lead
#  120 = Glass Reinforced Plastic (GRP)
# GGDM 3.0 has and no other has this:
#  25 = Coral Head
# 160 = Earthen

def transform_MCC(settings, row, field_in, target_field, src_fc, target_fc):
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


# TDS does not have any PLT
# GGDM 2.1 does not have 4 = Sewer, 5 = Pipeline Valve, 7 = Bubbler System
def transform_PLT(settings, row, field_in, target_field, src_fc, target_fc):
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


def transform_BST(settings, row, field_in, target_field, src_fc, target_fc):
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


# PPO
# transform_PPO
# GGDM 3 only has 49 = Grain, 70 = Milk, 123 = Wine
# GGDM 2.1 only has 73 = No Product, 825 = Bio-Solids   Save 825 Bio-Solids as 101 Sewage
# GGDM 3.0, GGDM 2.1 TDS 6.1 TDS 6 have: 170 = Beverage  ( It is not in TDS 4 or GGDM 2.1)
def transform_PPO(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if row[field_in] == 825:  # Bio-Solids
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], 'Bio-Solids'), 'OTH')
            settings.report_values['Physical Product (PPO) Bio-Solids converted to Sewage'] += 1
            return 101  # Sewage
        else:
            return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)
    else:
        return -999999


# transform_ZI014_PPO
# GGDM 3 is only one with 200 = Iron Ingot
def transform_ZI014_PPO(settings, row, field_in, target_field, src_fc, target_fc):
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


def transform_PRW(settings, row, field_in, target_field, src_fc, target_fc):
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


def transform_SSR(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if field_in in row:
            rtnval = row[field_in]
            if row[field_in] == 51 or (row[field_in] >= 77 and row[field_in] <= 82):

                report_transfomer_issue_or_increment_count(
                    settings, 'Roof shape (SSR) WARNING issued', "WARN",
                    "{0} Feature {1} has {2} attribute with value {3} = {4}. This information will be included in a new feature (appended in step 5) AL018 Building Superstructure".format(
                        src_fc, settings.sourcedb.get_f_code(), field_in,
                        row[field_in], strout(
                        settings.sourcedb.get_enum_text_for_index(field_in,
                                                              row[field_in]))))

                add_to_row_OTH(
                    settings, row, prepare_string_for_OTH_or_MEM(
                        field_in, row[field_in],
                         settings.sourcedb.get_enum_text_for_index(
                             field_in, row[field_in])), 'OTH')
                outinfo = "In step 5, we will append Building Superstructure AL018 based on this feature {0}. Please verify the appropriate records were added. ".format(
                    row['UFI'])

                # Prepare dictionary containing fields for new record
                outdvals = dict()
                outdvals['F_CODE'] = 'AL018'  # F_CODE for Building Superstructure
                outdvals['FCSubtype'] = 100087  # FCSubtype for Building Superstructure

                # Get geometry from current row data
                for geom in settings.targetdb.common_fields_geom_only:
                    if is_item_in_dict_nocase(geom, row):
                        outdvals[geom] = get_value_from_dict_nocase(geom, row)

                currentgeomtype = settings.targetdb.assume_geometry_type_from_featureclass_name(target_fc)
                if currentgeomtype == 'point':
                    outputfc = 'StructurePnt'
                elif currentgeomtype == 'polyline':
                    outputfc = 'StructureCrv'
                else:
                    outputfc = 'StructureSrf'

                outinfofields = ""
                if row[field_in] == 51:  # With Steeple
                    outinfofields = "{0} {1}".format(outinfofields, "  BSU:15 = Steeple")
                    outdvals['BSU'] = 15
                elif row[field_in] == 77:  # With Cupola
                    outdvals['BSU'] = 14
                    outinfofields = "{0} {1}".format(outinfofields, "  BSU:14 = Cupola")
                elif row[field_in] == 78:  # with turret
                    outdvals['BSU'] = 16
                    outinfofields = "{0} {1}".format(outinfofields, "  BSU:16 = Turret")
                elif row[field_in] == 79:  # with tower
                    outdvals['BSU'] = 13
                    outinfofields = "{0} {1}".format(outinfofields, "  BSU:13 = Tower")
                elif row[field_in] == 80:  # with minaret
                    outdvals['BSU'] = 5
                    outinfofields = "{0} {1}".format(outinfofields, "  BSU:5 = Minaret")
                elif row[field_in] == 81:  # with smokestack
                    outdvals['BSU'] = 1
                    outinfofields = "{0} {1}".format(outinfofields, "  BSU:1 = Chimney")
                elif row[field_in] == 82:  # Flat with parapet
                    outdvals['BSU'] = 7
                    outinfofields = "{0} {1}".format(outinfofields, "  BSU:7 = Parapet")
                    rtnval = 41  # flat

                settings.report_values['Derived Features - Preparing Instance'] += 1
                # Save the derived content to the targetDB for output later where the remaining fields will be defaulted.
                settings.targetdb.save_derived_record_for_later_output(
                    settings, outputfc, row['UFI'],
                   outdvals, "Derived from SSR value: {0}".format(
                        row[field_in]))

                if len(outinfofields) > 0:
                    report_transfomer_issue_or_increment_count(settings, outinfo, "WARN", outinfofields)
                rtnval = -999999
            return rtnval
    return -999999


# ZI024_HYP
def transform_ZI024_HYP(settings, row, field_in, target_field, src_fc, target_fc):
    # GGDM 2.2 and GGDM 2.1 may still have some settings of 3 = Ephemeral. These should be changed to 2 = Intermittent
    if 'ZI024_HYP' in row:
        if row['ZI024_HYP'] == 3:  # Ephemeral
            settings.report_values['Hydrologic Persistence (HYP) Ephemeral - Intermittent'] += 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZI024_HYP', row['ZI024_HYP'], 'Ephemeral'),
                           'OTH')
            return 2  # Intermittent
        else:
            return row['ZI024_HYP']
    elif 'HYP' in row:
        if row['HYP'] == 3:  # Ephemeral
            settings.report_values['Hydrologic Persistence (HYP) Ephemeral - Intermittent'] += 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('HYP', row['HYP'], 'Ephemeral'), 'OTH')
            return 2  # Intermittent
        else:
            return row['HYP']
    else:
        return -999999


def transform_HYP(settings, row, field_in, target_field, src_fc, target_fc):
    # GGDM 2.2 and GGDM 2.1 may still have some settings of 3 = Ephemeral. These should be changed to 2 = Intermittent
    if 'HYP' in row:
        if row['HYP'] == 3:  # Ephemeral
            settings.report_values['Hydrologic Persistence (HYP) Ephemeral - Intermittent'] += 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('HYP', row['HYP'], 'Ephemeral'), 'OTH')
            return 2  # Intermittent
        else:
            return row['HYP']
    elif 'ZI024_HYP' in row:
        if row['ZI024_HYP'] == 3:  # Ephemeral
            settings.report_values['Hydrologic Persistence (HYP) Ephemeral - Intermittent'] += 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZI024_HYP', row['ZI024_HYP'], 'Ephemeral'),
                           'OTH')
            return 2  # Intermittent
        else:
            return row['ZI024_HYP']
    else:
        return -999999


# Using the source row value containing the UID, and the memory saved entries,
# pull the representative restriction info data content and save it
def transform_get_from_restriction_info_table(settings, row, field_in, target_field, src_fc, target_fc):
    # Target field used, not field_in
    # create a temp dictionary containing the restriction info elements from this source feature
    # ZI002_UFI is the restriction info reference
    if settings.sourcedb.is_field_in_fclass(src_fc, 'F_CODE') and settings.sourcedb.is_field_in_fclass(src_fc,
                                                                                                       'ZI002_UFI'):
        if len(settings.saved_restrictioninfo) == 0:
            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)

        for irest in settings.saved_restrictioninfo:
            if 'UFI' in irest and target_field in irest and irest['UFI'] == row['ZI002_UFI']:
                settings.report_values['Restriction info from related table'] += 1
                return irest[target_field]
            if 'UFI' in irest and field_in is not None and field_in in irest and irest['UFI'] == row['ZI002_UFI']:
                settings.report_values['Restriction info from related table'] += 1
                return irest[field_in]
        # If we got out of the for loop, then we did not find this UFI and field in the saved restriction information
        report_transfomer_issue_or_increment_count(
            settings, "Unable to find matching UFI for Restriction Info", "WARN",
           "UFI: {0} not found in saved restriction information metadata matching field {1} value. Trying to set {2}".format(
               strout(row['ZI002_UFI']), 'ZI002_UFI', target_field))
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)
    else:
        return None


# Using the source row value containing the UID, and the memory saved entries,
# pull the representative entity collection data content and save it
def transform_get_from_entity_collection_table(settings, row, field_in, target_field, src_fc, target_fc):
    # ZI039S_UFI
    if settings.sourcedb.is_field_in_fclass(src_fc, 'F_CODE') and settings.sourcedb.is_field_in_fclass(src_fc,
                                                                                                       'ZI039S_UFI'):
        if len(settings.saved_entity_collection_metadata) == 0:
            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)

        for irest in settings.saved_entity_collection_metadata:
            if 'UFI' in irest and target_field in irest and irest['UFI'] == row['ZI039S_UFI']:
                settings.report_values['Metadata from related entity collection'] += 1
                return irest[target_field]
            if 'UFI' in irest and field_in is not None and field_in in irest and irest['UFI'] == row['ZI039S_UFI']:
                settings.report_values['Metadata from related entity collection'] += 1
                return irest[field_in]
        # If we got out of the for loop, then we did not find this UFI and field in the saved entity collection metadata
        report_transfomer_issue_or_increment_count(
            settings,
           "Unable to find matching UFI for Entity Collection Metadata", "WARN",
           "UFI: {0} not found in saved entity collection metadata matching field {1} value. Trying to set {2}".format(
               strout(row['ZI039S_UFI']), 'ZI039S_UFI', target_field))
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)
    else:
        report_transfomer_issue_or_increment_count(
            settings,
           "transform_get_from_entity_collection_table expected F_CODE and ZI039S_UFI on source feature",
           "ERROR",
           " Feature class {0} did not have expected content".format(src_fc))
        return None
    return None


def transformZI001_VSCfromenumtocodelist(settings, row, field_in, target_field, src_fc, target_fc):
    # if (settings.getsource_db_version() == "GGDM 3.0" and
    # settings.gettarget_db_version() == "TDS 6.1") or (src_fc == 'Saved Entity Collection'):
    if field_in is not None and field_in in row:
        return ZI001_VSCenumvalue2ZI001_VSCcodelist[row[field_in]]

    # ZI039S_UFI
    if settings.sourcedb.is_field_in_fclass(src_fc, 'F_CODE') and settings.sourcedb.is_field_in_fclass(src_fc,
                                                                                                       'ZI039S_UFI'):
        if len(settings.saved_entity_collection_metadata) == 0:
            return ZI001_VSCenumvalue2ZI001_VSCcodelist[
                settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)]

        for irest in settings.saved_entity_collection_metadata:
            if 'UFI' in irest and target_field is not None and \
                    target_field in irest and irest['UFI'] == row['ZI039S_UFI']:
                settings.report_values['Metadata from related entity collection'] += 1
                return ZI001_VSCenumvalue2ZI001_VSCcodelist[irest[target_field]]
            if 'UFI' in irest and field_in is not None and field_in in irest and irest['UFI'] == row['ZI039S_UFI']:
                settings.report_values['Metadata from related entity collection'] += 1
                return ZI001_VSCenumvalue2ZI001_VSCcodelist[irest[field_in]]

        return ZI001_VSCenumvalue2ZI001_VSCcodelist[
            settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)]
    else:
        report_transfomer_issue_or_increment_count(settings,
                                                   "transformZI001_VSCfromenumtocodelist expected F_CODE and ZI039S_UFI on source feature",
                                                   "ERROR",
                                                   " Feature class {0} did not have expected content.".format(src_fc))
        return None

    return None


def transform_ESC(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if row[field_in] == 3:  # Water
            settings.report_values['Elevation Surface Category (ESC)'] += 1
            return 6  # Tidal Water
    elif settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if row[field_in] == 6:  # Tidal Water
            settings.report_values['Elevation Surface Category (ESC)'] += 1
            return 3  # Water
    else:
        return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


def transform_Industry_F_CODE(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if row[field_in] == 'AF021':  # Bucket Elevator
            settings.report_values['Feature code (F_CODE) changed'] = settings.report_values[
                                                                          'Feature code (F_CODE) changed'] + 1
            fclass_static['translated_F_CODE'] = 'AF020'
            return 'AF020'  # Conveyor
        else:
            return row[field_in]
    elif settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if row[field_in] == 'AF020' and 'CVT' in row and row['CVT'] == 4:  # Conveyor with CVT = 4
            fclass_static['translated_F_CODE'] = 'AF021'
            return 'AF021'  # Bucket Elevator
        else:
            return row[field_in]
    else:
        return row[field_in]


def transform_WRDB_F_CODE(settings, row, field_in, target_field, src_fc, target_fc):
    # This transform is being called on any feature that might have the 'DMF' field extension added to it in order
    # to signify that the data may be actually an AON feature.
    #
    # if the input F_CODE AND DMF Feature Count is > 1, verify this is an AON F_CODe, and if not change it.
    # DFDDAlphaCode	AKA
    # BH010 -> NM010	AqueductAON (Area) 
    # BH020 -> NM020	NavigableCanalAON (Area) 
    # BI010 -> NU010	CisternAON (Area) 
    # BH030 -> NM030	DitchAON (Area) 
    # BH082 -> NM082	InlandWaterbodyAON (Area) 
    # BH170 -> NM170	NaturalPoolAON (Area) 
    # AQ113 -> NU113	PipelineAON (Area) 
    # AQ116 -> NU116	PumpingStationAON (Area) 
    # BH140 -> NM140	RiverAON (Area) 
    # BH160 -> NM160	SabkhaAON (Area) 
    # AM070 -> NU070	StorageTankAON (Area) 
    # AK170 -> NU170	SwimmingPoolAON (Area) 
    # BH230 -> NM230	WaterWellAON (Area)
    fmapping = {
        'BH010': 'NM010',
        'BH020': 'NM020',
        'BI010': 'NU010',
        'BH030': 'NM030',
        'BH082': 'NM082',
        'BH170': 'NM170',
        'AQ113': 'NU113',
        'AQ116': 'NU116',
        'BH140': 'NM140',
        'BH160': 'NM160',
        'AM070': 'NU070',
        'AK170': 'NU170',
        'BH230': 'NM230'}

    if src_fc[-3:] == "Srf" and target_fc[-6:] == "AONExt":
        # If the DMF value is greater than 1, then change the feature code and 'move' the feature to the AONExt feature class,
        # Otherwise, leave this feature in the original feature class.
        # First of all, the feature class MUST be a surface feature, and the DMF attribute must be in the row and > 1 for any conversion to work.
        if 'DMF' in row and row['DMF'] > 1 and row[field_in] in fmapping:
            # Convert this FCODE to the AON equivalent
            settings.report_values['Feature code (F_CODE) changed'] += 1
            fclass_static['translated_F_CODE'] = fmapping[row[field_in]]
            settings.report_values['WRDB Area of Numerous Detected, Feature Code Mapped'] += 1
            return fmapping[row[field_in]]

        elif 'DMF' in row and row['DMF'] > 1:
            report_transfomer_issue_or_increment_count(
                settings, "  Illegal value DMF greater than 1, Feature dropped",
                "ERROR",
                "Illegal value DMF {0} > 1 on FC: {1}. This feature will be dropped.".format(row['DMF'], src_fc))
            # This mapping does not apply to the target feature class, drop it
            return None  # The Feature code of None will trigger the row to not be retained
        else:
            report_transfomer_issue_or_increment_count(
                settings, "  DMF value is missing or 0",
                "WARN",
                "Feature in feature class {0} is not valid for the target feature class {1} with this feature code {2}. DMF missing or 0. This may be okay if in a rule set moving features to a new feature class". \
                    format(src_fc, target_fc, row['F_CODE']))
            # This mapping does not apply to the target feature class, drop it
            return None  # The Feature code of None will trigger the row to not be retained
    else:
        # If this is not a Srf -> AONExt mapping
        if src_fc[-3:] == "Srf" and 'DMF' in row and row['DMF'] > 1 and row[field_in] in fmapping:
            # This F_CODE was transformed to the AONExt fcode, this feature needs to be DROPPED from the Srf feature
            settings.report_values[
                'WRDB Area of Numerous Detected, Feature not retained in original feature class'] += 1
            return None  # The Feature code of None will trigger the row to not be retained

        elif 'DMF' in row and row['DMF'] > 1:
            report_transfomer_issue_or_increment_count(
                settings, "  Illegal value DMF cannot be greater than 1.",
                "ERROR",
                "DMF value {0} indicates this should be in a different feature class - Area of Numerous".format(
                    row['DMF']))
            return row[field_in]

    return row[field_in]


fc_SubtypeXform = (
    'Saved Entity Collection', 'Target Entity Collection', 'ENTITY_COLLECTION_METADATA_T', 'DATASET_T', 'IndustryPnt',
    'IndustryCrv', 'IndustrySrf', 'HypsographyPnt', 'HypsographyCrv', 'HypsographySrf', 'StructurePnt', 'StructureCrv',
    'StructureSrf', 'CollectionSrf', 'MetadataSrf', 'ResourceSrf')


def transform_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc):
    fmapping = {
        'BH010': 'NM010',
        'BH020': 'NM020',
        'BI010': 'NU010',
        'BH030': 'NM030',
        'BH082': 'NM082',
        'BH170': 'NM170',
        'AQ113': 'NU113',
        'AQ116': 'NU116',
        'BH140': 'NM140',
        'BH160': 'NM160',
        'AM070': 'NU070',
        'AK170': 'NU170',
        'BH230': 'NM230'
    }
    fcsubmap = {
        100077: 211,
        100139: 210,
        100179: 206,
        100182: 207,
        100295: 200,
        100297: 201,
        100298: 203,
        130384: 204,
        100314: 208,
        100318: 209,
        100320: 205,
        100326: 212,
        100329: 202
    }
    # This bit is all for WRDB extensions to GGDM 2.2, where they put AON feature content within standard features.
    # First of all, this will only work for certain feature codes, AND
    # the feature must be a certain geometry, AND the DMF attribute has to be greater than 1 to bother with this mapping.
    if settings.getsource_db_version() == "GGDM 2.2" and src_fc[-3:] == "Srf" and target_fc[-6:] == "AONExt":
        if ('DMF' in row and row['DMF'] > 1 and row[
            field_in] in fmapping):  # DMF is > 1 and there is a mapping for this subtype
            # If we have a subtype map and the target is an AONExt feature class, we can map this
            if row[field_in] in fcsubmap:
                settings.report_values['WRDB Area of Numerous Detected, Subtype Mapped'] += 1
                return fcsubmap[row[field_in]]
            else:
                # This mapping does not apply to the AON target
                return None
        else:
            return None  # this subtype does not apply


    # Surface to Surface feature
    elif settings.getsource_db_version() == "GGDM 2.2" and src_fc[-3:] == "Srf" and target_fc[
                                                                                    -6:] != "AONExt" and 'DMF' in row and \
            row['DMF'] > 1 and row[field_in] in fmapping:
        # If we have a subtype map and the target is an AONExt feature class, we can map this
        if row[field_in] in fcsubmap:
            # This subtype was mapped to a different feature class, This mapping does not apply
            return None
        else:
            return row[field_in]

    elif src_fc in fc_SubtypeXform or target_fc in fc_SubtypeXform:
        if src_fc == 'ENTITY_COLLECTION_METADATA_T' and target_fc == "CollectionSrf":
            return transformMetadataSrfCollectionSrf_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc)
        elif src_fc == 'MetadataSrf' and target_fc == "CollectionSrf":
            return transformMetadataSrfCollectionSrf_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc)
        elif src_fc == 'Saved Entity Collection' and target_fc == "Target Entity Collection":
            return transformMetadataSrfCollectionSrf_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc)
        elif src_fc == 'CollectionSrf' and target_fc == "MetadataSrf":
            return transform_CollectionSrfMetadataSrf_FCSubtype(settings, row, field_in, target_field, src_fc,
                                                                target_fc)
        elif src_fc == 'DATASET_T' and target_fc == 'DatasetSrf':
            return 121591  # Dataset FCSubtype
        elif src_fc == 'ResourceSrf' and target_fc == 'DatasetSrf':
            return 121591  # Dataset FCSubtype
        elif src_fc in ('IndustryPnt', 'IndustryCrv', 'IndustrySrf'):
            return transform_Industry_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc)
        elif src_fc in ('HypsographyPnt', 'HypsographyCrv', 'HypsographySrf'):
            return transform_Hypsography_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc)
        elif src_fc in ('StructurePnt', 'StructureCrv', 'StructureSrf'):
            return transform_Structure_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc)
        else:
            return row[field_in]
    else:
        if field_in in row:
            return row[field_in]
        else:
            report_transfomer_issue_or_increment_count(settings,
                                                       "Expected field is required for transform FCSubtype, but field not present in data source",
                                                       "ERROR",
                                                       "Transform of FCSubtype requires field {0} from source {1}, to target{2}.".format(
                                                           field_in, src_fc, target_fc))
            return None
    return None


def transform_Industry_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc):
    curfcode = settings.sourcedb.get_f_code()
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if curfcode == 'AF021':  # Bucket Elevator
            return 100026  # Conveyor
    elif settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if curfcode == 'AF020' and 'CVT' in row and row['CVT'] == 4:
            return 100027  # Bucket Elevator
    else:
        return row[field_in]
    return row[field_in]


def transform_CVT(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() == 'GGDM 3.0':
        if row['F_CODE'] == 'AF021':
            return 4  # CVT must be Bucket if this was originally a Bucket Elevator
        elif field_in in row:
            return row[field_in]
        else:
            return -999999
    else:
        if field_in in row:
            return row[field_in]
        else:
            return -999999


## At some point, if GGDM were to merge Benchmark and Geodetic Point into Survey point, we may also need this logic:
##            if row['SUY'] == 2:  # Benchmark
##                settings.report_values['Feature code (F_CODE) changed'] = settings.report_values['Feature code (F_CODE) changed'] + 1
##                fclass_static['translated_F_CODE'] = 'ZB020'
##                return 'ZB020'   # Benchmark
##            elif row['SUY'] == 5: # Geodetic Point
##                settings.report_values['Feature code (F_CODE) changed'] = settings.report_values['Feature code (F_CODE) changed'] + 1
##                fclass_static['translated_F_CODE'] = 'ZB060'
##                return 'ZB060'   # Geodetic Point
##            elif row['SUY'] == 3: # Cadastral Control Point
##                settings.report_values['Feature code (F_CODE) changed'] = settings.report_values['Feature code (F_CODE) changed'] + 1
##                fclass_static['translated_F_CODE'] = 'ZB035'
##                return 'ZB035'   # Control POint
##            else:

def transform_Hypsography_F_CODE(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if row[field_in] == 'ZB035' or row[field_in] == 'ZB020' or row[field_in] == 'ZB060':
            # Control Point, Benchmark, Geodetic Point
            settings.report_values['Feature code (F_CODE) changed'] = settings.report_values[
                                                                          'Feature code (F_CODE) changed'] + 1
            fclass_static['translated_F_CODE'] = 'ZB050'
            return 'ZB050'  # Survey Point
    elif settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if row[field_in] == 'ZB050':  # Survey Point
            report_transfomer_issue_or_increment_count(settings,
                                                       "Survey point no longer handled, changing to a Control Point",
                                                       "WARN",
                                                       "Survey Point Type SUY {0} in source {1} {2} not handled in target. Saving as Control Point".format(
                                                           row['SUY'], src_fc, settings.sourcedb.get_f_code()))
            settings.report_values['Feature code (F_CODE) changed'] = settings.report_values[
                                                                          'Feature code (F_CODE) changed'] + 1
            fclass_static['translated_F_CODE'] = 'ZB035'
            return 'ZB035'  # Control Point
    else:
        return row[field_in]


def transform_Hypsography_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc):
    curfcode = settings.sourcedb.get_f_code()
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if curfcode == 'ZB035' or curfcode == 'ZB020' or curfcode == 'ZB060':  # Control Point, Benchmark, Geodetic Point
            return 177997  # Survey Point
    elif settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if curfcode == 'ZB050':  # Survey Point
            return 100466  # Control Point
    else:
        return row[field_in]


# transform_Structure_F_CODE
def transform_Structure_F_CODE(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if row[field_in] == 'AL371':  # Manor House
            settings.report_values['Feature code (F_CODE) changed'] = settings.report_values[
                                                                          'Feature code (F_CODE) changed'] + 1
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('F_CODE', 'AL371', 'Manor House'), 'OTH')
            fclass_static['translated_F_CODE'] = 'AL013'
            return 'AL013'  # Building

    return row[field_in]


def transform_Structure_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if settings.sourcedb.get_f_code() == 'AL371':  # Manor House
            return 100083  # Building

    return row[field_in]


# We use a different feature code for this metadata in the GGDM 3.0
def transform_CollectionSrfMetadataSrf_F_CODE(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        settings.report_values['Feature code (F_CODE) changed'] = settings.report_values[
                                                                      'Feature code (F_CODE) changed'] + 1
        return 'ZI039'  # Entity Collection Metadata
    if field_in not in row:
        return 'ZI039'
    return row[field_in]


def transform_CollectionSrfMetadataSrf_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        return 132721  # Entity Collection Metadata
    elif field_in not in row:
        return 132721
    return row[field_in]


#         # If the ZI001_SRT value maps to ZI002_SSY, then this is a spatial source and the feature code is ZI040
# Otherwise it is a non-spatial source an ZI041
def transformMetadataSrfCollectionSrf_F_CODE(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if is_row_spatialmd(row, settings.getsource_db_version()):
            settings.report_values['Feature code (F_CODE) changed'] = settings.report_values[
                                                                          'Feature code (F_CODE) changed'] + 1
            return 'ZI040'
        else:
            settings.report_values['Feature code (F_CODE) changed'] = settings.report_values[
                                                                          'Feature code (F_CODE) changed'] + 1
            return 'ZI041'
    else:
        return row[field_in]


def transformMetadataSrfCollectionSrf_FCSubtype(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if is_row_spatialmd(row, settings.getsource_db_version()):
            return 133101
        else:
            return 133102
    else:
        return row[field_in]


# This transforms SRT from codelist to enum or enum to codelist.
# Source type. This one is complicated due to many changes over revisions of GGDM and TDS.
def transform_ZI001_SRT(settings, row, field_in, target_field, src_fc, target_fc):
    # If the target is GGDM and source is GGDM 2.1 or TDS 4.0, try to look up the source type from saved entity collection metadata
    if settings.gettarget_db_version() == "GGDM 3.0" and (
            settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]):
        value = -999999
        ssydict = search_for_meta_in_dictionary(settings.saved_entity_collection_metadata, 'ZI001_SSY')
        ssddict = search_for_meta_in_dictionary(settings.saved_entity_collection_metadata, 'ZI001_SSD')
        if len(ssydict) > 0:
            if len(ssydict) == 1:
                settings.report_values['Source type (SRT)'] = settings.report_values['Source type (SRT)'] + 1
                value = ZI001_SRTcodelist2ZI001_SRTenumvalue[ssydict['ZI001_SSY'][0]]
            else:
                value = 999
                add_to_row_OTH(settings, row,
                               prepare_string_for_OTH_or_MEM('ZI001_SSY', strout(ssydict['ZI001_SSY']), ''), 'OTH')
        else:
            nspdict = search_for_meta_in_dictionary(settings.saved_entity_collection_metadata, 'ZI001_NSP')
            nsddict = search_for_meta_in_dictionary(settings.saved_entity_collection_metadata, 'ZI001_NSD')
            if len(nspdict) > 0:
                if len(nspdict) == 1:
                    settings.report_values['Source type (SRT)'] = settings.report_values['Source type (SRT)'] + 1
                    value = ZI001_SRTcodelist2ZI001_SRTenumvalue[nspdict['ZI001_NSP'][0]]
                else:
                    value = 999
                    add_to_row_OTH(settings, row,
                                   prepare_string_for_OTH_or_MEM('ZI001_NSP', strout(nspdict['ZI001_NSP']), ''), 'OTH')
        return value

    elif settings.gettarget_db_version() == "GGDM 3.0" and \
            (settings.getsource_db_version() in ["GGDM 2.2", "TDS 6.1", "TDS 7.1", "TDS 7.0"]):
        if 'SRT' in row:
            settings.report_values['Source type (SRT)'] = settings.report_values['Source type (SRT)'] + 1
            value = ZI001_SRTcodelist2ZI001_SRTenumvalue[row['SRT']] if row['SRT'] is not None else -999999
            if value == "999":
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('SRT', row['SRT'], ''), 'OTH')
            return value
        elif 'ZI001_SRT' in row:
            settings.report_values['Source type (SRT)'] = settings.report_values['Source type (SRT)'] + 1
            value = ZI001_SRTcodelist2ZI001_SRTenumvalue[row['ZI001_SRT']] if row['ZI001_SRT'] is not None else -999999
            if value == "999":
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('ZI001_SRT', row['ZI001_SRT'], ''), 'OTH')
            return value
        else:
            return -999999

    elif settings.getsource_db_version() == "GGDM 3.0" and field_in is not None and field_in in row:
        if settings.gettarget_db_version() == "GGDM 2.2" or \
                settings.gettarget_db_version() in ["TDS 6.1", "TDS 7.0", "TDS 7.1", "TDS 6.0"]:
            settings.report_values['Source type (SRT)'] = settings.report_values['Source type (SRT)'] + 1
            value = ZI001_SRTenumvalue2ZI001_SRTcodelist(row[field_in]) if row[
                                                                               field_in] is not None else 'noInformation'
        else:
            value = 'noInformation'
        return value
    return -999999


# 
def transform_ZI001_SDP(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if settings.sourcedb.get_f_code() == 'ZI040':  # Spatial metadata
            settings.report_values['Source description (SDP)'] += 1
            if 'ZI001_SSN' in row:
                return row['ZI001_SSN']  # Source Information : Spatial Source Description
            else:
                return 'No Information'
        elif settings.sourcedb.get_f_code() == 'ZI041':
            settings.report_values['Source description (SDP)'] += 1
            if 'ZI001_NSN' in row:
                return row['ZI001_NSN']  # Source Information : Non-spatial Source Description
            else:
                return 'No Information'
        else:
            return 'No Information'
    else:
        if field_in is not None and field_in in row and row[field_in] is not None:
            return row[field_in]
        else:
            return 'No Information'
    return 'No Information'


# Source date and time. This one is complicated due to many changes over revisions of GGDM and TDS.
def transform_ZI001_SDV(settings, row, field_in, target_field, src_fc, target_fc):
    # If the target is GGDM and source is GGDM 2.1 or TDS 4.0, try to look up the source type from saved entity collection metadata
    if settings.gettarget_db_version() == "GGDM 3.0" and (
            settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]):
        value = "noInformation"

        ssddict = search_for_meta_in_dictionary(settings.saved_entity_collection_metadata, 'ZI001_SSD')
        if len(ssddict) > 0:
            if len(ssddict) == 1:
                settings.report_values['Source date and time (SDV)'] += 1
                value = ssddict['ZI001_SSD'][0]
            else:
                value = "Other"
                add_to_row_OTH(settings, row,
                               prepare_string_for_OTH_or_MEM('ZI001_SSD', strout(ssddict['ZI001_SSD']), ''), 'OTH')
        else:

            nsddict = search_for_meta_in_dictionary(settings.saved_entity_collection_metadata, 'ZI001_NSD')
            if len(nsddict) > 0:
                if len(nsddict) == 1:
                    settings.report_values['Source date and time (SDV)'] += 1
                    value = nsddict['ZI001_NSD'][0]
                else:
                    value = 999
                    add_to_row_OTH(settings, row,
                                   prepare_string_for_OTH_or_MEM('ZI001_NSD', strout(nsddict['ZI001_NSD']), ''), 'OTH')
        return value

    elif settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if settings.sourcedb.get_f_code() == 'ZI040' and 'ZI001_SSD' in row:  # Spatial metadata
            settings.report_values['Source date and time (SDV)'] += 1
            return row['ZI001_SSD']  # Source Information : Spatial Source Date and Time
        elif settings.sourcedb.get_f_code() == 'ZI041' and 'ZI001_NSD' in row:
            settings.report_values['Source date and time (SDV)'] += 1
            return row['ZI001_NSD']  # Source Information : Non-spatial Source Date and Time
        else:
            return 'noInformation'

    elif field_in is not None and field_in in row:
        return row[field_in]

    return 'noInformation'


def is_row_spatialmd(row, db_version):
    if 'ZI001_SRT' in row:
        invalue = row['ZI001_SRT']
        if db_version == "GGDM 3.0":
            if invalue <> -999 and invalue <> -999999 and invalue in ZI001_SRTenumvalue2ZI001_SSYcodelist:
                return True
            elif invalue <> -999 and invalue <> -999999 and invalue in ZI001_SRTenumvalue2ZI001_NSPcodelist:
                return False
            else:
                return None
        elif db_version == "GGDM 2.2":
            # Codelist values - textual return none if no information or if we can't find the value
            if invalue == "noInformation" or invalue == "other" or invalue == "No Information":
                return None
            # Is the codelist value in SSY? If so then we have a spatial metadata row
            for vvv in ZI001_SRTenumvalue2ZI001_SSYcodelist:
                if ZI001_SRTenumvalue2ZI001_SSYcodelist[vvv] == invalue:
                    return True
            # Is the codelist value in NSP? If so, then we have a non-spatial metadata row
            for vvv in ZI001_SRTenumvalue2ZI001_NSPcodelist:
                if ZI001_SRTenumvalue2ZI001_NSPcodelist[vvv] == invalue:
                    return False
            return None
    return None


def transform_ZI001_SPS(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if settings.sourcedb.get_f_code() == 'ZI040':  # this is Spatial Md
            settings.report_values['Source spatial (SPS) boolean was derived'] += 1
            return 1001
        else:
            settings.report_values['Source spatial (SPS) boolean was derived'] += 1
            return 1000
    return settings.targetdb.field_set_to_default('ZI001', 'ZI001_SPS', target_fc)


def transform_ZI001_SSY(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        # target_fc needs to be CollectionSrf, source_fc needs to MetadataSrf
        testval = is_row_spatialmd(row, settings.getsource_db_version())
        if testval is None:  # unknown - either Other or No Information
            return 'noInformation'
        elif testval:
            settings.report_values['Source Type - Spatial (SSY)'] += 1
            if settings.getsource_db_version() == "GGDM 3.0" and 'ZI001_SRT' in row:
                return ZI001_SRTenumvalue2ZI001_SSYcodelist[row['ZI001_SRT']]
            elif settings.getsource_db_version() == "GGDM 2.2" and 'ZI001_SRT' in row:
                return row['ZI001_SRT']  # GGDM 2.2 is a codelist value, and so is GGDM 2.1
        else:
            return 'noInformation'


def transform_ZI001_SSD(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        # target_fc needs to be CollectionSrf, source_fc needs to MetadataSrf
        testval = is_row_spatialmd(row, settings.getsource_db_version())
        if testval is None:  # unknown - either Other or No Information
            return 'noInformation'
        elif testval == True and 'ZI001_SDV' in row:
            settings.report_values['Source Date - Spatial (SSD)'] += 1
            return row['ZI001_SDV']
        else:
            return 'noInformation'


def transform_ZI001_SSN(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        # target_fc needs to be CollectionSrf, source_fc needs to MetadataSrf
        testval = is_row_spatialmd(row, settings.getsource_db_version())
        if testval is None:  # unknown - either Other or No Information
            return 'No Information'
        elif testval == True and 'ZI001_SDP' in row:
            settings.report_values['Source Description - Spatial (SSN)'] += 1
            return row['ZI001_SDP']
        else:
            return 'No Information'


# Underbridge Clearance can be derived from some BSC values
def transform_UBC_from_BSC(settings, row, field_in, target_field, src_fc, target_fc):
    if 'BSC' in row and 'UBC' not in row and field_in == 'UBC':
        if row['BSC'] == 5 or row['BSC'] == 28 or row['BSC'] == 30:  # 5=Floating or 28 Viaduct, or 30 Submersible
            add_to_row_MEM(settings, row, prepare_string_for_OTH_or_MEM('UBC', 'UBC derived from BSC', ''), 'MEM')
            settings.report_values['Underbridge Clearance derived from Bridge Structure'] += 1
            return 0
        elif row['BSC'] <> 999 and row['BSC'] <> -999999:
            add_to_row_MEM(settings, row, prepare_string_for_OTH_or_MEM('UBC', 'UBC derived from BSC', ''), 'MEM')
            settings.report_values['Underbridge Clearance derived from Bridge Structure'] += 1
            return 0.5
    return -999999


def transform_ZI001_NSP(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        # target_fc needs to be CollectionSrf, source_fc needs to MetadataSrf
        testval = is_row_spatialmd(row, settings.getsource_db_version())
        if testval is None:  # unknown - either Other or No Information
            return 'noInformation'
        elif testval == False:
            settings.report_values['Source Type - Non-Spatial (NSP)'] += 1
            if settings.getsource_db_version() == "GGDM 3.0" and 'ZI001_SRT' in row:
                return ZI001_SRTenumvalue2ZI001_NSPcodelist[row['ZI001_SRT']]
            elif settings.getsource_db_version() == "GGDM 2.2" and 'ZI001_SRT' in row:
                return row['ZI001_SRT']  # GGDM 2.2 is a codelist value, and so is GGDM 2.1
        else:
            return 'noInformation'


def transform_ZI001_NSD(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        # target_fc needs to be CollectionSrf, source_fc needs to MetadataSrf
        testval = is_row_spatialmd(row, settings.getsource_db_version())
        if testval is None:  # unknown - either Other or No Information
            return 'noInformation'
        elif testval == False and 'ZI001_SDV' in row:
            settings.report_values['Source Date - Non-Spatial (NSD)'] += 1
            return row['ZI001_SDV']
        else:
            return 'noInformation'


def transform_ZI001_NSN(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        # target_fc needs to be CollectionSrf, source_fc needs to MetadataSrf
        testval = is_row_spatialmd(row, settings.getsource_db_version())
        if testval is None:  # unknown - either Other or No Information
            return 'No Information'
        elif testval == False and 'ZI001_SDP' in row:
            settings.report_values['Source Description - Non-Spatial (NSN)'] += 1
            return row['ZI001_SDP']
        else:
            return 'No Information'


# INVL	9   Interval Closure Type Code	Less-than or Equal Semi-interval	The unbounded interval (-infinity, maximumValue].
# [Description] The value supplied for the lower endpoint, if any, is not meaningful.
# INVL	8   Interval Closure Type Code	Less-than Semi-interval	The right half-open unbounded interval (-infinity, maximumValue).
# [Description] The value supplied for the lower endpoint, if any, is not meaningful.
# INVL	7   Interval Closure Type Code	Greater-than or Equal Semi-interval	The unbounded interval [minimumValue, +infinity).
# [Description] The value supplied for the upper endpoint, if any, is not meaningful.
# INVL	6   Interval Closure Type Code	Greater-than Semi-interval	The left half-open unbounded interval (minimumValue, +infinity).
# [Description] The value supplied for the upper endpoint, if any, is not meaningful.
# INVL	5   Interval Closure Type Code	Closed Interval	The bounded interval [minimimValue, maximumValue].
# INVL	4   Interval Closure Type Code	Greater-than to Less-than-or-equal Interval	The left half-open bounded interval (minimimValue, maximumValue].
# INVL	3   Interval Closure Type Code	Greater-than-or-equal to Less-than Interval	The right half-open bounded interval [minimimValue, maximumValue).
# INVL	2   Interval Closure Type Code	Open Interval	The bounded open interval (minimumValue, maximumValue).
# INVL	1   Interval Closure Type Code	Single Value	A single value.   [Description] Usually the same value is provided for both interval endpoints.

def transform_VLM_fromInterval(settings, row, field_in, target_field, src_fc, target_fc):
    if 'VLMC' in row:
        if row['VLMC'] is not None and row['VLMU'] is not None and row['VLML'] is not None:
            # VLMC plus VLMU and VLML
            settings.report_values['Interval - Convert interval to single value'] += 1
            if row['VLMU'] == -999999.0 and row['VLML'] == -999999.0:
                return -999999.0
            elif row['VLMC'] == 9 or row['VLMC'] == 8:
                return row['VLMU']
            elif row['VLMC'] == 7 or row['VLMC'] == 6:
                return row['VLML']
            elif row['VLMC'] == 5 or row['VLMC'] == 4 or row['VLMC'] == 3 or row['VLMC'] == 2:
                return row['VLMU'] - ((row['VLMU'] - row['VLML']) / 2)
            elif row['VLMC'] == 1:
                return row['VLML']

    return -999999


def transform_VLM_from_nonInterval(settings, row, field_in, target_field, src_fc, target_fc):
    if 'VLM' in row:
        settings.report_values['Interval - Convert single value to interval'] += 1
        if row['VLM'] is not None:
            if target_field == 'VLMU':
                return row['VLM']
            elif target_field == 'VLML':
                return row['VLM']
            elif target_field == 'VLMC':
                return 1  # Interval closure type: Single value
    if target_field == 'VLMC':
        return 1  # No Information is not a valid value for the interval closure - argh!
    else:
        return -999999


def transform_VTI_fromInterval(settings, row, field_in, target_field, src_fc, target_fc):
    if 'VTIC' in row:
        if row['VTIC'] is not None and row['VTIU'] is not None and row['VTIL'] is not None:
            # VTIC plus VTIU and VTIL
            settings.report_values['Interval - Convert interval to single value'] += 1
            if row['VTIU'] == -999999.0 and row['VTIL'] == -999999.0:
                return -999999.0
            elif row['VTIC'] == 9 or row['VTIC'] == 8:
                return row['VTIU']
            elif row['VTIC'] == 7 or row['VTIC'] == 6:
                return row['VTIL']
            elif row['VTIC'] == 5 or row['VTIC'] == 4 or row['VTIC'] == 3 or row['VTIC'] == 2:
                return row['VTIU'] - ((row['VTIU'] - row['VTIL']) / 2)
            elif row['VTIC'] == 1:
                return row['VTIL']

    return -999999


def transform_VTI_toInterval(settings, row, field_in, target_field, src_fc, target_fc):
    if 'VTI' in row:
        settings.report_values['Interval - Convert single value to interval'] += 1
        if row['VTI'] is not None:
            if target_field == 'VTIU':
                return row['VTI']
            elif target_field == 'VTIL':
                return row['VTI']
            elif target_field == 'VTIC':
                return 1  # Interval closure type: Single value
    if target_field == 'VTIC':
        return 1  # No Information is not a valid value for the interval closure - argh!
    else:
        return -999999


# This Pump Present PMPR attribute is only found in GGDM 3.0.  If the source feature has a PMP attribute
# with a non-default value present then PMPR, the PMPR boolean can be derived. It is also
# a good idea to print out a warning because in GGDM 3.0, the pump can be represented as a separate
# feature.
def transform_PMPR(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() == "GGDM 3.0":
        settings.report_values['Pump type detected and may impact pump present'] += 1
        if 'PMP' in row:
            if not is_value_unknown(row['PMP'], settings.sourcedb.geodb_type_unknown, True):
                report_transfomer_issue_or_increment_count(
                    settings,
                   "Please set up the new Pump feature to capture Pump details",
                   "WARN",
                   "Attribute Pump Type (PMP) detected in source data {0}, {1}. Note GGDM 3.0 has a separate Pump feature and pump details may be dropped unless moved to the new feature.". \
                   format(src_fc, settings.targetdb.get_f_code()))
                return 1001  # True, PMP attribute is set to a non-default value, ergo a pump is present
        elif 'ZI024_PMP' in row:
            if not is_value_unknown(row['ZI024_PMP'], settings.sourcedb.geodb_type_unknown, True):
                report_transfomer_issue_or_increment_count(
                    settings,
                   "Please set up the new Pump feature to capture Pump details",
                   "WARN",
                   "Attribute Pump Type (ZI024_PMP) detected in source data {0}, {1}. Note GGDM 3.0 has a separate Pump feature and pump details may be dropped unless moved to the new feature.". \
                   format(src_fc, settings.targetdb.get_f_code()))
                return 1001  # True, PMP attribute is set to a non-default value, ergo a pump is present 
    if field_in is not None and field_in in row:
        return row[field_in]
    return -999999


# If the Pump Type PMP attribute is detected, and not default, we need to print a warning if the target is GGDM 3.0
# to tell the operator that the Pump feature is available and should be used.
def transform_PMP(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() == "GGDM 3.0":
        if field_in in row and not is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True):
            if field_in == 'PMP':
                report_transfomer_issue_or_increment_count(
                    settings, 'Pump type may be derived from pump present',
                   "WARN",
                   "Attribute Pump Type (PMP) detected in source data {0}, {1}. Note GGDM 3.0 has a separate Pump feature and pump details may be dropped unless moved to the new feature.". \
                   format(src_fc, settings.targetdb.get_f_code()))
                return row[field_in]
            elif field_in == 'ZI024_PMP':
                report_transfomer_issue_or_increment_count(
                    settings, 'Pump type may be derived from pump present',
                   "WARN",
                   "Attribute Pump Type (ZI024_PMP) detected in source data {0}, {1}. Note GGDM 3.0 has a separate Pump feature and pump details may be dropped unless moved to the new feature.". \
                   format(src_fc, settings.targetdb.get_f_code()))
                return row[field_in]
            elif field_in == 'PMPR':
                return row[field_in]

    elif settings.getsource_db_version() == "GGDM 3.0":
        if 'PMPR' in row:
            settings.report_values['Pump type may be derived from pump present'] += 1

            # If we see PMPR in the row and it is set to true, then we know we have a pump of some type
            if row['PMPR'] == 1001 and is_value_unknown(row[field_in], settings.sourcedb.geodb_type_unknown, True):
                report_transfomer_issue_or_increment_count(settings,
                                                           'Attribute Pump Present (PMPR) detected in source data',
                                                           "WARN",
                                                           "Attribute Pump Present (PMPR) detected in source data {0}, {1} and set to true. Setting PMP or ZI024_PMP to Other, but we do not have a pump type. Note GGDM 3.0 has a separate Pump feature and pump details may be dropped unless moved to the new feature.". \
                                                           format(src_fc, settings.targetdb.get_f_code()))
                return 999  # Other, but we can't set the type because we don't know what it is.
    return -999999


def transform_SUYfromCPA(settings, row, field_in, target_field, src_fc, target_fc):
    # SUY is present on Control Point in GGDM 3.0.  It can be derived from SUY values (mostly)
    if settings.gettarget_db_version() == "GGDM 3.0":
        if 'CPA' in row:
            if not is_value_unknown(row['CPA'], settings.sourcedb.geodb_type_unknown, True):
                settings.report_values['Survey point type (SUY) derived from control point type (CPA)'] = \
                    settings.report_values['Survey point type (SUY) derived from control point type (CPA)'] + 1
                curfcode = settings.sourcedb.get_f_code()
                if curfcode == 'ZB020':  # If SOURCE row F_CODE is Benchmark feature, set benchmark value
                    if row['CPA'] == 5:  # Source is 5 Vertical Position  (Benchmark feature)
                        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CPA', 5, 'Vertical Position'),
                                       'OTH')
                    elif row['CPA'] == 2 or row[
                        'CPA'] == 3:  # Source is either 3 = Horizontal Position and Benchmark; or 2 Horizontal Position (Benchmark feature)
                        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CPA', 2, 'Horizontal Position'),
                                       'OTH')
                    return 2  # Benchmark
                elif curfcode == 'ZB035':  # If source row is Control Point feature, set Cadastral Control Point
                    if row['CPA'] == 5:  # Source is 5 Vertical Position  (Control point feature)
                        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CPA', 5, 'Vertical Position'),
                                       'OTH')
                    elif row['CPA'] == 2 or row[
                        'CPA'] == 3:  # Source is either 3 = Horizontal Position and Benchmark; or 2 Horizontal Position (Control point feature)
                        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CPA', 2, 'Horizontal Position'),
                                       'OTH')
                    return 3  # Cadastral Control Point
                elif curfcode == 'ZB060':  # If source row is Geodetic Point feature, set Geodetic Point Point
                    if row['CPA'] == 5:  # Source is 5 Vertical Position  (Geodetic point feature)
                        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CPA', 5, 'Vertical Position'),
                                       'OTH')
                    elif row['CPA'] == 2 or row[
                        'CPA'] == 3:  # Source is either 3 = Horizontal Position and Benchmark; or 2 Horizontal Position (Geodetic point feature)
                        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('CPA', 2, 'Horizontal Position'),
                                       'OTH')
                    return 5  # Geodetic Control Point
            return row['CPA']
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)


def transform_CPAfromSUY(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() == "GGDM 3.0":
        if 'SUY' in row:
            if row['SUY'] == 999 or row['SUY'] == -999999:
                return row['SUY']

            if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
                if row['SUY'] != 3:  # Cadastral Control Point
                    settings.report_values[
                        'Transform processing: Control point type (CPA) derived from survey point (SUY)'] += 1

                    report_transfomer_issue_or_increment_count(
                        settings,
                        "This target DB does not accept this type of control point",
                        "WARN",
                        "Control point type {0}. Source {1} {2}. Will save to Control Point with SUY in OTH." \
                            .format(row['SUY'], src_fc,
                                    settings.sourcedb.get_f_code()))
                    add_to_row_OTH(
                        settings, row, prepare_string_for_OTH_or_MEM('SUY', row['SUY'],
                                                                     settings.sourcedb.get_enum_text_for_index(
                                                                         'SUY', row['SUY'])), 'OTH')
                    return 999
            else:
                if row['SUY'] != 3 and row['SUY'] != 5 and row['SUY'] != 2:
                    settings.report_values[
                        'Transform processing: Control point type (CPA) derived from survey point (SUY)'] += 1

                    report_transfomer_issue_or_increment_count(settings,
                                                               "This target DB does not accept this type of control point",
                                                               "WARN",
                                                               "Control point type {0}. Source {1} {2}. Will save to Control Point with SUY in OTH." \
                                                               .format(row['SUY'], src_fc,
                                                                       settings.sourcedb.get_f_code()))
                    add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('SUY', row['SUY'],
                                                                                settings.sourcedb.get_enum_text_for_index(
                                                                                    'SUY', row['SUY'])), 'OTH')
                    return 999
            return row['SUY']
    return check_for_and_transform_missing_enum_index(settings, row, field_in, target_field, src_fc, target_fc)


def transform_LDC(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        return row[field_in]
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)


# Only applies to 2.1 as the target. ISS is not in 4.0
# only applies if ISS is not in the source content.
def transform_ISS(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.gettarget_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if 'LMC' in row:
            settings.report_values['Isolated derived from landmark'] += 1
            return row['LMC']
        elif field_in is not None and field_in in row:
            return row[field_in]

    if field_in is not None and field_in in row:
        return row[field_in]
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)


def transform_LMC(settings, row, field_in, target_field, src_fc, target_fc):
    if settings.getsource_db_version() in ["TDS 4.0", "GGDM 2.1"]:
        if 'ISS' in row and 'LMC' not in row:
            settings.report_values['Landmark derived from isolated'] += 1
            return row['ISS']
        elif field_in in row:
            return row[field_in]
        else:
            return -999999
    if field_in is not None and field_in in row:
        return row[field_in]
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc)


def transform_SINU_PSN(settings, row, field_in, target_field, src_fc, target_fc):
    if 'SINU' in row and 'PSN' not in row:
        return row['SINU']
    elif 'PSN' in row and 'SINU' not in row:
        return row['PSN']
    elif field_in in row:
        return row[field_in]
    return -999999


def transform_WPUP_PUD(settings, row, field_in, target_field, src_fc, target_fc):
    if 'WPUP' in row and 'PUD' not in row:
        return row['WPUP']
    elif 'PUD' in row and 'WPUP' not in row:
        return row['PUD']
    elif field_in in row:
        return row[field_in]
    return -999999


def transform_WDWL_DWL(settings, row, field_in, target_field, src_fc, target_fc):
    if 'WDWL' in row and 'DWL' not in row:
        return row['WDWL']
    elif 'DWL' in row and 'WDWL' not in row:
        return row['DWL']
    elif field_in in row:
        return row[field_in]
    return -999999


def transform_WCSD_CDI(settings, row, field_in, target_field, src_fc, target_fc):
    if 'WDWL' in row and 'DWL' not in row:
        return row['WDWL']
    elif 'DWL' in row and 'WDWL' not in row:
        return row['DWL']
    elif field_in in row:
        return row[field_in]
    return -999999


def transform_MEL(settings, row, field_in, target_field, src_fc, target_fc):
    if 'MDSP' in row and 'MEL' not in row:
        if not is_value_unknown(row['MDSP'], settings.sourcedb.geodb_type_unknown, True):
            settings.report_values['Medical specialty transform'] += 1

            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('MDSP', 0, row['MDSP']), 'OTH')
            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                          "Put results dict here", False)
    elif 'MEL' in row:
        return check_for_and_transform_missing_enum_index(settings, row, 'MEL', target_field, src_fc, target_fc)
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                  "Put results dict here", False)


def transform_MFY(settings, row, field_in, target_field, src_fc, target_fc):
    if 'MDFC' in row and 'MFY' not in row:
        if not is_value_unknown(row['MDFC'], settings.sourcedb.geodb_type_unknown, True):
            settings.report_values['Medical facility transform'] += 1

            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('MDFC', 0, row['MDFC']), 'OTH')
            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                          "Put results dict here", False)
    elif 'MFY' in row:
        return check_for_and_transform_missing_enum_index(settings, row, 'MFY', target_field, src_fc, target_fc)
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                  "Put results dict here", False)


def transform_MDFC(settings, row, field_in, target_field, src_fc, target_fc):
    if 'MDY' in row and 'MDFC' not in row:
        if not is_value_unknown(row['MDY'], settings.sourcedb.geodb_type_unknown, True):
            settings.report_values['Medical facility transform'] = settings.report_values[
                                                                       'Medical facility transform'] + 1

            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('MDY', row['MDY'],
                                                                        settings.sourcedb.get_enum_text_for_index('MDY',
                                                                                                                  row[
                                                                                                                      'MDY'])),
                           'OTH')
            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                          "Put results dict here", False)
    elif 'MDFC' in row:
        return check_for_and_transform_missing_enum_index(settings, row, 'MDFC', target_field, src_fc, target_fc)
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                  "Put results dict here", False)


def transform_MDSP(settings, row, field_in, target_field, src_fc, target_fc):
    if 'MEL' in row and 'MDSP' not in row:
        if not is_value_unknown(row['MEL'], settings.sourcedb.geodb_type_unknown, True):
            settings.report_values['Medical specialty transform'] = settings.report_values[
                                                                        'Medical specialty transform'] + 1

            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('MEL', row['MEL'],
                                                                        settings.sourcedb.get_enum_text_for_index('MEL',
                                                                                                                  row[
                                                                                                                      'MEL'])),
                           'OTH')
            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                          "Put results dict here", False)
    elif 'MDSP' in row:
        return check_for_and_transform_missing_enum_index(settings, row, 'MDSP', target_field, src_fc, target_fc)
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                  "Put results dict here", False)


def transform_ZI010_ACC(settings, row, field_in, target_field, src_fc, target_fc):
    if 'ZI010_ACC' in row:
        if settings.gettarget_db_version() == "GGDM 3.0":
            if not is_value_unknown(row['ZI010_ACC'], settings.sourcedb.geodb_type_unknown, True):
                report_transfomer_issue_or_increment_count(settings,
                                                           'Transform no target: Accuracy category (ZI010_ACC) present in source but we have no defined transform content',
                                                           "WARN",
                                                           "Source data uses ZI010_ACC value {0} on {1} {2}. There is no target location for this. Data will be lost." \
                                                           .format(row['ZI010_ACC'], src_fc,
                                                                   settings.sourcedb.get_f_code()))
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                  "Put results dict here", False)


def transform_combine_FURL(settings, row, field_in, target_field, src_fc, target_fc):
    fout = ""
    if 'FURL' in row and not is_value_unknown(row['FURL'], settings.sourcedb.geodb_type_unknown, True):
        fout = "<url=" + row[field_in] + ">"
    if 'FURL2' in row and not is_value_unknown(row['FURL2'], settings.sourcedb.geodb_type_unknown, True):
        settings.report_values['WRDB Extension - FURL multiple value'] += 1
        fout = fout + "<url=" + row['FURL2'] + ">"
    if 'FURL3' in row and not is_value_unknown(row['FURL3'], settings.sourcedb.geodb_type_unknown, True):
        settings.report_values['WRDB Extension - FURL multiple value'] += 1
        fout = fout + "<url=" + row['FURL3'] + ">"
    if 'FURL4' in row and not is_value_unknown(row['FURL4'], settings.sourcedb.geodb_type_unknown, True):
        settings.report_values['WRDB Extension - FURL multiple value'] += 1
        fout = fout + "<url=" + row['FURL4'] + ">"
    if fout == "":
        fout = "noInformation"
    return fout


# transform_pull_wrdb_extension

def transform_pull_wrdb_extension(settings, row, field_in, target_field, src_fc, target_fc):
    # SOURCE is always None so far, therefore field_in is going to be None
    if target_field == 'ZI024_WTNO' and 'ZI024_NTR' in row and not is_value_unknown(row['ZI024_NTR'],
                                                                                    settings.sourcedb.geodb_type_unknown,
                                                                                    True):
        settings.report_values['WRDB Extension - Water Nitrate'] += 1
        return row['ZI024_NTR']
    elif target_field == 'ZI024_AWOX' and 'ZI024_AWO2' in row and not is_value_unknown(row['ZI024_AWO2'],
                                                                                       settings.sourcedb.geodb_type_unknown,
                                                                                       True):
        settings.report_values['WRDB Extension - Water Flow Minimum'] += 1
        return row['ZI024_AWO2']
    elif target_field == 'ZI024_AWOI' and 'ZI024_AWO3' in row and not is_value_unknown(row['ZI024_AWO3'],
                                                                                       settings.sourcedb.geodb_type_unknown,
                                                                                       True):
        settings.report_values['WRDB Extension - Water Flow Maximum'] += 1
        return row['ZI024_AWO3']
    elif target_field == 'STNB' and 'STRH' in row and not is_value_unknown(row['STRH'],
                                                                           settings.sourcedb.geodb_type_unknown, True):
        settings.report_values['WRDB Extension - Strahler Number'] += 1
        return row['STRH']
    return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                  "Put results dict here", False)


def transform_millions_to_ones(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if row[field_in] is not None and row[field_in] != 0:
            return row[field_in] / 1000000
        else:
            return row[field_in]
    else:
        return -999999


def transform_thousands_to_ones(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if row[field_in] is not None and row[field_in] != 0:
            return row[field_in] / 1000
        else:
            return row[field_in]
    else:
        return -999999


def transform_slope_degree_to_percent(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if row[field_in] is not None and row[field_in] != 0:
            return (math.tan(row[field_in])) * 100
        else:
            return row[field_in]
    else:
        return -999999.0


def transform_hydro_country_to_GE4(settings, row, field_in, target_field, src_fc, target_fc):
    if 'Country' in row:
        if row['Country'] in CountryNameToGE4:
            if row['Country'].lower() == 'other':
                add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM('Country', row['Country'], ""), 'OTH')
                return "Other"
            else:
                tname = CountryNameToGE4[row['Country']]
                return tname
    return 'No Information'


def transform_no_target_save_to_oth(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:
        if row[field_in] is not None:
            add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], ""), 'OTH')
    return


def transform_with_value_mapping(settings, row, field_in, target_field, src_fc, target_fc):
    rtnval = settings.mapping_specification.get_target_value_mapping(settings.current_mapping_group, target_field)
    if rtnval is not None:
        # {'IWT': {'sourec attribute' : { 'Lake_type': { 'value mapping': {1:1, 2:4, 3:999}, 'save to oth': {3:'name'}, 'operator': 'NA' } } } }
        if target_field in rtnval:
            srcattrset = rtnval[target_field]['source attribute']
            targetsetop = dict()
            done = False
            for srcattrtolookfor in srcattrset:
                if srcattrtolookfor in row and not done:

                    # A value mapping entry should contain a dictionary listing the source to target values
                    if 'value mapping' in srcattrset[srcattrtolookfor]:
                        valmap = srcattrset[srcattrtolookfor]['value mapping']
                    else:
                        valmap = None

                    # The operator should be an and/or value indicating whether we can map if any of the values map, or if we have to have multiple
                    if 'operator' in srcattrset[srcattrtolookfor]:
                        operator = srcattrset[srcattrtolookfor]['operator'].lower()
                    else:
                        operator = 'or'

                    # The operator will be a type of dictionary and each operator will get a list of 'results' 
                    # This is all going to work with one and only one operator just now
                    if operator not in targetsetop:
                        targetsetop[operator] = list()

                    # If the save to oth flag is set, then we expect to see a dictionary of source values and names that will be saved to oth
                    # This will be activated FIRST, so that we can save to OTH and in addition, map the value to a target
                    if 'save to oth' in srcattrset[srcattrtolookfor]:
                        savetoothvalues = srcattrset[srcattrtolookfor]['save to oth']
                    else:
                        savetoothvalues = dict()

                    #### Do the actions now
                    # Start with the save to oth if that is set up
                    if row[srcattrtolookfor] in savetoothvalues:
                        add_to_row_OTH(settings, row,
                                       prepare_string_for_OTH_or_MEM(srcattrtolookfor, row[srcattrtolookfor],
                                                                     savetoothvalues[row[srcattrtolookfor]]), 'OTH')

                    # Next, see if there is mapping. if there is a mapping save the result to the operator dictionary list, if not - save an indicator that we tried
                    if valmap is not None:
                        if row[srcattrtolookfor] in valmap:
                            targetsetop[operator].append(valmap[row[srcattrtolookfor]])
                            # Note that if the operator is 'or' and we find a mapping, we are done immediately
                            if operator == 'or' or operator == 'na':
                                done = True
                                return valmap[row[srcattrtolookfor]]
                        else:
                            report_transfomer_issue_or_increment_count(settings,
                                                                       "transform_with_value_mapping did not find specified target mapping",
                                                                       "WARN",
                                                                       "transform_with_value_mapping did not find specified target mapping from source value {2}, mapping group {0}, target field {1}. Setting to target default.".format(
                                                                           settings.current_mapping_group, target_field,
                                                                           row[srcattrtolookfor]))
                            return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field,
                                                                          target_fc, "Put results dict here", False)

                    else:
                        targetsetop[operator].append('no attribute')

            # If the oeprator is 'and' we need to verify that all of the values we looked for are present and the same results were specified
            if 'and' in targetsetop:
                if len(targetsetop['and']) > 0 and targetsetop['and'][0] != 'no attribute':
                    success = all(
                        targetsetop['and'][i] == targetsetop['and'][i + 1] for i in range(len(targetsetop['and']) - 1))
                    if success:
                        return targetsetop['and'][0]

    else:
        report_transfomer_issue_or_increment_count(settings,
                                                   "transform_with_value_mapping did not find specified transform rule set",
                                                   "WARN",
                                                   "transform_with_value_mapping did not find specified transform rule set, mapping group {0}, target field {1}".format(
                                                       settings.current_mapping_group, target_field))
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                      "Put results dict here", False)


def transform_set_to_specified_value(settings, row, field_in, target_field, src_fc, target_fc):
    rtnval = settings.mapping_specification.get_target_value_specific_default(settings.current_mapping_group,
                                                                              target_field)
    if rtnval is not None:
        return rtnval
    else:
        return settings.targetdb.field_set_to_default(settings.targetdb.get_f_code(), target_field, target_fc,
                                                      "Put results dict here", False)
    return


def transform_MDSP_MEL(settings, row, field_in, target_field, src_fc, target_fc):
    if field_in in row:  # Be sure the field is in the row to start with MDSP is source and MEL is target
        # The transformation needs to look at the target enum values and see if the source string is any of these
        if row[field_in] == '' or row[field_in].lower() == 'no information' or row[field_in].lower() == 'noinformation':
            return -999999
        for iinum in settings.targetdb.field_information['GGDM 3.0']['SummaryDomainValues']['MEL']:
            if settings.targetdb.field_information['GGDM 3.0']['SummaryDomainValues']['MEL'][iinum] == row[field_in]:
                return iinum
        # If we have not found a value and returned it, set OTH and return 999
        add_to_row_OTH(settings, row, prepare_string_for_OTH_or_MEM(field_in, row[field_in], ""), 'OTH')
        return 999


##################  Function definition within dictionary for later lookup/call
call_transform_function = dict()
call_transform_function['transform_with_value_mapping'] = transform_with_value_mapping
call_transform_function['transform_set_to_specified_value'] = transform_set_to_specified_value
call_transform_function['check_for_and_transform_missing_enum_index'] = check_for_and_transform_missing_enum_index
call_transform_function['transform_APT'] = transform_APT
call_transform_function['transform_ASU'] = transform_ASU
call_transform_function['transform_BOT'] = transform_BOT
call_transform_function['transform_BSC'] = transform_BSC
call_transform_function['transform_BST'] = transform_BST
call_transform_function['transform_CIW_to_ZI016_WTC'] = transform_CIW_to_ZI016_WTC
call_transform_function['transform_COS'] = transform_COS
call_transform_function['transform_CPAfromSUY'] = transform_CPAfromSUY
call_transform_function['transform_decimetre_to_metre'] = transform_decimetre_to_metre
call_transform_function['transform_default_value_string'] = transform_default_value_string
call_transform_function['transform_report_issue'] = transform_report_issue
call_transform_function['transform_ESC'] = transform_ESC
call_transform_function['transform_extend_shape_to_include_source'] = transform_extend_shape_to_include_source
call_transform_function['transform_FFN'] = transform_FFN
call_transform_function['transform_FFN_932_and_TRS'] = transform_FFN_932_and_TRS
call_transform_function['transform_find_dataset_id'] = transform_find_dataset_id
call_transform_function['transform_find_entity_collection_id'] = transform_find_entity_collection_id
call_transform_function['transform_find_entity_collection_id_by_shape'] = transform_find_entity_collection_id_by_shape
call_transform_function['transform_find_restriction_info_id'] = transform_find_restriction_info_id
call_transform_function['transform_GE4_from_IC2'] = transform_GE4_from_IC2
call_transform_function['transform_GE4_from_IC2_Dataset'] = transform_GE4_from_IC2_Dataset
call_transform_function['transform_get_from_entity_collection_table'] = transform_get_from_entity_collection_table
call_transform_function['transformZI001_VSCfromenumtocodelist'] = transformZI001_VSCfromenumtocodelist
call_transform_function['transform_get_from_restriction_info_table'] = transform_get_from_restriction_info_table
call_transform_function['transform_gramspercm_to_mgperlitre'] = transform_gramspercm_to_mgperlitre
call_transform_function['transform_GWS'] = transform_GWS
call_transform_function['transform_HYP'] = transform_HYP
call_transform_function['transform_Hypsography_F_CODE'] = transform_Hypsography_F_CODE
call_transform_function['transform_IC2_from_GE4'] = transform_IC2_from_GE4
call_transform_function['transform_Industry_F_CODE'] = transform_Industry_F_CODE
call_transform_function['transform_ISS'] = transform_ISS
call_transform_function['transform_LMC'] = transform_LMC
call_transform_function['transform_LOC'] = transform_LOC
call_transform_function['transform_LUN'] = transform_LUN
call_transform_function['transform_LVN'] = transform_LVN
call_transform_function['transform_MCC'] = transform_MCC
call_transform_function['transform_MDFC'] = transform_MDFC
call_transform_function['transform_MDSP'] = transform_MDSP
call_transform_function['transform_MEL'] = transform_MEL
call_transform_function['transform_MEM'] = transform_MEM
call_transform_function['transform_metre_to_decimetre'] = transform_metre_to_decimetre
call_transform_function['transform_MFY'] = transform_MFY
call_transform_function['transform_mgperlitre_to_gramspercm'] = transform_mgperlitre_to_gramspercm
call_transform_function['transform_missing_enum_index_with_warnings'] = transform_missing_enum_index_with_warnings
call_transform_function['transform_MZN'] = transform_MZN
call_transform_function['transform_OTH'] = transform_OTH
call_transform_function['transform_placeholder_return_None'] = transform_placeholder_return_None
call_transform_function['transform_placeholder_return_source'] = transform_placeholder_return_source
call_transform_function['transform_PLT'] = transform_PLT
call_transform_function['transform_PMP'] = transform_PMP
call_transform_function['transform_PMPR'] = transform_PMPR
call_transform_function['transform_PPO'] = transform_PPO
call_transform_function['transform_PRW'] = transform_PRW
call_transform_function['transform_RIN_ROI_from_RTN_ROI'] = transform_RIN_ROI_from_RTN_ROI
call_transform_function['transform_RTN_ROI_from_RIN_ROI'] = transform_RTN_ROI_from_RIN_ROI
call_transform_function[
    'transform_set_EntityCollection_field_to_default'] = transform_set_EntityCollection_field_to_default
call_transform_function['transform_set_FUFI_to_UFI_when_needed'] = transform_set_FUFI_to_UFI_when_needed
call_transform_function['transform_set_geodbversion'] = transform_set_geodbversion
call_transform_function['transform_set_schemaversion'] = transform_set_schemaversion
call_transform_function['transform_set_to_default'] = transform_set_to_default
call_transform_function['transform_set_UFI'] = transform_set_UFI
call_transform_function['transform_SRE'] = transform_SRE
call_transform_function['transform_SSR'] = transform_SSR
call_transform_function['transform_SUYfromCPA'] = transform_SUYfromCPA
call_transform_function['transform_TRS'] = transform_TRS
call_transform_function['transform_TRS_to_LUN'] = transform_TRS_to_LUN
call_transform_function['transform_TRS_to_PTG'] = transform_TRS_to_PTG
call_transform_function['transform_TSSM'] = transform_TSSM
call_transform_function['transform_TWS'] = transform_TWS
call_transform_function['transform_VEG'] = transform_VEG
call_transform_function['transform_VLM_fromInterval'] = transform_VLM_fromInterval
call_transform_function['transform_VLM_from_nonInterval'] = transform_VLM_from_nonInterval
call_transform_function['transform_VTI_fromInterval'] = transform_VTI_fromInterval
call_transform_function['transform_VTI_toInterval'] = transform_VTI_toInterval
call_transform_function['transform_WWM'] = transform_WWM
call_transform_function['transform_ZI001_NSD'] = transform_ZI001_NSD
call_transform_function['transform_ZI001_NSN'] = transform_ZI001_NSN
call_transform_function['transform_ZI001_NSP'] = transform_ZI001_NSP
call_transform_function['transform_ZI001_SDP'] = transform_ZI001_SDP
call_transform_function['transform_ZI001_SDV'] = transform_ZI001_SDV
call_transform_function['transform_ZI001_SRT'] = transform_ZI001_SRT
call_transform_function['transform_ZI001_SSD'] = transform_ZI001_SSD
call_transform_function['transform_ZI001_SSN'] = transform_ZI001_SSN
call_transform_function['transform_ZI001_SSY'] = transform_ZI001_SSY
call_transform_function['transform_ZI010_ACC'] = transform_ZI010_ACC
call_transform_function['transform_ZI013_FMM'] = transform_ZI013_FMM
call_transform_function['transform_ZI013_PIG'] = transform_ZI013_PIG
call_transform_function['transform_ZI014_PPO'] = transform_ZI014_PPO
call_transform_function['transform_ZI024_HYP'] = transform_ZI024_HYP
call_transform_function['transform_ZI037_REL'] = transform_ZI037_REL
call_transform_function['transformCPScodelist2CPSenumvalue'] = transformCPScodelist2CPSenumvalue
call_transform_function['transformCPSenumeration2CPScodelist'] = transformCPSenumeration2CPScodelist
call_transform_function['transformCreateUFI'] = transformCreateUFI
call_transform_function['transformDFU_to_DWT'] = transformDFU_to_DWT
call_transform_function['transformDWT_to_DFU'] = transformDWT_to_DFU
call_transform_function['transformEQCtoEQS'] = transformEQCtoEQS
call_transform_function['transformEQSinteger2EQCcodelist'] = transformEQStoEQC
call_transform_function['transformEQStoEQC'] = transformEQStoEQC
call_transform_function['transformETScodelist2ETSenumvalue'] = transformETScodelist2ETSenumvalue
call_transform_function['transformETSenumeration2ETScodelist'] = transformETSenumeration2ETScodelist
call_transform_function['transformHZDcodelist2HZDenumvalue'] = transformHZDcodelist2HZDenumvalue
call_transform_function['transformHZDenumeration2HZDcodelist'] = transformHZDenumeration2HZDcodelist
call_transform_function['transformMoveOTHOnlyIfMetaAttribute'] = transformMoveOTHOnlyIfMetaAttribute
call_transform_function['transformRCGcodelist2RCGenumvalue'] = transformRCGcodelist2RCGenumvalue
call_transform_function['transformRCGenumeration2RCGcodelist'] = transformRCGenumeration2RCGcodelist
call_transform_function['transformVDTcodelist2VDTenumvalue'] = transformVDTcodelist2VDTenumvalue
call_transform_function['transformVDTenumeration2VDTcodelist'] = transformVDTenumeration2VDTcodelist
call_transform_function['transformZI001_SRTcodelist2ZI001_SRTenumvalue'] = transformZI001_SRTcodelist2ZI001_SRTenumvalue
call_transform_function[
    'transformZI001_SRTenumeration2ZI001_SRTcodelist'] = transformZI001_SRTenumeration2ZI001_SRTcodelist
call_transform_function['transformZI001_SRTfromcodelisttoenum'] = transformZI001_SRTcodelist2ZI001_SRTenumvalue
call_transform_function['transformZI001_VSCcodelist2ZI001_VSCenumvalue'] = transformZI001_VSCcodelist2ZI001_VSCenumvalue
call_transform_function['transformZI001_VSCenumvalue2ZI001_VSCcodelist'] = transformZI001_VSCenumvalue2ZI001_VSCcodelist
call_transform_function['transformZI001_VSCfromcodelisttoenum'] = transformZI001_VSCcodelist2ZI001_VSCenumvalue
call_transform_function[
    'transformZI004_RCGenumeration2ZI004_RCGcodelist'] = transformZI004_RCGenumeration2ZI004_RCGcodelist
call_transform_function['transformZI004_RCGfromcodelisttoenum'] = transformZI004_RCGfromcodelisttoenum
call_transform_function['transformZI016_WTC'] = transformZI016_WTC
call_transform_function['transformZI016_WTC_CIW'] = transformZI016_WTC_CIW
call_transform_function['transformZVH_VDTenumeration2ZVH_VDTcodelist'] = transformZVH_VDTenumeration2ZVH_VDTcodelist
call_transform_function['transform_CollectionSrfMetadataSrf_F_CODE'] = transform_CollectionSrfMetadataSrf_F_CODE
call_transform_function['transformMetadataSrfCollectionSrf_F_CODE'] = transformMetadataSrfCollectionSrf_F_CODE
call_transform_function['transform_LDC'] = transform_LDC
call_transform_function['transform_UBC_from_BSC'] = transform_UBC_from_BSC
call_transform_function['transform_ZI001_SPS'] = transform_ZI001_SPS
call_transform_function['transform_Industry_FCSubtype'] = transform_Industry_FCSubtype
call_transform_function['transform_Hypsography_FCSubtype'] = transform_Hypsography_FCSubtype
call_transform_function['transform_Structure_FCSubtype'] = transform_Structure_FCSubtype
call_transform_function['transform_CollectionSrfMetadataSrf_FCSubtype'] = transform_CollectionSrfMetadataSrf_FCSubtype
call_transform_function['transformMetadataSrfCollectionSrf_FCSubtype'] = transformMetadataSrfCollectionSrf_FCSubtype
call_transform_function['transform_FCSubtype'] = transform_FCSubtype
call_transform_function['transform_SaveSomeRestrictionInfoToOTH'] = transform_SaveSomeRestrictionInfoToOTH
call_transform_function['transform_SaveRelatedECInformation'] = transform_SaveRelatedECInformation
call_transform_function['transform_saveECUFIToOther'] = transform_saveECUFIToOther
call_transform_function['transform_saveDatasetURItoOther'] = transform_saveDatasetURItoOther
call_transform_function['transform_put_source_to_OTH'] = transform_put_source_to_OTH
call_transform_function['find_matchrestrictioninfoCCNCDR'] = find_matchrestrictioninfoCCNCDR
call_transform_function['transform_CVT'] = transform_CVT
call_transform_function[
    'check_for_and_transform_missing_enum_index_target_diff'] = check_for_and_transform_missing_enum_index_target_diff
call_transform_function['transform_WRDB_F_CODE'] = transform_WRDB_F_CODE
call_transform_function['transform_combine_FURL'] = transform_combine_FURL
call_transform_function['transform_pull_wrdb_extension'] = transform_pull_wrdb_extension
call_transform_function['transform_millions_to_ones'] = transform_millions_to_ones
call_transform_function['transform_thousands_to_ones'] = transform_thousands_to_ones
call_transform_function['transform_slope_degree_to_percent'] = transform_slope_degree_to_percent
call_transform_function['transform_hydro_country_to_GE4'] = transform_hydro_country_to_GE4
call_transform_function['transform_no_target_save_to_oth'] = transform_no_target_save_to_oth
call_transform_function['transform_MDSP_MEL'] = transform_MDSP_MEL
